/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */
/***********\
*           *
*  KlO: EXO *
*  BODY     *
*           *
\***********/

/* Exo stand for "EXternal Objects", the clean Object layer on top of klone
 */

#include "EXTERN.h"
#include <stdio.h>
#include "klone.h"
#include "kl_atom.h"
#include "kl_list.h"
#include "kl_number.h"
#include "kl_string.h"
#include "kl_func.h"
#include "kl_hash.h"
#include "kl_stream.h"
#include "kl_link.h"
#include "klgeneric.h"
#include "INTERN.h"
#include "kl_exo.h"

int 			KlExoDestructor();
KlO 			KlExoMethodAdd();
unsigned int            KlGenericFunctionDeclare();
KlO                     KlCheckMethodValidity();
KlAtom                  KlClassSlotName();
KlO			KlDocumentationGet();
KlList			KlExoLocalSlots();
KlO			KlExoClassDeclareBuiltInSlots();
KlO			KlExoDescribeInstance();
KlO	 		KlExoDescribeMethod();
KlO			KlExoDescribeClass();
KlExoClass              KlExoClassMake();
KlO                     KlAccessClassChildren();
KlExoClass		KlRefreshObsoleteClass();
KlO			KlExoMakeInstance();
KlSubr                  KlMakeInstanceFromMetaclassSubr;
KlSlotHooks             KlSlotHooksMake();
KlO 			KlRemoveSlotHooks();


/* KlExoClassMake 
 * raw C primitive to create a class
 */

KlExoClass
KlExoClassMake(name, father, newslots, nnewslots, inits, metaclass,
	       documentations, documentation, get_hooks, put_hooks, 
	       slot_properties)
    KlAtom name;
    KlExoClass father;
    KlAtom *newslots;
    int nnewslots;
    KlO *inits;				/* slot init values (:initform) */
    KlExoClass metaclass;
    KlO *documentations;		/* slot docs */
    KlO documentation;			/* class doc */
    KlO *get_hooks;			/* slot hook on get (:get) */
    KlO *put_hooks;			/* slot hook on put (:put) */
    KlO *slot_properties;		/* properties (plist) of slots */
{
    /* alloc the class + room for class slots at the end */
    KlExoClass obj = (KlExoClass)
	KlOMakeOfSize(KlExoClassType,
		      sizeof(struct _KlExoClass) +
		      KLSO * (metaclass->slots->size - KlExoClassBuiltInSlots));
    KlHash slots = KlHashAlloc(nnewslots + father->slots->size);
    KlHash methods = KlHashAlloc(father->methods->size);
    int i, nslots = father->slots->size;
    int modified_slots = 0;
    int actual_newslots;
    UInt offset;
    int curslot;
    KlExoClass old_class;
    int hooks = (get_hooks && put_hooks); /* allow these to be null */

    slots->ref_counted = 0;
    methods->ref_counted = KlHashRefV;
    KlIncRef(obj->exoclass = metaclass);

    KlIncRef(obj->name = name);		/* insertion in class tree */
    obj->son = 0;
    KlIncRef(obj->father = father);
    if (father->son) {
	obj->brother = father->son;
    } else {
	obj->brother = 0;
    }
    KlIncRef(father->son = obj);

    obj->slots = slots;			/* slots inheritance */
    KlIncRef(slots);
    KlHashFORBEGIN(father->slots, cell) {
	if (KlPI_INT(cell->object)) {	/* normal slot */
	    KlHashPut(slots, cell->key, cell->object);
	} else {			/* copy slothook */
	    KlSlotHooksMake(obj, cell->key, 
			    ((KlSlotHooks) cell->object)->offset, 
			    ((KlSlotHooks) cell->object)->get, 
			    ((KlSlotHooks) cell->object)->put);
	}
    } KlHashFOREND;
    curslot = nslots;
    for (i = 0; i < nnewslots; i++) {
	if ((UInt) (KlHashGet(slots, newslots[i], KlPI_INVALID_VALUE)) 
	    != KlPI_INVALID_VALUE) {
	    /* check if there was not already a slot of this name */
	    modified_slots++;
	} else {
	    KlHashPut(slots, newslots[i], curslot);
	    curslot++;
	}
    }

    actual_newslots = nnewslots - modified_slots;
    /* inits */
#ifdef DEBUG				/* avoid warning for malloc(0) */
    if (!(nslots + actual_newslots)) {
	obj->init = (KlO *) Malloc(KLSO);
	obj->slot_properties = (KlO *) Malloc(KLSO);
    } else
#endif					/* DEBUG */
	{
	    obj->init = (KlO *) Malloc((nslots + actual_newslots) * KLSO);
	    obj->slot_properties = (KlO *) Malloc((nslots + actual_newslots) 
						  * KLSO);
	}
    KlCopyNObjects(obj->father->init, obj->init, nslots); /* inherited */
    KlCopyNObjects(obj->father->slot_properties, obj->slot_properties, nslots);
    /* add new or redefine olders */
    curslot = nslots;
    for (i = 0; i < nnewslots; i++) {
	if ((offset =
	     (UInt) KlHashGet(father->slots, newslots[i], KlPI_INVALID_VALUE))
	    != KlPI_INVALID_VALUE) {
	    /* redefine old slot */
	    if (inits[i]) {		/* if no init, inherit from parent */
		KlDecRef(obj->init[KlExoActualOffset(offset)]);
		KlIncRef(obj->init[KlExoActualOffset(offset)] = inits[i]);
	    }
	    if (slot_properties[i]) {
		KlDecRef(obj->slot_properties[KlExoActualOffset(offset)]);
		KlIncRef(obj->slot_properties[KlExoActualOffset(offset)] = 
			 slot_properties[i]);
	    }
	} else {
	    /* declare new slot */
	    KlIncRef(obj->init[curslot] = inits[i] ? inits[i] : NIL);
	    KlIncRef(obj->slot_properties[curslot] = slot_properties[i] ?
		slot_properties[i] : NIL);

	    curslot++;
	}
	/* add here the :get and :put keywords, if present */
	if (hooks && (get_hooks[i] || put_hooks[i])) {
	    offset = (UInt) KlHashGet(slots, newslots[i], KlPI_INVALID_VALUE);
	    if (KlPI_INT(offset)) {	/* create one, if there was not */
		KlSlotHooksMake(obj, newslots[i], offset,
				get_hooks[i] ? get_hooks[i] : NIL,
				put_hooks[i] ? put_hooks[i] : NIL);
	    } else {			/* update existing one */
		if (get_hooks[i]) {
		    KlDecRef(((KlSlotHooks) offset)->get);
		    KlIncRef(((KlSlotHooks) offset)->get = get_hooks[i]);
		}
		if (put_hooks[i]) {
		    KlDecRef(((KlSlotHooks) offset)->put);
		    KlIncRef(((KlSlotHooks) offset)->put = put_hooks[i]);
		}
		/* check that we did not remove it */
		if (((KlSlotHooks) offset)->put == NIL
		    && ((KlSlotHooks) offset)->get == NIL)
		    KlRemoveSlotHooks(obj, newslots[i], offset);
	    }
	}
    }    

    obj->methods = methods;		/* methods inheritance */
    KlIncRef(methods);
    methods->ref_counted = KlHashRefV;
    KlHashFORBEGIN(father->methods, cell) {
	KlHashPut(methods, cell->key, cell->object);
    } KlHashFOREND;

    /* exoHooks */
    obj->exoHooks = father->exoHooks;
    /* initialisation of class slots */
    if (metaclass != KlExoClassMeta)
	KlExoInitializeClassSlots(obj);

    /* grab the redefined class if any, before setting the atom */
    if (!((((KlO) (old_class = (KlExoClass) name->c_val)) != KlUndef)
	  && KlIsAnExoClass(old_class)))
	old_class = 0;
    /* Ok, sets the name value to the class */
    KlSend_setq(name, obj);

    /* if we redefine a class, get the old one out of the class tree */
    if (old_class) {
        KlList subclasses = (KlList) KlAccessClassChildren (old_class);
	KlExoClassPrune(old_class);

        /* initialize new class with the methods of old class */
        KlHashFORBEGIN(old_class->methods, cell) {
	    KlHashPut(obj->methods, cell->key, cell->object);
        } KlHashFOREND;
	
        /* update subtree : propagate the redefinition to  the subclasses */
	if (subclasses != (KlList) NIL) {
	    for (i= (subclasses->size - 1) ; i>=0 ; i--) {
		KlEval(KlExoDescribeClass(subclasses->list[i]));
	    }
        }
    }

    /* sets the documentations */
    KlExoClassDocumentationSet(obj, documentation, newslots,
			       nnewslots, documentations);

    return obj;
}

/* KlDefClass_BEGIN */

/* KlDefClass
 * klone interface to KlExoClassMake
 */

KlO
KlDefClass(argc, argv)
    int argc;
    KlAtom *argv;
{
    KlExoClass exoclass;
    KlExoClass father;
    int nnewslots;
    int oldslots;
    KlO *newslots;
    KlList arglist;
    KlO *inits;
    KlO *slot_properties;
    KlO *get_hooks;
    KlO *put_hooks;
    int i, j;
    KlO documentation;
    KlO *documentations;
    KlO metaclassname;
    KlExoClass metaclass = 0;

    if (argc < 3)
	return KlBadNumberOfArguments(argc);
    KlMustBeSymbol(argv[0], 0);

    if (argv[1] == (KlAtom) NIL) {	/* general syntax check */
	father = KlExoClassRoot;
    } else {
	if (KlIsAList(argv[1])) {
	    KlList clist = (KlList) argv[1];
	    KlMustBeOrEval(father, clist->list[0], KlExoClass,
			   KlIsAnExoClass(father),
			   KlError(KlE_NO_CLASS, clist->list[0]));
    
	} else {
	    KlMustBeOrEval(father, argv[1], KlExoClass, KlIsAnExoClass(father),
			   KlError(KlE_NO_CLASS, argv[1]));
	}
    }

    /* prevent using an obsolete father */
    father = KlRefreshObsoleteClass(father);

    oldslots = father->slots->size;

    arglist = (KlList) argv[2];
    KlMustBeList(arglist, 2);
    if (KlTrueP(arglist)) {		/* parses slot descriptions */
	nnewslots = arglist->size;
	newslots = KlAlloca(nnewslots);
	inits = KlAlloca(nnewslots);
	slot_properties = KlAlloca(nnewslots);
	documentations = KlAlloca(nnewslots);
	get_hooks = KlAlloca(nnewslots);
	put_hooks = KlAlloca(nnewslots);

	for (i = 0; i < nnewslots; i++) { /* parses each slot desc */

	    if (KlIsAList(arglist->list[i])) { /* options (name options...) */
		KlList arg = ((KlList)(arglist->list[i]));
		if (!(arg->size) || !KlIsASymbol(arg->list[0]))
		    /* the list is empty or its first element is not an Atom */
		    KlError2(KlE_BAD_CLASS_DEF, argv[0], arg);
		/* retrieve the slot name */
		newslots[i] = (KlO) KlAtomOrKey(arg->list[0]);

		/* now each slot desc keywords */
		KlParseKeywords(arg->size, arg->list, 1);

		/* no value = 0, keep the previous def if any */
		inits[i] = KlKeyVal(KlK_initform, 0);
		/* this is false: Common lisp specifies initforms are only
		   evaluated at make-instance, not at defclass
		inits[i] = KlEvalOrNull(KlKeyVal(KlK_initform, 0));
		*/
		documentations[i] =
		    KlEvalOrNull(KlKeyVal(KlK_documentation, 0));
		get_hooks[i] = KlEvalOrNull(KlKeyVal(KlK_get, 0));
		put_hooks[i] = KlEvalOrNull(KlKeyVal(KlK_put, 0));
		slot_properties[i] = KlEvalOrNull(KlKeyVal(KlK_properties, 0));
		/* be tolerant with obsolete calls */
		KlCheckUnvalidKeywords(arg->size, arg->list, 1);

	    } else {			/* no slot options, slot name alone */
		/* the slot description is not a list, it must be an Atom */
		if (!KlIsASymbol(arglist->list[i]))
		    KlError2(KlE_BAD_CLASS_DEF, argv[0], arglist->list[i]);
		newslots[i] = (KlO) KlAtomOrKey(arglist->list[i]);
		inits[i] = 0;
		documentations[i] = 0;
		get_hooks[i] = 0;
		put_hooks[i] = 0;
		slot_properties[i] = 0;
	    }
	    /* check slots are not defined twice */
	    for (j = 0; j < i; j++) {
		if (newslots[j] == newslots[i]) {
		    KlError2(KlE_BAD_CLASS_DEF, argv[0], newslots[i]);
		}
	    }
	}
    } else {				/* no slots at all */
	nnewslots = 0;
	newslots = 0;
	inits = 0;
	documentations = 0;
	slot_properties = 0;
	get_hooks = 0;
	put_hooks = 0;
    }

    /* class keywords */
    KlParseKeywords(argc, argv, 3);
    documentation = KlKeyVal(KlK_documentation, 0);
    metaclassname = KlKeyVal(KlK_metaclass, 0);
    KlCheckUnvalidKeywords(argc, argv, 3);

    if (metaclassname) {		/* check metaclass validity */
	KlMustBeOrEval(metaclass, metaclassname, KlExoClass,
		       KlIsAnExoClass(metaclass),
		       KlError(KlE_NO_CLASS, metaclassname));
	KlExoClassDeclareBuiltInSlots(metaclass);
    }
    if (!metaclass) {
	metaclass = father->exoclass;
    }

    /* call  KlExoClassMake for doing the actual job */
    exoclass = KlExoClassMake(argv[0], father, newslots, nnewslots,
			      inits, metaclass, documentations, documentation,
			      get_hooks, put_hooks, slot_properties);
    
    return (KlO) exoclass;
}

/* KlExoInitializeClassSlots 
 * does the equivalent of make-instance for Exo to ExoClasses:
 * initializes the class slots from the metaclass description
 */

KlExoInitializeClassSlots(exoclass)
    KlExoClass exoclass;
{
    int i;

    for (i = KlExoClassBuiltInSlots;
	 i < exoclass->exoclass->slots->size;
	 i++) {
	KlIncRef(KlCastExo(exoclass)->slots[i] =
		 KlSend_eval(exoclass->exoclass->init[i]));
    }
    if (KlExoHasHook(exoclass, KlSelEHInit))
	KlApplyMethod1(KlExoMethodNamed(exoclass, KlA_init),
		       KlA_init, exoclass);
}

/* KlExoClassBootstrapMake
 * we need a way to create the primitive classes, this is it
 */

KlExoClass
KlExoClassBootstrapMake()
{
    KlExoClass obj = (KlExoClass) KlOMakeOfSize(KlExoClassType,
						sizeof(struct _KlExoClass));

    KlOZero(obj, sizeof(struct _KlExoClass));
    return obj;
}

/* KlExoClassDeclareBuiltInSlots
 * fills the default metaclass (or any class to be used as metaclass) with
 * built-in slots: (not to be used, though...)
 * :name
 * :father CAUTION: not refcounted! DO NOT SET!!!
 * :son
 * :brother
 * :methods
 * :slots
 * ::*inits CAUTION: not a klone object!!! (hence the ::), accessing will crash
 * ::exohooks CAUTION: not a klone object!!! (hence the ::)
 * metaclass
 *
 * the functions tests the presence of the built-in slot ::*inits to see if
 * the class has already been processed. 
 */

KlO
KlExoClassDeclareBuiltInSlots(exoclass)
    KlExoClass exoclass;
{
    KlHash slots = exoclass->slots;
    int i;

    /* if already a metaclass, no adjustements needed */
    if (KlIsAMetaclass(exoclass)) {
	return (KlO) exoclass;
    }
    /* not a son of a metaclass? then we cannot add the slots since
       they would end up at a bad offset*/
    if (slots->size) {
	KlError2(KlE_BAD_CLASS_DEF, KlK_metaclass, exoclass);
    }
    /* then we must fill in (bootstrap) a new class */
    KlHashPut(slots, KlK_name, 0);
    KlHashPut(slots, KlK_father, 1);
    KlHashPut(slots, KlK_son, 2);
    KlHashPut(slots, KlK_brother, 3);
    KlHashPut(slots, KlK_methods, 4);
    KlHashPut(slots, KlK_slots, 5);
    KlHashPut(slots, KlK_STAR_inits, 6);
    KlHashPut(slots, KlK_exo_hooks, 7);
    KlHashPut(slots, KlK_properties, 8);

    ASSERT(KlExoClassBuiltInSlots == 9);
    exoclass->init = (KlO *)
	Realloc(exoclass->init, KlExoClassBuiltInSlots *KLSO);
    exoclass->slot_properties = (KlO *)
	Realloc(exoclass->slot_properties, KlExoClassBuiltInSlots *KLSO);
    for (i = 0; i < KlExoClassBuiltInSlots; i++) {
	KlIncRef(exoclass->init[i] = NIL);
	KlIncRef(exoclass->slot_properties[i] = NIL);
    }
    KlExoMethodAdd(exoclass, KlA_make_instance,
		   KlMakeInstanceFromMetaclassSubr, KlSelEHMakeInstance);
    return (KlO) exoclass;
}

/* KlExoClassFree
 * removing an already pruned class
 * function called by the GC
 */

KlO
KlExoClassFree(exoclass)
    KlExoClass exoclass;
{
    if (KlExoHasHook(exoclass, KlSelEHFree)) {
	KlO result;
	int normal;
	KlGCMark();
	KlCatch(KlA_ALL,
		KlApplyMethod1(KlExoMethodNamed(exoclass, KlA_free),
			       KlA_free, exoclass),
		result, normal);
	KlGC();
	if (exoclass->reference_count) { /* the destructor pointed to it */
	    exoclass->reference_count |= 1; /* escape its fate, clean up */
	    return (KlO) exoclass;
	}

    } 

    /* actual imediate obliteration */
    KlDecRefList(exoclass->slots->size, exoclass->init);
    KlDecRefList(exoclass->slots->size, exoclass->slot_properties);
    Free(exoclass->init);
    Free(exoclass->slot_properties);

    KlDecRefNonNull(exoclass->exoclass);
    KlDecRefNonNull(exoclass->name);
    KlDecRef(exoclass->father);
    KlDecRef(exoclass->son);
    KlDecRef(exoclass->brother);
    KlDecRefNonNull(exoclass->methods);

    /* remove slothooks, we must do it since they are not garbaged */
    KlHashFORBEGIN(exoclass->slots, cell) {
	KlRemoveSlotHooks(exoclass, cell->key, cell->object);
    } KlHashFOREND;
    KlDecRefNonNull(exoclass->slots);

    if (exoclass->exoclass != KlExoClassMeta) {
	int i;

	for (i = KlExoClassBuiltInSlots;
	     i < exoclass->exoclass->slots->size;
	     i++) {
	    KlDecRef(KlCastExo(exoclass)->slots[i]);
	}
    }
    Free(exoclass);
    return (KlO)  exoclass;
}

/* KlExoClassPrune:
 * to remove a class from the class tree
 */

static char *KlObsoleteClassPrefix = "ObsoleteClass:";
static int KlObsoleteClassPrefixLen = 14; /* strlen of the above */

KlExoClassPrune(exoclass)
    KlExoClass exoclass;
{
    KlAtom name = exoclass->name;
    ASSERT (exoclass->father);

    /* remove documentations */
    KlExoClassDocumentationClear(exoclass);

    /* prefix the name by "ObsoleteClass:", if not already done */
    if (strncmp(name->p_name, KlObsoleteClassPrefix,
		 KlObsoleteClassPrefixLen)) {
	char *prefixed_name = (char *) Malloc(strlen(name->p_name) + 1 +
					      KlObsoleteClassPrefixLen);
	strcpy(stpcpy(prefixed_name, KlObsoleteClassPrefix),
	       name->p_name);
	KlDecRef(name);
	KlIncRef(exoclass->name = KlIntern(prefixed_name));
	Free(prefixed_name);
    }
    KlDecRef(exoclass);
    if (exoclass->father->son == exoclass) {
	exoclass->father->son = exoclass->brother; /* ref compensated */
    } else {
	KlExoClass elder_brother = exoclass->father->son;
	while (elder_brother->brother != exoclass) {
	    elder_brother = elder_brother->brother;
	    ASSERT(elder_brother);
	}
	elder_brother->brother = exoclass->brother;
    }
    exoclass->brother = 0;
    exoclass->exoHooks |= KlSelEHOBSOLETE; /* set obsolete bit */
}

/* KlRefreshObsoleteClass
 * gets the class replacing the obsolete one
 */

KlExoClass
KlRefreshObsoleteClass(exoclass)
    KlExoClass exoclass;
{
    char *name;
    KlAtom newname;
    KlExoClass newclass;

    if (!(exoclass->exoHooks & KlSelEHOBSOLETE))
	return exoclass;		/* still alive */

    ASSERT(strlen(exoclass->name->p_name) > KlObsoleteClassPrefixLen);
    name = (char *) Malloc(strlen(exoclass->name->p_name)
			   + 1 - KlObsoleteClassPrefixLen);
    strcpy(name, exoclass->name->p_name + KlObsoleteClassPrefixLen);
    newname = KlIntern(name);
    Free(name);
    if ((((KlO) (newclass = (KlExoClass) newname->c_val)) != KlUndef)
	  && KlIsAnExoClass(newclass)) {
	return newclass;
    } else {
	return (KlExoClass) KlError(KlE_NO_CLASS, exoclass->name);
    }
}

/* KlExoMake
 * Raw function to create an instance
 */

KlExo
KlExoMake(exoclass)
    KlExoClass exoclass;
{
    int nslots = exoclass->slots->size;
    KlExo obj = (KlExo) KlOMakeOfSize(KlExoType,
				      (sizeof(struct _KlExo)
				       + sizeof(KlO) * (nslots - 1)));
    KlIncRef(obj->exoclass = exoclass);
    bzero(&(obj->slots[0]), nslots * sizeof(KlO));
    return obj;
}

/* applying a class is buildiung an instance with elements
 */

KlO
KlExoClassApply(exoclass, list)
    KlExoClass exoclass;
    KlList list;
{
    return (KlO) KlExoMakeInstance(list->size, list->list);
}

/*****************************************************************************\
* 			Class documentation management                        *
\*****************************************************************************/
/* manages the doc associated to the class symbol and to the slots of the
 * form class.slot, e.g.: Point, Point.x, Point.y
 */

KlExoClassDocumentationSet(exoclass, documentation, slots, nslots,
			   documentations)
    KlExoClass exoclass;
    KlO documentation;			/* class doc */
    KlAtom *slots;			/* local slots */
    int nslots;				/* number of slots */
    KlO *documentations;		/* slot docs */
{
    int i;

    if (documentation) {
	KlDocumentationSet(exoclass->name, documentation);
    } else {
	/* if no documentation, we must set the documentation string to nil,
	 * because the symbol may have already a documentation string */
        KlDocumentationClear(exoclass->name);
    }
    for (i = 0; i < nslots; i++) {
	if (documentations[i]) {
	    KlDocumentationSet(KlInternDotPrefixed
			       (exoclass->name->p_name, slots[i]->p_name),
			       documentations[i]);
	}
    }    
}

KlExoClassDocumentationClear(exoclass)
    KlExoClass exoclass;
{
    KlList localSlots = (KlList) KlExoLocalSlots(exoclass);
    int i;

    KlDocumentationClear(exoclass->name);
    for (i=0 ; i< localSlots->size; i++) {
	KlDocumentationClear(KlInternDotPrefixed(exoclass->name->p_name,
				((KlAtom) localSlots->list[i])->p_name)); 
    }
}


/*****************************************************************************\
*                              Class Facilities                               *
\*****************************************************************************/

/* KlExoClassLocalMethods
 *
 * Return a list local methods
*/

KlList
KlExoClassLocalMethods (exoclass)
     KlExoClass exoclass;
{
    KlList list =KlListNMake(0);

    /* create a list of local methods */
    KlHashFORBEGIN (exoclass->methods, cell) {
	if (KlExoMethodIsLocal(cell->key, exoclass))
	    KlListAppend (list, cell->key);
    } KlHashFOREND;
    return list;
}

/* convenience function */

int
KlExoHooksEqual(exoclass1, exoclass2, slot)
    KlExoClass exoclass1;
    KlExoClass exoclass2;
    KlO slot;
{
    UInt offset1, offset2;
    offset1 = (UInt) KlHashGet(exoclass1->slots, slot, KlPI_INVALID_VALUE);
    offset2 = (UInt) KlHashGet(exoclass2->slots, slot, KlPI_INVALID_VALUE);

    if (offset1 == offset2)
	return 1;			/* both not hooked */
    if (KlPI_INT(offset1) || KlPI_INT(offset2))
	return 0;			/* one is hooked the other not */
    /* both hooked */
    if (((KlSlotHooks) offset1)->get == ((KlSlotHooks) offset2)->get
	&& ((KlSlotHooks) offset1)->put == ((KlSlotHooks) offset2)->put)
	return 1;
    return 0;
}

/* KlExoSlotIsLocal
 *
 * Test wether a slot is local (return 1)
 *      or is inherited (return 0)
*/

KlExoSlotIsLocal (slot,exoclass)
     KlAtom slot;
     KlExoClass exoclass;
{
    UInt offset;

    if (!exoclass->father)		/* Object has no inherited slots */
	return 1;

    /* the class has a superclass */
    if ((offset = (UInt) KlHashGet(exoclass->father->slots,slot, 
				  KlPI_INVALID_VALUE))
	== KlPI_INVALID_VALUE)
	/* there is no slot with the same name, in the superclass */
	return  1;

    /* there is a slot with the same name in the superclass */
    /* Verify that the slot options are the same */
    return 
	(
	 /* test initform slot option */
	 (exoclass->init[KlExoActualOffset(offset)]
	  != exoclass->father->init[KlExoActualOffset(offset)])
	 ||
         /* test properties */
	 (exoclass->slot_properties[KlExoActualOffset(offset)]
	  != exoclass->father->slot_properties[KlExoActualOffset(offset)])
	 ||
	 /* test documentation slot option */
	 ((KlDocumentationGet(KlInternDotPrefixed(exoclass->name->p_name, 
						  slot->p_name)))
	  !=
	  (KlDocumentationGet(KlInternDotPrefixed(exoclass->father
						  ->name->p_name,
						  slot->p_name))))
	 || !KlExoHooksEqual(exoclass, exoclass->father, slot)
	 );
}

/* KlExoClassSlotsList
 *
 * Return a list of slots (local and inherited of a given class
 *
*/

KlList
KlExoClassSlotsList(exoclass)
    KlExoClass exoclass;
{
    KlList list;
    if (KlIsAMetaclass(exoclass)) {
	int offset;
	list = KlListNMake(exoclass->slots->size - KlExoClassBuiltInSlots);
	KlHashFORBEGIN(exoclass->slots, cell) {
	    offset = (int)(cell->object) - KlExoClassBuiltInSlots;
	    if (offset >= 0) {		/* skip built-ins */
		KlIncRef(list->list[offset] = cell->key);
	    }
	} KlHashFOREND;
	
    } else {				/* normal class */
	list = KlListNMake(exoclass->slots->size);
	KlHashFORBEGIN(exoclass->slots, cell) {
	    KlIncRef(list->list[KlExoActualOffset(cell->object)] = cell->key);
	} KlHashFOREND;
    }
    return  list;
}

/*
 * KlAccessClassSots
 * 
 * Return a list of  slots for a given class.
 *
 * klone-interface: class-slots 
*/
/*
KlO
KlAccessClassSlots(exoclass)
    KlExoClass exoclass;
{
  KlMustBeExoClass(exoclass, 0);
  return (KlO) KlExoClassSlotsList(exoclass);
}
*/

/* KlExoLocalSlots
 * 
 * Return a list of local slots for a given class.
*/

KlList
KlExoLocalSlots(exoclass) 
     KlExoClass exoclass; 
{ 
  KlList result = KlListNMake (0);
  KlList classSlotsList = KlExoClassSlotsList(exoclass);
  int i;
  for (i=0 ; i< classSlotsList->size; i++) {
    if (KlExoSlotIsLocal(classSlotsList->list[i], exoclass)) {
       KlListAppend (result,classSlotsList->list[i]);
     }
  }
  return result;
}


/*
 * KlExoRemoveClass
 *
 * Remove a class
*/

KlO
KlExoRemoveClass (exoclass)
     KlExoClass exoclass;
{
    int i;
    KlList localSlots;
    
    KlMustBeExoClass(exoclass, 0);
  
    /* remove Class or Object or already removed class is FORBIDDEN */
    if ((exoclass == KlExoClassMeta)
	|| (exoclass == KlExoClassRoot)
	|| (exoclass->exoHooks & KlSelEHOBSOLETE)
	) {
	return NIL;
    }

    localSlots = KlExoLocalSlots(exoclass);
    /* remove documentation string if exists */
    KlDocumentationClear(exoclass->name);
    
    /* For each slot, delete documentation string if exists */
    for (i=0 ; i< localSlots->size; i++) {
	KlAtom slot= (KlAtom) localSlots->list[i];
	KlDocumentationClear(KlInternDotPrefixed(exoclass->name->p_name,
						 slot->p_name));
    }
        
    if (exoclass->son) {
	/* a son exist, propagate to class son */
	KlExoClass next, brother = exoclass->son->brother;
	KlExoRemoveClass(exoclass->son);
	/* propagate to all subclasses */
	while (brother) {
	    next = brother->brother;
	    KlExoRemoveClass(brother);
	    brother = next;
	}
    }

    if ((((KlAtom) exoclass->name)->c_val != KlUndef) 
	&&  ((KlExoClass) exoclass->name->c_val == exoclass)) {
	KlAtomUnbind (exoclass->name);
    }

    KlExoClassPrune (exoclass);

    return TRU;
}

/*****************************************************************************\
* 				  instances                                   *
\*****************************************************************************/

/* make-instance
 * klone interface to KlExoMake
 * WARNING: if a put slothook is present on a slot, it will NOT be applied to
 * init values!
 * if you must perform side effects, use a "init" exoHook afterwards
 * Rationale: it is not possible to perform side effects while building the
 * object since some fields might be null pointers and trigger an error!
 */

KlO
KlExoMakeInstance(argc, argv)
    int argc;
    KlAtom *argv;
{
    KlExoClass exoclass;
    KlExo obj;
    int i;
    KlAtom slot;
    int offset;
    KlExpr method;
    KlO *newslots = 0;
    int nnewslots;
    
    if (!argc)
	return KlBadNumberOfArguments(argc);

    KlMustBeOrEval(exoclass, argv[0], KlExoClass, KlIsAnExoClass(exoclass),
		   KlError(KlE_NO_CLASS, argv[0]));

    if (exoclass->exoHooks & KlSelEHMakeInstance) {
	return KlApplyMethodN(KlHashGet(exoclass->methods,
					KlA_make_instance,
					NIL),
			      KlA_make_instance, argc, argv);
    }
    obj = KlExoMake(exoclass);
    if (!(argc & 1))
	return KlBadNumberOfArguments(argc);
    for (i = 1; i < argc; i += 2) {	/* init fields specified in args */
	slot = KlAtomOrKey(argv[i]);
	KlMustBeSymbol(slot, i);
	offset = (UInt) KlHashGet(obj->exoclass->slots, slot, 
				  KlPI_INVALID_VALUE);
	if (KlPI_INT(offset)) {		/* normal slot */
	    KlIncRef(obj->slots[offset] = (KlO) argv[i+1]);
	} else if (KlPI_INVALID(offset)) { /* not found */
	    if (exoclass->exoHooks & KlSelEHInit) {
		KlList dummy_list;
		/* store them to pass to function */
		if (!newslots) {
		    dummy_list = KlListNMake(1); /* allocs on stack */
		    dummy_list->size = -1;
		    nnewslots = 1;
		    newslots = dummy_list->list;
		    newslots[0] = (KlO) obj;
		}
		nnewslots += 2;
		dummy_list->list = newslots = (KlO *)
		    Realloc(newslots, KLSO * nnewslots);
		newslots[nnewslots - 2] = (KlO) slot;
		newslots[nnewslots - 1] = (KlO) argv[i+1];
	    } else {			/* neither slot nor method */
		return KlError2(KlE_NO_ELEMENT, slot, obj->exoclass->name);
	    }
	} else {			/* hooked */
	    KlIncRef(obj->slots[((KlSlotHooks) offset)->offset] =
		     (KlO) argv[i+1]);
	}
    }
    /* then, fill with inits slots that were not initialized by arguments */
    for (i = 0; i < exoclass->slots->size; i++) {
	if (!obj->slots[i]) {
	    KlIncRef(obj->slots[i] = KlSend_eval(exoclass->init[i]));
	}
    }

    /* if present, execute the "init" exoHook 
     * normally init is called without arguments (only self object), but if 
     * non-existing slots were passed as arguments, no error is signalled and
     * they are handled to the init routine, a bit like keywords.
     * WARNING: these can be atoms, not actually keywords
     */
    if (exoclass->exoHooks & KlSelEHInit) {
	if (newslots) {
	    KlApplyMethodN(KlExoMethodNamed(obj, KlA_init), KlA_init,
			   nnewslots, newslots);
	} else {
	    KlApplyMethod1(KlExoMethodNamed(obj, KlA_init), KlA_init, obj);
	}
    }
    
    return (KlO) obj;
}

/* *:make-instance
 * raw klone interface to KlExoMake
 */

KlO
KlExoRawMakeInstance(argc, argv)
    int argc;
    KlAtom *argv;
{
    KlExoClass exoclass;
    KlExo obj;
    int i;
    
    if (!argc)
	return KlBadNumberOfArguments(argc);

    KlMustBeOrEval(exoclass, argv[0], KlExoClass, KlIsAnExoClass(exoclass),
		   KlError(KlE_NO_CLASS, argv[0]));

    if (!argc)
	return KlBadNumberOfArguments(argc);

    KlMustBeOrEval(exoclass, argv[0], KlExoClass, KlIsAnExoClass(exoclass),
		   KlError(KlE_NO_CLASS, argv[0]));
    obj = KlExoMake(exoclass);
    for (i = 0; i < exoclass->slots->size; i++) {
	KlIncRef(obj->slots[i] = NIL);
    }
    return (KlO) obj;
}

/* for now, we prevent people from instanciating metaclasses, this is not 
 * supported 
 */

KlO
KlMakeInstanceFromMetaclass(argc, argv)
    int argc;
    KlAtom *argv;
{
    return KlError2(KlE_NO_METACLASS_INSTANCE, argv[0], KlListKl(argc, argv));
}

KlO
KlExoFree(obj)
    KlExo obj;
{
    KlO *p = &(obj->slots[0]);
    KlO *last = p + obj->exoclass->slots->size;

    if (KlExoHasHook(obj, KlSelEHFree)) {
	KlO result;
	int normal;
	KlGCMark();
	KlCatch(KlA_ALL,
		KlApplyMethod1(KlExoMethodNamed(obj, KlA_free), KlA_free, obj),
		result, normal);
	KlGC();
	if (obj->reference_count) {	/* the destructor pointed to it */
	    obj->reference_count |= 1;	/* escape its fate, clean up */
	    return (KlO) obj;
	}
    }
    /* actual imediate obliteration */
    while (p < last) {
	KlDecRef(*p);
	p++;
    }
    KlDecRef(obj->exoclass);
    Free(obj);
    return (KlO) obj;
}

/*****************************************************************************\
* 			    access to class fields                            *
\*****************************************************************************/

#define KlAccessClassField(name, field) \
    KlO name(exoclass) KlExoClass exoclass; \
    {  KlMustBeExoClass(exoclass, 0); \
       return (exoclass->field ? (KlO) exoclass->field : NIL);}

#ifdef OBSOLETE
KlAccessClassField(KlAccessClassSon, son);
KlAccessClassField(KlAccessClassBrother, brother);
#endif /* OBSOLETE */

/*
 * KlAccessClassName
 *
 * Get and Set the class name.
 * Klone function: (class-name class [new-name])
*/

KlO
KlAccessClassName(argc, argv)
    int argc;
    KlO *argv;
{
    KlExoClass exoclass;

    KlNumberOfArgumentsCheck(argc < 1, argc)
    KlMustBeExoClass(argv[0], 0);
    exoclass = (KlExoClass) argv[0];

    switch (argc) {
    case 1:
	/* get the class name */
	return (KlO) exoclass->name;
	break;
    case 2:
	/* set the class name */
	KlMustBeAtom(argv[1], 1);
	{
	    KlAtom new_name = (KlAtom) argv[1];

	    KlDecRef(exoclass->name);
	    KlAtomUnbind(exoclass->name);
	    KlIncRef(exoclass->name = new_name);
	    KlSend_setq(new_name, exoclass);
	    return (KlO) exoclass;
	}
	break;
    default:
	return KlBadNumberOfArguments((char *) argc);
    }
}
     
/* slots is a hash table of C integers, which are not klone objects,
 * so we cannot return it just as is...
 */

KlO
KlAccessClassSlots(exoclass)
    KlExoClass exoclass;
{
    KlList list;

    KlMustBeExoClass(exoclass, 0);
    if (KlIsAMetaclass(exoclass)) {	/* metaclass */
	UInt offset;
	list = KlListNMake((exoclass->slots->size -
			    KlExoClassBuiltInSlots) * 2);
	KlHashFORBEGIN(exoclass->slots, cell) {
	    offset = (UInt)KlExoActualOffset(cell->object) - 
		KlExoClassBuiltInSlots;
	    if (offset >= 0) { /* skip built-ins */
		KlIncRef(list->list[offset * 2] = cell->key);
		KlIncRef(list->list[offset * 2 + 1] =
			 exoclass->init[((int)(cell->object))]);
	    }
	} KlHashFOREND;
	
    } else {				/* normal class */
	UInt offset;
	list = KlListNMake(exoclass->slots->size * 2);
	KlHashFORBEGIN(exoclass->slots, cell) {
	    offset = KlExoActualOffset(cell->object);
	    KlIncRef(list->list[offset * 2] = cell->key);
	    KlIncRef(list->list[offset * 2 + 1] = exoclass->init[offset]);
	} KlHashFOREND;
    }
    return (KlO) list;
}

/* 
 * KlAccessClassMethods
 * works also on base types
 */

KlO
KlAccessClassMethods(exoclass)
    KlExoClass exoclass;
{
    if (KlIsAnExoClass(exoclass)) {
	return exoclass->methods ? (KlO) exoclass->methods : NIL;
    } else {
	KlType type = KlFindType(exoclass);

	return KlTypeMethodsGet(type) ? (KlO) KlTypeMethodsGet(type) : NIL;
    }
}

/* KlAccessClassFather
 * works also on base types
 */

KlO
KlAccessClassFather(exoclass)
    KlExoClass exoclass;
{
    if (KlIsAnExoClass(exoclass)) {
	return exoclass->father ? (KlO) exoclass->father : NIL;
    } else {
	KlType type = KlFindType(exoclass);

	return KlTypeFatherGet(type) ? (KlO) KlTypeFatherGet(type) : NIL;
    }
}

/* KlAccessClassChildren
 * works also on base types
 */

KlO
KlAccessClassChildren(exoclass)
    KlExoClass exoclass;
{
    if (KlIsAnExoClass(exoclass)) {
	if (exoclass->son) {
	    KlExoClass son = exoclass->son;
	    KlList children_list = KlListNMake(0);
	    while (son) {
		KlListAppend(children_list, son);
		son = son->brother;
	    }
	    return (KlO) children_list;
	}
    } else {
	KlType type = (exoclass == (KlExoClass) NIL ?
		       0 : KlFindType(exoclass));
	KlList list = KlListNMake(0);
	KlType t;

	for (t = KlTypes; t; t = KlTypeNext(t)) {
	    if (KlTypeFatherGet(t) == type) {
		KlListAppend(list, t);
	    }
	}
	return (KlO) list;
    }
    return NIL;
}

/* KlExoClassSlotNumber
 * from an offset, gives the slot name
 */

KlO
KlExoClassSlotNumber(exoclass, offset)
    KlExoClass exoclass;
    UInt offset;
{
    KlHashFORBEGIN(exoclass->slots, cell) {
	if (offset == KlExoActualOffset(cell->object)) {
	    return cell->key;
	}
    } KlHashFOREND;
    return NIL;
}


/*****************************************************************************\
* 		  access to slots: put, get, (but no delete)                  *
\*****************************************************************************/

KlO
KlExoGet(obj, slot, value)
    KlExo obj;
    KlAtom slot;
    KlO value;
{
    int offset;
    KlExpr method;

    if (KlIsAKeyword(slot))
	slot = ((KlKeyword)(slot))->atom;

    if (KlExoHasHook(obj, KlSelEHGet)) {
	return (KlO)
	    KlApplyMethod3(KlExoMethodNamed(obj, KlA_get),
			   KlA_get, obj, slot, value);
    }

    offset = (UInt) KlHashGet(obj->exoclass->slots, slot, KlPI_INVALID_VALUE);
    if (KlPI_INT(offset)) {		/* normal slot */
	return obj->slots[offset];
    } else if (KlPI_INVALID(offset)) {	/* not found */
	return KlExecuteGetDefault(obj, slot, value);
    } else {				/* hooked */
	if (((KlSlotHooks) offset)->get == NIL) { /* no hook */
	    return obj->slots[((KlSlotHooks) offset)->offset];
	} else if (KlIsASubr(((KlSlotHooks) offset)->get)) {
	    /* optimisation */
	    return (*(((KlSubr) ((KlSlotHooks) offset)->get)->body))
		(obj, slot, obj->slots[((KlSlotHooks) offset)->offset]);
	} else {
	    return KlApplyV(((KlSlotHooks) offset)->get,
			    3, obj, slot, 
			    obj->slots[((KlSlotHooks) offset)->offset]);
	}
    }
}

KlO
KlExoPut(obj, slot, value)
    KlExo obj;
    KlAtom slot;
    KlO value;
{
    int offset;
    KlExpr method;

    if (KlIsAKeyword(slot))
	slot = ((KlKeyword)(slot))->atom;

    if (KlExoHasHook(obj, KlSelEHPut)) {
	return (KlO)
	    KlApplyMethod3(KlExoMethodNamed(obj, KlA_put),
			   KlA_put, obj, slot, value);
    }

    offset = (UInt) KlHashGet(obj->exoclass->slots, slot, KlPI_INVALID_VALUE);
    if (KlPI_INT(offset)) {		/* normal slot */
	KlDecRefNonNull(obj->slots[offset]);
	KlIncRef(obj->slots[offset] = value);
	return (KlO) obj;
    } else if (KlPI_INVALID(offset)) {	/* not found */
	return KlError2(KlE_NO_ELEMENT, slot, obj->exoclass->name);
    } else {				/* hooked */
	KlO *actual_slot = &(obj->slots[((KlSlotHooks) offset)->offset]);
	KlDecRef(*actual_slot);
	if (((KlSlotHooks) offset)->put == NIL) { /* no hook */
	    KlIncRef(*actual_slot = value);
	} else if (KlIsASubr(((KlSlotHooks) offset)->put)) {
					/* optimisation */
	    KlIncRef(*actual_slot =
		     (*(((KlSubr) ((KlSlotHooks) offset)->put)->body))
		     (obj, slot, value));
	} else {
	    KlIncRef(*actual_slot = 
		     KlApplyV(((KlSlotHooks) offset)->put, 3, obj, slot, value));
	}
	return value;
    }
}

KlO
KlExoRawCopy(obj)
    KlExo obj;
{ 
    KlO *p = &(obj->slots[0]);
    KlO *last = p + obj->exoclass->slots->size;
    KlExo clone = KlExoMake(obj->exoclass);
    KlO *q = &(clone->slots[0]);

    while (p < last) {
	KlIncRef(*q = *p);
	p++, q++;
    }
    /* if present, execute the "init" exoHook */
    if (obj->exoclass->exoHooks & KlSelEHInit) {
	KlApplyMethod1(KlExoMethodNamed(obj, KlA_init), KlA_init, obj);
    }
    return (KlO) clone;
}

KlO
KlExoCopy(obj)
    KlExo obj;
{ 
    if (KlExoHasHook(obj, KlSelEHCopy)) {
	return KlApplyMethod1(KlExoMethodNamed(obj, KlA_copy), KlA_copy, obj);
    } else {
	return KlExoRawCopy(obj);
    }
}

/*****************************************************************************\
* 			     raw access to slots                              *
\*****************************************************************************/

/* (*:slot obj offset [val]) get/put slot
 * get/sets raw slots
 */

KlO
KlExoRawSlot(argc, argv)
    int argc;
    KlO *argv;
{
    UInt offset;
    KlExo obj;
    KlNumber slot;

    if (argc < 1 || argc > 3)
	return KlBadNumberOfArguments(argc);
    obj = (KlExo) argv[0];
    slot = (KlNumber) argv[1];
     
    KlMustBeExo(obj, 0);
    if (KlIsANumber(slot)) {
	offset = (UInt) slot->number;
    } else if (KlIsASymbol(slot)) {	/* computes offset */
	offset = (UInt) KlHashGet(obj->exoclass->slots, slot, 
				  KlPI_INVALID_VALUE);
	if (offset != KlPI_INVALID_VALUE && !KlPI_INT(offset))
	    offset = ((KlSlotHooks) offset)->offset;
    } else {
	KlMustBeNumber(slot, 1);
    }
    
    if (offset < obj->exoclass->slots->size) {
	if (argc == 3) {		/* put */
	    KlDecRef(obj->slots[offset]);
	    KlIncRef(obj->slots[offset] = argv[2]);
	    return argv[2];
	} else {			/* get */
	    return obj->slots[offset];
	}
    } else {
	return KlError2(KlE_NO_ELEMENT, slot, obj->exoclass->name);
    }
}

/* (*:slot-offset class name) returns offset
 */

KlO
KlExoRawSlotOffset(exoclassname, slot)
    KlExoClass exoclassname;
    KlAtom slot;
{
    UInt offset;
    KlExoClass exoclass;

    if (KlIsAnExo(exoclassname) && !KlIsAnExoClass(exoclassname))
	exoclassname = ((KlExo)exoclassname)->exoclass;
    KlMustBeOrEval(exoclass, exoclassname, KlExoClass,
		       KlIsAnExoClass(exoclass),
		       KlError(KlE_NO_CLASS, exoclassname));
    KlMustBeSymbol(slot, 1);
    if ((offset = (UInt) KlHashGet(exoclass->slots, slot, 
				   KlPI_INVALID_VALUE)) != KlPI_INVALID_VALUE) {
	return (KlO) KlNumberMake(KlExoActualOffset(offset));
    } else {
	return KlError2(KlE_NO_ELEMENT, slot, exoclass->name);
    }
}

/* (*:system-method class name) returns the ExoHook from Klone
 */

KlO
KlSystemMethodKl(exoclassname, slot)
    KlExoClass exoclassname;
    KlAtom slot;
{
    int offset;
    KlExoClass exoclass;

    if (KlIsAnExo(exoclassname) && !KlIsAnExoClass(exoclassname))
	exoclassname = ((KlExo)exoclassname)->exoclass;
    KlMustBeOrEval(exoclass, exoclassname, KlExoClass,
		       KlIsAnExoClass(exoclass),
		       KlError(KlE_NO_CLASS, exoclassname));
    KlMustBeSymbol(slot, 1);
    return KlHashGet(exoclass->methods, slot, NIL);
}
   

/*****************************************************************************\
* 				   methods                                    *
\*****************************************************************************/

/* KlExoMethodAdd
 * raw routine to add a method to a class
 */

KlO
KlExoMethodAdd(exoclass, selector, method, exoHooksFlags)
    KlExoClass exoclass;
    KlAtom selector;
    KlO method;
    unsigned int exoHooksFlags;
{
    KlO oldmethod = KlHashGet(exoclass->methods, selector, NIL);

    KlHashPut(exoclass->methods, selector, method);
    if (exoHooksFlags)
	if (method == NIL) {		/* remove */
	    exoclass->exoHooks &= ~exoHooksFlags;
	} else {			/* add */
	    exoclass->exoHooks |= exoHooksFlags;
	}

    if (exoclass->son) {
	KlExoMethodPropagate(exoclass->son, selector, oldmethod, method,
			     exoHooksFlags);
    }
#ifdef KLPROFILER
    if (KlIsAnExpr(oldmethod)) {
	KlDecRef(((KlExpr) oldmethod)->profiling_name);
	((KlExpr) oldmethod)->profiling_name = 0;
    }
    if (KlIsAnExpr(method)) {
	KlDecRef(((KlExpr) method)->profiling_name);
	KlIncRef(((KlExpr) method)->profiling_name =
		 (KlO) KlListPairMake(selector, exoclass));
    }
    
#endif

    return (KlO) method;
}

/* recurse on children to inherit method
 * stops recursing as soon as the method changes
 */

KlExoMethodPropagate(exoclass, selector, old, newer, exoHooksFlags)
    KlExoClass exoclass;
    KlAtom selector;
    KlO old;
    KlO newer;
    unsigned int exoHooksFlags;
{
    
    if (exoclass->brother) {
	KlExoMethodPropagate(exoclass->brother, selector, old, newer
			     , exoHooksFlags);
    }
    if (KlHashGet(exoclass->methods, selector, NIL) == old) {
	KlHashPut(exoclass->methods, selector, newer);
	if (exoHooksFlags)
	    if (newer == NIL) {		/* remove */
		exoclass->exoHooks &= ~exoHooksFlags;
	    } else {			/* add */
		exoclass->exoHooks |= exoHooksFlags;
	    }
	if (exoclass->son) {
	    KlExoMethodPropagate(exoclass->son, selector, old, newer
				 , exoHooksFlags);
	}
    }
}

/* KlTypeMethodAdd
 * raw routine to add a method to a primitive klone type
 */

KlO
KlTypeMethodAdd(exoclass, selector, method)
    KlType exoclass;			/* primitive klone type */
    KlAtom selector;
    KlO method;
{
    KlO old_method = NIL;
    KlType t;

    method = KlCheckMethodValidity(method, selector, 0);
    if (KlTypeMethodsGet(exoclass)) {
	old_method = KlHashGet(KlTypeMethodsGet(exoclass), selector, NIL);
	KlHashPut(KlTypeMethodsGet(exoclass), selector, method);
    } else {
	KlHash ht = KlHashAlloc(0);	/* we must create a table */

	KlTypeMethodsSet(exoclass, ht);
	KlIncRef(ht);
	ht->ref_counted = KlHashRefV;
	KlHashPut(ht, selector, method);
    }
    
    /* inheritance: for all derived types, */
    if (old_method != NIL) {			/* look for same method */
	for (t = KlTypes; t; t = KlTypeNext(t)) {
	    if (KlTypeFatherGet(t) == exoclass) {
		if (KlTypeMethodsGet(t)
		    && KlHashGet(KlTypeMethodsGet(t), selector, NIL)
		    == old_method) {
		    KlTypeMethodAdd(t, selector, method); /* recurse */
		}
	    }
	}
    } else {				/* look for no method */
	for (t = KlTypes; t; t = KlTypeNext(t)) {
	    if (KlTypeFatherGet(t) == exoclass) {
		if (!(KlTypeMethodsGet(t)
		      && (KlHashGet(KlTypeMethodsGet(t), selector, NIL)
			  != old_method)
		      )) {
		    KlTypeMethodAdd(t, selector, method); /* recurse */
		}
	    }
	}
    }
    return (KlO) method;
}

/* add-method:
 * external interface to KlTypeMethodAdd or KlExoMethodAdd
 */

KlO
KlAddMethodKl(exoclass, selector, method)
    KlExoClass exoclass;
    KlAtom selector;
    KlO method;
{
    unsigned int flags;

    KlMustBeAtom(selector, 1);
    KlCheckMethodValidity(method, selector, 0);
    flags = KlGenericFunctionDeclare(selector);
    if (KlIsAnExoClass(exoclass)) {
	return KlExoMethodAdd(exoclass, selector, method, flags);
    } else {
	KlType type = KlFindType(exoclass);
	return KlTypeMethodAdd(type, selector, method);
    }
}

/* defmethod klone interface
 * (defmethod selector-name ((variable class) args...) [documentation] body)
 */

KlO
KlDefMethod(argc, argv)
    int argc;
    KlO *argv;
{
    KlList list, lambdalist;
    int i;
    KlExoClass exoclass;
    KlType kltype = 0;
    KlO lambda;
    KlAtom selector;
    KlO *call_list;
    unsigned int flags;

    if (argc < 2)
	return KlBadNumberOfArguments(argc);
    selector = (KlAtom) argv[0];
    KlMustBeAtom(selector, 0);
    
    flags = KlGenericFunctionDeclare(selector);
    KlMustBeList(argv[1], 1);
    list = (KlList) argv[1];
    if (!list->size
	|| !KlIsAList(list->list[0])
	|| ((KlList) (list->list[0]))->size != 2
	|| !KlIsAnAtom(((KlList) (list->list[0]))->list[0])) {
	return KlError(KlE_BAD_DEFUN, selector);
    }
    exoclass = (KlExoClass) ((KlList) (list->list[0]))->list[1];
    if (!KlIsAnExoClass(exoclass)) {
	exoclass = (KlExoClass) KlSend_eval(exoclass);
	if (!KlIsAnExoClass(exoclass)) {
	    if (KlIsAType(exoclass)) {
		kltype = (KlType) exoclass;
	    } else {
		KlError(KlE_NO_CLASS, ((KlList) (list->list[0]))->list[1]);
	    }
	}
    }
    /* look for documentation string, and set it to class.selector */
    if (argc > 3 && argv[2]->type == KlStringType) {
	char *name;

	if (KlIsAnExoClass(exoclass)) {
	    name = exoclass->name->p_name;
	} else {
	    name = KlTypeCName(kltype);
	}
	KlDocumentationSet(KlInternDotPrefixed(name, selector->p_name),
			   argv[2]);
	argv++;
	argc--;
    }
    /* builds lambda list call */
    lambdalist = KlListNMake(list->size);
    KlListStore(lambdalist, 0, ((KlList) (list->list[0]))->list[0]);
    for (i = 1; i < list->size; i++) {
	KlListStore(lambdalist, i, list->list[i]);
    }
    call_list = KlAlloca(argc - 1);
    call_list[0] = (KlO) lambdalist;
    for (i = 1; i < argc - 1; i++) {
	call_list[i] = argv[i + 1];
    }

    lambda = KlLambda(argc - 1, call_list);
    if (kltype) {
	KlTypeMethodAdd(kltype, selector, lambda);
    } else {
	KlExoMethodAdd(exoclass, selector, lambda, flags);
    }
    return lambda;
}

/* KlCheckMethodValidity
 * checks that a method is either an Expr or a Subr
 * returns the method if Ok, and else:
 * if silent = 0, error
 * if silent = 1, returns 0
 */

KlO
KlCheckMethodValidity(method, selector, silent)
    KlExpr method;
    KlAtom selector;
    int silent;				/* shall we just return result */
{
#ifndef CHECK_METHOD_TYPES
    if (KlIsUndefinedMethod(method,KlSelExecute )) {
#else /* !CHECK_METHOD_TYPES */
    if (!KlIsAnExpr(method) && !KlIsASubr(method) && KlTrueP(method)) {
#endif /* !CHECK_METHOD_TYPES */
	if (silent)
	    return 0;
	else
	    return KlError2(KlE_NOT_A_METHOD, method, selector);
    }
    return (KlO) method;
}

/*****************************************************************************\
* 			      KlExoMethodRemove                               *
\*****************************************************************************/

/* removes the definition of a method
 * complex for handling correctly a dyunamic situation
 */

/* KlExoClassMethodRemove 
 * Removes the method of a class.
 */
KlO
KlExoClassMethodRemove(selector, exoclass)
     KlAtom selector;
     KlExoClass exoclass;
{
    KlO method;
    KlO inherited;
    KlAtom desc;
    int flags;

    if ((method = KlHashGet (exoclass->methods, selector, NIL))
	== NIL)
 	/* ERROR: There is no method for the specicified selector */
	return KlError2(KlE_NO_METHOD, selector, exoclass->name);	 

    if (!KlExoMethodIsLocal (selector, exoclass))
	/* The method is NOT local, to remove it is useless, in any case
	 *  the class still inherites from the method of its superclass.
	 *  So return NIL. */
	return NIL;

    flags = KlAtomIsAnExoHookableSelector(selector);

    if (KlExoMethodIsSpecialized(selector, exoclass, method, &inherited)) {
	/* the inherited form replaces the form to be removed */
	KlHashPut(exoclass->methods, selector, inherited);
	/* Propagate the inherited method to the sublasses, if any */
	if (exoclass->son) {
	    KlExoMethodPropagate(exoclass->son, selector,
				 method, inherited, flags);
	}
    } else {
	/* Remove the entry form the hastable of class methods */
	KlHashDeleteEqual(exoclass->methods,selector);
	exoclass->exoHooks &= ~flags;
	/* Propagate the deletion to the subclasses (if they exixt) */
	if (exoclass->son) {
	    KlExoRemoveMethodPropagate(exoclass->son, selector, method, flags);
	}
    }

    /* If documentation string for the method exists, remove it too. */
    KlDocumentationClear(KlInternDotPrefixed(exoclass->name->p_name,
					      selector->p_name));

    return TRU;  
}

/* KlTypeMethodRemove
 * Removes the method of a type.
 */
KlO
KlTypeMethodRemove(selector, type)
    KlAtom selector;
    KlType type;
{
    KlO method = KlHashGet (KlTypeMethodsGet(type),selector, NIL);
    KlAtom desc;
    KlO inherited;
  
    if (method == NIL)
	return KlError2(KlE_NO_METHOD, selector, KlTypeName(type));

    if (!KlExoTypeMethodIsLocal (selector, type)) {
	/* The method is NOT local, nothing to do, return Nil */
	return NIL;
    }
    /* The method is Local */
    if (KlExoTypeMethodIsSpecialized(selector, type, method, &inherited)){
	/* the inherited form may replace the form to remove */
	KlHashPut(KlTypeMethodsGet(type), selector, inherited);
	/* Propagate to subtypes */
	KlExoReplaceTypeMethodPropagate(type,selector,method,inherited);
    } else {
	/* remove it */
	KlHashDelete(KlTypeMethodsGet(type), selector);
	/* Propagate to subtypes */
	KlExoRemoveTypeMethodPropagate (type, selector, method);
    }
  
    /* If documentation string exists, remove it too. */
    KlDocumentationClear(KlInternDotPrefixed((KlTypeName(type))-> p_name,
					      selector->p_name)); 
    return TRU;  
}

/* KlExoMethodRemove
 *
 * Klone interface: (remove-method [class | type] selector)
 *                 (remove-method generic [class | type])
 *
 * calls either KlTypeMethodRemove or KlExoClassMethodRemove
 * TODO: 
 *     - si la fonction generique est vide (pas de methodes), la supprimer.
 */

KlO 
KlExoMethodRemove(arg1, arg2)
    KlO arg1, arg2;
{
    if (KlIsAnAtom(arg2)) {
	if (KlIsAnExoClass(arg1)) {
	    return KlExoClassMethodRemove((KlAtom) arg2, (KlExoClass) arg1);
	} else {
	    KlType type = KlFindType(arg1);
	    return KlTypeMethodRemove((KlAtom) arg2, type);
	}
    } else {
	KlMustBeGenericFunction(arg1, 0);
	if (KlIsAnExoClass(arg2)) {
	    return KlExoClassMethodRemove(((KlGenericFunction) arg1)->name,
					  (KlExoClass) arg2);
	} else {
	    KlType type = KlFindType(arg2);
	    return KlTypeMethodRemove(((KlGenericFunction) arg1)->name, type);
	}
    }
}
/******************************************** propagation of method deletion */
/* KlExoRemoveMethodPropagate
 *
 * Remove the method named 'selector' whose lambda is 'method' in the
 * class 'exoclass'.
 * Propagate to sublcasses, and to brother.
 */   
KlExoRemoveMethodPropagate(exoclass, selector, method, flags) 
    KlExoClass	exoclass;
    KlAtom 	selector;
    KlO		method;
    unsigned int flags;
{
    /* Propagate to next brother */
    if (exoclass->brother) {
	KlExoRemoveMethodPropagate(exoclass->brother, selector, method, flags);
    }
    if (KlHashGet(exoclass->methods, selector, NIL) == method) {
	/* the same method exists, then remove it */
	KlHashDelete(exoclass->methods, selector);
	exoclass->exoHooks &= ~flags;
	/* Propagate to subclasses */
	if (exoclass->son) {
	    KlExoRemoveMethodPropagate(exoclass->son, selector, method, flags);
	}
    }
}
    
/* KlExoRemoveTypeMethodPropagate
 *
 * Remove the method named 'selector' whose lambda is 'method' for the
 * subtypes of the type 'type'.
 * Propagate to subtypes.
*/
KlExoRemoveTypeMethodPropagate(type, selector, method)
    KlType	type;
    KlAtom  selector;
    KlO     method;
{
    KlType t;
    for (t = KlTypes; t ; t = KlTypeNext (t)) {
	if (KlTypeFatherGet (t) == type ) {
	    if (KlHashGet(KlTypeMethodsGet (t), selector, NIL) == method) {
		KlHashDelete(KlTypeMethodsGet(type), selector);
		KlExoRemoveTypeMethodPropagate (t, selector, method);
	    }
	}
    }
}
   
/* KlExoReplaceTypeMethodPropagate
 *
 * Replace the method named 'selector' whose lambda is 'method' by 
 * the new lambda 'inherited'
 * for the subtypes of the type 'type'.
 * Propagate to the subtypes.
*/ 
KlExoReplaceTypeMethodPropagate(type, selector, method, inherited)
    KlType	type;
    KlAtom  selector;
    KlO     method;
    KlO     inherited;
{
    KlType t;
    for (t = KlTypes; t ; t = KlTypeNext (t)) {
	if (KlTypeFatherGet (t) == type ) {
	    if (KlHashGet(KlTypeMethodsGet (t), selector, NIL) == method) {
		KlHashPut(KlTypeMethodsGet(type), selector, inherited);
		KlExoReplaceTypeMethodPropagate (t, selector, method,
						 inherited);
	    }
	}
    }
}		
		

/* KlExoMethodIsLocal
 *
 * Test wether a method is local (return 1)
 *      or is inherited (return 0)
 *
*/
int
KlExoMethodIsLocal(selector,exoclass)
     KlAtom selector;
     KlExoClass exoclass;
{
    if (exoclass->father) {
	return (KlHashGet(exoclass->methods,selector,NIL) 
		!=
		KlHashGet(exoclass->father->methods,selector,NIL)
		);
    } else {
	return 1; 
    }
}

/* KlExoTypeMethodIsLocal
 *
 * Test wether a method is local (return 1)
 *      or is inherited (return 0)
 */
int
KlExoTypeMethodIsLocal(selector,type)
        KlAtom selector;
        KlType type;
{
    KlType father=KlTypeFatherGet(type);
  
    if (father) {
	return (KlHashGet(KlTypeMethodsGet(type),selector,NIL)
		!=
		KlHashGet(KlTypeMethodsGet(father),selector,NIL)
		);
    } else {
	return 1;
    }
}


/* KlExoMethodIsSpecialized
 * 
 * Test wether a method is a specilized form of an inherited method, 
 * then returns the inherited form of the method, 
 * or 0 otherwise.
 */


KlExoMethodIsSpecialized(selector, exoclass, method, inherited)
        KlAtom selector;
        KlExoClass exoclass;
	KlO  method;
        KlO  *inherited;
{
    return (exoclass->father 
	    && ((*inherited= KlHashGet(exoclass->father->methods,selector,NIL))
		!=NIL)
	    && (method != *inherited));
}


/* KlExoTypeMethodIsSpecialized
 *
 * Test wether a method is a specilized form of an inherited method, 
 * then returns the inherited form of the method, 
 * or 0 otherwise.
 */


KlExoTypeMethodIsSpecialized(selector, type, method, inherited)
        KlAtom selector;
        KlType type;
	KlO  method;
        KlO  *inherited;
{
    KlType father=KlTypeFatherGet(type);
  
    return (father
	    && ((*inherited= KlHashGet(KlTypeMethodsGet(father),selector,NIL))
		!=NIL)
	    && (method != *inherited));
}

#ifdef OBSOLETE
/* The following procedure has several disadvantages:
   + No test about the method existance
   + Do not remove the entry from methods tables
   + execute useless code when the method to remove is not local (the same
   method exists in the superclass)
The file "smool.c" contains a more complete implementation of this procedure.
Laurent.
*/
KlO
KlExoMethodRemove(exoclass, selector)
    KlExoClass exoclass;
    KlAtom selector;
{
    KlExpr method;

    KlMustBeAtom(selector, 1);

    if (KlIsAnExoClass(exoclass)) {
	unsigned int flags = KlAtomIsAnExoHookableSelector(selector);

	KlExoMethodAdd(exoclass, selector, NIL, flags);	/* remove */
	if (exoclass->father		/* re-propagate inherited */
	    && (method = (KlExpr)
		KlHashGet((exoclass->father)->methods, selector, NIL))
	    != (KlExpr) NIL) {
	    KlExoMethodAdd(exoclass, selector, method, flags);
	}
	    
    } else {
	KlType type = KlFindType(exoclass); /* remove */
	KlTypeMethodAdd(type, selector, NIL); /* propagate */
	if (KlTypeFatherGet(type)
	    && KlTypeMethodsGet(KlTypeFatherGet(type))
	    && (method = (KlExpr)
		KlHashGet(KlTypeMethodsGet(KlTypeFatherGet(type)),
			  selector, NIL))
	    != (KlExpr) NIL) {
	    KlTypeMethodAdd(KlTypeFatherGet(type), selector, method);
	}
    }
    /* TODO: remove generic if was the only method left */
    return NIL;    
}
#endif /* OBSOLETE */

/*****************************************************************************\
* 			    method for base types                             *
\*****************************************************************************/

KlExpr
KlMethodOfType(type, selector)
    KlType type;
    KlAtom selector;
{
    KlExpr method;

    if (KlTypeMethodsGet(type) &&
	((method = (KlExpr)
	  KlHashGet(KlTypeMethodsGet(type), selector, NIL))
	 != (KlExpr) NIL)
	) {
	return method;
    } else {
	return 0;
    }
}
	

/*****************************************************************************\
* 			      generic functions                               *
\*****************************************************************************/

KlGenericFunction
KlGenericFunctionMake(name)
    KlAtom name;			/* name of the selector */
{
    KlGenericFunction object = (KlGenericFunction)
	KlOMake(KlGenericFunctionType);

    object->name = name;		/* no incref needed */
    return object;
}

KlO
KlGenericFunctionPrint(object, stream)
    KlGenericFunction object;			/* name of the selector */
    KlStream stream;
{
    KlSPrintf(stream, "{^ GenericFunction %s}", object->name->p_name);
    return (KlO) object;
}

/* KlGenericFunctionDeclare
 * for a name (selector), declare a generic function of this name.
 * If a function already existed of this name, do some footwork to make this
 * function called by the generic function as the default case
 * This induces runtime overhead, but is necessary.
 */

unsigned int
KlGenericFunctionDeclare(selector)
    KlAtom selector;
{
    unsigned int flags;
    
    if (!(flags = KlAtomIsAnExoHookableSelector(selector))) {
	if (selector->c_val == KlUndef
	    || !KlIsAGenericFunction(selector->c_val)) {
	    /* save previous function */
	    if (selector->c_val != KlUndef && KlIsAFunction(selector->c_val)) {
		KlGenericFunctionCatchallSet(selector, selector->c_val);
	    }
	    /* create generic function if the first element wasn't one */
	    KlSend_setq(selector, KlGenericFunctionMake(selector));
	}
    }
    return flags;
}

/* KlExoApplyMethodReturn
 * The low-level routine to actually executing a method once found
 */

#define KlExoApplyMethodReturn(WEAME, WEAMe, WEAMsel, WEAMexo, WEAMclass, WEAMlist) \
{ \
    struct _KlGenFuncContext WEAMgfc; \
    KlO WEAMresult; \
 \
    WEAMgfc.selector = WEAMsel; \
    WEAMgfc.exoclass = WEAMclass; \
    WEAMgfc.list = WEAMlist; \
    WEAMgfc.obj = WEAMexo; \
    WEAMgfc.method = (KlO) WEAMe; \
    KlStackFramePushSpecialHold(KlSFID_genfunc, (KlO) &WEAMgfc, WEAMexo); \
 \
    WEAMresult = WEAME; \
 \
    KlStackFramePopSpecial(); \
    return WEAMresult; \
}

/* KlGenericFunctionExecute
 * must look to find which method (function) to apply instead
 * Exo : look in class for method
 * others : look into a hashtable found at
 *          entry KlSelMethods into the type, which maintains methods for 
 *          all primitive types
 */

/*ARGSUSED*/
KlO
KlGenericFunctionExecute(func, list)
    KlGenericFunction func;		/* the function */
    KlList list;			/* call list */
{
    KlAtom selector = func->name;

    if (list->size > 1) {
	KlExpr method;
	KlExo obj = (KlExo) KlSend_eval(list->list[1]);
	KlExoClass exoclass;
	KlList call;

	/* find-method: get method for object */
	if (KlIsAnExo(obj)) {		/* for regular ExOs */
	    exoclass = obj->exoclass;
		
	    if ((method = (KlExpr)
		 KlHashGet(exoclass->methods, selector, NIL))
		== (KlExpr) NIL) {
		return KlError2(KlE_NO_METHOD, selector, exoclass->name);
	    }
	} else if (method = KlMethodOfType(obj->type, selector)) {
	    exoclass = (KlExoClass) obj->type;
	} else {
	    return KlError2(KlE_NO_METHOD, selector, obj->type);
	}

	/* Execute: evaluates args */
	KlListNMakeZ(call, list->size);
	{
	    KlO *source = list->list + 2;
	    KlO *last = list->list + list->size;
	    KlO *dest = call->list;

	    KlIncRef(*dest++ = (KlO) func);
	    KlIncRef(*dest++ = (KlO) obj);
	    while (source < last) {
		KlIncRef(*dest++ = KlSend_eval(*source));
		source++;
	    }
	}

	KlExoApplyMethodReturn(KlSend_apply(method, call),
			       method, selector, obj, exoclass, call);
    } else {
	/* we come here when there is a previously existing function */
	/* hack: we get the list type method since () is a list... */
	KlO func = KlHashGet(KlTypeMethodsGet(KlListType),
			     selector, NIL);

	return KlApply(KlListKl(1, &func));
    }
}

/* apply is like execute but do not evaluate arguments
 */

KlO
KlGenericFunctionApply(func, list)
    KlGenericFunction func;		/* the function */
    KlList list;			/* call list */
{
    KlAtom selector = func->name;

    if (list->size > 1) {
	KlExpr method;
	KlExo obj = (KlExo) list->list[1];
	KlO result;
	KlExoClass exoclass;

	/* find-method: get method for object */
	if (KlIsAnExo(obj)) {		/* for regular ExOs */
	    exoclass = obj->exoclass;
		
	    if ((method = (KlExpr)
		 KlHashGet(exoclass->methods, selector, NIL))
		== (KlExpr) NIL) {
		return KlError2(KlE_NO_METHOD, selector, exoclass->name);
	    }
	} else if (method = KlMethodOfType(obj->type, selector)) {
	    exoclass = (KlExoClass) obj->type;
	} else {
	    return KlError2(KlE_NO_METHOD, selector, obj->type);
	}

	KlExoApplyMethodReturn(KlSend_apply(method, list),
			       method, selector, obj, exoclass, list);
    } else {
	return NIL;			/* never reached (normally) */
    }
}

/* the KlApplyMethodN functions:
 * used to "apply" a method but setting up the stack to be able to call 
 * call-next-method in them
 */

KlO
KlApplyMethodN(func, selector, argc, argv)
    KlO func;
    KlAtom selector;
    int argc;
    KlO *argv;
{
    KlList call = KlListNMake(argc + 1);

    KlListStore(call, 0, func);
    KlCopyNObjects(argv, call->list + 1, argc);

    KlExoApplyMethodReturn(KlApply(call),
			   func, selector, ((KlExo) argv[0]),
			   ((KlExo) argv[0])->exoclass, call);
}

/****************************************** VARARGS version, very convenient */
/* KlApplyMethodV(func, selector, number_of_args, args...) */

KlO
#if __STDC__
KlApplyMethodV(KlO func, KlAtom selector, int size, ...)
#else
KlApplyMethodV(va_alist)
    va_dcl
#endif
{
    va_list argList;
    KlList list;
    KlO *p;

#if __STDC__
    va_start(argList, size);
#else
    int size;
    KlO func;
    KlAtom selector;

    va_start(argList);
    func = va_arg(argList, KlO);
    selector = va_arg(argList, KlAtom);
    size = va_arg(argList, int);
#endif
    list = KlListNMake(size+1);
    p = list->list;
    KlIncRef(*p++ = func);
    for (;size > 0; size--) {
	*p = va_arg(argList, KlO);
	KlIncRef(*p++);
    }
    va_end(argList);
    KlExoApplyMethodReturn(KlApply(list),
			   func, selector, ((KlExo) (list->list[0])),
			   ((KlExo) (list->list[0]))->exoclass, list);
}

/* KlCallAnotherMethod
 * The call-next-method and call-super-method underlying function
 */

/*ARGSUSED*/
KlO
KlCallAnotherMethod(next, argc, argv)
    int next;				/* 1 = call-next-method */
    int argc;
    KlO *argv;
{
    KlAtom selector;
    KlExoClass class_orig;
    KlList call;
    KlExo obj;
    KlO method;
    KlO curmethod;
    KlExoClass exoclass;
    KlType type;
    int stackptr = KlStackPtr;
    KlGenFuncContext cur_gfc = (KlGenFuncContext) /* find previous method */
	KlStackFrameLookForTag(KlSFID_genfunc);

    if (!cur_gfc) return KlError0(KlE_NO_PREVIOUS_METHOD);

    obj = cur_gfc->obj;
    if (next) {
	selector = cur_gfc->selector;
	call = cur_gfc->list;
	curmethod = cur_gfc->method;
    } else {
	KlNumberOfArgumentsCheck((argc < 1) , argc);
	KlMustBeAtom (argv[0], 0);
	selector = (KlAtom) argv[0];
	call = (KlList) KlListCopy(cur_gfc->list);
	KlDecRefNonNull(call->list[0]);
	KlIncRef(call->list[0] = (KlO) selector);
	curmethod = KlIsAnExo(obj) 
	    ? KlHashGet(obj->exoclass->methods, selector, NIL)
	    : (KlO) KlMethodOfType(obj->type, selector);	    
    }

    if (KlIsAnExoClass(cur_gfc->exoclass)) {		/* ExO */
	exoclass = (class_orig = cur_gfc->exoclass)->father;
	for (;;) {
	    if (!exoclass) goto no_next_method;
	    method = KlHashGet(exoclass->methods, selector, NIL);
	    if (method != curmethod)
		if (method == NIL) goto no_next_method;
		else break;
	    exoclass = exoclass->father;
	}
    } else {				/* base type */
	type = (KlType) (class_orig = cur_gfc->exoclass);
	for (;;) {
	    if (!(type = KlTypeFatherGet(type))) goto no_next_method;
	    method = (KlO) KlMethodOfType(type, selector);
	    if (method != curmethod)
		if (method == 0) goto no_next_method;
		else break;
	    type = KlTypeFatherGet(type);
	}
	exoclass = (KlExoClass) type;
    }

    if (next) {
	if (argc) {			
	    int i;
	    /* get passed call in the stack to modify method args */
	    ASSERT(((Int) KlStack[KlStackPtr]) == KlSFID_subr);
	    ASSERT(KlIsAList(KlStack[KlStackPtr - KlSFO_call]));
	    if (argc != ((KlList) KlStack[KlStackPtr - KlSFO_call])->size - 1)
		return KlBadNumberOfArguments(argc);
	    /* We set the call list to new arguments */
	    call = (KlList) KlListNMake(argc + 1);
	    KlIncRef(call->list[0] =  (KlO) selector);
	    obj = (KlExo) argv[0];
	    for (i = 0; i < argc; i++)
		KlIncRef(call->list[i+1] = argv[i]);
	}
    } else {
	if (argc > 1) {
	    /* get passed call in the stack to modify method args */
	    int i;

	    ASSERT(((Int) KlStack[KlStackPtr]) == KlSFID_subr);
	    ASSERT(KlIsAList(KlStack[KlStackPtr - KlSFO_call]));

	    call = KlListNMake(argc);
	    for (i=0 ; i< argc ; i++) 
		KlIncRef( call->list[i]= argv[i] );
	    obj = (KlExo) argv[1];
	}
    }
    KlExoApplyMethodReturn(KlSend_apply(method, call),
			   method, selector, obj, exoclass, call);

 no_next_method:
#ifdef OLD_ERROR_ON_NO_NEXT_METHOD
    return KlError3(KlE_NO_NEXT_METHOD, selector, class_orig, call);
#else /* !OLD_ERROR_ON_NO_NEXT_METHOD */
    return NIL;
#endif /* !OLD_ERROR_ON_NO_NEXT_METHOD */
}


/* call-next-method
 * used to get the next method up in the hierarchy and apply to args
 * current selector and class are stored in the globals KlA_CurGenFuncSel
 * and KlA_CurGenFuncClass (*current-generic-function-selector* and
 * *current-generic-function-class* at klone level);
 */

KlO
KlCallNextMethod(argc, argv)
    int argc;
    KlO *argv;
{
    return KlCallAnotherMethod(1, argc, argv);
}

/*
 * KlCallSuperMethod
 *
 * KlCallSuperMethod is used to call the next method which selector is given
 * by argv[0].
 * (call-next-method ...)  
 *    == (call-super-method *current-generic-function-selector* ...)
*/
KlO
KlCallSuperMethod(argc, argv)
    int argc;
    KlO *argv;
{
    return KlCallAnotherMethod(0, argc, argv);
}

/* KlGenericFunctionCatchallSet
 * sets a function to be called as this method for any primitive klone type
 * by convention, the "Catch-All" method will be the one defined on the Any
 * type.
 */

KlO
KlGenericFunctionCatchallSet(selector, func)
    KlAtom selector;
    KlO func;
{
    KlType t;
    KlO oldfunc = func;
    KlO old_catchall = KlHashGet(KlTypeMethodsGet(KlAnyType), selector, NIL);

    if (!(func = KlCheckMethodValidity(func, selector, 1))) {
	return KlError2(KlE_CANNOT_DECLARE_GENERIC, selector, oldfunc);
    }
    for (t = KlTypes; t; t = KlTypeNext(t)) {
	if (KlHashGet(KlTypeMethodsGet(t), selector, NIL) == old_catchall) {
	    KlHashPut(KlTypeMethodsGet(t), selector, func);
	}
    }
    return func; 
}

/* creates type[KlSelMethods] hashtable of new types
 */

KlCatchallMethodsInit(type)
    KlType type;
{
    KlHash ht;
    
    if (KlHashInitialized) {
	KlIncRef(ht = KlHashAlloc(0));
	ht->ref_counted = KlHashRefV;
	KlTypeMethodsSet(type, ht);
    }
}

/*****************************************************************************\
* 				   Printing                                   *
\*****************************************************************************/

/* printing 
 */

KlO
KlExoClassPrint(obj, stream)
    KlExoClass obj;
    KlO stream;
{
    if (KlExoHasHook(obj, KlSelEHPrint)) {
	KlApplyMethod2(KlExoMethodNamed(obj, KlA_print),
		       KlA_print, obj, stream);
    } else {
	KlSPrintf(stream, "{^ Class 0x%x ", obj);
	KlSPrint(obj->name, stream);
	KlSPuts("}", stream);
    }

    return (KlO) obj;
}

/* printing an object can be redefined by the method print-object
 */

KlO
KlExoPrint(obj, stream)
    KlExo obj;
    KlO stream;
{
    KlO method;
    int i;

    if (++KlPrintLevel > KlMaxPrintLevel) {
	KlSPuts("...", stream);
    } else {
	if (KlExoHasHook(obj, KlSelEHPrint)) {
	    KlApplyMethod2(KlExoMethodNamed(obj, KlA_print),
			   KlA_print, obj, stream);
	} else {
	    KlO *slotnames = KlAlloca(obj->exoclass->slots->size);
	    KlO *get_hooks = KlAlloca(obj->exoclass->slots->size);
	    KlO *p = obj->slots;
	    KlO *last = p + obj->exoclass->slots->size;
	    /* sort slots by date */
	    KlHashFORBEGIN(obj->exoclass->slots, cell) {
		UInt actual_offset = KlExoActualOffset(cell->object);
		slotnames[actual_offset] = cell->key;
		if (KlPI_INT(cell->object)) { /* no hook */
		    get_hooks[actual_offset] = 0;
		} else {		/* hook */
		    if (((KlSlotHooks) cell->object)->get == NIL)
			get_hooks[actual_offset] = 0;
		    else
			get_hooks[actual_offset] =
			    ((KlSlotHooks) cell->object)->get;
		}
	    } KlHashFOREND;
	    
	    if (KlPrintReadably) {
		KlSPuts("{^ ExO", stream);
		KlSPrintf(stream, " 0x%x", obj);
	    } else {
		KlSPuts("{", stream);
		KlSPrint(obj->exoclass->name, stream);
	    }
	    while (p < last) {
		if (((KlAtom)(*slotnames))->p_name[0] != '&'
		    || KlPrintReadably) {
		    KlSPuts(" ", stream);
		    KlSPuts(((KlAtom)(*slotnames))->p_name, stream);
		    KlSPuts(" ", stream);
		    if (*get_hooks) {
			/* KlSPrint is a macro, we must protect first arg from
			 *  a double evaluation
			 */
			KlO slot_value = 
			    KlApplyV(*get_hooks, 3, obj, *slotnames, *p);
			KlSPrint(slot_value, stream);
		    } else {
			KlSPrint(*p, stream);
		    }
		}
		p++, slotnames++; get_hooks++;
	    }
	    
	    KlSPuts("}", stream);
	}
    }
    KlPrintLevel--;
    return (KlO) obj;
}


/*****************************************************************************\
*                             Method Facilities                               *
\*****************************************************************************/

/*
 * KlFindMethod
 *
 * Returns the lambda of a method, according to the selector and the class
 * or type of the method.
 * works also on types
 *
 * tolerates an alternate new syntax, i.e:
 * (find-method [class | type] selector)
 * (find-method generic [class | type])
 */

KlO
KlFindMethod (exoclass, selector)
	KlExoClass exoclass;
	KlAtom selector;
{
    KlMustBeAtom(selector,0);

    if (KlIsAGenericFunction(exoclass)) { /* second form */
	KlGenericFunction gen = (KlGenericFunction) exoclass;
	exoclass = (KlExoClass) selector;
	selector = gen->name;
    }

    if (KlIsAnExoClass(exoclass)) {	/* class */
	return KlHashGet(exoclass->methods,selector,NIL);
    } else {				/* type */
    	KlType type = KlFindType(exoclass);
        return KlHashGet(KlTypeMethodsGet(type), selector,NIL);
    }
}

/*****************************************************************************\
* 				  utilities                                   *
\*****************************************************************************/

KlO
KlClassOf(obj)
    KlExo obj;
{
    KlMustBeExo(obj, 0);
    return (KlO) obj->exoclass;
}

/*****************************************************************************\
* 			       Hookable methods                               *
\*****************************************************************************/

KlO
KlExoDelete(obj, slot)
    KlExo obj;
    KlAtom slot;
{
    if (KlExoHasHook(obj, KlSelEHDelete)) {
	return KlApplyMethod2(KlExoMethodNamed(obj, KlA_delete),
			      KlA_delete, obj, slot);
    } else {
	return CFAPPLY(KlUndefinedMethod(KlSelDelete), (obj, slot));
    }    
}

KlO
KlExoInsert(obj, slot, value)
    KlExo obj;
    KlAtom slot;
    KlO value;
{
    if (KlExoHasHook(obj, KlSelEHInsert)) {
	return (KlO)
	    KlApplyMethod3(KlExoMethodNamed(obj, KlA_insert),
			   KlA_insert, obj, slot, value);
    } else {
	return CFAPPLY(KlUndefinedMethod(KlSelInsert), (obj, slot, value));
    }    
}

KlO
KlExoEval(obj)
    KlExo obj;
{
    if (KlExoHasHook(obj, KlSelEHEval)) {
	return KlApplyMethod1(KlExoMethodNamed(obj, KlA_eval), KlA_eval, obj);
    } else {
	return (KlO) obj;
    }    
}

KlO
KlExoExecute(obj, arg)
    KlExo obj;
    KlO arg;
{
    if (KlExoHasHook(obj, KlSelEHExecute)) {
	return KlApplyMethod2(KlExoMethodNamed(obj, KlA_execute),
			      KlA_execute, obj, arg);
    } else {
	return CFAPPLY(KlUndefinedMethod(KlSelExecute), (obj, arg));
    }    
}

KlO
KlExoApply(obj, arg)
    KlExo obj;
    KlO arg;
{
    if (KlExoHasHook(obj, KlSelEHApply)) {
	return KlApplyMethod2(KlExoMethodNamed(obj, KlA_apply),
			      KlA_apply, obj, arg);
    } else if (KlExoHasHook(obj, KlSelEHExecute)) {
	return KlApplyMethod2(KlExoMethodNamed(obj, KlA_execute),
			      KlA_execute, obj, arg);
    } else {
	return CFAPPLY(KlUndefinedMethod(KlSelApply), (obj, arg));
    }    
}

KlO
KlExoEqual(obj, arg)
    KlExo obj;
    KlExo arg;
{
    if (KlExoHasHook(obj, KlSelEHEqual)) {
	return KlApplyMethod2(KlExoMethodNamed(obj, KlA_equal),
			      KlA_equal, obj, arg);
    } else {
	return obj == arg ? TRU : NIL;
    }    
}

KlO
KlExoSetq(obj, arg)
    KlExo obj;
    KlO arg;
{
    if (KlExoHasHook(obj, KlSelEHSetq)) {
	return KlApplyMethod2(KlExoMethodNamed(obj, KlA_setq),
			      KlA_setq, obj, arg);
    } else {
	return CFAPPLY(KlUndefinedMethod(KlSelSetq), (obj, arg));
    }    
}

KlO
KlExoAdd(argc, argv)
    int argc;
    KlExo *argv;
{
    if (KlExoHasHook(argv[0], KlSelEHAdd)) {
	return KlApplyMethodN(KlExoMethodNamed(argv[0], KlA_add),
			      KlA_add, argc, argv);
    } else {
	return CFAPPLY(KlUndefinedMethod(KlSelAdd), (argc, argv));
    }    
}

KlO
KlExoNth(obj, i, value)
    KlExo obj;
    UInt i;
    KlO value;
{
    if (KlExoHasHook(obj, KlSelEHNth)) {
	/* nth system method called with last arg = :nil for get */
	return KlApplyMethod3(KlExoMethodNamed(obj, KlA_nth),
			      KlA_nth, obj, KlNumberMake(i),
			      (value ? value : (KlO) KlK_nil));
    } else {
	if (value) {			/* put */
	    if (i < obj->exoclass->slots->size) {
		KlDecRefNonNull(obj->slots[i]);
		KlIncRef(obj->slots[i] = value);
		return (KlO) obj;
	    } else {
		return KlError2(KlE_NO_ELEMENT, KlNumberMake(i),
				obj->exoclass->name);
	    }
	} else {			/* get */
	    if (i < obj->exoclass->slots->size) {
		return obj->slots[i];
	    } else {
		return NIL;
	    }
	}
    }
}

KlO
KlExoLength(obj)
    KlExo obj;
{
    if (KlExoHasHook(obj, KlSelEHLength)) {
	return KlApplyMethod1(KlExoMethodNamed(obj, KlA_length),
			      KlA_length, obj);
    } else {
	return CFAPPLY(KlUndefinedMethod(KlSelLength), (obj));
    }    
}

KlO
KlExoHash(obj)
    KlExo obj;
{
    if (KlExoHasHook(obj, KlSelEHHash)) {
	return (KlO)
	    ((KlNumber) KlApplyMethod1(KlExoMethodNamed(obj, KlA_hash),
				       KlA_hash, obj))
		->number;
    } else {
	return (KlO) obj;
    }    
}

KlO
KlExoDolist(obj, var, argc, argv)
    KlExo obj;
    KlO var;
{
    KlList list = KlListKl(argc, argv);

    if (KlExoHasHook(obj, KlSelEHDolist)) {
	return KlApplyMethod3(KlExoMethodNamed(obj, KlA_dolist),
			      KlA_dolist, obj, var, list);
    } else {
	return CFAPPLY(KlUndefinedMethod(KlSelDolist), (obj, var,
		       KlNumberMake(argc), list));
    }    
}

KlO
KlExoDohash(obj, var, val, argc, argv)
    KlExo obj;
    KlO var;
    KlO val;
{
    KlList list = KlListKl(argc, argv);

    if (KlExoHasHook(obj, KlSelEHDohash)) {
	return KlApplyMethod4(KlExoMethodNamed(obj, KlA_dohash),
			      KlA_dohash, obj, var, val, list);
    } else {
	return CFAPPLY(KlUndefinedMethod(KlSelDohash), (obj, var, val,
		       KlNumberMake(argc), list));
    }    
}

int
KlExoCompare(obj, obj2)
    KlExo obj;
    KlO obj2;
{
    KlNumber res;
    if (KlExoHasHook(obj, KlSelEHCompare)) {
	res = (KlNumber) KlApplyMethod2(KlExoMethodNamed(obj, KlA_compare),
			      KlA_compare, obj, obj2);
    } else {
	res = (KlNumber) CFAPPLY(KlUndefinedMethod(KlSelCompare), (obj, obj2));
    }    
    return (int) res->number;
}

/*****************************************************************************\
* 			 Describe support for classes                         *
\*****************************************************************************/

/*
 * KlExoClassOptions
 *
 * Returns a plist of class local options or NIL.
 *
*/

KlList
KlExoClassOptions(exoclass)
        KlExoClass exoclass;
{
  int size=0;
  KlO doc ;
  KlO metaclass = 0;
    
  if ((doc = (KlO) KlDocumentationGet( exoclass->name)) != NIL)
      size +=2;
  if (KlExoMetaclassClassOptionIsLocal(exoclass) &&
      (metaclass = (KlO) exoclass->exoclass->name))
    size += 2;
  if (size) {
    int     raw=0;
    KlList  list;
    
    list= KlListNMake(size);
    
    if (doc != NIL) {
      KlIncRef(list->list[raw++]= (KlO) KlK_documentation);
      KlIncRef(list->list[raw++]= (KlO) doc);
    }
    if (metaclass) {
      KlIncRef(list->list[raw++]= (KlO) KlK_metaclass);
      KlIncRef(list->list[raw++]= (KlO) metaclass);
    }
    return  list;
  } else {
    return (KlList) NIL;
  }
}


/*
 * KlExoSlotOptions(exoclass, slot)
 *
 * Returns a p-list of slot local options
*/

KlList
KlExoSlotOptions(exoclass, slot)
    KlExoClass exoclass;
    KlAtom slot;
{
    
    UInt offset = (UInt) KlHashGet(exoclass->slots, slot, KlPI_INVALID_VALUE);
    
    if  (offset != KlPI_INVALID_VALUE) {
	/* the slot exists */
	KlList list = KlListNMake(0);
	KlO option;

	/* Retrieve documentation string */
	if ((option = KlDocumentationGet(
	    KlInternDotPrefixed(exoclass->name->p_name, slot->p_name))) != NIL) {
	    KlListAppend2(list, KlK_documentation, option);
	}
	/* Retrieve the :initform option */
	if ((
	    /* slot not in father, test init[offset] different from NIL  */
	    (!(exoclass->father) 
	     || 
	     ((exoclass->father)
	      && ((int) KlHashGet(exoclass->father->slots, slot, -1)<0)))
	    && (exoclass->init[KlExoActualOffset(offset)] != NIL)) 
	    ||
	    (/* slot redefined, tests init[offset] differs from superclass */
		(exoclass->father) && 
		((UInt) KlHashGet(exoclass->father->slots, slot, 
				  KlPI_INVALID_VALUE) != KlPI_INVALID_VALUE) &&
		(exoclass->init[KlExoActualOffset(offset)]
		 != exoclass->father->init[KlExoActualOffset(offset)]))
	    ){
	    KlListAppend2(list, KlK_initform,
			  exoclass->init[KlExoActualOffset(offset)]);
	}
	/* test the get and put hooks */
	if (KlPI_PTR(offset)) {
	    KlListAppendV(list, 4, KlK_get, ((KlSlotHooks) offset)->get,
			  KlK_put, ((KlSlotHooks) offset)->put);
	}
	/* test the slot_properties */
	if (KlTrueP(exoclass->slot_properties[KlExoActualOffset(offset)])) {
	    KlListAppend2(list, KlK_properties,
			  exoclass->slot_properties[KlExoActualOffset(offset)]);
	}

	return list;
    } else {
	/* ERROR: the given argument is not a slot of the given class */
	return (KlList) KlError2(KlE_NO_ELEMENT, slot, exoclass->name);
    }
}

/****************************************************************************\
 *                              Describe				    * 
\****************************************************************************/

/*
 * KlExoClassOptionIsLocal
 *
 * Test wether a class option is local (returns 1)
 * or inherited (return 0).
 *
 * tested class options are: constructor, destructor and metaclass.
*/

#ifdef COMMENT
/* this code causes warning on solaris brain-dead compiler... */
#define KlExoClassOptionIsLocal(procedure, option) \
procedure (exoclass) \
	KlExoClass	exoclass; \
{ \
  if (exoclass->father) { \
    return (exoclass->option != exoclass->father->option); \
  } else { \
    return 1; \
  } \
}
KlExoClassOptionIsLocal(KlExoMetaclassClassOptionIsLocal, exoclass);
#endif /* COMMENT */

/* expanded version of the above */
KlExoMetaclassClassOptionIsLocal(exoclass) 
    KlExoClass	exoclass; 
{ 
    if (exoclass->father) { 
	return (exoclass->exoclass != exoclass->father->exoclass); 
    } else { 
	return 1; 
    } 
}

/*
 * KlExoDescribe (argc, argv)
 *
*/

KlO
KlExoDescribe(argc, argv)
    int argc;
    KlO *argv;
{
  
    KlNumberOfArgumentsCheck((argc == 0) || (argc > 2) , argc);
    if (argc == 1) {
	if (KlIsAnExoClass(argv[0])) {
	    /* describe a class */
	    return KlExoDescribeClass (argv[0]);
	} else {
	    KlMustBeExo(argv[0],0);
	    /* describe an instance */
	    return KlExoDescribeInstance (argv[0]);
	}
    } else {
	/* describe a method ? */
	return KlExoDescribeMethod(argv[0], argv[1]);
    }
}


/*
 * KlExoDescribeClass (exoclass)
 *
 * Warning: we do not print the names but the classes themselves if they are
 * obsolete, since the obsolete name do not point to the class (otherwise it 
 * could never be freed)
 */

KlO
KlExoDescribeClass (exoclass)
     KlExoClass exoclass;
{
    KlList localSlots = KlExoLocalSlots(exoclass);
    KlList classOptions = KlExoClassOptions(exoclass);
    KlList result = KlListNMake ( 4 + classOptions->size);
    int i;
  
    /* function name */
    KlListStore (result, 0, KlIntern("defclass"));
  
    /* class name */
    if (exoclass->exoHooks & KlSelEHOBSOLETE)
	KlListStore (result, 1, exoclass);
    else
	KlListStore (result, 1, exoclass->name);
  
    /* superclass */
    if(exoclass->father) {
	if (exoclass->father->exoHooks & KlSelEHOBSOLETE)
	    KlListStore (result, 2, exoclass->father);
	else
	    KlListStore (result, 2, exoclass->father->name);
    } else {
	KlListStore (result, 2, NIL);
    }

    /* slots */

    if (localSlots->size) {	
	/* there are local slots */
	KlList resultSlots = KlListNMake (localSlots->size);

	for (i=0; i < localSlots->size; i++) {
	    /* describe a slot */
	    KlList slotOptions = KlExoSlotOptions(exoclass,
						  localSlots->list[i]);
      
	    if (slotOptions->size) {
		/* there are slot options */
		int j;
		KlList resultSlotOptions = KlListNMake (slotOptions->size +1);
	
		KlListStore (resultSlotOptions, 0, localSlots->list[i]);
		for (j=0 ; j < slotOptions->size; j++) {
		    KlListStore (resultSlotOptions, j+1, slotOptions->list[j]);
		}
		KlListStore (resultSlots, i, resultSlotOptions);
	    } else {
		/* no slot options */
		KlListStore (resultSlots, i, localSlots->list[i]);
	    }
	}
	KlListStore (result, 3, resultSlots);
    } else {
	/* No local Slots */
	KlListStore (result, 3, NIL);
    }

    /* class options */

    for (i = 0 ; i < classOptions->size; i++) {
	KlListStore (result, 4+i, classOptions->list[i]);
    }

    return (KlO) result;
}
 
/*
 * KlExoDescribeInstance (obj)
 *
*/

KlO
KlExoDescribeInstance (obj)
        KlExo obj;
{
    KlList result = KlListNMake (2);

    /* function name */
    KlListStore (result, 0, KlIntern("make-instance"));
  
    /* class name */
    KlListStore (result, 1, obj->exoclass->name);

    /* slots values if different than :initform values */
    KlHashFORBEGIN(obj->exoclass->slots, cell) {
	UInt offset = KlExoActualOffset((UInt) cell->object);
    
	if (obj->slots[offset] != obj->exoclass->init[offset]) {
	    /* slot value differs from init value of the defclass function: */
	    KlListAppend (result, KlKeywordMakeFromAtom( cell->key));
	    KlListAppend (result,  obj->slots[offset]);
	}				/* otherwise dont mention it */
    } KlHashFOREND;

    return (KlO) result;
}

/*
 * KlExoDescribeMethod(selector, exoclass)
 *
*/

KlO
KlExoDescribeMethod(selector, exoclass)
	KlAtom selector;
	KlExoClass exoclass;
{
    KlAtom doc, classOrType;
    KlExpr lambda = 0;
    Int 	i, offset=0;
  
    /* Colas: we tolerate the (describe class method) syntax too,
     * for consistency
     */
    if (KlIsAnExoClass(selector)) {
	KlO swap = (KlO) selector;
	selector = (KlAtom) exoclass;
	exoclass = (KlExoClass) swap;
    }
    if (KlIsAGenericFunction(selector)) {
	selector = ((KlGenericFunction) selector)->name;
    }

    KlMustBeAtom (selector, 0);
    if (KlIsAnExoClass(exoclass)) {
	classOrType = exoclass->name;
	lambda=  (KlExpr) KlHashGet(exoclass->methods,selector,0);
	if (!lambda )
	    return KlError2(KlE_NO_METHOD, selector, exoclass->name);
    }
    else {
	KlType type = KlFindType(exoclass);
	classOrType = KlTypeName(type); 
	lambda= (KlExpr) KlHashGet(KlTypeMethodsGet(type),selector,0);
	if (!lambda)
	    return KlError2(KlE_NO_METHOD, selector,type);
    }
  
    /* Retrieve documentation if exists */
    if ((doc=(KlAtom) KlDocumentationGet(KlInternDotPrefixed(classOrType->p_name, selector->p_name))) != (KlAtom) NIL)
	offset= 1;
  
    {
	KlList result= KlListNMake(3 +lambda->body_size + offset); 
    
	/* function name */
	KlListStore (result, 0, KlIntern("defmethod"));
    
	/* selector name */
	KlListStore (result, 1, selector);
    
	/* called parameters */
    
	{
	    KlList calledParam = (KlList) KlUnparseLambdaList(lambda);
	    KlList firstParam = KlListNMake (2);
      
	    KlListStore (firstParam, 0 , calledParam->list[0]);
	    KlListStore (firstParam, 1, classOrType);
	    KlListStore (calledParam, 0, firstParam);
      
	    KlListStore (result, 2, calledParam);
	}
    
	/* if documentation, store it */
	if (offset) {
	    KlListStore (result, 3 , doc); 
	}
    
	/*  body */
	for (i=0 ; i < lambda->body_size ; i++) {
	    KlListStore ( result, 3 + offset + i , lambda->body[i]);
	}
    
	return (KlO) result;
    }
}

/*
 *
 * KlExoSlotExistsP
 *
 * Predicat to test slot existance
*/

KlO
KlExoSlotExistsP(exo, slot)
     KlExo exo;
     KlAtom slot;
{
    KlMustBeExo(exo, 0);
    KlMustBeAtom(slot, 1);
    return ((UInt) KlHashGet(exo->exoclass->slots, slot, KlPI_INVALID_VALUE) ==
	    KlPI_INVALID_VALUE ? TRU : NIL);
}

#ifdef SMOOL
/*****************************************************************************\
* 			       SMOOL-ONLY part                                *
\*****************************************************************************/
/*****************************************************************************\
* This is code that is only included with SML. Klone would provide the same    *
* functionality as klone autoloadable code                                     *
* (SMOOL is an obsolete name for a SML-specific OO extension)                  *
\*****************************************************************************/
/*****************************************************************************\
*                                                                             *
*  		  @@@@@@@  @       @   @@@@@     @@@@@   @                    *
*  		 @       @ @@     @@  @     @   @     @  @                    *
*  		 @         @ @   @ @ @       @ @       @ @                    *
*  		  @@@@@@@  @  @ @  @ @       @ @       @ @                    *
*  		         @ @   @   @ @       @ @       @ @                    *
*  		 @       @ @       @  @     @   @     @  @                    *
*  		  @@@@@@@  @       @   @@@@@     @@@@@   @@@@@@@              *
*                                                                             *
\*****************************************************************************/
/*
** For ease of management, ISM-designed (by Laurent Fontvielle) code 
** is integrated here for automatic inclusion before being moved 
** gradually to kl_exo.c main body.
** 
** This code is only included if SMOOL (or ISM) is #defined
*/

extern KlO		KlDocumentationGet();
extern KlList		KlExoLocalSlots();
extern KlO		KlExoClassDeclareBuiltInSlots();
extern KlO              KlAccessClassChildren();

/***************************************************************************

  CLASS DEFINITION, REDEFINITION

****************************************************************************/

/* Defclass
   
   layer #1, compatible with C A.P.I.

*/

KlO
KlExoDefClass(name, fatherName, slots, nslots,
	      inits, documentations,
	      documentation, metaclassName)
     KlO name;
     KlO fatherName;
     KlO *slots;
     int nslots;
     KlO *inits;
     KlO *documentations;
     KlO documentation;
     KlO metaclassName;
{
    KlExoClass exoclass;
    KlExoClass father;
    int i, j;
    KlExoClass metaclass = 0;

    KlMustBeSymbol(name, 0);

    /* redefine Class or Object is FORBIDDEN */
    if ((((KlAtom) name)->c_val != KlUndef) &&
	KlIsAnExoClass(((KlAtom) name)->c_val) &&
	((((KlExoClass)((KlAtom) name)->c_val) == KlExoClassMeta) || 
	 (((KlExoClass)((KlAtom)  name)->c_val) == KlExoClassRoot))) {
	return NIL;
    }
    /* if no class father, then superclass is Object by default */
    if (fatherName == (KlO) NIL) {	
	father = KlExoClassRoot;
    } else {
	/* if class father, then check it */
	if (KlIsAList(fatherName)) {
	    KlList clist = (KlList) fatherName;
	    KlMustBeOrEval(father, clist->list[0], KlExoClass,
			   KlIsAnExoClass(father),
			   KlError(KlE_NO_CLASS, clist->list[0]));
    
	} else {
	    KlMustBeOrEval(father, fatherName, KlExoClass,
			   KlIsAnExoClass(father),
			   KlError(KlE_NO_CLASS, fatherName));
	}
    }


    if (nslots) {			/* parses slot descriptions */
	for (j= 0; j < i; j++) {	/* check slots are not defined twice */
	    if (slots[j] == slots[i]) {
		KlError2(KlE_BAD_CLASS_DEF, name, slots[i]);
	    }
	} 
    }
    
    if (metaclassName) {		/* check metaclass validity */
	KlMustBeOrEval(metaclass, metaclassName, KlExoClass,
		       KlIsAnExoClass(metaclass),
		       KlError(KlE_NO_CLASS, metaclassName));
	KlExoClassDeclareBuiltInSlots(metaclass);
    }
    if (!metaclass) {
      /* if no metclass specified, then the metaclass is the metaclass of the superclass 
	 by default. If the superclass is Object, the metaclass is Class */
	metaclass = father->exoclass;
    }

    /* call  KlExoClassMake for doing the actual job*/
    exoclass = (KlExoClass) KlExoClassMake
                     (name, father, slots, nslots,
		      inits, metaclass, documentations, documentation,
		      0, 0		/* no slot hooks for now */
		      ,0		/* nor slot properties */
		      );
    
    return (KlO) exoclass;
}

/*****************************************************************************\
*                              Slot Facilities                                *
\*****************************************************************************/


/*
 *  KlExoSlotValue(slotName)
 *
 * KLONE INTERFACE: slot-get
 *
 */
KlO
  KlExoSlotValue(slotName)
KlAtom slotName;
{
    UInt offset;
    KlGenFuncContext cur_gfc = (KlGenFuncContext) /* find method context*/
	KlStackFrameLookForTag(KlSFID_genfunc);
  
    if (!cur_gfc) {			/* we were not in a method */
	return KlError0(KlE_NO_PREVIOUS_METHOD);
    } 
  
    offset = (UInt) KlHashGet(cur_gfc->exoclass->slots, slotName, 
			     KlPI_INVALID_VALUE);
    if (KlPI_INT(offset)) {		/* normal slot */
	return cur_gfc->obj->slots[offset];
    } else if (KlPI_INVALID(offset)) {	/* not found */
	return KlExecuteGetDefault(cur_gfc->obj, slotName, KlE_NO_ELEMENT);
    } else {				/* hooked */
	if (KlIsASubr(((KlSlotHooks) offset)->get)) {
					/* optimisation */
	    return (*(((KlSubr) ((KlSlotHooks) offset)->get)->body))
		(cur_gfc->obj, slotName, 
		 cur_gfc->obj->slots[((KlSlotHooks) offset)->offset]);
	} else if (((KlSlotHooks) offset)->get == NIL) {
	    return cur_gfc->obj->slots[((KlSlotHooks) offset)->offset];
	} else {
	    return KlApplyV(((KlSlotHooks) offset)->get,
			    3, cur_gfc->obj, slotName, 
			    cur_gfc->obj->slots[((KlSlotHooks) offset)->offset]);
	}
    }
}

/*
 * KlExoSlotSetValue(slotName, newValue)
 *
 * KLONE INTERFACE: slot-put
 */
KlO
  KlExoSlotSetValue(slotName, newValue)
KlAtom slotName;
KlO newValue;
{
    UInt offset;
    KlGenFuncContext cur_gfc = (KlGenFuncContext) /* find method context*/
	KlStackFrameLookForTag(KlSFID_genfunc);
  
    if (!cur_gfc) {			/* we were not in a method */
	return KlError0(KlE_NO_PREVIOUS_METHOD);
    } 
    offset = (UInt) KlHashGet(cur_gfc->exoclass->slots, slotName, 
			      KlPI_INVALID_VALUE);
    if (KlPI_INT(offset)) {		/* normal slot */
	KlDecRefNonNull(cur_gfc->obj->slots[offset]);
	KlIncRef(cur_gfc->obj->slots[offset] = newValue);
    } else if (KlPI_INVALID(offset)) {	/* not found */
	return KlError2(KlE_NO_ELEMENT, slotName, cur_gfc->obj->exoclass->name);
    } else {				/* hooked */
	KlO *actual_slot = 
	    &(cur_gfc->obj->slots[((KlSlotHooks) offset)->offset]);
	KlDecRef(*actual_slot);
	if (KlIsASubr(((KlSlotHooks) offset)->put)) {
					/* optimisation */
	    KlIncRef(*actual_slot =
		     (*(((KlSubr) ((KlSlotHooks) offset)->put)->body))
		     (cur_gfc->obj, slotName, newValue));
	} else if (((KlSlotHooks) offset)->put == NIL) {
	    KlIncRef(*actual_slot = newValue);
	} else {
	    KlIncRef(*actual_slot = 
		     KlApplyV(((KlSlotHooks) offset)->put, 3, cur_gfc->obj, 
			      slotName, newValue));
	}
	return newValue;
    }
}

/*****************************************************************************\
* 				 KlSMLExoInit                                 *
\*****************************************************************************/

KlSMLExoInit()
{
    KlDeclareSubr(KlExoSlotValue,"slot-get", 1);
    KlDeclareSubr(KlExoSlotSetValue,"slot-put", 2);
}
/*****************************************************************************\
* 				  SMOOL END                                   *
\*****************************************************************************/
#endif /* SMOOL */

/*****************************************************************************\
* 			       slot properties                                *
\*****************************************************************************/
/* a simple form: get or set
 * KlSlotProperties(class, slot) returns property. if none set, returns it.
 *     WARNING: may be returning NIL, in which case you must set it to A nil
 *     list (or better, vector) before adding into it
 * KlSlotProperties(class, slot, plist) sets property to it 
 */

KlO
KlSlotProperties(argc, argv)
    int argc;
    KlO *argv;
{
    KlExoClass exoclass;
    KlAtom slot;
    UInt offset;

    KlNumberOfArgumentsCheck(argc < 2 || argc > 3, argc)
	KlMustBeExoClass(argv[0], 0);
    exoclass = (KlExoClass) argv[0];
    KlMustBeAtom(argv[1], 1);
    slot = (KlAtom) argv[1];
    offset = (UInt) KlHashGet(exoclass->slots, slot, KlPI_INVALID_VALUE);

    if  (argc == 2) {
	/* get the properties */
	return exoclass->slot_properties[KlExoActualOffset(offset)];
    } else {
	/* set the properties */
	KlMustBeEvenList(((KlList) argv[2]), 2);
	KlDecRef(exoclass->slot_properties[KlExoActualOffset(offset)]);
	KlIncRef(exoclass->slot_properties[KlExoActualOffset(offset)] = argv[2]);
	return (KlO) argv[2];
    }
}

/**************************************************************************\
* 				  SlotHooks                                *
\**************************************************************************/
/* get slot hooks, if any set */

KlO
KlGetSlotHooks(exoclass, slot)
    KlExoClass exoclass;
    KlAtom slot;
{
    UInt offset;
    if (KlIsAKeyword(slot))
	slot = ((KlKeyword)(slot))->atom;
    KlMustBeExoClass(exoclass, 0);
    KlMustBeAtom(slot, 1);

    offset = (UInt) KlHashGet(exoclass->slots, slot, KlPI_INVALID_VALUE);
    if (KlPI_INT(offset)) {		/* normal slot */
	return NIL;
    } else if (KlPI_INVALID(offset)) {	/* not found */
	return KlError2(KlE_NO_ELEMENT, slot, exoclass);
    } else {				/* hooked */
	return (KlO) KlListPairMake(((KlSlotHooks) offset)->get,
				    ((KlSlotHooks) offset)->put);
    }
}

KlO
KlDefSlotHooks(exoclass, slot, get, put)
    KlExoClass exoclass;
    KlAtom slot;
    KlO get;
    KlO put;
{
    UInt offset;
    if (KlIsAKeyword(slot))
	slot = ((KlKeyword)(slot))->atom;
    KlMustBeExoClass(exoclass, 0);
    KlMustBeAtom(slot, 1);
    if (KlIsUndefinedMethod(get, KlSelApply) && get != NIL)
	KlMustBeFunction(get, 2);
    if (KlIsUndefinedMethod(put, KlSelApply) && put != NIL)
	KlMustBeFunction(put, 3);
    if (KlIsAMetaclass(exoclass)) {
	/* we forbid hooks on metaclass slots, too many problems */
	KlBadArgument(exoclass, 0, "non-metaclass class");
    }

    offset = (UInt) KlHashGet(exoclass->slots, slot, KlPI_INVALID_VALUE);
    if (get == NIL && put == NIL)	/* remove hooks, see below */
	return KlRemoveSlotHooks(exoclass, slot, offset);
    if (KlPI_INT(offset)) {		/* normal slot */
	KlSlotHooksMake(exoclass, slot, offset, get, put);
    } else if (KlPI_INVALID(offset)) {	/* not found */
	return KlError2(KlE_NO_ELEMENT, slot, exoclass);
    } else {				/* update existing hooks */
	KlDecRef(((KlSlotHooks) offset)->get);
	KlIncRef(((KlSlotHooks) offset)->get = get);
	KlDecRef(((KlSlotHooks) offset)->put);
	KlIncRef(((KlSlotHooks) offset)->put = put);
    }
    return NIL;
}

KlSlotHooks
KlSlotHooksMake(exoclass, slot, offset, get, put)
    KlExoClass exoclass;
    KlAtom slot;
    UInt offset;
    KlO get;
    KlO put;
{
    KlSlotHooks hooks;			/* create a hookslot struct */
    KlMustBeSlotHook(get, slot);
    KlMustBeSlotHook(put, slot);
    hooks = (KlSlotHooks) Malloc(sizeof(struct _KlSlotHooks));
    hooks->offset = offset;
    KlIncRef(hooks->get = get);
    KlIncRef(hooks->put = put);
    KlHashPut(exoclass->slots, slot, hooks);
    return hooks;
}

KlO
KlRemoveSlotHooks(exoclass, slot, offset)
    KlExoClass exoclass;
    KlAtom slot;
    UInt offset;
{
    if (KlPI_INT(offset)) {		/* normal slot, no need */
    } else if (KlPI_INVALID(offset)) {	/* not found */
	return KlError2(KlE_NO_ELEMENT, slot, exoclass);
    } else {				/* hooked, remove */
	UInt raw_offset = ((KlSlotHooks) offset)->offset;
	KlDecRef(((KlSlotHooks) offset)->get);
	KlDecRef(((KlSlotHooks) offset)->put);
	Free((KlSlotHooks) offset);
	KlHashPut(exoclass->slots, slot, raw_offset);
    }
    return NIL;
}

KlMustBeSlotHook(func, slot)
    KlExpr func;
    KlO slot;
{
    if (!(func == (KlExpr) NIL
	|| (KlIsALambda(func) && func->arity == 3)
	|| (KlIsACFunc(func) && ((KlSubr) func)->arity == 3))) {
	KlError2(KlE_NOT_A_METHOD, func, slot);
    }
}

/************************************************************** link hook */
/* a very useful hook, "link", to point to another KlO via an object slot
 * without incrementing the refcount of the pointed object
 */
KlO
KlLinkSlotHookGet(obj, slot, value)
    KlExo obj;
    KlAtom slot;
    KlLink value;
{
    return KlIsALink(value) ? KlLinkDereference(value) : (KlO) value;
}

KlO
KlLinkSlotHookPut(obj, slot, value)
    KlExo obj;
    KlAtom slot;
    KlLink value;
{
    return (KlO) KlLinkMake(value);
}


/*****************************************************************************\
* 				  TYPE INIT                                   *
\*****************************************************************************/

KlExoInit()
{
					/* exos */
    KlDeclareType(&KlExoType, "ExObject", 0);
    KlDeclareIsTrait(KlExoType, KlTrait_exo);

    KlDeclareMethod1(KlExoType, KlSelPrint, KlExoPrint);
    KlDeclareMethod1(KlExoType, KlSelGet, KlExoGet);
    KlDeclareMethod1(KlExoType, KlSelPut, KlExoPut);
    KlDeclareMethod1(KlExoType, KlSelDelete, KlExoDelete);
    KlDeclareMethod1(KlExoType, KlSelInsert, KlExoInsert);
    KlDeclareMethod1(KlExoType, KlSelFree, KlExoFree);
    KlDeclareMethod1(KlExoType, KlSelCopy, KlExoCopy);
    KlDeclareMethod1(KlExoType, KlSelEval, KlExoEval);
    KlDeclareMethod1(KlExoType, KlSelExecute, KlExoExecute);
    KlDeclareMethod1(KlExoType, KlSelApply, KlExoApply);
    KlDeclareMethod1(KlExoType, KlSelEqual, KlExoEqual);
    KlDeclareMethod1(KlExoType, KlSelSetq, KlExoSetq);
    KlDeclareMethod1(KlExoType, KlSelAdd, KlExoAdd);
    KlDeclareMethod1(KlExoType, KlSelNth, KlExoNth);
    KlDeclareMethod1(KlExoType, KlSelLength, KlExoLength);
    KlDeclareMethod1(KlExoType, KlSelHash, KlExoHash);
    KlDeclareMethod1(KlExoType, KlSelDolist, KlExoDolist);
    KlDeclareMethod1(KlExoType, KlSelDohash, KlExoDohash);
    KlDeclareMethod1(KlExoType, KlSelCompare, (KlMethod) KlExoCompare);

					/* classes */
    KlDeclareSubType(&KlExoClassType, "ExClass", KlExoType, 0);

    KlDeclareMethod1(KlExoClassType, KlSelPrint, KlExoClassPrint);
    KlDeclareMethod1(KlExoClassType, KlSelFree, KlExoClassFree);
    KlDeclareMethodUndefined(KlExoClassType, KlSelCopy);
    KlDeclareMethod1(KlExoClassType, KlSelApply, KlExoClassApply);
    KlDeclareMethod1(KlExoClassType, KlSelExecute, KlExoClassApply);

					/* generic functions */
    KlA_GenericFunction =
	KlDeclareType(&KlGenericFunctionType, "GenericFunction",
		      sizeof(struct _KlGenericFunction));
    /* child, but we dont inherit anything */
    KlTypeFatherSet(KlGenericFunctionType, KlFunctionType);

    KlDeclareMethod1(KlGenericFunctionType, KlSelExecute,
	KlGenericFunctionExecute);
    KlDeclareMethod1(KlGenericFunctionType, KlSelApply,
	KlGenericFunctionApply);
    KlDeclareMethod1(KlGenericFunctionType, KlSelPrint,
	KlGenericFunctionPrint);

    KlDeclareFSubr(KlDefClass, "defclass", NARY);
    KlDeclareSubr(KlExoRemoveClass,"remove-class", 1);
    KlA_make_instance = (KlAtom)
	KlDeclareSubr(KlExoMakeInstance, "make-instance", NARY);
    KlDeclareSubr(KlExoRawMakeInstance, "*:make-instance", NARY);
    KlDeclareSubr(KlExoRawCopy, "*:copy", 1);
    KlDeclareSubr(KlAddMethodKl, "add-method", 3);
    KlDeclareFSubr(KlDefMethod, "defmethod", NARY);
    KlDeclareSubr(KlCallNextMethod, "call-next-method", NARY);
    KlDeclareSubr(KlCallSuperMethod, "call-super-method", NARY);

    KlDeclareSubr(KlClassOf, "class-of", 1);
    KlDeclareSubr(KlAccessClassName, "class-name", NARY);
    KlDeclareSubr(KlAccessClassFather, "class-father", 1);
#ifdef OBSOLETE
    KlDeclareSubr(KlAccessClassSon, "class-son", 1);
    KlDeclareSubr(KlAccessClassBrother, "class-brother", 1);
#endif /* OBSOLETE */
#ifdef OLD_KLONE_PACKAGE
    KlDeclareSubr(KlExoRawMakeInstance, "klone:make-instance", NARY);
    KlDeclareSubr(KlExoRawCopy, "klone:copy", 1);
    KlDeclareSubr(KlExoRawSlot, "klone:slot", NARY);
    KlDeclareSubr(KlExoRawSlotOffset, "klone:slot-offset", 2);
    KlDeclareSubr(KlGenericFunctionMake, "klone:defgeneric", 1);
#endif

    KlDeclareSubr(KlAccessClassChildren, "class-children", 1);
    KlDeclareSubr(KlAccessClassMethods, "class-methods", 1);
    KlDeclareSubr(KlAccessClassSlots, "class-slots", 1);
    KlDeclareSubr(KlExoMethodRemove, "remove-method", 2);

    KlDeclareSubr(KlFindMethod,"find-method", 2);
    KlDeclareSubr(KlExoRawSlot, "*:slot", NARY);
    KlDeclareSubr(KlExoRawSlotOffset, "*:slot-offset", 2);
    KlDeclareSubr(KlSystemMethodKl, "*:system-method", 2);
    KlDeclareSubr(KlGenericFunctionMake, "*:defgeneric", 1);

    KlDeclareSubr(KlExoSlotExistsP,"slot-exists-p", 2);
    KlDeclareSubr(KlExoDescribe, "describe", NARY);

    KlDeclareSubr(KlGetSlotHooks, "slothooks", 2);
    KlDeclareSubr(KlDefSlotHooks, "defslothooks", 4);
    KlDeclareSubr(KlLinkSlotHookGet, "slothook-linkget", 3);
    KlDeclareSubr(KlLinkSlotHookPut, "slothook-linkput", 3);
    KlDeclareSubr(KlSlotProperties, "slot-properties", NARY);

    KlMakeInstanceFromMetaclassSubr = (KlSubr)
	KlSubrMake(KlSubrType, KlMakeInstanceFromMetaclass, NARY);
    
    /* creates bootstrap classes KlExoClassRoot & KlExoClassMeta */

    KlExoClassRoot = KlExoClassBootstrapMake();
    KlExoClassMeta = KlExoClassBootstrapMake();
    KlExoClassRoot->exoclass = KlExoClassMeta;
    KlExoClassMeta->exoclass = KlExoClassMeta;
    KlExoClassRoot->name = KlA_Object;
    KlExoClassMeta->name = KlA_Class;

    KlExoClassMeta->father = KlExoClassRoot;
    KlIncRef(KlExoClassRoot->son = KlExoClassMeta);

    KlIncRef(KlExoClassRoot->slots = KlHashAlloc(0));
    KlIncRef(KlExoClassRoot->methods = KlHashAlloc(0));
    KlIncRef(KlExoClassMeta->slots = KlHashAlloc(0));
    KlIncRef(KlExoClassMeta->methods = KlHashAlloc(0));

    KlExoClassRoot->slots->ref_counted = 0;
    KlExoClassRoot->methods->ref_counted = KlHashRefV;
    KlExoClassMeta->slots->ref_counted = 0;
    KlExoClassMeta->methods->ref_counted = KlHashRefV;

    KlSend_setq(KlA_Object, KlExoClassRoot);
    KlSend_setq(KlA_Class, KlExoClassMeta);
    KlA_Object->type = KlConstantType;
    KlA_Class->type = KlConstantType;

    KlExoClassDeclareBuiltInSlots(KlExoClassMeta);

    KlListInsert(KlA_error_correctors->c_val,
		 KlNumberMake(0),
		 KlIntern("default-exo-system-methods-loader"));
    KlListInsert(KlA_error_correctors->c_val,
		 KlNumberMake(0),
		 KlIntern("default-metaclass-instancier-loader"));

#ifdef SMOOL
    KlSMLExoInit();
#endif /* SMOOL */
}

