/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */
/**************\
*              *
*  KlO  Exo    *
*  DEFINITIONS *
*              *
\**************/
/* Exo = EXternal Objects for Klone
 * They implement a simple, clean extensible (but slower) OO layer in klone
 */

#ifndef INCLUDE_Kl_EXO_H
#define INCLUDE_Kl_EXO_H

/* type */

typedef struct _KlExoClass {
    KlKLONE_HEADER;
    struct _KlExoClass *exoclass;		/* for reflexivity */
    KlAtom name;
    struct _KlExoClass *father;		/* class tree */
    struct _KlExoClass *son;
    struct _KlExoClass *brother;
    KlHash methods;
    KlHash slots;
    KlO *init;				/* inits for slots */
    KlO *slot_properties;		/* array of slot properties */
    unsigned int exoHooks;
					/* <== class slots begin here */
}          *KlExoClass;

typedef struct _KlExo {
    KlKLONE_HEADER;
    KlExoClass exoclass;
    KlO slots[1];
}     *KlExo;

typedef struct _KlGenericFunction {	/* select on first element */
    KlKLONE_HEADER;
    KlAtom name;			/* name of the selector */
}                 *KlGenericFunction;

typedef struct _KlGenFuncContext {	/* method activation context */
    KlAtom selector;
    KlExoClass exoclass;		/* KlExoClass or KlType */
    KlList list;
    KlExo obj;
    KlO method;				/* the method itself */
} *KlGenFuncContext;

typedef struct _KlSlotHooks {		/* cell of slothooks */
    UInt offset;			/* actual offset of slot */
    KlO get;				/* KlExpr or KlSubr, but with arity 3 */
    KlO put;				/* KlExpr or KlSubr, but with arity 3 */
} *KlSlotHooks;

/* exported functions */

EXT KlExo KlExoMake();
EXT KlO KlExoPrint();
EXT KlO KlExoEqual();
EXT KlO KlExoGet();
EXT KlO KlExoPut();
EXT KlGenericFunction KlGenericFunctionMake();
EXT KlO KlGenericFunctionCatchallSet();
EXT KlO KlApplyMethodN();
#if __STDC__
EXT KlO KlApplyMethodV(KlO func, KlAtom selector, int size, ...);
#else
EXT KlO KlApplyMethodV _ANSI_ARGS_(VARARGS);
#endif
/* exported objects */

EXT int KlExoClassBuiltInSlots 
    INIT((sizeof(struct _KlExoClass) - sizeof(struct _KlExo)) / KLSO + 1);
				    
EXT KlExoClass KlExoClassRoot;
EXT KlExoClass KlExoClassMeta;

/* methods */

EXT KlType KlExoType;
EXT KlType KlExoClassType;
EXT KlType KlGenericFunctionType;

/* macros */

#define KlIsAnExo(obj) KlHasTrait(obj, KlTrait_exo)
#define KlMustBeExo(o, n) KlArgumentMustHaveTrait(o, n, KlTrait_exo)
#define KlIsAnExoClass(obj) ((obj)->type == KlExoClassType)
#define KlMustBeExoClass(o, n) KlArgumentMustBe(o, n, KlExoClassType)

#define KlIsAGenericFunction(obj) ((obj)->type == KlGenericFunctionType)
#define KlMustBeGenericFunction(o, n) \
        KlArgumentMustBe(o, n, KlGenericFunctionType)

#define KlCastExo(obj) ((KlExo)(obj))
#define KlCastExoClass(obj) ((KlExoClass)(obj))

#define KlExoHasHook(object, flag) \
    ((object)->exoclass->exoHooks & (flag))

#define KlExoMethodNamed(object, name) \
    KlHashGet((object)->exoclass->methods, name, NIL)

#define KlApplyMethod1(func, selector, arg1) \
    KlApplyMethodN(func, selector, 1, &arg1)
#define KlApplyMethod2(func, selector, arg1, arg2) \
     KlApplyMethodV(func, selector, 2, arg1, arg2)
#define KlApplyMethod3(func, selector, arg1, arg2, arg3) \
     KlApplyMethodV(func, selector, 3, arg1, arg2, arg3)
#define KlApplyMethod4(func, selector, arg1, arg2, arg3, arg4) \
     KlApplyMethodV(func, selector, 4, arg1, arg2, arg3, arg4)

/* actual offset of slot from the value of the class->slots hashtable */
#define KlExoActualOffset(o) \
    (KlPI_INT(o) ? ((UInt) (o)) : ((KlSlotHooks) (o))->offset)

/* tests if a class is a metaclass */
/* we test the presence of the reserved slot ::*inits */
#define KlIsAMetaclass(exoclass) \
    ((UInt) KlHashGet((exoclass)->slots, KlK_STAR_inits, KlPI_INVALID_VALUE) \
        != KlPI_INVALID_VALUE)

#endif					/* INCLUDE_Kl_EXO_H */
