/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */
/**************\
*              *
*  KlO  WString *
*  BODY        *
*              *
\**************/

#if defined(__NUTC__)
#include <limits.h>
#endif
#include "EXTERN.h"
#include <ctype.h>
#include "klone.h"
#include "kl_number.h"
#include "kl_atom.h"
#include "kl_list.h"
#include "klgeneric.h"
#include "kl_string.h"
#include "klwregexp.h"
#include "kl_stream.h"
#include "INTERN.h"
#ifdef WCHARS
#include "kl_wstring.h"
#include "EXTERN.h"

KlWString KlWStringMakeFromArray();
KlO KlWStringCoerce();
WREchar *Kl_wregsub();
static char KlWStringPrintFormat[8];

#ifndef MB_LEN_MAX
#define MB_LEN_MAX WREsize
#endif 


#ifdef KOALA
/* braindead emulation for test purposes. these are false, just for tests */
int towupper(int c) {return toupper(c);}
int towlower(int c) {return tolower(c);}
int mbslen(char *s) {return strlen(s);}
int wcscpy(WREchar *s1, WREchar *s2) {while (*s2) {*s1++ = *s2++;} *s1 = *s2;}
int wcsncpy(WREchar *s1, WREchar *s2, int n)
    {while (*s2 && n--) {*s1++ = *s2++;} *s1 = *s2;}
int wcsncat(WREchar *s1, WREchar *s2, int n) {
    return wcsncpy(s1 + wcslen(s1), s2, n);}
int wcslen(WREchar *s) {int n = 0; while (*s) n++; return n;}
int wcsncmp(WREchar *s1, WREchar *s2, int n) {
    while (*s2 && n--) {if (*s1 > *s2) return 1; else if (*s1 < *s2) return -1;
    s1++; s2++;}}
WREchar * wcschr(WREchar *s, int n) {
    while (*s) {if (*s == n) return s; s++;} return 0;}
#endif /* KOALA */


/*****************************************************************************\
* 				 WIDE STRINGS                                 *
\*****************************************************************************/
/* This is here just to enable me to test the regexp code on wchars
 * for me, I just want to have the conversion with lists of 16-bit integers and 
 * normal strings
 */

KlWString
KlWStringMakeFromArray(size, array, slen)
    int size;
    WREchar *array; 
    int slen;
{
    KlWString object;

    object = (KlWString) KlOMake(KlWStringType);
    object->string = (WREchar *) Malloc(size * WREsize + WREsize);
    if (array != (WREchar *) NULL) wcsncpy(object->string, array, size);
    /* bcopy(array, object->string, size * WREsize); */
    object->string[size] = 0;
    object->size = size;
    object->length = slen;
    return object;
}

KlO KlWStringCopy(KlWString obj)
{
  return (KlO) KlWStringMakeFromArray(KlWStringLength(obj), obj->string,
				      obj->length);
}

KlO KlWStringLengthKl(KlWString obj)
{
  return (KlO) KlNumberMake(KlWStringLength(obj));
}

KlO KlWStringGet(KlWString obj, KlNumber key, KlO def)
{
  if (KlIsANumber(key)) {		/* array */
    int i, l;
    i = key->number;
    l = KlWStringLength(obj);
    if (i < l) {
      if (i < 0) {
	if (l) {
	  i = l - 1;
	} else {
	  return KlExecuteGetDefault(obj, key, def);
	}
      }
      return (KlO) KlNumberMake(obj->string[i]);
    } else {
      return KlExecuteGetDefault(obj, key, def);
    }
  } else {
    KlMustBeNumber(key, 1);		/* expansion? */
    return KlExecuteGetDefault(obj, key, def);
  }
}

KlO KlWStringPut(KlWString obj, KlNumber key, KlNumber val)
{
  if (KlIsANumber(key)) {		/* array */
    if (KlIsANumber(val)) {
      KlWString tmp;
      tmp = (KlWString) KlList2WString
	(KlListPut((KlList) KlWString2List(obj), key, val));
      obj->size = tmp->size;
      obj->length = tmp->length;
      obj->string = (WREchar *) Realloc(obj->string,
					obj->size * WREsize + WREsize);
      wcscpy (obj->string, tmp->string);
    } else {
      KlWString wstring = (KlWString)KlWStringCoerce(KlWStringType, (KlO) val);
      int inserted_len = KlWStringLength(wstring);
      int i = key->number;
      int l = KlWStringLength(obj);
      
      if (i < 0) {			/* <0 means append to end */
	i = l;
      }
      if ((i + inserted_len) > l) {
	obj->string = (WREchar *) Realloc(obj->string,
					  (i + inserted_len) * WREsize + WREsize);
	obj->size = i + inserted_len;
	obj->length = i * MB_LEN_MAX + wstring->length;
	if (i > l) {		/* fill with blanks */
	  int j = l;
	  KlWString blank = (KlWString)KlWStringCoerce(KlWStringType,
						       KlStringMake(" "));
	  while (j++ < i ) 
	    wcsncat (obj->string, blank->string, blank->size);
	}
      }
      /* wcsncpy (obj->string + i, wstring->string, wstring->size); */
      bcopy (wstring->string, obj->string + i, wstring->size * WREsize);
      obj->string[obj->size] = 0;
    }
  } else {
    KlMustBeNumber(key, 1);
  }
  return (KlO) obj;
}

KlO KlWStringInsert(KlWString obj, KlNumber key, KlNumber val)
{
  int i, l;
  WREchar *p, *q, *start;
  int inserted_len;
  KlWString wstring;

  KlMustBeNumber(key, 1);
  i = key->number;
  l = KlWStringLength(obj);
  if (i < 0 || i >= l) {		/* after end = Put */
    return KlWStringPut(obj, key, val);
  }
  if (KlIsANumber(val)) {
    wstring = (KlWString) KlList2WString
      (KlListInsert((KlList) KlWString2List(obj), key, val));
    obj->size = wstring->size;
    obj->length = wstring->length;
    obj->string = (WREchar *) Realloc(obj->string,
				      obj->size * WREsize + WREsize);
    wcsncpy (obj->string, wstring->string, obj->size);
    obj->string[obj->size] = 0;
  } else {
    wstring = (KlWString)KlWStringCoerce(KlWStringType, (KlO) val);
    inserted_len = KlWStringLength(wstring);
    obj->string = (WREchar *) Realloc(obj->string,
				      (l + inserted_len) * WREsize + WREsize);
    obj->size = l + inserted_len;
    obj->length += wstring->length;
    start = obj->string + i;
    p = obj->string + l;
    q = p + inserted_len;
    
    while (p >= start) {
      *q-- = *p--;
    }
    bcopy(wstring->string, obj->string + i, wstring->size * WREsize);
    obj->string[obj->size] = 0;
  }    
  return (KlO) obj;
}

KlO KlWStringDelete(KlWString obj, KlNumber key)
{
  int i, l;
  WREchar *p, *end;
  
  if (!obj->string) {
    return (KlO) obj;
  }
  if (KlIsANumber(key)) {		/* array */
    i = key->number;
    l = KlWStringLength(obj);
    if (i < 0) {
      i = (l ? l - 1 : 0);
    } 
    p = obj->string + i;
    end = obj->string + l;
    while (p < end) {
      *p = *(p+1);
      p++;
    }
    if (l && i < l)
      obj->size = obj->size - 1;
  } else {
    KlMustBeNumber(key, 1);
  }
  return (KlO) obj;
}

KlO KlWStringEqual(KlWString object1, KlWString object2)
{
    if (!KlIsAWString(object1) && KlIsAString(object1))
	KlMustBeWStringOrConvert (object1, 0);
    else
	return NIL;

    if (!KlIsAWString(object2) && KlIsAString(object2))
	KlMustBeWStringOrConvert (object2, 1);
    else
	return NIL;
  
    if (object1->size == object2->size)
	if (wcsncmp (object1->string, object2->string, object1->size) == 0)
	    return (KlO) object1;
    return NIL;
}

int KlWStringCompare(KlWString object1,KlWString object2)
{
  KlMustBeWStringOrConvert ((KlO) object1, 0);
  KlMustBeWStringOrConvert ((KlO) object2, 1);
  
  return wcsncmp (object1->string, object2->string, object1->size);
}

KlO KlWStringToUpper(KlWString string)
{
  KlWString result;
  int i;

  KlMustBeWString(string, 0);
  result = KlWStringMakeFromArray(KlWStringLength(string), string->string,
				  string->length);
  for (i = 0; i <= KlWStringLength(string); i++) {
    result->string[i] = towupper(string->string[i]);
  }
  return (KlO) result;
}

KlO KlWStringToLower(KlWString string)
{
  KlWString result;
  int i;

  KlMustBeWString(string, 0);
  result = KlWStringMakeFromArray(KlWStringLength(string), string->string,
				  string->length);
  for (i = 0; i <= KlWStringLength(string); i++) {
    result->string[i] = towlower(string->string[i]);
  }
  return (KlO) result;
}

/* KlWStringAdd
 * concatenates strings
 */
static char *strings_temp_buffer;
static int strings_temp_buffer_size = 0;

KlO
KlWStringAdd(argc, argv)
    int argc;
    KlWString argv[];
{
    int required_length = KlWStringLength(argv[0]), i;
    int required_slength = argv[0]->length;
    char *p;

    /* check args & needed length */
    for (i = 1; i < argc; i++) {
	KlMustBeWString(argv[i], i);
	required_length += KlWStringLength(argv[i]);
	required_slength += argv[i]->length;
    }

    /* verify we have room in temp space */
    if (required_length >= strings_temp_buffer_size) {
	if (!strings_temp_buffer) {
	    strings_temp_buffer = (char *)
		Malloc(strings_temp_buffer_size = 256 * WREsize);
	} else {
	    strings_temp_buffer_size = required_length + 1;
	    strings_temp_buffer = (char *)
		Realloc(strings_temp_buffer, strings_temp_buffer_size * WREsize);
	}
    }
    /* do the concatenation */
    strings_temp_buffer[0] = '\0';
    for (i = 0, p = strings_temp_buffer; i < argc; i++) {
	int l;
	bcopy(argv[i]->string, p, l = KlWStringLength(argv[i]) * WREsize);
	p += l;
    }
    return (KlO) KlWStringMakeFromArray(required_length, strings_temp_buffer,
					required_slength);
}

KlO KlWStringPosition (KlWString obj, KlWString ref)
{
  WREchar *p, *string;
  int length;
  
  KlMustBeWString(ref, 1);
  if (KlIsANumber(obj)) {
    if (p = (WREchar *) wcschr(((KlWString)ref)->string, ((KlNumber)obj)->number))
      return (KlO) KlNumberMake(p - ((KlWString) ref)->string);
    else
      return NIL;
  } else {
    KlMustBeWStringOrConvert(obj, 0);
    string = ((KlWString) ref)->string;
    length = KlWStringLength(obj);
    for (p = string; *p; p++)
      if (!wcsncmp(obj->string, p, length))
	return (KlO) KlNumberMake(p - string);
    return NIL;
  }
}

KlO KlWStringDolist(string, var, argc, argv)
     KlWString string;
     KlO var;
     int argc;
     KlO *argv;
{
  KlGCMark();
  if (KlIsASymbol(string)) {	/* no null chars embedded,
				   and not modifiable */
    WREchar *p = string->string;
    while (*p) {
      KlSend_setq(var, KlNumberMake(*p));
      KlProgn(argc, argv);
      KlGC();
      p++;
    }		
  } else {
    int i;
    for (i = 0; i < string->size; i++) {
      KlSend_setq(var, KlNumberMake(string->string[i]));
      KlProgn(argc, argv);
      KlGC();
    }
  }
}

/* subseq
 * extracts a sub-sequence of a Wide String
 * (subseq wstring start [end])
 * start is inclusive (begins at 0), end is exclusive, defaults to end.
 */

KlO KlWStringSubseq (KlWString wstring, unsigned int from, unsigned int to, 
		     int argc)
{
  WREchar *p;
  WREchar *q;
  int l;
  int i, j;
  unsigned int size;
  KlWString s;

  KlMustBeWString(wstring, 0);

  p = wstring->string;
  l  = KlWStringLength(wstring);
  if (argc == 2) to = l;
  if ((size = to - from) < 0) {
    s = KlWStringMakeFromArray(0, 0, 0);
    return (KlO) s;
  }

  s = KlWStringMakeFromArray(size, p, size * MB_LEN_MAX);
  q = s->string;
  
  if (from >= 0 && to <= l) {	/* normal case */
    bcopy(p + from, q, size * WREsize);
  } else {
    for (i = from, j = 0; i < to; i++, j++) {
      q[j] = (i < l && i >= 0) ? p[i] : ' ';
    }
  }
  q[size] = 0;
  return (KlO) s;
}

/* nconc
 * physically concatenates rest to first argument
 * this works for wide strings
 */

KlO KlWStringNconc(argc, argv)
     int argc;
     KlWString *argv;
{
  KlWString tmp = (KlWString) KlWStringAdd(argc, argv);
  KlWString obj = argv[0];
  obj->size = tmp->size;
  obj->length = tmp->size;  
  obj->string = (WREchar *) Realloc(obj->string,
				    obj->size * WREsize + WREsize);
  wcsncpy (obj->string, tmp->string, obj->size);
  obj->string[obj->size] = 0;
  return (KlO) obj;
}

KlO KlWStringNth(obj, i, value)
    KlWString obj;
    int i;
    KlNumber value;
{
    if (i >= KlWStringLength(obj))
	return NIL;
    if (value) {
	obj->string[i] = value->number;
	return (KlO) obj;
    } else {
	return (KlO) KlNumberMake(obj->string[i]);
    }
}

/********************************************************** conversions/casts */
/* x2y make a new object (copy) */

KlO
KlWString2List(s)
KlWString s;
{
    KlList result;
    KlO *pl;
    WREchar *ps, *pe;

    KlMustBeWString(s, 0);
    result = KlListNMake(s->size);
    
    for (pl = result->list, ps = s->string, pe = s->string + s->size; ps < pe;)
      KlIncRef(*pl++ = (KlO) KlNumberMake(*ps++));

    return (KlO) result;
}

KlO
KlList2WString(l)
KlList l;
{
    KlWString s;
    KlNumber *pl;
    WREchar *ps, *pe;

    KlMustBeList(l, 0);
    s = (KlWString) KlOMake(KlWStringType);
    s->size = l->size;
    s->length = l->size * MB_LEN_MAX;
    ps = s->string = (WREchar *) Malloc(s->size * WREsize + WREsize);
    for (pe = ps + l->size, pl = ((KlNumber *) (l->list)); ps < pe;)
      *ps++ = (*pl++)->number;
    *ps = 0;
    return (KlO) s;
}

KlO
KlWString2String(s)
KlWString s;
{
    KlString result;
    WREchar *ps, *pe;
    char *p;
    int len;

    KlMustBeWString(s, 0);
    result = KlStringNMake(s->length);
     
    len = (int) wcstombs(result->string, s->string, s->length);
    if (len < s->length) {
	s->length = len;
	KlModStringSetLength(result, len);
    }
    result->string[s->length] = '\0';
    return (KlO) result;
}

KlO
KlString2WString(s)
KlString s;
{
  KlWString result;

  KlMustBeString(s, 0);
  result = (KlWString) KlOMake(KlWStringType);
  result->string = (WREchar *) Malloc(mbslen(s->string) * WREsize + WREsize);
#ifdef _AIX
  result->string = (WREchar *) Malloc(mbslen(s->string) * WREsize + WREsize);
#else
  result->string = (WREchar *) Malloc(s->size * WREsize + WREsize);
#endif
  result->size = mbstowcs(result->string, s->string, s->size);
  if (result->size == (unsigned int) -1)
    KlError2(KlE_NO_COERCION, s, KlWStringType);
  result->length = s->size;
  result->string[result->size] = 0;
  return (KlO) result;
}

/* Casts modifies in place the object */

KlString 
KlStringCast(s)
    KlString s;
{
    KlMustBeString(s, 0);
    if (KlIsAWString(s))		/* dont touch regexps, atoms... */
	s->type = KlStringType;
    return s;
}

KlWString 
KlWStringCast(s)
    KlWString s;
{
    KlMustBeString(s, 0);
    if (s->type == KlStringType)	/* dont touch regexps, atoms... */
	s->type = KlWStringType;
    return s;
}

/****************************************************************** printing */

KlO
KlWStringPrint(s, stream)
KlWString s;
KlStream stream;
{
    WREchar *p, *end;

    if (KlPrintReadably) {		/* {WString #( chars... ) [string]} */
	KlSPuts("{WString #(", stream);
	for (p = s->string, end = p + s->size; p < end; p++) {
	    if (p != s->string)
		KlSPutc(' ', stream);
	    KlSPrintf(stream, "%d", *p);
	}
#if defined(sun)
	KlSPrintf(stream, ", length: %d,", s->size);
	KlSPrintf(stream, " [%ws]", s->string);
#else
	KlSPrintf(stream, ", length: %d,", s->size);
	KlSPrintf(stream, " [%S]", s->string);
#endif
	KlSPuts(")}", stream);
    } else {				/* "<charsinhex>..." */
	/*
	  for (p = s->string, end = p + s->size; p < end; p++) {
	  if (*p < 0xff)
	  KlSPutc(*p, stream);
	  else
	  KlSPrintf(stream, KlWStringPrintFormat, *p);
	  }
	  */
#if defined(sun)
	KlSPrintf(stream, "%ws", s->string);
#else
	KlSPrintf(stream, "%S", s->string);
#endif
    }
}

/* non-optimized version */
KlO
KlWStringCoerce(totype, obj)
    KlType totype;
    KlO obj;
{
    if (KlIsAWString(obj)) {
	return obj;
    } else if (KlIsAList(obj)) {
	return (KlO) KlList2WString(obj);
    } else if (KlIsAString(obj)) {
	return (KlO) KlString2WString(obj);
    } else {				/* ? --> string --> list --> wstring */
	KlO s = KlStringCoerce(KlStringType, obj);
	if (s)
	    return (KlO) KlList2WString(KlListCoerce(KlListType, s));
    }
    return 0;
}

/*****************************************************************************\
* 	  interface to the regexp package for wide strings: WREregexps          *
\*****************************************************************************/
/* uses the modified Henry Spencer's REGEXP package: klwregexp.c
 */

/********************************************* private data type definitions */

typedef struct _KlWRegexp {
    KlKLONE_HEADER;
    WREchar *string;			/* printable form, wstring compatible */
    unsigned int size;
    WREregexp *prog;
    KlWString matched_string;		/* last regexec-ed string */
}        *KlWRegexp;

KlWRegexp KlWRegcomp();
/* defined now in kl_wstring.h
KlO KlWRegexec();
KlWString KlWRegsub();

KlType KlWRegexpType;
*/

/* #define KlIsAWRegexp(obj) ((obj)->type == KlWRegexpType) */
#define KlMustBeWRegexp(o, n) KlArgumentMustBe(o, n, KlWRegexpType)
/********************************************************************** body */
/* type management 
 */

KlWRegexp
KlWRegexpMake(expr, size)
    WREchar *expr;
    int size;				/* in bytes, not in WREchars */
{
    KlWRegexp object;
    WREregexp *prog = WREregcomp(expr);

    object = (KlWRegexp) KlOMake(KlWRegexpType);
    object->prog = prog;
    object->size = size / WREsize;
    object->string = (WREchar *) Malloc(size + WREsize);
    object->matched_string = 0;
    bcopy(expr, object->string, size + WREsize);
    return (KlWRegexp) object;
}

KlO
KlWRegexpFree(obj)
    KlWRegexp obj;
{
    Free(obj->string);
    KlDecRef(obj->matched_string);
    free(obj->prog->startp);		/* lowercase free intentional */
    if (obj->prog->o_parnums) {
	free(obj->prog->o_parnums);
	free(obj->prog->c_parnums);
    }
    free(obj->prog);			/* lowercase free intentional */
    Free(obj);
    return (KlO) obj;
}

KlO
KlWRegexpGet(obj, key, def)
    KlWRegexp obj;
    KlNumber key;
    KlO def;
{
   int offset;

    if (KlIsANumber(key)) {
	offset = key->number;
	if (offset >= 0 && offset < obj->prog->nsubexp && obj->matched_string
	    && (obj->prog->startp[offset] || obj->prog->endp[offset])) {
	    KlList result = KlListNMake(2);
	    KlListStore(result, 0, KlNumberMake(obj->prog->startp[offset]
						- obj->matched_string->string));
	    KlListStore(result, 1, KlNumberMake(obj->prog->endp[offset]
						- obj->matched_string->string));
	    return (KlO) result;
	} else if (offset == -1) {
	    return obj->matched_string ? (KlO) obj->matched_string : NIL;
	} else {
	    return KlExecuteGetDefault(obj, key, def);
	}
    } else if (key == (KlNumber) KlA_depth) {
	return (KlO) KlNumberMake(obj->prog->nsubexp - 1);
    } else {
	KlMustBeNumber(key, 1);
    }
}

/* wregcomp
 */

KlWRegexp
KlWRegcomp(expr)
    KlWString expr;
{
    KlMustBeWStringOrConvert(expr, 0);
    return KlWRegexpMake(expr->string, expr->size * WREsize);
}

/* wregexec
 * wregexec wregexp source [source-offset]
 */

KlO
KlWRegexec(argc, argv)
    int argc;
    KlWString *argv;
{
    
    int offset;
    KlWString re, source;

    switch (argc) {
    case 2:
	offset = 0;
	break;
    case 3:
	KlMustBeNumber(argv[2], 2);
	offset = ((KlNumber)argv[2])->number;
	if (offset < 0) offset = 0;
	break;
    default:
	return KlBadNumberOfArguments(argc);
    }

    re = argv[0];
    source = argv[1];
    KlMustBeWRegexp(re, 0);
    KlMustBeWStringOrConvert(source, 1);
    KlDecRef(((KlWRegexp)re)->matched_string);

    if (offset > KlWStringLength(source))
	offset = KlWStringLength(source);

    return Klwregexec(((KlWRegexp)re)->prog, source->string + offset)
	? (KlIncRef(((KlWRegexp)re)->matched_string = source), TRU)
	    : (((KlWRegexp)re)->matched_string = 0, NIL);
}

/* KlWRegsub
 * KlWRegsub wregexp source [source-offset]
 */

KlWString
KlWRegsub(argc, argv)
    int argc;
    KlWString *argv;
{
    int offset;
    WREchar *dest;
    KlWString result;
    KlWRegexp re;

    switch (argc) {
    case 2:
	offset = 0;
	break;
    case 3:
	KlMustBeNumber(argv[2], 2);
	offset = ((KlNumber)argv[2])->number;
	break;
    default:
	return (KlWString) KlBadNumberOfArguments(argc);
    }
    re = (KlWRegexp) argv[0];
    KlMustBeWRegexp(re, 0);
    if (KlIsANumber(argv[1])) {
	int no = ((KlNumber) argv[1])->number;
	KlWRegexp obj = (KlWRegexp) re;
	if (no >= 0 && no < obj->prog->nsubexp && obj->matched_string
	    && (obj->prog->startp[no] || obj->prog->endp[no])) {
	    return KlWStringMakeFromArray
		(obj->prog->endp[no] - obj->prog->startp[no],
		 obj->prog->startp[no],
		 (obj->prog->endp[no] - obj->prog->startp[no]) * MB_LEN_MAX);
	} else {
	    return (KlWString) KlWStringMakeFromArray(0, 0, 0);
	}
    } else {
	int size;
	KlWString source = argv[1];
	KlMustBeWStringOrConvert(source, 1);
	if (offset > KlWStringLength(source))
	    offset = KlWStringLength(source);
	dest = Kl_wregsub(((KlWRegexp)re)->prog, source->string + offset,
			  &size);
	result = KlWStringMakeFromArray(size, dest, size * MB_LEN_MAX);
	Free(dest);
	return result;
    }
}

/*
 * Kl_wregsub - perform substitutions after a regexp match
 */

#define	WREUCHARAT(p)	((int)*(WREchar *)(p))

WREchar *
Kl_wregsub(prog, source, sizep)
    WREregexp *prog;
    WREchar *source;
    int *sizep;				/* will be set to size of dest */
{
    register WREchar *src;
    register WREchar *dst;
    register WREchar c;
    register int no;
    register int len;
    WREchar *dest = (WREchar *) Malloc(12 * WREsize);
    WREchar *lastdest = dest+11;		/* last allowable - 1 */

    if (prog == NULL || source == NULL) {
	Kl_regerror("NULL parm to regsub");
	/*NOTREACHED*/
    }
    if (WREUCHARAT(prog->program) != WREREGMAGIC) {
	Kl_regerror("damaged regexp fed to regsub");
	/*NOTREACHED*/
    }

    src = source;
    dst = dest;
    while ((c = *src++) != 0) {
	if (c == '&')
	    no = 0;
	else if (c == '\\' && '0' <= *src && *src <= '9')
	    no = *src++ - '0';
	else
	    no = -1;

	if (no < 0) {			/* Ordinary character. */
	    if (dst + 1 >= lastdest) {
		WREchar *olddest = dest;
		int newlength = ((lastdest - dest) + 5) * 2 - 4;
		dest = (WREchar *) Realloc(dest, newlength * WREsize);
		lastdest = dest + (newlength - 1);
		dst = dest + (dst - olddest);
	    }
	    *dst++ = c;
	} else if (prog->startp[no] != NULL && prog->endp[no] != NULL) {
	    len = prog->endp[no] - prog->startp[no];
	    if (dst + len >= lastdest) {
		WREchar *olddest = dest;
		int newlength = ((lastdest - dest) + 5) * 2 - 4;

		while (newlength <= (dst - dest) + len)
		    newlength = ((newlength) + 5) * 2 - 4;
		dest = (WREchar *) Realloc(dest, newlength * WREsize);
		lastdest = dest + (newlength - 1);
		dst = dest + (dst - olddest);
	    }
	    
	    (void) WREstrncpy(dst, prog->startp[no], len);
	    dst += len;
	    if (dst > dest && *(dst-1) == '\0') {	/* strncpy hit NUL. */
		Kl_regerror("damaged match string");
		/*NOTREACHED*/
	    }
	}
    }
    *sizep = (dst - dest);
    *dst = '\0';
    return dest;
}

#ifdef DEBUG
/* some debugging code from Klone */
extern int wregnarrate;
extern void Klwregdump();

KlO
KlWRegdump(obj)
    KlWRegexp obj;
{
    KlMustBeWRegexp(obj, 0);
    Klwregdump(obj->prog);
    return (KlO) obj;
}

#endif /* DEBUG */


/* match
 * implemented with the preceding functions
 */

KlWString
KlWStringMatchAux(prog, offset)
    WREregexp *prog;
    int offset;
{
    KlWString str;
    int len;

    if (offset >= 0 && offset < prog->nsubexp
	&& (prog->startp[offset] || prog->endp[offset])) {
	len = prog->endp[offset] - prog->startp[offset];
	str = KlWStringMakeFromArray(len, prog->startp[offset], 
				     len * MB_LEN_MAX);
    } else {
	len = 0;
	str = KlWStringMakeFromArray(0, 0, 0);
    }
    str->string[len] = '\0';
    return str;
}

KlWString
KlWStringMatch(argc, argv)
    int argc;
    KlWString argv[];
{
    int i;
    KlList kl_list;
    WREregexp *prog;
    int free_prog = 0;
    KlWString source;

    if (argc < 2)
	return (KlWString) KlBadNumberOfArguments(argc);
    if (KlIsAWRegexp(argv[0])) {
	prog = ((KlWRegexp) argv[0])->prog;
    } else {
	KlWString re = argv[0];
	KlMustBeWStringOrConvert(re, 0);
	/* compile reg. expr */
	prog = WREregcomp(re->string);
	free_prog = 1;
    }
    source = argv[1];
    KlMustBeWStringOrConvert(source, 1);

    if (Klwregexec(prog, source->string)) {
	switch (argc) {
	case 2:
	    if (free_prog)
		free(prog);
	    return (KlWString) source;

	case 3: {
	    KlWString klstr;
	    KlMustBeNumber(argv[2], 2);
	    klstr = KlWStringMatchAux(prog, ((KlNumber) argv[2])->number);
	    if (free_prog)
		free(prog);
	    return klstr;
	}

	default:
	    KlListNMakeZ(kl_list, argc - 2);
	    for (i = 2; i < argc; i++) {
		KlMustBeNumber(argv[i], i);
		KlListStore(kl_list, i - 2,
			    KlWStringMatchAux(prog,
					     ((KlNumber) argv[i])->number));
	    }
	}
    } else {
	kl_list = (KlList) NIL;
    }
    if (free_prog)
	free(prog);
    return (KlWString) kl_list;
}


/*****************************************************************************\
* 				  TYPE INIT                                   *
\*****************************************************************************/

KlWStringInit()
{
    KlDeclareSubType(&KlWStringType, "WString", KlStringType, 
		     sizeof(struct _KlWString));
    KlDeclareMethod1(KlWStringType, KlSelPrint, KlWStringPrint);
    KlDeclareMethod1(KlWStringType, KlSelGet, KlWStringGet);
    KlDeclareMethod1(KlWStringType, KlSelPut, KlWStringPut);
    KlDeclareMethod1(KlWStringType, KlSelInsert, KlWStringInsert);
    KlDeclareMethod1(KlWStringType, KlSelDelete, KlWStringDelete);
    KlDeclareMethod1(KlWStringType, KlSelCopy, KlWStringCopy);
    KlDeclareMethod1(KlWStringType, KlSelLength, KlWStringLengthKl);
    KlDeclareMethod1(KlWStringType, KlSelEqual, KlWStringEqual);
    KlDeclareMethod1(KlWStringType, KlSelCompare, (KlMethod) KlWStringCompare);
    KlDeclareMethod1(KlWStringType, KlSelAdd, KlWStringAdd);
    KlDeclareMethod1(KlWStringType, KlSelDolist, (KlMethod) KlWStringDolist);
    KlDeclareMethod1(KlWStringType, KlSelNth, KlWStringNth);

    KlDeclareSubType(&KlWRegexpType, "WRegexp", KlWStringType,
		     sizeof(struct _KlWRegexp));

    KlDeclareMethod1(KlWRegexpType, KlSelFree, KlWRegexpFree);
    KlDeclareMethod1(KlWRegexpType, KlSelGet, KlWRegexpGet);
    KlDeclareMethodUndefined(KlWRegexpType, KlSelPut);
    KlDeclareMethodUndefined(KlWRegexpType, KlSelDelete);
    KlDeclareMethodUndefined(KlWRegexpType, KlSelInsert);

    KlDeclareSubr(KlWRegcomp, "wregcomp", 1);
    KlDeclareSubr(KlWRegexec, "wregexec", NARY);
    KlDeclareSubr(KlWRegsub, "wregsub", NARY);
    KlDeclareSubr(KlWStringMatch, "wmatch", NARY);
  
    KlDeclareSubr(KlStringCast, "string!", 1);
    KlDeclareSubr(KlWStringCast, "wstring!", 1);

    KlDeclareCoerce(KlListType, KlWStringType, KlWStringCoerce);
    KlDeclareCoerce(KlStringType, KlWStringType, KlWStringCoerce);
    KlDeclareCoerce(KlWStringType, KlListType, KlListCoerce);
    KlDeclareCoerce(KlWStringType, KlStringType, KlStringCoerce);

#ifdef DEBUG
    KlDeclareSubr(KlWString2List, "ws2l", 1);
    KlDeclareSubr(KlList2WString, "l2ws", 1);
    KlDeclareSubr(KlWRegdump, "wregdump", 1);
    KlActivePointerToIntMake("wregnarrate", &wregnarrate);
#endif /* DEBUG */

    sprintf(KlWStringPrintFormat, "<%%0%dx>", WREsize);
}

#endif /* WCHARS */
