/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/*****************************************************************************\
*                                                                             *
* KL FORMAT                                                                   *
*                                                                             *
* Internal format primitives to avoid problems using NLS                      *
*                                                                             *
\*****************************************************************************/

#ifdef ISM
#include <errno.h>
#include <float.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


double
KlAtof (nptr)
     char *nptr;
{
  char *s;
  char decimal_char = '.';
  double number = 0.0;
  int dot_found = 0;
  int digit_found = 0;
  int sign;
  long int exponent = 0;

  if (nptr == NULL)
    goto empty;

  s = nptr;
  while (isspace(*s))
    ++s;

  if (*s == '\0')
    goto empty;

  sign = *s == '-' ? -1 : 1;
  if (*s == '-' || *s == '+')
    ++s;

  for (;; ++s)
    {
      if (isdigit (*s))
	{
	  digit_found = 1;
	  if (number > DBL_MAX * 0.1)
	    ++exponent;
	  else
	    number = (number * 10.0) + (*s - '0');
	  if (dot_found)
	    --exponent;
	}
      else if (!dot_found && (wchar_t) *s == decimal_char)
	dot_found = 1;
      else
	break;
    }

  if (!digit_found)
    goto noconv;

  if (tolower(*s) == 'e')
    {
      int save = errno;
      char *end;
      long int exp;

      errno = 0;
      ++s;
      exp = strtol(s, &end, 10);
      if (errno) {
	if (errno == ERANGE) {
	  if (exp < 0)
	    goto underflow;
	  else
	    goto overflow;
 	} else
 	  goto noconv;
      } else 
 	if (end == s)
	  end = (char *) s - 1;
      errno = save;
      s = end;
      exponent += exp;
    }
  
  
  if (number == 0.0)
    return 0.0;

  if (exponent < 0)
    {
      if (number < DBL_MIN * pow(10.0, (double) -exponent))
	goto underflow;
    }
  else if (exponent > 0)
    {
      if (number > DBL_MAX * pow(10.0, (double) -exponent))
	goto overflow;
    }

  number *= pow(10.0, (double) exponent);

  return number * sign;

 overflow:
  errno = ERANGE;
  return HUGE_VAL * sign;

 underflow:
  errno = ERANGE;
  return 0.0;

 noconv:
  errno = EINVAL;
  return 0.0;
 
 empty:
  /* No error set */
  return 0.0;
}

#endif	/* ISM */
