/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */
/*****************************************************************************\
*                                                                             *
* KLSESI                                                                      *
* BODY                                                                        *
*                                                                             *
\*****************************************************************************/
/* Klone Safe Extension System (SES) Interface
 */

#ifdef KLSES

#ifdef SYSV
#include <sys/types.h>
#endif					/* SYSV */
#include <sys/file.h>
#include <fcntl.h>
#include <sys/time.h>
#ifdef NEED_SELECT_H
#include <sys/select.h>
#endif
#include <errno.h>

#include "EXTERN.h"
#include "klone.h"
#include "kl_number.h"
#include "kl_string.h"
#include "kl_atom.h"
#include "kl_list.h"
#include "kl_stream.h"
#include "kl_func.h"
#include "INTERN.h"
#include "ses.h"

#include "klsesi.h"

#ifdef SHM
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#endif /* SHM */

#ifdef SES_PATH
static char *KlSesPath = SES_PATH;
#else
static char *KlSesPath = ".:/usr/local/lib/klone/SES";
#endif

#ifndef SES_BUFSIZE
#define SES_BUFSIZE 16 * 1024		/* 16k buffer default value for socket*/
#endif

#define ECHECK(err) (errno == err)
#define ESET(val) errno = val

#ifdef SYSV
#define KlNDELAY O_NDELAY
#else
#define KlNDELAY FNDELAY
#endif

KlAtom KlA_Ses_loadpathname;
KlAtom KlA_SesCurrentModule;
KlString KlS_PATH;
KlAtom KlA_SesRetValue;
KlO KlSesUserErrorHandler;

/* internal functions forward declarations */

SES_PS_Header KlSesModuleNewRequest _ANSI_ARGS_((
    KlSesModule module, unsigned int type, 
    void *func_id, KlSesReqId reqtype,
    unsigned int size, int arity));
void KlSesReturnRequest _ANSI_ARGS_((
    KlSesModule module, unsigned int type, unsigned int func_id, 
    unsigned int size, int arity, char *data));
void KlSesDecodeRequest();
void KlSesReadAndHandle();
void KlSesListOfBuiltinTypesInit();
KlO KlSesFlush _ANSI_ARGS_((KlSesModule module));
KlO KlSesWait _ANSI_ARGS_((KlSesReqId reqid));
KlO KlSesSend _ANSI_ARGS_((int argc, KlO *argv));
KlO KlSesSync _ANSI_ARGS_((KlSesModule module));
KlO KlSesErrorHandlerSet _ANSI_ARGS_((KlSesModule module, KlO handler));
KlO KlSesReturn _ANSI_ARGS_((KlO value));
KlO KlSesReturnError _ANSI_ARGS_((int argc, KlO *argv));

#define TODO_NOT_YET_IMPLEMENTED(name) \
    fprintf(stderr, "NOT_YET_IMPLEMENTED: %s\n", name);return NIL;

/* codecs, common to both sides */
#define SESCODECPAR ,module
#define SESCODECPARD KlSesModule module;
#define SES__CODEC_ERROR(e, t) KlSesCodecError(module, e, t)
#include "sescodecs.c"			

#define KlSesRequestPending() \
    (SES_PS_HeaderEnd(KlSes_soc_inptr) <= KlSes_soc_inend \
     && (KlSes_soc_inptr + ((SES_PS_Header) KlSes_soc_inptr)->length) \
         <= KlSes_soc_inend)

/*****************************************************************************\
* 				 KlSesModule                                  *
\*****************************************************************************/

KlSesModule
KlSesModuleMake(name, fd)
    KlAtom name;
    FILE *fd;
{
    int i;
    KlSesModule module = (KlSesModule)  KlOMake(KlSesModuleType);

    KlIncRef(module->name = name);
    module->fd = fd;
    KlIncRef(module->func_symbols = (KlListOfAtom) KlListNMake(0));
    module->ready = 0;
    module->dead = 0;
    module->ready_callback = 0;

    /* add it to the list(array) of all modules */
    for (i = 0; i < KlSesModules->size; i++)
	if (!KlSesModules->list[i])	{	/* found a hole in array */
	    KlIncRef(KlSesModules->list[i] = (KlO) module);
	    module->id = i;			/* id is index in list */
	    goto done;
	}
    KlListAppend(KlSesModules, module);	/* no hole, append */
    module->id = KlSesModules->size - 1;
done:
    {   /* allocate output buffer for sockets */
	int buf_size = SES_BUFSIZE - sizeof(void *); /* room for malloc info */

	module->outbuf = (uchar *) Malloc(buf_size);
	module->outptr = module->outbuf;
	module->outlimit = module->outbuf + buf_size;
    }
    KlSesListOfBuiltinTypesInit(module);

    return module;
}

KlO
KlSesModuleFree(module)
    KlSesModule module;
{
    int i;
    KlDecRef(module->name);
    if (module->fd) {
	close(module->fd);
	module->fd = 0;
    }
    KlDecRef(module->ready_callback);
    KlSesModules->list[module->id] = 0;	/* remove from list */
    if (module->id == (KlSesModules->size - 1)) /* was last, shorten list */
	KlSesModules->size--;
    for (i = 0; i < module->func_symbols->size; i++) { /* removes func_symbols */
	if (KlIsAnAtom(module->func_symbols->list[i]))
	    KlAtomUnbind(module->func_symbols->list[i]);
    }
    KlDecRef(module->func_symbols);
    for (i = 0; i < module->ids->size; i++) { /* removes ids */
	if (KlIsAnAtom(module->ids->list[i]))
	    KlAtomUnbind(module->ids->list[i]);
    }
    KlDecRef(module->ids);
    Free(module->outbuf);
    return (KlO) module;
}

void
KlSesModuleRename(module, name)
    KlSesModule module;
    KlAtom name;
{
    /* look if this name exists */
    int i;
    for (i = 0; i < KlSesModules->size; i++)
	if (KlSesModules->list[i] 
	    && module != (KlSesModule) KlSesModules->list[i]
	    && name == ((KlSesModule) KlSesModules->list[i])->name) {
	    /* this name exist, remove the old module! */
	    KlGCMark();
	    KlAtomSetq(name, NIL);
	    KlSesModuleFree((KlSesModule) KlSesModules->list[i]);
	    KlGC();
	}
    KlDecRef(module->name);		/* remove temp. name */
    KlIncRef(module->name = name);
}

char *
KlSesModuleId2Name(id)
    int id;
{
    if (id >=0
	&& id < KlSesModules->size
	&& KlSesModules->list[id])
	return  ((KlSesModule) KlSesModules->list[id])->name->p_name;
    else
	return "???";
}

KlSesModule
KlSesModuleOrName(module, pos)
    KlSesModule module;
    int pos;
{
    KlSesModule mod;

    if (KlIsAString(module)
	&& NIL != (KlO) (mod = KlSesModuleFromName((KlString) module))) {
	return mod;
    }
    KlMustBeSesModule(module, pos);
    return module;
}

KlSesModule
KlSesModuleFromName(extname)
    KlString extname;
{
    int i;

    for (i = 0; i < KlSesModules->size; i++) {
	if (strequal(KlSesModuleId2Name(i), extname->string)) {
	    return (KlSesModule) KlSesModules->list[i];
	}
    }
    return (KlSesModule) NIL;
}

KlO
KlSesModuleListFunctions(modname)
    KlSesModule modname;
{
    KlSesModule module = KlSesModuleOrName(modname, 0);
    return (KlO) module->func_symbols;
}

KlO
KlSesModuleListTypes(modname)
    KlSesModule modname;
{
    KlSesModule module = KlSesModuleOrName(modname, 0);
    return (KlO) module->ids;
}

KlO
KlSesModulePID(modname)
    KlSesModule modname;
{
    KlSesModule module = KlSesModuleOrName(modname, 0);
    return (KlO) KlNumberMake(module->pid);
}

KlO
KlSesModulesGet(data)
    void *data;
{
    KlList l = KlListNMake(0);
    int i;
    for (i = 0; i < KlSesModules->size; i++) {
	if (KlSesModules->list[i])
	    KlListAppend(l, KlSesModules->list[i]);
    }
    return (KlO) l;
}


/*****************************************************************************\
* 				  KlSesFunc                                   *
\*****************************************************************************/

/* argument_types must be malloced by caller, but not freed. will be freed by
 * KlSesFuncFree automatically
 */
KlSesFunc
KlSesFuncMake(module, name, func, arity, return_type, argument_types)
    KlSesModule module;
    KlAtom name;
    void *func;		
    int arity;		
    int return_type;			
    unsigned int *argument_types;
{
    KlSesFunc object = (KlSesFunc)  KlOMake(KlSesFuncType);
    KlIncRef(object->module = module);
    object->func = func;
    object->arity = arity;
    object->return_type = return_type;
    object->argument_types = argument_types;
    /* then append to list of func names. absolute positions are meaningless */
    KlListPut(module->func_symbols, KlNumberMake(-1), name);
    KlAtomSetq(name, object);
    return object;
}

KlO
KlSesFuncFree(object)
    KlSesFunc object;
{
    KlDecRef(object->module);
    free(object->argument_types);
    Free(object);
    return (KlO) object;
}

/* calling/applying a function will send a synchronous request 
 * altough "network"-wise inefficient, it provides an easy-to-use interface
 */

/* KlSesFuncDoApply: the low-level function common to Apply and Execute */

KlO
KlSesFuncDoApply(func, argc, argv)
    KlSesFunc func;
    int argc;
    KlO *argv;
{
    SES_PS_Header request;
    unsigned int args_size = 0;				/* padded, bytes */
    uchar *ptr;
    KlSesModule module = func->module;
    unsigned int *argtypes = func->argument_types;
    KlSesType *idlist = module->ids->list;
    int i;
    KlSesReqId reqid;
    
    KlNumberOfArgumentsCheck(func->arity != argc, argc);
    for (i = 0; i < argc; i++) {
	args_size += (*(idlist[argtypes[i]]->length_check))
	    (argv[i], module, argtypes[i], i);
    }
    /* allocate request, append args */
    request = KlSesModuleNewRequest(module, SES_HT_Out_Call, 
				    (void *) func->func, 
				    func->return_type == SES_None ? 0
				    : (reqid = KlSesReqIdMake(module, 
							      0, 
							      func->return_type)
					),
				    args_size, argc);
    ptr = SES_PS_HeaderEnd(request);
    for (i = 0; i < argc; i++) {
	ptr = (*(idlist[argtypes[i]]->encode))
	    (ptr, argv[i], module, argtypes[i]);
    }

    if (func->return_type != SES_None) { /* if retval expected, block */
	return KlSesWait(reqid); /* wait does a flush */
    } else {
	return NIL;
    }  
}   

KlO
KlSesFuncApply(func, list)
    KlSesFunc func;
    KlList list;
{
    return KlSesFuncDoApply(func, list->size - 1, list->list + 1);
}

/* Execute evaluates arguments, then call Apply */

KlO
KlSesFuncExecute(func, list)
    KlSesFunc func;
    KlList list;
{
    int argc = list->size - 1;
    KlO *argv = list->list + 1;

    if (argc) {				/* evaluate arguments */
	KlO *eval_list = KlAlloca(argc);
	KlO *dest = eval_list;
	KlO *last = argv + argc;

	while (argv < last) {
	    *dest = KlSend_eval(*argv);
	    dest++, argv++;
	}
	return KlSesFuncDoApply(func, argc, eval_list);
    } else {
	return KlSesFuncDoApply(func, 0, 0);
    }
}

/*****************************************************************************\
* 				   KlSesId                                    *
\*****************************************************************************/

/********************************************************************** type */


KlSesType
KlSesTypeMake(module, name, id, close_callback, length_check, encode, decode)
    KlSesModule module;
    KlAtom name;
    int id;		
    void *close_callback;
    int (*length_check) 
	_ANSI_ARGS_((KlO obj, KlSesModule module, unsigned int id, int n));
    uchar *(*encode) 
	_ANSI_ARGS_((uchar *p, KlO obj, KlSesModule module, unsigned int id));
    uchar *(*decode) 
	_ANSI_ARGS_((uchar *p, KlO *objp, KlSesModule module, unsigned int id));
{
    KlSesType kltype = (KlSesType) KlOMake(KlSesTypeType);

    kltype->module = module;		/* dont incref */
    kltype->id = id;
    KlIncRef(kltype->name = name);
    kltype->close_callback = close_callback;
    kltype->length_check = length_check;
    kltype->encode = encode;
    kltype->decode = decode;
    /* then add it at position id in list of types for this module */
    KlListPut(module->ids, KlNumberMake(id), kltype);
    return kltype;
}

KlO
KlSesTypeFree(kltype)
    KlSesType kltype;
{
    KlDecRef(kltype->name);
    /* dont mess with the list of types of the module. If we are freed, 
       it means the list is already gone */
    Free(kltype);
    return (KlO) kltype;
}

/****************************************************************** instance */

KlSesId
KlSesIdMake(module, id, data)	/* create an instance (handle) */
    KlSesModule module;
    int id;
    void *data;		
{
    KlSesId klid = (KlSesId) KlOMake(KlSesIdType);
    klid->module = module;
    klid->ID =  module->ids->list[id];
    klid->data = data;
    return klid;
}

/* Freeing sends back a trigger for the close_callback, if needed.
 * protocol is sending SES_HT_Out_Free with the callback and data
 */

KlO
KlSesIdFree(id)
    KlSesId id;
{
    if (id->ID->close_callback) {
	SES_PS_Header request = 
	    KlSesModuleNewRequest(id->module, SES_HT_Out_Free,
				  id->ID->close_callback, 0,
				  SES_PS_Ptr_padlen, 1);
	uchar *ptr = SES_PS_HeaderEnd(request);
	SES__EncodePtr(ptr, id->data);
    }
    Free(id);
    return (KlO) id;
}

/*****************************************************************************\
* 				  KlSesReqId                                  *
\*****************************************************************************/


KlSesReqId
KlSesReqIdMake(module, name, type)
    KlSesModule module;
    KlAtom name;		/* or 0 */
    unsigned int type;
{
    KlSesReqId obj = (KlSesReqId) KlOMake(KlSesReqIdType);
    obj->module = module;
    obj->name = name;
    obj->value = 0;
    obj->valuetype = type;
    obj->previous = KlSesReqIdLast;
    KlIncRef(KlSesReqIdLast = obj);
    return obj;
}

KlO
KlSesReqIdFree(obj)
    KlSesReqId obj;
{
    KlDecRef(obj->value);
    KlDecRef(obj->previous);
    Free(obj);
    return (KlO) obj;
}


/*****************************************************************************\
* 				     init                                     *
\*****************************************************************************/
/***************************************************************** KlSesLoad */
/* Loading an extension is forking and executing the extension as a standalone
 * executable, setting the ways to communicate later (pipes and optional 
 * shared memory)
 * A module is returned which is kept in the list of loaded modules, so you 
 * dont need to keep a reference onto it.
 * The extension will initialize itself and initialise the Klone side by 
 * uploading all relevant info via the file descriptor
 * 
 * The only thing to do is to make sure you call KlSesHandler() when something
 * is waiting on the KlSes_fd_main_r filedescriptor
 * for instance by XtAddInput for the Xt Toolkit
 *
 * (ses:load module-name [:handler expression])
 * blocks till init is done, or if handler is given, returns and will evaluate
 * the expression when init is done.
 *
 * If compiled with -DSES_DEBUG, loading ":foo" will load 
 * "ddd -x .gdbinit_ses .foo"
 */

/* todo: implement the :timeout keyword */

KlO
KlSesLoad(argc, argv)
    int argc;
    KlString *argv;
{
    KlString extname;
    KlO handler;
    KlNumber timeout;
    KlString path, oldpath;
    KlSesModule module;
    int fdt[2];
    char klone_ses_args[80];
    int pid;
#ifdef SESDEBUG
    int ddd = 0;
#endif  /* SESDEBUG */

    KlParseKeywords(argc, argv, 1);
    handler = KlKeyVal(KlK_handler, 0);
    timeout = (KlNumber) KlKeyVal(KlK_timeout, 0);
    KlCheckUnvalidKeywords(argc, argv, 1);
    extname = argv[0];
    KlMustBeString(extname, 0);
#ifdef SESDEBUG
    if (extname->string[0] == ':')
	ddd = 1, extname = KlStringMake(extname->string + 1);
#endif  /* SESDEBUG */
    if (NIL != (KlO) (module = KlSesModuleFromName(extname)))
	return (KlO) module;

    path = (KlString) KlA_Ses_loadpathname->c_val;
    KlMustBeString(path, 0);	/* change PATH */
    oldpath = (KlString) KlGetenv(KlS_PATH);
#ifdef SESDEBUG
    { /* prepend to current path, to be able to run ddd */
	KlString argv[3];
	argv[0] = path;
	argv[1] = KlStringMake(":");
	argv[2] = (KlString) KlGetenv(KlS_PATH);
	path = (KlString) KlStringAdd(3, argv);
    }    
#endif  /* SESDEBUG */

    KlPutenv(KlS_PATH, path);
    
    pipe(fdt);				/* pipe main -> extension */
    module = KlSesModuleMake(extname, fdt[1]);

    KlSesInitComms(module->id);

    sprintf(klone_ses_args, "KLONE_SES_ARGS=%d:%d:%d:%d:%d:%d:%d:%d",
	    KlSesVersion, KlSesTransportSHM, 
#ifdef SHM
	    KlSes_shm_id, KlSes_sem_id,
#else /* !SHM */
	    0, 0,
#endif /* !SHM */
	    fdt[0], KlSes_fd_main_w, getpid(),  module->id);
    Kl_putenv(klone_ses_args, 0);

    if (!(pid = VFORK())) {			/* son */
	/* todo: close uneeded fds, keep fdt[0] and KlSes_fd_main_w and the
	   contents of KlA_keep_streams_exec */
	close(fdt[1]);
	close(KlSes_fd_main_r);
#ifndef SESDEBUG
	{
	    char *args[2];
	    args[0] = extname->string;
	    args[1] = 0;
	    execvp(*args, args);
	}
#else  /* SESDEBUG */
	{
	    char *args[5]; 
	    if (ddd) {
		args[0] = "ddd"; 
		args[1] = "-x";
		args[2] = ".gdbinit_ses";
		args[3] = extname->string; 
		args[4] = 0;
		fprintf(stderr, "SES: execing %s %s %s %s\n", args[0], args[1],
			args[2], args[3]);
	    } else {
		args[0] = extname->string;
		args[1] = 0;
		fprintf(stderr, "SES: execing %s\n", args[0]);
	    }
	    execvp(*args, args);
	    fprintf(stderr, "SES: could not find %s\n", args[0]);
	}
#endif /* SESDEBUG */
	/* an error! send a message on pipe and die */
	KlSesReturnRequest(module, SES_HT_Out_Send, SES_HFI_NotFound, 0, 0, 0);
	close(KlSes_fd_main_w);
	KlExecError(0, extname->string);
    }
#ifdef SESDEBUG
    fprintf(stderr, "forked extension %s\n", argv[0]);
#endif

    module->pid = pid;
    
    KlPutenv(KlS_PATH, oldpath);	/* clean */
    Kl_putenv(klone_ses_args, 1);
    close(fdt[0]);

    if (handler) {
	KlIncRef(module->ready_callback = handler);
    } else {
	/* now we block till extension initializes or dies */
	while (!module->ready && !module->dead) {
	    KlSesReadAndHandle();
	}
    }
    return (KlO) module;
}

/* unloading a module */

KlO
KlSesModuleUnload(modname)
    KlSesModule modname;
{
    KlSesModule module = KlSesModuleOrName(modname, 0);

    if (!module) 
	return NIL;
    KlDecRef(module->name);
    if (module->fd) {			/* avoid latency */
	close(module->fd);
	module->fd = 0;
    }
    /* cut circular references here */
    KlDecRef(module->func_symbols);
    module->func_symbols = 0;
    KlDecRef(module->ids);
    module->ids = 0;
    KlDecRef(module);
    KlSesModules->list[module->id] = 0;	/* make hole, dont compact list */
    if (module->id == (KlSesModules->size - 1)) /* was last, shorten list */
	KlSesModules->size--;

    return (KlO) module;
}

/*****************************************************************************\
* 				   requests                                   *
\*****************************************************************************/
/* Allocates a new request, and fills the header. The body should be filled 
 * afterwards by the caller
 */

SES_PS_Header
KlSesModuleNewRequest(module, type, func_ptr, reqid, size, arity)
    KlSesModule module;
    unsigned int type;			/* SES_HT_... major # of request */
    void *func_ptr;			/* id (minor) of request */
    KlSesReqId reqid;
    unsigned int size;			/* size in bytes to reserve */
    int arity;				/* arity of request */
{
    SES_PS_Header request;
    /* pad size to granularity (debug code, normally size is already padded) */
    size = SES_padded(size);

					/* reserve size more bytes */
    if (module->outlimit - module->outptr </* remaining space too small*/
	(SES_PS_Header_padlen + size)) {	/* we must realloc */
	int newsize = module->outlimit - module->outbuf;
	uchar *oldbuf = module->outbuf;
	while ((newsize - (module->outlimit - module->outptr))
	       < (SES_PS_Header_padlen + size)) {
	    newsize = 2 * newsize + sizeof(void *);
	}
	module->outbuf = (uchar *) realloc(oldbuf, newsize);
	module->outptr = module->outbuf + (module->outptr - oldbuf);
	module->outlimit = module->outbuf + newsize;
    }
					/* encode header */
    request = (SES_PS_Header) module->outptr;
    request->type = type;
    request->module_id = module->id;
    request->func.ptr = func_ptr;
    request->length = SES_PS_Header_padlen + size;
    request->arity = arity;
    request->reqid = reqid; /* default: dont expect answer */
					/* compute checksum */
    request->header_checksum = SES__HeaderChecksum(request);
					/* advance pointer */
    module->outptr += SES_PS_Header_padlen + size;

    return request;
}

/* sends back to oneself a request. only used when EXEC-ing the extension 
 * failed, we fake a return code from a virtual extension
 */

void
KlSesReturnRequest(module, type, func_id, size, arity, data)
    KlSesModule module;
    unsigned int type;			/* SES_HT_... major # of request */
    unsigned int func_id;		/* id (minor) of request */
    unsigned int size;			/* size in bytes to reserve */
    int arity;				/* arity of request */
    char *data;				/* the appended data. can be 0 */
{
    SES_PS_Header request = (SES_PS_Header) 
	Malloc(sizeof(struct _SES_PS_Header) + size);

    request->type = type;
    request->func.id = func_id;
    request->length = SES_PS_Header_padlen + size;
    request->module_id = module->id;
    request->reqid = 0;
    request->arity = arity;
    request->header_checksum = SES__HeaderChecksum(request);
    if (data)
	bcopy(data, SES_PS_HeaderEnd(request), size);
#ifdef SESDEBUG
    fprintf(stderr, "seslib emul: fake request, type: %d, module: %d, length %d, seq: %d, func: %d, arity: %d, checksum: %d\n",
	    request->type , request->module_id 
	    , request->length , request->reqid , request->func.id
	    , request->arity, request->header_checksum);
#endif
    write(KlSes_fd_main_w, (uchar *) request, request->length);
}

/*****************************************************************************\
* 				   Answers                                    *
\*****************************************************************************/

/**************************************************** decoding error handler */

KlSesCodecError(module, expected_type, provided_type)
    KlSesModule module;
    int expected_type;
    int provided_type;
{
    KlError3(KlE_SES_BAD_ARGUMENT, module->name,
	     module->ids->list[expected_type]->name,
	     module->ids->list[provided_type]->name);
}

KlO					/* seqnum */
KlSesSend(argc, argv)
    int argc;
    KlO *argv;
{

    TODO_NOT_YET_IMPLEMENTED("KlSesSend");
}

KlO					/* return value of the call */
KlSesWait(reqid)
    KlSesReqId reqid;
{
    KlSesReqId req;
    KlMustBeSesReqId(reqid, 0);
    KlSesFlush(reqid->module);
    while (!reqid->value) {
	KlSesReadAndHandle();
    }
    for (req = KlSesReqIdLast; req; req = req->previous) {
	if (req == reqid) {
	    req->previous = 0;
	    KlDecRef(reqid);
	    break;
	}
    }
    return reqid->value;
}

KlO					/* undefined */
KlSesFlush(module)
    KlSesModule module;
{
    uchar *buffer = module->outbuf;
    int bytes_to_send = module->outptr - module->outbuf;
    while (bytes_to_send) {
	int sent;
	if (-1 == (sent = 
		   write(module->fd, buffer, bytes_to_send))) { /* error */
	    KlError1(KlE_SES_IO_ERROR, module->name);
	}
	bytes_to_send -= sent;
	buffer += sent;
    }
    module->outptr = module->outbuf;
    return NIL;
}

KlO					/* undefined */
KlSesSync(module)
    KlSesModule module;
{
    KlSesReqId reqid = KlSesReqIdMake(module, 0, SES_None);
    SES_PS_Header request = KlSesModuleNewRequest(module, SES_HT_Out_Sync, 0,
						  reqid, 0, 0);
    KlSesFlush(module);
    while (!reqid->value) {
	KlSesReadAndHandle();
    }
    return NIL;
}

KlO					/* previous handler or () */
KlSesErrorHandlerSet(module, handler)
    KlSesModule module;
    KlO handler;
{
    KlO oldhandler = KlSesUserErrorHandler;
    KlSesUserErrorHandler = handler;
    return oldhandler;
}

KlO					/* undefined */
KlSesReturn(value)
    KlO value;
{
    TODO_NOT_YET_IMPLEMENTED("KlSesReturn");
}

KlO					/* undefined */
KlSesReturnError(argc, argv)
    int argc;
    KlO *argv;
{
    TODO_NOT_YET_IMPLEMENTED("KlSesReturnError");
}


/*****************************************************************************\
* 				    codecs                                    *
\*****************************************************************************/
/* codecs of built-in types 
 * These are used in the Klone side, to provide dynamic encoding/decoding of 
 * typed arguments and return values to declared functions
 *  - length_check must check data and return the used length
 *  - encode must stream the obj to the buffer ptr and return next ptr
 *  - decode must decode data and create an obj which is returned through the
 *    pointer argument *objp, and return ptr to next object data
 * The actual work is performed on C data by the codecs in sescodecs.c,
 * these codecs provide a Klone/C interface to them
 */

/****************************************************************** SES_None */
int
KlSesCodec_None_length_check(KlO obj, KlSesModule module, unsigned int id, int n)
{
    return 0;
}
uchar *
KlSesCodec_None_encode(uchar *ptr, KlO obj, KlSesModule module, unsigned int id)
{
    return ptr;
}
uchar *
KlSesCodec_None_decode(uchar *ptr, KlO *objp, KlSesModule module, unsigned int id)
{
    *objp = NIL;
    return ptr;
}

/******************************************************************* SES_Int */
int
KlSesCodec_Int_length_check(KlO obj, KlSesModule module, unsigned int id, int n)
{
    KlMustBeNumber(obj, n);
    return SES_PS_Int_padlen;
}
uchar *
KlSesCodec_Int_encode(uchar *ptr, KlO obj, KlSesModule module, unsigned int id)
{
    return SES__EncodeInt(ptr, ((KlNumber) obj)->number);
}
uchar *
KlSesCodec_Int_decode(uchar *ptr, KlO *objp, KlSesModule module, unsigned int id)
{
    int number;
    ptr = SES__DecodeInt(ptr, &number, module);
    *objp = (KlO) KlNumberMake(number);
    return ptr;
}

/**************************************************************** SES_String */
int
KlSesCodec_String_length_check(KlO obj, KlSesModule module, unsigned int id, int n)
{
    KlMustBeString(obj, n);
    return SES_PS_String_padlen(KlStringLength(obj));
}
uchar *
KlSesCodec_String_encode(uchar *ptr, KlO obj, KlSesModule module, unsigned int id)
{
    return SES__EncodeString(ptr, KlStringLength(obj), ((KlString) obj)->string);
}
uchar *
KlSesCodec_String_decode(uchar *ptr, KlO *objp, KlSesModule module, unsigned int id)
{
    char *string;
    int len;
    ptr = SES__DecodeString(ptr, &len, &string, module);
    *objp = (KlO) KlStringMakeFromBytes(len, string);
    return ptr;
}

/**************************************************************** SES_Atom */
int
KlSesCodec_Atom_length_check(KlO obj, KlSesModule module, unsigned int id, int n)
{
    KlMustBeSymbol(obj, n);
    return SES_PS_Atom_padlen(KlStringLength(obj));
}
uchar *
KlSesCodec_Atom_encode(uchar *ptr, KlO obj, KlSesModule module, unsigned int id)
{
    return SES__EncodeAtom(ptr, KlStringLength(obj), ((KlString) obj)->string);
}
uchar *
KlSesCodec_Atom_decode(uchar *ptr, KlO *objp, KlSesModule module, unsigned int id)
{
    char *string;
    int len;
    ptr = SES__DecodeAtom(ptr, &len, &string, module);
    *objp = (KlO) KlInternBytes(len, string);
    return ptr;
}

/******************************************************************** SES_Id */
int
KlSesCodec_Id_length_check(KlO obj, KlSesModule module, unsigned int id, int n)
{
    KlMustBeSesId(obj, n);
    if (((KlSesId) obj)->ID->id != id)
	KlBadArgument(obj, n, module->ids->list[id]->name->p_name);
    return SES_PS_Id_padlen;
}
uchar *
KlSesCodec_Id_encode(uchar *ptr, KlO obj, KlSesModule module, unsigned int id)
{
    return SES__EncodeId(ptr, id, ((KlSesId) obj)->data);
}
uchar *
KlSesCodec_Id_decode(uchar *ptr, KlO *objp, KlSesModule module, unsigned int id)
{
    void *data;
    ptr = SES__DecodeId(ptr, id, &data, module);
    *objp = (KlO) KlSesIdMake(module, id, data);
    return 0;
}

/*************************************************************** SES_Boolean */
int
KlSesCodec_Boolean_length_check(KlO obj, KlSesModule module, unsigned int id, int n)
{
    if (obj != TRU && KlTrueP(obj))
	KlBadArgument(obj, n, "boolean");
    return SES_PS_Boolean_padlen;
}
uchar *
KlSesCodec_Boolean_encode(uchar *ptr, KlO obj, KlSesModule module, unsigned int id)
{
    return SES__EncodeBoolean(ptr, obj == TRU ? 1 : 0);
}
uchar *
KlSesCodec_Boolean_decode(uchar *ptr, KlO *objp, KlSesModule module, unsigned int id)
{
    int number;
    ptr = SES__DecodeBoolean(ptr, &number, module);
    *objp = (number ? TRU : NIL);
    return ptr;
}

/******************************************************************* SES_Real */
int
KlSesCodec_Real_length_check(KlO obj, KlSesModule module, unsigned int id, int n)
{
    KlMustBeNumber(obj, n);
    return SES_PS_Real_padlen;
}
uchar *
KlSesCodec_Real_encode(uchar *ptr, KlO obj, KlSesModule module, unsigned int id)
{
    return SES__EncodeReal(ptr, KlNumberRealValue(obj));
}
uchar *
KlSesCodec_Real_decode(uchar *ptr, KlO *objp, KlSesModule module, unsigned int id)
{
    double x;
    ptr = SES__DecodeReal(ptr, &x, module);
    *objp = (KlO) KlRealMake(x);
    return ptr;
}

/******************************************************************* SES_Ptr */
int
KlSesCodec_Ptr_length_check(KlO obj, KlSesModule module, unsigned int id, int n)
{
    KlMustBeNumber(obj, n);
    return SES_PS_Ptr_padlen;
}
uchar *
KlSesCodec_Ptr_encode(uchar *ptr, KlO obj, KlSesModule module, unsigned int id)
{
    return SES__EncodePtr(ptr, ((KlNumber) obj)->number);
}
uchar *
KlSesCodec_Ptr_decode(uchar *ptr, KlO *objp, KlSesModule module, unsigned int id)
{
    Int number;
    ptr = SES__DecodePtr(ptr, &number, module);
    *objp = (KlO) KlNumberMake(number);
    return ptr;
}

/*********************************************** init list of built-in types */

void
KlSesListOfBuiltinTypesInit(module)
    KlSesModule module;
{
    KlNumber klsize = KlNumberMake(SES_MAX_Builtins);
    KlListOfSesType kllt = (KlListOfSesType) KlListNMakeKl(1, &klsize);
    KlSesType *lt = kllt->list;
    KlIncRef(module->ids = kllt);
    lt[SES_None] = KlSesTypeMake(module, KlIntern("None"), SES_None, 0,
				 KlSesCodec_None_length_check,
				 KlSesCodec_None_encode,
				 KlSesCodec_None_decode);
    lt[SES_Int] = KlSesTypeMake(module, KlIntern("Int"), SES_Int, 0,
				 KlSesCodec_Int_length_check,
				 KlSesCodec_Int_encode,
				 KlSesCodec_Int_decode);
    lt[SES_String] = KlSesTypeMake(module, KlIntern("String"), SES_String, 0,
				 KlSesCodec_String_length_check,
				 KlSesCodec_String_encode,
				 KlSesCodec_String_decode);
    lt[SES_Atom] = KlSesTypeMake(module, KlIntern("Atom"), SES_Atom, 0,
				 KlSesCodec_Atom_length_check,
				 KlSesCodec_Atom_encode,
				 KlSesCodec_Atom_decode);
    lt[SES_Boolean] = KlSesTypeMake(module, KlIntern("Boolean"), SES_Boolean, 0,
				 KlSesCodec_Boolean_length_check,
				 KlSesCodec_Boolean_encode,
				 KlSesCodec_Boolean_decode);
    lt[SES_Real] = KlSesTypeMake(module, KlIntern("Real"), SES_Real, 0,
				 KlSesCodec_Real_length_check,
				 KlSesCodec_Real_encode,
				 KlSesCodec_Real_decode);
    lt[SES_Ptr] = KlSesTypeMake(module, KlIntern("Ptr"), SES_Ptr, 0,
				 KlSesCodec_Ptr_length_check,
				 KlSesCodec_Ptr_encode,
				 KlSesCodec_Ptr_decode);
    lt[SES_ERROR] = KlSesTypeMake(module, KlIntern("ERROR"), SES_ERROR, 0,
				 KlSesCodec_None_length_check,
				 KlSesCodec_None_encode,
				 KlSesCodec_None_decode);;
}

/* called to declare id type dynamically from extension */
void
KlSesIdDeclareType(module, name, id, close_callback)
    KlSesModule module;
    KlAtom name;
    int id;		
    void *close_callback;
{
    KlSesType kltype = KlSesTypeMake(module, name, id, close_callback, 
				     KlSesCodec_Id_length_check,
				     KlSesCodec_Id_encode,
				     KlSesCodec_Id_decode);
}

/*****************************************************************************\
* 				   handler                                    *
\*****************************************************************************/
/* when something arrives (select) on the file descriptor of extensions, 
 * KlSes_fd_main_r, call this handler on it, as a XtInputCallbackProc
 *
 * For efficiency, it gobbles up as much as it can at a time and processes it
 *
 * Protocol is described in ses.h
 * For requests, the minor number (SES_HFI_...) determines the function:
 */

void
KlSesHandler(closure, source, id)
    void *closure;
    int source;
    unsigned long *id;			/* XtInputId */
{
    if (KlSesReadWhatsAvailable()) {
	while (KlSesRequestPending()){
	    KlSesDecodeRequest(KlSes_soc_inptr);
	}
    } else {				/* select activated, but empty => EOF */
	/* extension crashed. todo: see if can happen on a shared pipe */
	KlError1s(KlE_SES_IO_ERROR, "common pipe from extensions");
    }
}

/* blocks till something comes, then parses it
 * can be used in a while loop while waiting for some condition, as it 
 * parses only one request at a time.
 */

void
KlSesReadAndHandle()
{
    if (KlSesRequestPending()) {
	KlSesDecodeRequest(KlSes_soc_inptr);
    } else {
	fd_set fds;
	FD_ZERO(&fds);
	FD_SET(KlSes_fd_main_r, &fds);
	KlSesMainPreSelect();
	select(KlSes_fd_main_r+1, &fds, NULL, NULL, NULL);
	if (FD_ISSET(KlSes_fd_main_r, &fds)) {
	    if (KlSesReadWhatsAvailable()) {
		if (KlSesRequestPending()) {
		    KlSesDecodeRequest(KlSes_soc_inptr);
		}
	    } else {		/* select activated, but empty => EOF */
		/* extension crashed. todo: see this can happen on a shared pipe */
		KlError1s(KlE_SES_IO_ERROR, "common pipe from extensions");
	    }
	}
    }
}

int
KlSesReadWhatsAvailable()
{
    int size;

retry:
    if (size = (KlSes_soc_inend - KlSes_soc_inptr)) { 
	/* move remnants at start, use memmove as may overlap */
	memmove(KlSes_soc_inbuf, KlSes_soc_inptr, size);
    }
    KlSes_soc_inptr = KlSes_soc_inbuf;
    KlSes_soc_inend = KlSes_soc_inptr + size;
    if (KlSes_soc_inlimit - KlSes_soc_inend < sizeof (struct _SES_PS_Header)) {
	/* no room for a least another request, realloc */
	uchar *oldbuf = KlSes_soc_inbuf;
	size = 2 * (KlSes_soc_inlimit - KlSes_soc_inbuf) + sizeof(void *);
	KlSes_soc_inbuf = (uchar *) Realloc(oldbuf, size);
	KlSes_soc_inptr = KlSes_soc_inbuf + (KlSes_soc_inptr - oldbuf);
	KlSes_soc_inlimit = KlSes_soc_inbuf + size;
	KlSes_soc_inend = KlSes_soc_inbuf + (KlSes_soc_inend - oldbuf);
    }
    /* append to the end of remnants already in buffer */
    size = read(KlSes_fd_main_r, KlSes_soc_inend, 
		KlSes_soc_inlimit - KlSes_soc_inend);
    if (size < 0) {			/* connection problem */
	if (!ECHECK(EINTR))		/* weird problem, quit */
	    KlError1s(KlE_SES_IO_ERROR, "common pipe from extensions");
	else				/* normal interrupt, retry */
	    goto retry;
    }
    KlSes_soc_inend += size;
    return size;
}

/* processes a complete incoming request (it MUST be complete, check by 
 * a call to KlSesRequestPending() before.
 * first, discriminates on request types
 * then on function IDs
 */

void
KlSesDecodeRequest(request)
    SES_PS_Header request;
{
    KlSesModule module = KlSesId2Module(request->module_id);
    /* todo: better validity checks, and how to recover? */
    if (request->header_checksum != SES__HeaderChecksum(request)) {
	KlError1(KlE_SES_INTERNAL_ERROR, module->name);
    }
    KlSes_soc_inptr += request->length; /* goto next request */
    switch (request->type) {
    case SES_HT_In_Result:		/* result from a previous call */
    {   KlSesReqId reqid = request->reqid;
    if (reqid) {		/* was expected */
	uchar *ptr = SES_PS_HeaderEnd(request);
	KlO result;
#ifdef DEBUG
	if (!KlObjectIsValid(reqid))
	    KlError1(KlE_SES_INTERNAL_ERROR, module->name);
#endif	    
	(*(module->ids->list[reqid->valuetype]->decode))
	    (ptr, &result, module, reqid->valuetype);
	if (reqid->name)
	    KlSend_setq(reqid->name, result);
	KlIncRef(reqid->value = result);
    } /* else not expected, discard (should not happen) */
    }
    break;
    case SES_HT_In_Send:		/* extension-pushed message */
	switch (request->func.id) {
	    
	case SES_HFI_ReadEvalText:	/* eval, no return expected */
	{
	    uchar *ptr = SES_PS_HeaderEnd(request);
	    char *text;
	    int textlen;
	    int stackptr = KlStackPtr;
 
	    SES__DecodeString(ptr, &textlen, &text, module);

	    KlStackFramePush(1, &KlA_SesCurrentModule, NIL, NIL);
	    KlSend_setq(KlA_SesCurrentModule, module);
	    KlExecuteStringNoReturn(text);
	    KlStackFramePopNormal(stackptr);
	}
	break;

	case SES_HFI_DeclareType:	/* extension declare new opaque type */
	{
	    uchar *ptr = SES_PS_HeaderEnd(request);
	    int id;
	    char *name;
	    int namelen;
	    void *close_callback;

	    ptr = SES__DecodeInt(ptr, &id, module);
	    ptr = SES__DecodeString(ptr, &namelen, &name, module);
	    ptr = SES__DecodePtr(ptr, &close_callback, module);
	    KlSesIdDeclareType(KlSesModules->list[request->module_id],
			       KlInternBytes(namelen, name),
			       id, close_callback);
	}
	break;

	case SES_HFI_DeclareFunction: /* extension declare new function */
	{
	    uchar *ptr = SES_PS_HeaderEnd(request);
	    void *func;
	    int arity;
	    int retype;
	    char *name; int namelen;
	    int i, n_types;
	    unsigned int *argument_types, *p;

	    ptr = SES__DecodePtr(ptr, &func, module);
	    ptr = SES__DecodeString(ptr, &namelen, &name, module);
	    ptr = SES__DecodeInt(ptr, &arity, module);
	    ptr = SES__DecodeInt(ptr, &retype, module);
	    n_types = request->arity - 4;
	    argument_types = (unsigned int *) Malloc(n_types *
						     sizeof(unsigned int));
	    if (retype >= module->ids->size || !module->ids->list[retype]) {
		KlError3(KlE_SES_BAD_ARGUMENT, module->name, 
			 KlStringMake("valid type id"), KlNumberMake(retype));
	    }
	    p = argument_types;
	    for (i = 0; i < n_types; i++) {
		ptr = SES__DecodeInt(ptr, p, module);
		if (*p >= module->ids->size || !module->ids->list[*p]) {
		    KlError3(KlE_SES_BAD_ARGUMENT, module->name, 
			     KlStringMake("valid type id"), KlNumberMake(*p));
		}
		p++;
	    }
	    KlSesFuncMake(KlSesModules->list[request->module_id],
			  KlInternBytes(namelen, name),
			  func, n_types, retype, argument_types);
	}
	break;

	case SES_HFI_Ready:		/* end of Inits, you can use it now */
	{
	    KlO rcb;
	    module->ready = 1;
	    if (rcb = module->ready_callback) {
		int stackptr = KlStackPtr;
		KlStackFramePush(1, &KlA_SesCurrentModule, NIL, NIL);
		KlSend_setq(KlA_SesCurrentModule, module);
		KlSend_eval(rcb);
		KlStackFramePopNormal(stackptr);
		KlDecRef(rcb);
		module->ready_callback = 0;
	    }
	}
	break;
	}
	break;
    case SES_HT_In_Call:
	/* TODO */
	break;

    case SES_HT_In_Error:		/* error in previous call or send */
	/* what to do? we must call the error handler for it to decide */
    {
	uchar *ptr = SES_PS_HeaderEnd(request);
	char *text;
	int textlen;
	int stackptr = KlStackPtr;
	KlList errlist; KlO result;
	int normal;
	int call_standard_handler = 1;
	int discard_sends = 0;
	    
 
	SES__DecodeString(ptr, &textlen, &text, module);

	if (request->reqid) { /* set value to 'ERROR */
	    if (request->reqid->name)
		KlSend_setq(request->reqid->name, KlA_ERROR);
	    KlIncRef(request->reqid->value = (KlO) KlA_ERROR);
	}

	KlStackFramePush(1, &KlA_SesCurrentModule, NIL, NIL);
	KlSend_setq(KlA_SesCurrentModule, module);
	KlCatch(KlA_ALL, KlParseString(textlen, text, 0), result, normal);
	errlist = (KlList) result;
	if (normal && KlIsAList(errlist) && errlist->size > 0) {	
	    if (KlSesUserErrorHandler == (KlO) KlK_nil) {
		normal = 1; result = (KlO) KlK_nil;
	    } else {
		KlCatch(KlA_ALL, KlApplyN(KlSesUserErrorHandler, 
					  errlist->size, errlist->list),
			result, normal);
	    }
	}
	if (!normal) {
	    /* couldnt parse or exec err ? ==> default error message */
	    errlist = KlListTripletMake(KlE_SES_BAD_ERROR, module->name,
					KlStringMake(text));
	    _KlError(3, errlist->list);
	    goto depop_error;
	}
	if (result == (KlO) KlK_true) { /* discard events */
	    discard_sends = 1;
	} else if (result == (KlO) KlK_nil) {/* dont discard */
	} else {		/* dont even call error handler */
	    call_standard_handler = 0;
	}
	{
	    SES_PS_Header err_reply = 
		KlSesModuleNewRequest(module, SES_HT_Out_Error,
				      (void *) (discard_sends ? 2 : 1),
				      0, 0, 0);
	    KlSesFlush(module);
	}
	if (call_standard_handler)
	    _KlError(errlist->size, errlist->list);
    depop_error:
	KlStackFramePopNormal(stackptr);
    }
    /* dont forget to update the requiring reqid if it was a Call */
    break;
    case SES_HT_In_Sync:
	/* TODO */
	break;
    default:
	;
    }
}

/*****************************************************************************\
* 				    comms                                     *
\*****************************************************************************/
#ifdef COMMENTED_OUT
/* stream pipes are just Unix pipes, but full duplex bi-directional.
 * this is a wrapper function masking the differences between BSD4.2 and SYSVR4
 */

#ifndef SVR4				/* otherwise is just a macro: pipe() */
#include <sys/socket.h>
int 
KlStreamPipe(fdpair)
    int fdpair[];
{
    return(socketpair(AF_UNIX, SOCK_STREAM, 0, fdpair));
}
#endif /* !SVR4 */
#endif /* COMMENTED_OUT */

/*****************************************************************************\
* 				  KlSesInit                                   *
\*****************************************************************************/

/* init the communication layer on first use */

int
KlSesInitComms(id)
int id;					/* id of current module (unused) */
{
    if (KlSes_fd_main_w)
	return 0;				/* already initialized */
    {					/* init main pipe extensions -> main */
	int fds[2];
	pipe(fds);
	KlSes_fd_main_r = fds[0];
	KlSes_fd_main_w = fds[1];
    }
    /* put file descriptor extension->main in non-blocking mode */
    {
	int file_mode;
	fcntl(KlSes_fd_main_r, F_GETFL, &file_mode);
	file_mode |= KlNDELAY;	/* set non-blocking flag */
	fcntl(KlSes_fd_main_r, F_SETFL, file_mode);
    }

#ifdef SHM
    /* WARNING: incomplete! */
    /* init shared memory */
    KlSes_shm_id = shmget(IPC_PRIVATE, KlSes_shm_size, 
			  IPC_CREAT | SHM_R | SHM_W); /* create, u+rw */
    if (KlSes_shm_id == -1)
	return KlError1s(KlE_SES_IO_ERROR, "cannot initialize shared memory"),0;
    KlSes_shm_ptr = shmat(KlSes_shm_id, 0, 0); /* map in memory */
    if (KlSes_shm_ptr == (char *) -1)
	return KlError1s(KlE_SES_IO_ERROR, "cannot map shared memory"),0;

    /* init semaphores */
    KlSes_sem_id = semget(IPC_PRIVATE, 1,
			  IPC_CREAT | SHM_R | SHM_W); /* create, u+rw */
    if (KlSes_sem_id == -1)
	return KlError1s(KlE_SES_IO_ERROR, "cannot initialize semaphores"), 0;
    /* register info into a file in a system-wide directory: /tmp/SES_SHM/
       so that shared memory can be cleared by next process launched if 
       the creator process seems dead */

    semctl(KlSes_sem_id, 0, SETVAL, 1);	/* initialize value to free */
    /* TODO : pnum ppnum KlSes_shm_id KlSes_sem_id */
#endif /* SHM */

    /* sockets */
    {
	/* allocate input buffer. output buffers are in each module */
	KlSes_soc_inbuf = (uchar *) Malloc(SES_BUFSIZE - sizeof(void *));
	KlSes_soc_inptr = KlSes_soc_inend = KlSes_soc_inbuf;
	KlSes_soc_inlimit = KlSes_soc_inbuf + SES_BUFSIZE - sizeof(void *);
    }

    /* tell application to add the fd into its main loop */
    if (KlSes_addinput)
	(*KlSes_addinput)(KlSes_fd_main_r);

    return 0;    
}

/* clean up connections */

KlSesEnd()
{
#ifdef SHM
    shmctl(KlSes_shm_id, IPC_RMID, 0);
    semctl(KlSes_sem_id, 0, IPC_RMID, 0);
#endif SHM
}

/* KlSesModulePidDied
 * clean-up routine on detection of extension termination
 * called by the SIGCHLD signal handler when a process dies
 * must return 0 if it was not a SES process
 */

int
KlSesModulePidDied(pid)
    int pid;
{
    int i;
    /* the module is created and initialized before the fork */
    for (i = 0; i < KlSesModules->size; i++) {
	if (KlSesId2Module(i)->pid == pid) {
	    /* Module died */
	    KlSesId2Module(i)->dead = 1;
	    KlDecRef(KlSesId2Module(i));/* Free will be triggered later */
	    KlSesId2Module(i) = 0;
	    return 1;
	}
    }
    return 0;
}


#ifdef ISM
/* function to add the connection handler to the main SML loop */
void
KlISM_RegisterSESmainloop(fd)
    int fd;
{
    /* add the handler to SML main loop */
    sml_CreateStreamHandler(fd, sml_ReadMask, KlSesHandler, (void *) 0);
}
#endif /* ISM */

/* You MUST also call this function just before doing the select: it will 
 * flush and process pending requests if any are left
 */

void
KlSesMainPreSelect()
{
    KlSesModule *modulep = (KlSesModule *) KlSesModules->list;
    KlSesModule *last_modulep = modulep + KlSesModules->size;
    while (modulep < last_modulep) {
	if (*modulep && ((*modulep)->outptr > (*modulep)->outbuf)) {
	    KlSesFlush(*modulep);
	}
	modulep++;
    }
    while (KlSesRequestPending()){
	KlSesDecodeRequest(KlSes_soc_inptr);
    }
}

/* Inits done on startup */

KlSesInit()
{
    KlIncRef(KlSesModules = KlListNMake(0));

    /* type declarations */
    KlDeclareType(&KlSesModuleType, "SesModule", sizeof(struct _KlSesModule));
    KlDeclareMethod1(KlSesModuleType, KlSelFree, KlSesModuleFree);

    KlDeclareType(&KlSesFuncType, "SesFunc", sizeof(struct _KlSesFunc));
    KlDeclareMethod1(KlSesFuncType, KlSelFree, KlSesFuncFree);
    KlDeclareMethod1(KlSesFuncType, KlSelApply, KlSesFuncApply);
    KlDeclareMethod1(KlSesFuncType, KlSelExecute, KlSesFuncExecute);

    KlDeclareType(&KlSesTypeType, "SesType", sizeof(struct _KlSesType));
    KlDeclareMethod1(KlSesTypeType, KlSelFree, KlSesTypeFree);

    KlDeclareType(&KlSesIdType, "SesId", sizeof(struct _KlSesId));
    KlDeclareMethod1(KlSesIdType, KlSelFree, KlSesIdFree);

    KlDeclareType(&KlSesReqIdType, "SesReqId", sizeof(struct _KlSesReqId));
    KlDeclareMethod1(KlSesReqIdType, KlSelFree, KlSesReqIdFree);

    /* API functions */
    KlDeclareSubr(KlSesLoad, "ses:load", NARY);
    KlDeclareSubr(KlSesModuleUnload, "ses:unload", 1);
    KlDeclareSubr(KlSesModuleListFunctions, "ses:module-functions", 1);
    KlDeclareSubr(KlSesModuleListTypes, "ses:module-types", 1);
    KlDeclareSubr(KlSesModulePID, "ses:module-pid", 1);
    KlDeclareSubr((KlMethod) KlSesModuleFromName, "ses:name-to-module", 1);
    KlDeclareSubr(KlSesSend, "ses:send", NARY);
    KlDeclareSubr(KlSesWait, "ses:wait", 2);
    KlDeclareSubr(KlSesFlush, "ses:flush", 1);
    KlDeclareSubr(KlSesSync, "ses:sync", 1);
    KlDeclareSubr(KlSesErrorHandlerSet, "ses:error-handler", 2);
    KlDeclareSubr(KlSesReturn, "ses:return", 1);
    KlDeclareSubr(KlSesReturnError, "ses:error", NARY);

    /* variables */
    KlActiveMake("*ses:modules*", KlSesModulesGet, 0, 0);
    KlA_SesCurrentModule = KlIntern("*ses:current-module*");
    KlA_SesRetValue = KlIntern("ses::retvalue");
    KlSesUserErrorHandler = (KlO) KlK_nil;

    /* setting up execution mecanisms: path */
    KlA_Ses_loadpathname = KlIntern("*ses:load-pathname*");
    {
	char *s;
	if (!(s = (char *) getenv("SES_PATH")))
	    s = KlSesPath;
	KlAtomSetq(KlA_Ses_loadpathname, KlStringMake(s));
    }
    KlIncRef(KlS_PATH = KlStringMake("PATH"));

#ifdef ISM
    KlSes_addinput = KlISM_RegisterSESmainloop;
#else
#endif
}

#endif /* !KLSES */
