/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */
/*************\
*             *
* KLSESI      *
* DEFINITIONS *
*             *
\*************/

#ifndef INCLUDE_Kl_SESI_H
#define INCLUDE_Kl_SESI_H

#ifdef SHM
#include <sys/ipc.h>
#include <sys/shm.h>
#endif 

#include "kclist.h"
    
/* types */

KlTypedefListOf(struct _KlSesType *, KlListOfSesType);
KlTypedefListOf(KlAtom, KlListOfAtom);

typedef struct _KlSesModule {		/* the module descriptor */
    KlKLONE_HEADER;
    KlAtom name;			/* the name of the module */
    FILE *fd;				/* fd to write into main->ext */
    int pid;				/* pid of the extension */
    int id;				/* number of the extension < 256 */
    KlListOfAtom func_symbols;		/* list of KlSesFuncs names, freeform */
    KlListOfSesType ids;	        /* KlSesIds, indexed by ID - builtins */
    char ready;				/* inits done, ready to be used */
    char dead;				/* pid died */
    KlO ready_callback;			/* when ready, optional */
    /* socket-specific */
					/* --- output --- */
    uchar *outbuf;			/* buffer */
    uchar *outlimit;			/* its allocated size */
    uchar *outptr;			/* current (start) pointer in it */
}        *KlSesModule;

typedef struct _KlSesFunc {		/* a function (wrapper) of a module */
    KlKLONE_HEADER;
    KlSesModule module;			/* where does it belongs? */
    void *func;				/* ID of this function in module */
    int arity;				/* fixed arity */
    int return_type;		        /* can be None */
    unsigned int *argument_types;       /* arity types of arguments */
} *KlSesFunc;

typedef struct _KlSesType {		/* an opaque type (wrapper) */
    KlKLONE_HEADER;
    KlSesModule module;			/* where does it belongs? */
    int id;				/* its extension-side type ID */
    KlAtom name;			/* external name */
    void *close_callback;		/* closing callback ID, can be 0 */
				/* codec functions */
    int (*length_check) 
	_ANSI_ARGS_((KlO obj, KlSesModule module, unsigned int id, int pos));
    uchar *(*encode) 
	_ANSI_ARGS_((uchar *p, KlO obj, KlSesModule module, unsigned int id));
    uchar *(*decode) 
	_ANSI_ARGS_((uchar *p, KlO *objp, KlSesModule module, unsigned int id));
} *KlSesType;

typedef struct _KlSesId {		/* an opaque type (wrapper) instance */
    KlKLONE_HEADER;
    KlSesModule module;			/* where does it belongs? */
    KlSesType ID;			/* its extension-side type ID */
    void *data;				/* closure, opaque data */
} *KlSesId;

typedef struct _KlSesReqId {		/* opaque type to request ID */
    KlKLONE_HEADER;
    KlSesModule module;			/* where does it belongs? */
    KlAtom name;			/* setq value to (0 = temp store) */
    KlO value;				/* 0, then filled on reception */
    unsigned int valuetype;		/* KlSesType id expected */
    struct _KlSesReqId *previous;	/* linked list of pending requests */
} *KlSesReqId;

/* exported functions */

#define KlIsASesModule(obj)  ((obj)->type == KlSesModuleType)
#define KlMustBeSesModule(obj, n)  KlArgumentMustBe(obj, n, KlSesModuleType)

#define KlIsASesFunc(obj) ((obj)->type == KlSesFuncType)
#define KlMustBeSesFunc(obj, n) KlArgumentMustBe(obj, n, KlSesFuncType)

#define KlIsASesType(obj) ((obj)->type == KlSesTypeType)
#define KlMustBeSesType(obj, n) KlArgumentMustBe(obj, n, KlSesTypeType)

#define KlIsASesId(obj) ((obj)->type == KlSesIdType)
#define KlMustBeSesId(obj, n) KlArgumentMustBe(obj, n, KlSesIdType)

#define KlIsASesReqId(obj) ((obj)->type == KlSesReqIdType)
#define KlMustBeSesReqId(obj, n) KlArgumentMustBe(obj, n, KlSesReqIdType)

#define KlSesId2Module(n) ((KlSesModule) KlSesModules->list[n])

KlO KlSesModuleFree _ANSI_ARGS_((KlSesModule module));
void KlSesHandler _ANSI_ARGS_((void *closure, int source, unsigned long *id));
KlSesModule KlSesModuleFromName _ANSI_ARGS_((KlString extname));
char *KlSesModuleId2Name _ANSI_ARGS_((int id));
KlSesReqId KlSesReqIdMake _ANSI_ARGS_((KlSesModule module, KlAtom name,
				       unsigned int type));
void KlSesMainPreSelect();

/* methods */

/* objects */

EXT KlType KlSesModuleType;
EXT KlType KlSesFuncType;
EXT KlType KlSesTypeType;
EXT KlType KlSesIdType;
EXT KlType KlSesReqIdType;

/* the single pipe ext->main. main->ext is a field of modules */
EXT int KlSes_fd_main_r;		/* main read here */
EXT int KlSes_fd_main_w INIT(0);	/* exts write there */
EXT void (*KlSes_addinput)() INIT(0);	/* application-supplied hook to add
					   input fd (only arg) to mainloop,
					   which must call KlSesHandler() */
#define KlSesHandlerBufferSize 1024	/* size of fifo buffer */

#ifdef SHM
/* shared memory. infos can be get at runtime via the ipcs unix program */
EXT int KlSes_shm_id INIT(0);		/* id of shared segment */
EXT char* KlSes_shm_ptr;		/* start of zone in memory */
EXT int KlSes_shm_size INIT(SHMMAX);	/* size in bytes */

/* semaphores, to gain write access to the shared memory portion */
#define KlSes_sem_number 1		/* number of semaphores in the set */
EXT int KlSes_sem_id INIT(0);		/* number of semaphores */
#endif /* !SHM */

/* sockets connection-type specific parameters */
					/* --- input --- */
EXT uchar *KlSes_soc_inbuf;		/* buffer */
EXT uchar *KlSes_soc_inlimit;		/* its allocated size */
EXT uchar *KlSes_soc_inend;		/* end marker in it */
EXT uchar *KlSes_soc_inptr;		/* current (start) pointer in it */

/* modules */
EXT KlList KlSesModules;		/* list of all modules */
EXT KlList KlSesListOfBuiltinTypes;

EXT KlSesReqId KlSesReqIdLast INIT(0);	/* ptr to last waitable request */

/************************************************************** stream pipes */
#ifdef SVR4
/* in SVR4, normal pipes are bidirectional anyways */
#define KlStreamPipe(fd) pipe(fd) 
#else /* !SVR4 */
/* else we will have to emulate it via socketpair */
extern int KlStreamPipe();
#endif /* !SVR4 */

#endif					/* INCLUDE_Kl_SESI_H */
