/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */
/*****************************************************************************\
*                                                                             *
* 			     WPPB parser routines                             *
*                                                                             *
\*****************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>

#include "EXTERN.h"
#include "klone.h"
#include "kl_number.h"
#include "kl_atom.h"
#include "kl_list.h"
#include "kl_coll.h"
#include "kl_string.h"
#include "kl_stream.h"
#include "kl_func.h"
#include "klparser.h"

#ifdef WPPB

#ifdef KOALA
#ifdef ISM
#undef ISM
#endif
#endif /* KOALA */

#include "INTERN.h"
#include "klwppb.h"

KlO KlWPPBWriteNull();
extern KlActive KlA_stdout;
#ifdef ISM
int (*SmlCheckPointer) (void) = NULL;
KlType SmlClassMExprHide, SmlClassExprHide, SmlClassFExprHide;
#define  KlWPPBVersionOffset 36		/* offset of version byte in files */
#else
#define  KlWPPBVersionOffset 12		/* offset of version byte in files */
#endif

extern KlActive KlA_stdout;

/**************************************************************************\
* 				   Contexts                                *
\**************************************************************************/

KlWPPBContextSave(context, filename, buffer
#ifdef DEBUG
		  , last
#endif 
		  )
    KlWPPBContext context;
    char *filename;
    unsigned char *buffer;
#ifdef DEBUG
    unsigned char *last;
#endif
{
    context->old = KlWPPBCurrentContext;
    context->filename = filename;
    context->buffer = buffer;
    context->dicts = (KlO **) Malloc(KLSO * KlWPPBCurrentContextDefDicts);
    context->version = buffer[KlWPPBVersionOffset];
    context->parse_disabled = KlStdyy->parse_disabled; /* save old value */
    KlStdyy->parse_disabled = 1;
#ifdef DEBUG
    context->dictsizes = (int *) Malloc(KLSO * KlWPPBCurrentContextDefDicts);
    context->ndicts = 0;
    context->last = last;
#endif
    context->dictlimit = KlWPPBCurrentContextDefDicts;
    KlWPPBCurrentContext = context;
}

KlWPPBContextRestore(context)
    KlWPPBContext context;
{
    KlStdyy->parse_disabled = context->parse_disabled;
    Free(context->dicts);
#ifdef DEBUG
    Free(context->dictsizes);
#endif
    KlWPPBCurrentContext = context->old;
}

KlWPPBContextAddDict(context, size, list, id)
    KlWPPBContext context;
    int size;
    KlO *list;
    unsigned int id;
{
    if (id >= context->dictlimit) {
	context->dictlimit = (context->dictlimit + 1) * 2 - 1;
	context->dicts = (KlO **) Realloc(context->dicts, KLSO *
					      context->dictlimit);
#ifdef DEBUG
	context->dictsizes = (int *) Realloc(context->dictsizes, KLSO *
					      context->dictlimit);
#endif
    }
#ifdef DEBUG
    context->ndicts = Max(id+1, context->ndicts);
    context->dictsizes[id] = size;
#endif
    context->dicts[id] = list;    
}

/**************************************************************************\
* 				   parsing                                 *
\**************************************************************************/

#ifdef ISM
int KlWPPBDecrypt(unsigned char *buffer, int size)
{
  int i, j = 0;
  int key, position = 37;

  for (i = 0; i < size; i++)
    if (i == position) {
      key = KlIntBE32Read(&buffer[i]);
      position = key ^ KlIntBE32Read(&buffer[i+4]);
      i += 7;
    } else {
      buffer[j] = buffer[i];
      j++;
    }
  return j;
}
#endif /* ISM */

/* Load a (pre-opened) WPPB file
 * returns 1 = success, 0 = error
 */

int
KlLoadWPPBFile(filename, fp)
    char *filename;
    FILE *fp;
{
    struct stat buf;
    int status = 1;
    int version;
    int stackptr;

    if (!fstat(KlFp2Fd(fp), &buf) && (buf.st_mode & S_IFREG)) {
	/* gobble whole file into a buffer */
	KlGCMark();
	KlString string = KlStringNMake(buf.st_size);
	unsigned char *buffer = (unsigned char *) string->string;
	int size = fread(buffer, 1, buf.st_size, fp);
	unsigned char *p = buffer, *last = p + size;
	int i;
	struct _KlWPPBContext context;

	fclose(fp);			/* buffer is gobbled, file not needed */

#ifdef ISM
        /* there, we must decrypt the buffer in place and return size, which 
	   must be lesser or equal to original size */
	size = KlWPPBDecrypt(buffer, size);
	last = buffer + size;
#endif

#ifdef DEBUG
	KlWPPBContextSave(&context, filename, buffer, last);
#else
	KlWPPBContextSave(&context, filename, buffer);
#endif

	buffer[size] = '\0';
	if (size < 14) {		/* verify that header is there */
	    if (size == 0)
		KlLoadWPPBError(p, "empty file!%s", (long) "");
	    else
		KlLoadWPPBError(p, "file too short to be a .wlp (%d bytes)", 
				(long) size);
	    goto error;
	}
	/* check header info */
	if (KlWPPB_MAGIC != (i = KlIntBE32Read(p))) {
	    KlLoadWPPBError(p, "bad magic number %d", (long) i);
	    goto error;
	}
	p += 4;
#ifdef ISM
	p += 20;
	if (SmlCheckPointer != NULL)
	    if ((*SmlCheckPointer)() < (i = KlIntBE32Read(p))) 
		fprintf (stderr,
			 "Warning: File %s have bad Sml Version Number %d\n",
			 KlWPPBCurrentContext->filename, (long) i);
	p += 4;
#endif
	if (size != KlIntBE32Read(p) || size != buf.st_size) {
	    KlLoadWPPBError(p, "bad size %d", (long) size);
	    goto error;
	}
#ifdef DEBUG
	/* verify checksum */
	p += 4;
	{
	    int checksum;		/* check=0:file has not be checksumed */
	    if (checksum = KlIntBE32Read(p)) {
		/* remove checksum form count */
		int comp_checksum = - (*p + p[1] + p[2] + p[3]);
		unsigned char *q;
		for (q = buffer; q < last; q++)
		    comp_checksum += *q;
		comp_checksum += size;
		if (comp_checksum != checksum)
		    KlLoadWPPBError(p, "bad checksum, computed %d", 
				    (long) comp_checksum);
	    }
	}
	p += 5;				/* version read by KlWPPBContextSave */
#else					/* !DEBUG */
	p += 9;				/* checksum not checked, too slow */
#endif					/* !DEBUG */
	/* version read by KlWPPBContextSave */
	/* main loop: read items */
	stackptr = KlStackPtr;
	KlStackFramePush(1, &KlA_star_package, NIL, NIL);
	KlUnwindProtectStatement(status = KlWPPBParseBuffer(p, last),
				 KlApplyUnary, KlWPPBContextRestore, &context);
	KlStackFramePopNormal(stackptr);
	KlGC();
	return status;
    error:
	KlWPPBContextRestore(&context);
	KlGC();
	return 0;
    } else {
	return 0;
    }
}

int					/* 1 = success, 0 = error */
KlWPPBParseBuffer(p, last)
    unsigned char *p, *last;
{
    KlO obj;				/* discarded at toplevel */
    jmp_buf jbuf;
    int res;

    if (res = setjmp(KlWPPBCurrentContext->jbuf)) { /* EOF */
	return res - 1;			/* longjmp 2 = success, 1 = error */
    } else {				/* main loop */
	for (;;) {
#ifdef DEBUG
	    if (p >= last) {
		KlLoadWPPBError(p, "reading past buffer end%s!", "");
		return 1;
	    }
	    if (*p >= KlWPPBTableSize || !KlWPPBTable[*p]) {
		KlLoadWPPBError(p, "bad item code %d", (long) *p);
		return 0;
	    }
#endif /* DEBUG */
	    if (*p) {
		p = (*(KlWPPBTable[*p]))(p+1, &obj);
	    } else {			/* optim. for the REF frequent case */
		p += 2;
#ifdef DEBUG
		if (*(p-1) >= KlWPPBCurrentContext->ndicts)
		    KlLoadWPPBError(p, "reference to unknown dict %d\n", 
				    (long) *(p-1));
		if (*p >= KlWPPBCurrentContext->dictsizes[*(p-1)])
		    KlLoadWPPBError(p, "reference out of dict %d\n", 
				    (long) *(p-1));
#endif /* DEBUG */
		obj = KlWPPBCurrentContext->dicts[*(p-1)][*p];
	    }
	}
    }
}

#ifdef DEBUG
unsigned char *
KlWPPBParse(p, objp) 
    unsigned char *p;
    KlO *objp;
{
    static unsigned char *op;
    static  KlO oldobj;
    unsigned char *res;

    KlWPPBReadFunction f = KlWPPBTable[*p];
    res = (*f) ((p)+1, (objp));
    if (!(*objp) && (*p != KlWPPBType_Null))
	KlLoadWPPBError(p, "NULL pointer returned by a %s", 
			KlWPPBTableNames[*p]);
    op = p, oldobj = *objp;
    return res;
}
#endif /* DEBUG */
/**************************************************************************\
* 				    error                                  *
\**************************************************************************/

KlLoadWPPBError(ptr, pattern, data)
    unsigned char *ptr;
    char *pattern, *data;
{
    char temp[KlMAX_TEMP_STRING_SIZE];
    sprintf(temp, pattern, data);
    KlError3(KlE_BAD_WPPB, KlStringMake(KlWPPBCurrentContext->filename),
	     KlNumberMake(ptr - KlWPPBCurrentContext->buffer),
	     KlStringMake(temp));
}

/**************************************************************************\
* 			       table management                            *
\**************************************************************************/
/* KlDeclareWPPBRead
 * adds a new read function
 * Warning: name is not copied, just pointed to, so take care not to free
 * or modify name after calling this function.
 */

KlDeclareWPPBRead(desired_typenumber, name, parse_function)
    int desired_typenumber;
    char *name;
    KlWPPBReadFunction parse_function;
{
    if (desired_typenumber >= KlWPPBTableSize) {
	int i = KlWPPBTableSize;
	KlWPPBTableSize = desired_typenumber + 1;
	KlWPPBTable = (KlWPPBReadFunction *) Realloc(KlWPPBTable,
					   KLSO * KlWPPBTableSize);
	KlWPPBTableNames = (char **) Realloc(KlWPPBTableNames,
						KLSO * KlWPPBTableSize);
#ifdef DEBUG
	while (i <= desired_typenumber) {
	    KlWPPBTable[i++] = 0;	/* unused slots are filled with 0 */
	}
#endif /* DEBUG */
    }
#ifdef DEBUG
    if (KlWPPBTable[desired_typenumber])
	KlFatalError(0, "Overwriting an existing WPPB type!");
#endif /* DEBUG */
    KlWPPBTable[desired_typenumber] = parse_function;
    KlWPPBTableNames[desired_typenumber] = name;
}

/**************************************************************************\
* 			 Some built-in read functions                      *
\**************************************************************************/
/* A template for such read functions is:
unsigned char *
KlWPPBRead(p, objp)
    unsigned char *p;
    KlO *objp;
{
    return p;
}
 */
/******************************************************** NOT IMPLEMENTED */
unsigned char *
KlWPPBReadNotImplemented(p, objp)
    unsigned char *p;
    KlO *objp;
{
    KlLoadWPPBError(p, "WPPB unknown type: %s", KlWPPBTableNames[*(p-1)]);
    longjmp(KlWPPBCurrentContext->jbuf, 1);
} 

/**************************************************************** TRUE & NIL */
unsigned char *
KlWPPBReadNil(p, objp)
    unsigned char *p;
    KlO *objp;
{
    *objp = NIL;
    return p;
}

unsigned char *
KlWPPBReadTrue(p, objp)
    unsigned char *p;
    KlO *objp;
{
    *objp = TRU;
    return p;
}

/************************************************************** NULL pointer */

unsigned char *
KlWPPBReadNull(p, objp)
    unsigned char *p;
    KlO *objp;
{
    *objp = 0;
    return p;
}

/*********************************************************************** EOF */

unsigned char *
KlWPPBReadEOF(p, objp)
    unsigned char *p;
    KlO *objp;
{
    longjmp(KlWPPBCurrentContext->jbuf, 2);
}


/********************************************************************* XTYPE */
/* XTYPE: for types > 255
 */
unsigned char *
KlWPPBReadXType(p, objp)
    unsigned char *p;
    KlO *objp;
{
    unsigned int code = KlIntBE24Read(p);
#ifdef DEBUG
    if (p > KlWPPBCurrentContext->last) 
	KlLoadWPPBError(p, "reading past end of byffer by %d bytes",
			(long) (p - KlWPPBCurrentContext->last));
    if (KlWPPBTableSize < 256 || code < 256)
	KlLoadWPPBError(p-1, "bad item code 0%s", "");
    if (code >= KlWPPBTableSize || !KlWPPBTable[code])
	KlLoadWPPBError(p, "bad extended item code %d", (long) code);
#endif /* DEBUG */
    return (*(KlWPPBTable[code]))(p+3, objp);
}

/*********************************************************************** RAW */
/* RAW: normal parsing of a string
 *     <N>          4 byte           length of the string, null excluded
 *     <string>     N bytes          string to be parsed by Klone parser, NULL
 *                                   included
 */
unsigned char *
KlWPPBReadRaw(p, objp)
    unsigned char *p;
    KlO *objp;
{
    int length = KlIntBE32Read(p);
    *objp = KlParseString(length, p+4, 0);
    return p + 5 + length;
}

/*********************************************************************** SET */
/* A convenience function to set variable to value, both quoted:
 * <variable>   1 item
 * <value>      1 item
 */

unsigned char *
KlWPPBReadSet(p, objp)
    unsigned char *p;
    KlO *objp;
{
    KlO var, val;
    p = KlWPPBParse(p, &var);
    p = KlWPPBParse(p, &val);
    *objp = KlSend_setq(var, val);
    return p;
}

unsigned char *
KlWPPBReadSetFunc(p, objp)
    unsigned char *p;
    KlO *objp;
{
    KlO var, val;
    p = KlWPPBParse(p, &var);
    p = KlWPPBParse(p, &val);
    *objp = KlAddDefun(var, val);
    return p;
}

unsigned char *
KlWPPBReadSetUnparse(p, objp)
    unsigned char *p;
    KlO *objp;
{
    KlO var, val;
    KlCollection col = KlCollectionMake();
    p = KlWPPBParse(p, &var);
    p = KlWPPBParse(p, &val);
    KlCollectionAdd(col, KlA_setq);
    KlCollectionAdd(col, var);
    KlCollectionAdd(col, val);
    *objp = (KlO) col;
    return p;
}

/********************************************************************** EVAL */
/* Items read in WPPB files are just read, not evaluated. Eval is still needed
 * for constructs that could not be totally pre-evaluated at preparse time and
 * has one argument, the form to eval
 * <value>      1 item
 */

unsigned char *
KlWPPBReadEval(p, objp)
    unsigned char *p;
    KlO *objp;
{
    KlO val;
    p = KlWPPBParse(p, &val);
    *objp = KlSend_eval(val);
    return p;
}

unsigned char *
KlWPPBReadEvalUnparse(p, objp)
    unsigned char *p;
    KlO *objp;
{
    KlO val;
    KlCollection col = KlCollectionMake();
    p = KlWPPBParse(p, &val);
    if (val->type == KlListType) {	/* common case: {} == EVAL+LIST */
	int i;				/* convert list to col */
	for (i = 0; i < ((KlList) val)->size; i++) {
	    KlCollectionAdd(col, ((KlList) val)->list[i]);
	}
    } else {				/* dont know what it is, play it safe */
	KlCollectionAdd(col, KlA_identity);
	KlCollectionAdd(col, val);
    }
    *objp = (KlO) col;
    return p;
}

/************************************************************** WPPD_INCLUDE */
/* NOTE: loading the dict file MUST set its name to its value (list)
 */

extern KlString KlFileInPath();

unsigned char *
KlWPPBReadWPPDInclude(p, objp)
    unsigned char *p;
    KlO *objp;
{
    int id = KlIntBE16Read(p);
    int checksum;
    KlAtom name;

    p = KlWPPBParse(p + 2, (KlO *) &name);
    checksum = KlIntBE32Read(p);
    if (name->c_val == KlUndef || !KlIsAList(name->c_val)) { /* load it */
	KlString actual_pathname =
	    KlStringNMake(strlen(name->p_name)
			  + KlWPPD_EXT_len - KlWPPD_PACKAGE_LEN);
	char *q = stpcpy(actual_pathname->string, 
			 name->p_name + KlWPPD_PACKAGE_LEN);
	char temp_filename[KlMAX_TEMP_STRING_SIZE];
	FILE *fp = 0;

	stpcpy(q, KlWPPD_EXT);
        /* append atom to *wppd:list* list of loaded dicts */
	if ((actual_pathname = 
	     KlFileInPath(actual_pathname->string,
			  temp_filename, &fp))
	     && KlLoadWPPBFile(actual_pathname->string, fp)
	     && KlIsAList(name->c_val)
	     && ((KlList) (name->c_val))->list[0]
	     && KlIsAnInteger(((KlList) (name->c_val))->list[0])
	    ) {
	    KlList l = ((KlList) (KlA_WPPD_list->c_val));
	    int i;
	    if (NIL == (KlO) l)
		KlAtomSetq(KlA_WPPD_list, l = KlListNMake(0));
	    for (i = 0; i < l->size; i++)
		if (l->list[i] == (KlO) name)
		    goto done;
	    KlListAppend(l, name);
	    KlWPPBContextAddDict(KlWPPBCurrentContext, 
				 ((KlList)(name->c_val))->size, 
				 ((KlList)(name->c_val))->list,
				 id);
	} else {
	    KlError(KlE_CANNOT_LOAD_FILE, actual_pathname);
	}
    } else {				/* already loaded */
	KlWPPBContextAddDict(KlWPPBCurrentContext, 
			     ((KlList)(name->c_val))->size, 
			     ((KlList)(name->c_val))->list,
			     id);
    }
  done:
    return p + 4;
}

/************************************************************** WPPD_HERE */

unsigned char *
KlWPPBReadWPPDHere(p, objp)
    unsigned char *p;
    KlO *objp;
{
    int id = KlIntBE16Read(p);
    int size = KlIntBE16Read(p+2);
    KlList templist = KlListNMake(size);
    KlO *dict = templist->list, *last = dict + size;
    
    templist->size = -1;		/* no GC on elements */
    KlWPPBContextAddDict(KlWPPBCurrentContext, size, dict, id);
    p += 4;
    while (dict < last)
	p = KlWPPBParse(p, dict++);
    return p;
}

/*************************************************************** WPPD_REF */
/* short form */
unsigned char *
KlWPPBReadWPPDRef(p, objp)
    unsigned char *p;
    KlO *objp;
{
#ifdef DEBUG
    if (*p >= KlWPPBCurrentContext->ndicts)
	KlLoadWPPBError(p, "reference to unknown dict %d\n", (long) *p);
    if (*(p+1) >= KlWPPBCurrentContext->dictsizes[*p])
	KlLoadWPPBError(p, "reference out of dict %d\n", (long) *p);
#endif
    *objp = KlWPPBCurrentContext->dicts[*p][*(p+1)];
    return p+2;
}

/* long form */
unsigned char *
KlWPPBReadWPPDRefL(p, objp)
    unsigned char *p;
    KlO *objp;
{
#ifdef DEBUG
    unsigned int d = (unsigned int) KlIntBE16Read(p);
    if (d >= KlWPPBCurrentContext->ndicts)
	KlLoadWPPBError(p, "long reference to unknown dict %d\n", (long) d);
    if ((unsigned int) KlIntBE16Read(p+2) >= KlWPPBCurrentContext->dictsizes[d])
	KlLoadWPPBError(p, "long reference out of dict %d\n", (long) d);
#endif
    *objp = KlWPPBCurrentContext->dicts[(unsigned int) KlIntBE16Read(p)]
	[KlIntBE16Read(p+2)];
    return p+4;
}

/******************************************************************* ATOM */
/* ATOM: a Klone atom
 *     <N>          l bytes          length of the string, null excluded
 *     <hash>       4 bytes          hash value
 *     <string>     N bytes          name of the atom (string), NULL excluded
 * l is 1 for ATOM, 4 for AtomL
 */

#ifdef ISM
static int crypt_key = 7;

static void SmlDecrypt(int len, char *buffer)
{
  int i;
  int car, key = crypt_key;
  for (i=0; i<len; i++) {
    car = buffer[i];
    buffer[i] ^= key;
    key = car;
  }
}
#endif /* ISM */

unsigned char *
KlWPPBReadAtom(p, objp)
    unsigned char *p;
    KlO *objp;
{
    unsigned int hash = KlIntBE32Read(p+1);
#ifdef ISM
    SmlDecrypt(*p, p + 5);
#endif /* ISM */
    *objp = (KlO) KlInternBytesHash(*p, p+5, hash);
    return p + 5 + *p;
}

unsigned char *
KlWPPBReadAtomL(p, objp)
    unsigned char *p;
    KlO *objp;
{
    int l = KlIntBE32Read(p);
    unsigned int hash = KlIntBE32Read(p+1);
#ifdef ISM
    SmlDecrypt(*p, p + 8);
#endif /* ISM */
    *objp = (KlO) KlInternBytesHash(l, p+8, hash);
    return p + 8 + l;
}

/******************************************************************* STRING */
/* STRING: a Klone string
 *     <N>          l bytes          length of the string, null excluded
 *     <string>     N bytes          contents of the string, NULL excluded
 * l is 1 for String, 4 for StringL
 */

unsigned char *
KlWPPBReadString(p, objp)
    unsigned char *p;
    KlO *objp;
{
#ifdef ISM
    SmlDecrypt(*p, p + 1);
#endif /* ISM */
    *objp = (KlO) KlStringMakeFromBytes(*p, p+1);
    return p + 1 + *p;
}

unsigned char *
KlWPPBReadStringL(p, objp)
    unsigned char *p;
    KlO *objp;
{
    int l = KlIntBE32Read(p);
#ifdef ISM
    SmlDecrypt(l, p + 4);
#endif /* ISM */
    *objp = (KlO) KlStringMakeFromBytes(l, p+4);
    return p + 4 + l;
}

/******************************************************************* LIST */

unsigned char *
KlWPPBReadListAny(l, p, objp)
    int l;
    unsigned char *p;
    KlO *objp;
{
    KlO *po, *plast, elt;
    *objp = (KlO) KlListNMake(l);
    po = ((KlList)(*objp))->list;
    plast = po + l; 
    while (po < plast) {
	p = KlWPPBParse(p, &elt);
	KlIncRef(*po++ = elt);
    }    
    return p;
}

unsigned char *
KlWPPBReadList(p, objp)
    unsigned char *p;
    KlO *objp;
{
    return KlWPPBReadListAny(*p, p+1, objp);
}

unsigned char *
KlWPPBReadListM(p, objp)
    unsigned char *p;
    KlO *objp;
{
    int type = *(p+1);
    p = KlWPPBReadListAny(*p, p+2, objp);
    (*objp)->type = KlWPPBReadListTypes[type];
    return p;
}

unsigned char *
KlWPPBReadListML(p, objp)
    unsigned char *p;
    KlO *objp;
{
    int type = *(p+4);
    p = KlWPPBReadListAny(KlIntBE32Read(p), p+5, objp);
    (*objp)->type = KlWPPBReadListTypes[type];
    return p;
}

/************************************************************ QUOTED EXPR */

unsigned char *
KlWPPBReadQuotedExpr(p, objp)
    unsigned char *p;
    KlO *objp;
{
    int type = *p;
    KlO expr;
    p = KlWPPBParse(p+1, &expr);
    *objp = (KlO) KlAnyQuotedExprMake(KlWPPBReadQuotedExprSymbols[type],
				      expr,
				      KlWPPBReadQuotedExprEvals[type]);
    return p;
}

/**************************************************************** NUMBERS */

unsigned char *
KlWPPBReadIntS(p, objp)
    unsigned char *p;
    KlO *objp;
{
    *objp = (KlO) KlBuiltInNumbers[*p];
    return p+1;
}

unsigned char *
KlWPPBReadIntP(p, objp)
    unsigned char *p;
    KlO *objp;
{
    *objp = (KlO) KlNumberMake((Int) KlIntBE32Read(p));
    return p+4;
}

unsigned char *
KlWPPBReadIntN(p, objp)
    unsigned char *p;
    KlO *objp;
{
    *objp = (KlO) KlNumberMake(- (Int) KlIntBE32Read(p));
    return p+4;
}

#ifdef UNSIGNED
unsigned char *
KlWPPBReadIntU(p, objp)
    unsigned char *p;
    KlO *objp;
{
    *objp = (KlO) KlUnsignedMake((UInt) KlIntBE32Read(p));
    return p+4;
}
#endif /* UNSIGNED */

unsigned char *
KlWPPBReadReal(p, objp)
    unsigned char *p;
    KlO *objp;
{
    double x;
    KlParseReal(p+1, x);
    *objp = (KlO) KlRealMake(x);
    return p + *p + 2;
}

/***************************************************************** FUNCTIONS */
/* Exprs and siblings in fact
 */

unsigned char *
KlWPPBReadExpr(p, objp)
    unsigned char *p;
    KlO *objp;
{
    int n;
    KlO klo;
    KlAtom *pp, *plast;
    KlExpr obj = (KlExpr) KlOMake(KlExprType);
    *objp = (KlO) obj;

#ifdef ISM
    obj->type = SmlClassExprHide;
    if (*p == 1) obj->type = SmlClassFExprHide;
    else if (2 == *p) obj->type = SmlClassMExprHide;
#else
    if (*p == 1) obj->type = KlFExprType;
    else if (2 == *p) obj->type = KlMExprType;
#endif /* ISM */
    p++;
    obj->arity = *(p++);
    if (n = obj->arity) {
	obj->parameters = pp = (KlAtom *) Malloc(KLSO * n);
	for (plast = pp + n; pp < plast;) {
	    p = KlWPPBParse(p, &klo);
	    KlIncRef(*pp++ = (KlAtom) klo);
	}
    } else {
	obj->parameters = 0;
    }
    obj->body_size = n = KlIntBE16Read(p);
    p += 2;
    pp = n ? (KlAtom *) Malloc(KLSO * n) : 0;
    obj->body = (KlO *) pp;
    for (plast = pp + n; pp < plast;) {
	p = KlWPPBParse(p, &klo);
	KlIncRef(*pp++ = (KlAtom) klo);
    }
    if (*p++) {
	KlLambdaList ll =
	    obj->lambdalist =
	    (KlLambdaList) Malloc(sizeof(struct _KlLambdaList));
	ll->nfixed = *p++;
	ll->variable_arity = *p++;
	if (n = *p++) {
	    pp = (KlAtom *) Malloc(KLSO * (n+1));
	    ll->optionals = (KlO *) pp;
	    for (plast =  pp + n; pp < plast;) {
		p = KlWPPBParse(p, &klo);
		KlIncRef(*pp++ = (KlAtom) klo);
	    }
	    *pp = 0;
	} else {
	    ll->optionals = 0;
	}
	if (*p++) {
	    p = KlWPPBParse(p, &klo), KlIncRef(ll->rest = (KlAtom) klo);
	} else {
	    ll->rest = 0;
	}
	if (*p++) {
	    p = KlWPPBParse(p, &klo), KlIncRef(ll->whole = (KlAtom) klo);
	} else {
	    ll->whole = 0;
	}
	if (n = *p++) {
	    KlKeyDecls pk = ll->key = (KlKeyDecls)
		Malloc(sizeof(struct _KlKeyDecls) * (n+1));
	    KlKeyDecls pklast = pk + n;
	    while (pk < pklast) {
		p = KlWPPBParse(p, &klo), KlIncRef(pk->key = (KlKeyword) klo);
		p = KlWPPBParse(p, &klo), KlIncRef((pk++)->init = klo);
	    }
	    pk->key = 0;
	} else {
	    ll->key = 0;
	}
	if (n = *p++) {
	    pp = (KlAtom *) Malloc(KLSO * (n+1));
	    ll->aux = (KlO *) pp;
	    for (plast = pp + n; pp < plast;) {
		p = KlWPPBParse(p, &klo);
		if (klo)		/* may be NULL ? */
		    KlIncRef(*pp++ = (KlAtom) klo);
		else
		    *pp++ = 0;
	    }
	    *pp = 0;
	} else {
	    ll->aux = 0;
	}
    } else {
	obj->lambdalist = 0;
    }    
#ifdef KLPROFILER
    KlProfilerInitFunc(obj);
#endif
    
    return p;
}

#ifndef ISM			/* ? requested by Olivier... */
/*****************************************************************************\
* 			     Some Write Routines                              *
\*****************************************************************************/
/* not many, just for convenience since for writing no performance is needed
 */

/* THE main dispatch function is: (wppb:write obj stream)
 * here is a C function to call it
 * must return t if OK
 */

KlO
KlWPPBWrite(obj, stream)
    KlO obj;
    KlStream stream;
{
    KlMustBeStream(stream, 1);
    return obj ? KlApply2(KlA_WPPB_write, obj, stream)
	: KlWPPBWriteNull(obj, stream);
}

KlO
KlWPPBWriteNull(obj, stream)
    KlO obj;				/* 0 */
    KlStream stream;
{
    KlSPutc(KlWPPBType_Null, stream);
    return TRU;
}

/* (wppb:write-expr func stream) 
 * WARNING: if () is returned, the write was aborted, you MUST step back and 
 * write the func as a RAW item (normally t is retuned if OK)
 */

KlO
KlWPPBWriteExpr(obj, stream)
    KlExpr obj;
    KlStream stream;
{
#define PutByte(b) if (KlWriteByte(byte = (b), stream)) return NIL
    int byte, i;
    if (!KlIsALambda(obj))
	KlMustBeExpr(obj, 0);
    KlMustBeStream(stream, 1);
    
    KlWriteByte(KlWPPBType_Expr, stream); /* header */
    if (obj->type == KlFExprType) { 
	KlSPutc(1, stream);
    } else if (obj->type == KlMExprType) { 
	KlSPutc(2, stream);
    } else {
	KlSPutc(0, stream);
    }

    PutByte(obj->arity);		/* base args */
    for (i = 0; i < obj->arity; i++)
	KlWPPBWrite(obj->parameters[i], stream);
    PutByte((obj->body_size) >> 8);	/* body */
    PutByte((obj->body_size) & 0xff);
    for (i = 0; i < obj->body_size; i++)
	KlWPPBWrite(obj->body[i], stream);
    if (obj->lambdalist) {		/* fancy args &... */
	KlSPutc(1, stream);
	PutByte(obj->lambdalist->nfixed);
	PutByte(obj->lambdalist->variable_arity);
	PutByte(KlCountItems(obj->lambdalist->optionals, 1));
	for (i = 0; i < byte; i++)
	    KlWPPBWrite(obj->lambdalist->optionals[i], stream);
	if (obj->lambdalist->rest) {
	    KlSPutc(1, stream);
	    KlWPPBWrite(obj->lambdalist->rest, stream);
	} else {
	    KlSPutc(0, stream);
	}
	if (obj->lambdalist->whole) {
	    KlSPutc(1, stream);
	    KlWPPBWrite(obj->lambdalist->whole, stream);
	} else {
	    KlSPutc(0, stream);
	}
	PutByte(KlCountItems(obj->lambdalist->key, 
			     sizeof(struct _KlKeyDecls) / KLSO));
	for (i = 0; i < byte; i++) {
	    KlWPPBWrite(obj->lambdalist->key[i].key, stream);
	    KlWPPBWrite(obj->lambdalist->key[i].init, stream);
	}
	PutByte(KlCountItems(obj->lambdalist->aux, 1));
	for (i = 0; i < byte; i++)
	    KlWPPBWrite(obj->lambdalist->aux[i], stream);
    } else {
	KlSPutc(0, stream);
    }
    return TRU;
}

/* write a byte but aborts if number dont fit in a byte */
int
KlWriteByte(b, stream)
    unsigned int b;
    KlStream stream;
{
    return (b > 255) ? 1 : (KlSPutc(b, stream), 0);
}

/* count null-terminated items */
int
KlCountItems(p, size)
    KlO *p;
    int size;				/* in KLSO increments */
{
    int i;
    if (!p) return 0;
    for (i = 0; *p; i++) p += size;
    return i;
}

#endif /* ISM */

/*****************************************************************************\
* 				  deparsing                                   *
\*****************************************************************************/
/* deparsing is like parsing, but with more checks (speed is not important)
 * and instead of evaluating on the fly the read items, just dump them
 * into the output file
 * For modularity, we just use the parse routines, modifying only the 
 * side-effects ones such as EVAL and SETQ
 */

void KlWPPBUnparseBufferRestore();
struct KlWPPBTable_mod {
    int index;
    KlWPPBReadFunction func;
};
/* Unparse a (pre-opened) WPPB file into a (pre-opened) text file
 * formats optionally with the given pretty-printer
 * returns 1 = success, 0 = error
 */

int
KlUnparseWPPBFile(filename, fp, out, verbose, pp)
    char *filename;			/* name of input file */
    FILE *fp;				/* opened input file (fp) */
    KlStream *out;			/* klone stream, not fp */
    int verbose;			/* flag */
    KlO pp;				/* KlO or NULL=raw */
{
    struct stat buf;
    int status = 1;
    int version;
    int stackptr;
    int checksum = 0;

    if (!fstat(KlFp2Fd(fp), &buf) && (buf.st_mode & S_IFREG)) {
	/* gobble whole file into a buffer */
	KlGCMark();
	KlString string = KlStringNMake(buf.st_size);
	unsigned char *buffer = (unsigned char *) string->string;
	int size = fread(buffer, 1, buf.st_size, fp);
	unsigned char *p = buffer, *last = p + size;
	int i;
	struct _KlWPPBContext context;

	fclose(fp);			/* buffer is gobbled, file not needed */

#ifdef ISM
        /* there, we must decrypt the buffer in place and return size, which 
         must be lesser or equal to original size */
	size = KlWPPBDecrypt(buffer, size);
	last = buffer + size;
	buf.st_size = size;
#endif

#ifdef DEBUG
	KlWPPBContextSave(&context, filename, buffer, last);
#else
	KlWPPBContextSave(&context, filename, buffer);
#endif

	buffer[size] = '\0';
	if (size < (KlWPPBVersionOffset + 2)) { /* verify that header is there */
	    if (size == 0)
		KlLoadWPPBError(p, "empty file!%s", (long) "");
	    else
		KlLoadWPPBError(p, "file too short to be a .wlp (%d bytes)", 
				(long) size);
	    goto error;
	}
	/* check header info */
	if (KlWPPB_MAGIC != (i = KlIntBE32Read(p))) {
	    KlLoadWPPBError(p, "bad magic number %d", (long) i);
	    goto error;
	}
	p += 4;
	if (size != KlIntBE32Read(p) || size != buf.st_size) {
	    KlLoadWPPBError(p, "bad size %d", (long) size);
	    goto error;
	}
	/* here we check the checksum, we have time */
	p += 4;
	if (checksum = KlIntBE32Read(p)) { /* check=0 means file non checked */
	    /* remove checksum form count */
	    int comp_checksum = - (*p + p[1] + p[2] + p[3]);
	    unsigned char *q;
	    for (q = buffer; q < last; q++)
		comp_checksum += *q;
	    comp_checksum += size;
	    if (comp_checksum != checksum)
		KlLoadWPPBError(p, "bad checksum, computed %d", 
				(long) comp_checksum);
	}
	p += 5;				/* version read by KlWPPBContextSave */
	/* main loop: read items */
	stackptr = KlStackPtr;
	KlStackFramePush(1, &KlA_star_package, NIL, NIL);
	{
	    /* modify parse table not to do side-effects */
	    /* tracks mods to be undone by KlWPPBUnparseBufferRestore */
	    struct KlWPPBTable_mod KlWPPBTable_mods[3];
	    KlWPPBTable_mods[0].index = KlWPPBType_Eval;
	    KlWPPBTable_mods[0].func = KlWPPBReadEval;
	    KlWPPBTable_mods[1].index = KlWPPBType_Set;
	    KlWPPBTable_mods[1].func = KlWPPBReadSet;
	    KlWPPBTable_mods[2].func = 0;

	    KlWPPBTable[KlWPPBType_Eval] = KlWPPBReadEvalUnparse;
	    KlWPPBTable[KlWPPBType_Set] = KlWPPBReadSetUnparse;

	    KlUnwindProtectStatement
		(status = KlWPPBUnparseBuffer(p, last, out, verbose, pp),
		 KlWPPBUnparseBufferRestore, &context, KlWPPBTable_mods);
	}
	KlStackFramePopNormal(stackptr);
	KlGC();
	return status;
    error:
	KlWPPBContextRestore(&context);
	KlGC();
	return 0;
    } else {
	return 0;
    }
}

void
KlWPPBUnparseBufferRestore(contextp, table)
    struct _KlWPPBContext *contextp;
    struct KlWPPBTable_mod *table;
{
    int i;
    for (i = 0; table[i].func; i++) {
	KlWPPBTable[table[i].index] = table[i].func;
    }
    KlWPPBContextRestore(contextp);
}

/* unparsing a buffer */

int					/* 1 = success, 0 = error */
KlWPPBUnparseBuffer(p, last, out, verbose, pp)
    unsigned char *p, *last;
    KlStream out;
    int verbose;
    KlO pp;
{
    KlO obj;				/* discarded at toplevel */
    jmp_buf jbuf;
    int res;

    if (res = setjmp(KlWPPBCurrentContext->jbuf)) { /* EOF */
	if (verbose == 1)
	    fprintf(KlStreamWriteFd(KlStderr), "\n");
	return res - 1;			/* longjmp 2 = success, 1 = error */
    } else {				/* main loop */
	for (;;) {
	    if (p >= last) {
		KlLoadWPPBError(p, "reading past buffer end%s!", "");
		return 1;
	    }
	    if (*p >= KlWPPBTableSize || !KlWPPBTable[*p]) {
		KlLoadWPPBError(p, "bad item code %d", (long) *p);
		return 0;
	    }
	    obj = 0;
	    p = (*(KlWPPBTable[*p]))(p+1, &obj);
	    /* now, difference from parse: we print p on out */
	    /* we strip top-level evals */
	    if (!obj)
		continue;		/* some functions dont fill &obj */

	    /* simplifications */
	    if (KlIsACollection(obj)) {
		KlCollection col = (KlCollection) obj;
		if (col->list[0] == (KlO) KlA_identity)
		    obj = col->list[1];
		else 
		    obj = (KlO) KlListMake(col);
	    }
	    if (KlIsAList(obj)		/* setq lambda ==> defun */
		&& ((KlList)obj)->size == 3
		&& KlIsALambda(((KlList)obj)->list[2])
		&& ((KlList)obj)->list[0] == (KlO) KlA_setq) {
		KlAtom defun = 0;
		KlList ll = (KlList) KlListCoerce(KlListType, 
						  ((KlList)obj)->list[2]);
		KlAtom lambda = (KlAtom) ll->list[0];
		if (lambda == KlA_lambda) 
		    defun = KlIntern("defun");
		else if (lambda == KlA_lambdaq)
		    defun = KlIntern("defunq");
		else if (lambda == KlA_lambdam)
		    defun = KlIntern("defmacro");
		if (defun) {
		    int i;
		    KlList new = KlListNMake(ll->size + 1);
		    KlListStore(new, 0, defun);
		    KlListStore(new, 1, ((KlList)obj)->list[1]);
		    for (i = 1; i < ll->size; i++)
			KlListStore(new, i+1, ll->list[i]);
		    obj = (KlO) new;
		}
	    }

	    /* if verbose, print what we are doing */
	    switch (verbose) {
	    case 0: break;
	    case 2: {
		KlString s = (KlString) KlApply3
		    (KlIntern("print-format")->c_val, KlStringType,
		     KlStringMake(" -> %0\n"), obj);
		if (s->size > 70)
		    s->string[69] = '\n', s->string[70] = '\0';
		fprintf(KlStreamWriteFd(KlStderr), s->string);
		fflush(KlStreamWriteFd(KlStderr));
		break;
	    }
	    default:
		fprintf(KlStreamWriteFd(KlStderr), "."),
		fflush(KlStreamWriteFd(KlStderr));
	    }
	    if (pp) {			/* use pretty-printer */
		KlStackFramePushValuesV(1, KlA_stdout, out);
		KlApply1(pp, obj);
		KlStackFramePop();
	    } else {
		KlStackFramePushValuesV(1, KlA_print_readably, TRU);
		KlSend_print(obj, out);
		KlStackFramePop();
	    }
	    KlSPutc('\n', out);
	}
    }
}


/* High-level Klone interface to KlUnparseWPPBFile:
 * (wppb:unpreparse [:pretty pretty-printer] files...)
 * for each file, generates its unparsed form in a <filename>.unp file
 * to avoid clobbing original file if present
 */

KlO
KlWPPBUnparse(argc, argv)
    int argc;
    KlString *argv;
{
    KlString filename;
    KlString outname;
    KlString args[2];
    int i;
    FILE *in, *out;
    KlStream kl_out;
    KlO pretty = 0;
    int verbose = 0;

    for (i = 0; i < argc; i++) {
	filename = argv[i];
	if (KlIsAKeyword(filename)) {
	    if (filename == (KlString) KlK_print) {
		i++;
		pretty = KlTrueP(argv[i]) ? (KlO) argv[i] : 0;
		continue;
	    } else if (filename == (KlString) KlK_verbose) {
		i++;
		if (argv[i] == (KlString) KlIntern(":debug"))
		    verbose = 2;
		else 
		    verbose = KlTrueP(argv[i]);
		continue;
	    } else {
		return KlError2(KlE_INVALID_KEYWORD, argv[i],
				KlListMakeV(2, KlK_print, KlK_verbose));
	    }
	}
	KlMustBeString(filename, i);
	
	args[0] = filename;
	args[1] = KlStringMake(".unp");	/* do it each time, for GC? */
	outname = (KlString) KlStringAdd(2, args); /* concat */
	if (verbose)
	    fprintf(KlStreamWriteFd(KlStderr), "unpreparsing %s into %s\n",
		    filename->string, outname->string);
	if (!(in = fopen(filename->string, "r"))) 
	    return (KlO) KlOpenError(0, filename, KlK_input, 1);
	if (!(out = fopen(outname->string, "w"))) 
	    return (KlO) KlOpenError(0, outname, KlK_output, 1);
	kl_out = KlStreamMake(out, 2, outname->string);

	KlUnparseWPPBFile(filename->string, in, kl_out, verbose, pretty);
    }
    return NIL;
}

/**************************************************************************\
* 				     Init                                  *
\**************************************************************************/

KlWPPBInit()
{
    KlWPPB_EXT_len = strlen(KlWPPB_EXT);
    KlWPPD_EXT_len = strlen(WPPD_EXT);
    KlWPPB_EXTs_len = Max(KlWPPB_EXT_len, KlWPPD_EXT_len);
    KlWPPD_PACKAGE_LEN = strlen(KlWPPD_PACKAGE);

    KlWPPBTableSize = 1;
    KlWPPBTable = (KlWPPBReadFunction *) Malloc(KLSO * 1);
    KlWPPBTableNames = (char **) Malloc(KLSO * 1);
    KlWPPBTable[0] = 0;

    KlAtomSetq(KlA_WPPD_list, KlListNMake(0));
    KlA_WPPD_list->type = KlConstantType;

    KlWPPBReadListTypes[0] = KlListType;
    KlWPPBReadListTypes[1] = KlVectorType;
#ifdef LOCATORS
    KlWPPBReadListTypes[2] = KlLocatorType;
#else 
    KlWPPBReadListTypes[2] = KlListType;
#endif

    KlWPPBReadQuotedExprSymbols[0] = KlA_quote;	/* ' */
    KlWPPBReadQuotedExprEvals[0] = KlQuote;
    KlWPPBReadQuotedExprSymbols[1] = KlA_backquote; /* ` */
    KlWPPBReadQuotedExprEvals[1] = KlBackquote;
    KlWPPBReadQuotedExprSymbols[2] = KlA_unquote; /* , */
    KlWPPBReadQuotedExprEvals[2] = KlQuotedExprCommaEval;
    KlWPPBReadQuotedExprSymbols[3] = KlA_unquotesplicing; /* ,@ */
    KlWPPBReadQuotedExprEvals[3] = KlQuotedExprCommaEval;	

#ifdef DEBUG
    bzero(&KlWPPBCurrentContextData, sizeof(struct _KlWPPBContext));
#endif /* DEBUG */

    KlDeclareWPPBRead(KlWPPBType_XType, "XType", KlWPPBReadXType);
    KlDeclareWPPBRead(KlWPPBType_Raw, "Raw", KlWPPBReadRaw);
    KlDeclareWPPBRead(KlWPPBType_Set, "Set", KlWPPBReadSet);
    KlDeclareWPPBRead(KlWPPBType_SetFunc, "SetFunc", KlWPPBReadSetFunc);
    KlDeclareWPPBRead(KlWPPBType_Eval, "Eval", KlWPPBReadEval);
    KlDeclareWPPBRead(KlWPPBType_WPPDInclude, "WPPDInclude", 
		      KlWPPBReadWPPDInclude);
    KlDeclareWPPBRead(KlWPPBType_WPPDRef, "WPPDRef", KlWPPBReadWPPDRef);
    KlDeclareWPPBRead(KlWPPBType_WPPDRefL, "WPPDRefL", KlWPPBReadWPPDRefL);
    KlDeclareWPPBRead(KlWPPBType_Atom, "Atom", KlWPPBReadAtom);
    KlDeclareWPPBRead(KlWPPBType_AtomL, "AtomL", KlWPPBReadAtomL);
    KlDeclareWPPBRead(KlWPPBType_String, "String", KlWPPBReadString);
    KlDeclareWPPBRead(KlWPPBType_StringL, "StringL", KlWPPBReadStringL);
    KlDeclareWPPBRead(KlWPPBType_List, "List", KlWPPBReadList);
    KlDeclareWPPBRead(KlWPPBType_ListM, "ListM", KlWPPBReadListM);
    KlDeclareWPPBRead(KlWPPBType_ListML, "ListML", KlWPPBReadListML);
    KlDeclareWPPBRead(KlWPPBType_IntS, "IntS", KlWPPBReadIntS);
    KlDeclareWPPBRead(KlWPPBType_IntP, "IntP", KlWPPBReadIntP);
    KlDeclareWPPBRead(KlWPPBType_IntN, "IntN", KlWPPBReadIntN);
#ifdef UNSIGNED
    KlDeclareWPPBRead(KlWPPBType_IntU, "IntU", KlWPPBReadIntU);
#else
    KlDeclareWPPBRead(KlWPPBType_IntU, "IntU", KlWPPBReadNotImplemented);
#endif
    KlDeclareWPPBRead(KlWPPBType_Real, "Real", KlWPPBReadReal);
    KlDeclareWPPBRead(KlWPPBType_QuotedExpr, "QuotedExpr", KlWPPBReadQuotedExpr);
    KlDeclareWPPBRead(KlWPPBType_Expr, "Expr", KlWPPBReadExpr);
    KlDeclareWPPBRead(KlWPPBType_Nil, "Nil", KlWPPBReadNil);
    KlDeclareWPPBRead(KlWPPBType_True, "True", KlWPPBReadTrue);
    KlDeclareWPPBRead(KlWPPBType_WPPDHere, "WPPDHere", KlWPPBReadWPPDHere);
    KlDeclareWPPBRead(KlWPPBType_Null, "Null", KlWPPBReadNull);
    KlDeclareWPPBRead(KlWPPBType_EOF, "EOF", KlWPPBReadEOF);

#ifdef ISM
    KlDeclareSubType (&SmlClassExprHide, "HExpr", KlExprType, 0);
    KlDeclareSubType (&SmlClassFExprHide, "HFExpr", KlFExprType, 0);
    KlDeclareSubType (&SmlClassMExprHide, "HMacro", KlMExprType, 0);
#else /* !ISM */
    KlDeclareSubr(KlWPPBWriteExpr, "wppb:write-expr", 2);
#endif /* !ISM */

    KlDeclareSubr(KlWPPBUnparse, "wppb:unpreparse", NARY);
    KlDeclareFSubr(KlWPPBUnparse, "unpreparse", NARY);

}

#endif /* WPPB */
