/******************************************************************************\
*                                                                              *
* 			      WIDE-CHAR REGEXPS                                *
*  a modification of the standard regexp package by Henry Spencer to work with *
* 			  wide chars, 16 or 32 bits                            *
*                                                                              *
\******************************************************************************/
#ifdef WCHARS
/*
 * regcomp and regexec -- regsub and regerror are elsewhere
 *
 *	Copyright (c) 1986 by University of Toronto.
 *	Written by Henry Spencer.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. The author is not responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *         **** THIS IS A HEAVILY ALTERED VERSION!!! ****
 * Beware that some of this code is subtly aware of the way operator
 * precedence is structured in regular expressions.  Serious changes in
 * regular-expression syntax might require a total rethink.
 */

/*****************************************************************************\
* 				 ALTERATIONS                                  *
\*****************************************************************************/
/*
 * By: Colas Nahaboo (Bull Research) December 1995, for Klone distrib
 * colas@sophia.inria.fr
 * This is a **BIG** modification of Henry Spencer's code. these regexps works
 * on an array of WREchar that can be char, short, int
 * big enhancement: no more fixed limit to the number of parentheses
 */

#include <stdio.h>
#include "klwregexp.h"			/* COLAS (was: <regexp.h>) */
#include "klmachine.h"			/* COLAS (for OS-specific decls) */
/* #include "regmagic.h" */		/* COLAS (included in regexp.h) */

DECLARE_strchr;				/* COLAS */
static Klparnum_add();
static int Klparnum_find_o();
static int Klparnum_find_c();
/*
 * The "internal use only" fields in regexp.h are present to pass info from
 * compile to execute that permits the execute phase to run lots faster on
 * simple cases.  They are:
 *
 * regstart	WREchar that must begin a match; '\0' if none obvious
 * reganch	is the match anchored (at beginning-of-line only)?
 * regmust	string (pointer into program) that match must include, or NULL
 * regmlen	length of regmust string
 *
 * Regstart and reganch permit very fast decisions on suitable starting points
 * for a match, cutting down the work a lot.  Regmust permits fast rejection
 * of lines that cannot possibly match.  The regmust tests are costly enough
 * that regcomp() supplies a regmust only if the r.e. contains something
 * potentially expensive (at present, the only such thing detected is * or +
 * at the start of the r.e., which can involve a lot of backup).  Regmlen is
 * supplied because the test in regexec() needs it and regcomp() is computing
 * it anyway.
 */

/*
 * Structure for WREregexp "program".  This is essentially a linear encoding
 * of a nondeterministic finite-state machine (aka syntax charts or
 * "railroad normal form" in parsing technology).  Each node is an opcode
 * plus a "next" pointer, possibly plus an operand.  "Next" pointers of
 * all nodes except BRANCH implement concatenation; a "next" pointer with
 * a BRANCH on both ends of it is connecting two alternatives.  (Here we
 * have one of the subtle syntax dependencies:  an individual BRANCH (as
 * opposed to a collection of them) is never concatenated with anything
 * because of operator precedence.)  The operand of some types of node is
 * a literal string; for others, it is a node leading into a sub-FSM.  In
 * particular, the operand of a BRANCH node is the first node of the branch.
 * (NB this is *not* a tree structure:  the tail of the branch connects
 * to the thing following the set of BRANCHes.)  The opcodes are:
 */

/* definition	number	opnd?	meaning */
#define	END	0	/* no	End of program. */
#define	BOL	1	/* no	Match "" at beginning of line. */
#define	EOL	2	/* no	Match "" at end of line. */
#define	ANY	3	/* no	Match any one WREcharacter. */
#define	ANYOF	4	/* str	Match any WREcharacter in this string. */
#define	ANYBUT	5	/* str	Match any WREcharacter not in this string. */
#define	BRANCH	6	/* node	Match this alternative, or the next... */
#define	BACK	7	/* no	Match "", "next" ptr points backward. */
#define	EXACTLY	8	/* str	Match this string. */
#define	NOTHING	9	/* no	Match empty string. */
#define	STAR	10	/* node	Match this (simple) thing 0 or more times. */
#define	PLUS	11	/* node	Match this (simple) thing 1 or more times. */
#define	OPEN	20	/* no	Mark this point in input as start of #n. */
			/*	OPEN+1 is number 1, etc. */
#define	CLOSE	30	/* no	Analogous to OPEN. */

/*
 * Opcode notes:
 *
 * BRANCH	The set of branches constituting a single choice are hooked
 *		together with their "next" pointers, since precedence prevents
 *		anything being concatenated to any individual branch.  The
 *		"next" pointer of the last BRANCH in a choice points to the
 *		thing following the whole choice.  This is also where the
 *		final "next" pointer of each individual branch points; each
 *		branch starts with the operand node of a BRANCH node.
 *
 * BACK		Normal "next" pointers all implicitly point forward; BACK
 *		exists to make loop structures possible.
 *
 * STAR,PLUS	'?', and complex '*' and '+', are implemented as circular
 *		BRANCH structures using BACK.  Simple cases (one WREcharacter
 *		per match) are implemented with STAR and PLUS for speed
 *		and to minimize recursive plunges.
 *
 * OPEN,CLOSE	...are numbered at compile time.
 */

/*
 * A node is one WREchar of opcode followed by two WREchars of "next" pointer.
 * "Next" pointers are stored as two 8-bit pieces, high order first.  The
 * value is a positive offset from the opcode of the node containing it.
 * An operand, if any, simply follows the node.  (Note that much of the
 * code generation knows about this implicit relationship.)
 *
 * Using two bytes for the "next" pointer is vast overkill for most things,
 * but allows patterns to get big without disasters.
 */
#define	OP(p)	(*(p))
#define NEXT(p) (((*((p)+1)&0377)<<8) + (*((p)+2)&0377)) /* COLAS */
/* COLAS: The previous line was:
#define NEXT(p)	(((*((p)+1)&0377)<<8) + *((p)+2)&0377)
*/
#define	OPERAND(p)	((p) + 3)

/*
 * See regmagic.h for one further detail of program structure.
 */


/*
 * Utility definitions.
 */
#define	UCHARAT(p)	((int)*(WREchar *)(p))

#define	FAIL(m)	{ Kl_regerror(m); return(NULL); }
#define	IS_MULT(c)	((c) == '*' || (c) == '+' || (c) == '?')
#define	META	"^$.[()|?+*\\"

/*
 * Flags to be passed up and down.
 */
#define	HASWIDTH	01	/* Known never to match null string. */
#define	SIMPLE		02	/* Simple enough to be STAR/PLUS operand. */
#define	SPSTART		04	/* Starts with * or +. */
#define	WORST		0	/* Worst case. */

/*
 * Global work variables for regcomp().
 */
static WREchar *regparse;		/* Input-scan pointer. */
static int regnpar;		/* () count. */
static WREchar regdummy;
static WREchar *regcode;		/* Code-emit pointer; &regdummy = don't. */
static WREchar *regcode_start;
static long regsize;		/* Code size. */
static short *o_parnums;				/* list of offsets */
static int o_parnums_size;
static short *c_parnums;				/* list of offsets */
static int c_parnums_size;
#ifdef DEBUG
static int nsubexp;
#endif 
/*
 * Forward declarations for regcomp()'s friends.
 */
#ifndef STATIC
#define	STATIC	static
#endif
STATIC WREchar *reg();
STATIC WREchar *regbranch();
STATIC WREchar *regpiece();
STATIC WREchar *regatom();
STATIC WREchar *regnode();
STATIC WREchar *regnext();
STATIC void regc();
STATIC void reginsert();
STATIC void regtail();
STATIC void regoptail();
STATIC int WREstrcspn();

/*
 - regcomp - compile a regular expression into internal code
 *
 * We can't allocate space until we know how big the compiled form will be,
 * but we can't compile it (and thus know how big it is) until we've got a
 * place to put the code.  So we cheat:  we compile it twice, once with code
 * generation turned off and size counting turned on, and once "for real".
 * This also means that we don't allocate space until we are sure that the
 * thing really will compile successfully, and we never have to move the
 * code and thus invalidate pointers into it.  (Note that it has to be in
 * one piece because free() must be able to free it all.)
 *
 * Beware that the optimization-preparation code in here knows about some
 * of the structure of the compiled WREregexp.
 */
WREregexp *
WREregcomp(exp)
WREchar *exp;
{
    register WREregexp *r;
    register WREchar *scan;
    register WREchar *longest;
    register int len;
    int flags;

    if (exp == NULL)
	FAIL("NULL argument");

    /* First pass: determine size, legality. */
    o_parnums = c_parnums = 0;
    o_parnums_size = c_parnums_size = 0;
    regparse = exp;
    regnpar = 1;
    regsize = 0L;
    regcode_start = regcode = &regdummy;
    regc(WREREGMAGIC);
    if (reg(0, &flags) == NULL)
	return(NULL);

    /* Small enough for pointer-storage convention? */
    if (regsize >= 32767L)		/* Probably could be 65535L. */
	FAIL("regexp too big");

    /* Allocate space. */
    r = (WREregexp *)malloc(sizeof(WREregexp) + 
			    (unsigned)regsize * WREsize);
    if (r == NULL)
	FAIL("out of space");
    r->nsubexp = regnpar;			/* inits */
    r->startp = (WREchar **) malloc(sizeof(WREchar *) * regnpar * 2);
    r->endp = r->startp + regnpar;	/* we save one malloc this way */

    /* Second pass: emit code. */
    regparse = exp;
    regnpar = 1;
    regcode_start = regcode = r->program;
    regc(WREREGMAGIC);
    if (reg(0, &flags) == NULL)
	return(NULL);

    /* Dig out information for optimizations. */
    r->regstart = '\0';	/* Worst-case defaults. */
    r->reganch = 0;
    r->regmust = NULL;
    r->regmlen = 0;
    r->o_parnums = o_parnums;
    r->c_parnums = c_parnums;
    scan = r->program+1;			/* First BRANCH. */
    if (OP(regnext(scan)) == END) {		/* Only one top-level choice. */
	scan = OPERAND(scan);

	/* Starting-point info. */
	if (OP(scan) == EXACTLY)
	    r->regstart = *OPERAND(scan);
	else if (OP(scan) == BOL)
	    r->reganch++;

	/*
	 * If there's something expensive in the r.e., find the
	 * longest literal string that must appear and make it the
	 * regmust.  Resolve ties in favor of later strings, since
	 * the regstart check works with the beginning of the r.e.
	 * and avoiding duplication strengthens checking.  Not a
	 * strong reason, but sufficient in the absence of others.
	 */
	if (flags&SPSTART) {
	    longest = NULL;
	    len = 0;
	    for (; scan != NULL; scan = regnext(scan))
		if (OP(scan) == EXACTLY && WREstrlen(OPERAND(scan)) >= len) {
		    longest = OPERAND(scan);
		    len = WREstrlen(OPERAND(scan));
		}
	    r->regmust = longest;
	    r->regmlen = len;
	}
    }

    return(r);
}

/*
 - reg - regular expression, i.e. main body or parenthesized thing
 *
 * Caller must absorb opening parenthesis.
 *
 * Combining parenthesis handling with the base level of regular expression
 * is a trifle forced, but the need to tie the tails of the branches to what
 * follows makes it hard to avoid.
 */
static WREchar *
reg(paren, flagp)
int paren;			/* Parenthesized? */
int *flagp;
{
    register WREchar *ret;
    register WREchar *br;
    register WREchar *ender;
    register int parno = 0;		/* COLAS */
    int flags;

    *flagp = HASWIDTH;	/* Tentatively. */

    /* Make an OPEN node, if parenthesized. */
    if (paren) {
	/* COLAS
	   if (regnpar >= NSUBEXP)
	   FAIL("too many ()");
	   */
	parno = regnpar;
	regnpar++;
	if (parno < WRE_NSUBEXP_BUILTIN) {
	    ret = regnode(OPEN+parno);
	} else {
	    ret = regnode(OPEN);
	    if (ret)
		Klparnum_add(OPEN, parno, regcode - 3);
	}
    } else
	ret = NULL;

    /* Pick up the branches, linking them together. */
    br = regbranch(&flags);
    if (br == NULL)
	return(NULL);
    if (ret != NULL)
	regtail(ret, br);	/* OPEN -> first. */
    else
	ret = br;
    if (!(flags&HASWIDTH))
	*flagp &= ~HASWIDTH;
    *flagp |= flags&SPSTART;
    while (*regparse == '|') {
	regparse++;
	br = regbranch(&flags);
	if (br == NULL)
	    return(NULL);
	regtail(ret, br);	/* BRANCH -> BRANCH. */
	if (!(flags&HASWIDTH))
	    *flagp &= ~HASWIDTH;
	*flagp |= flags&SPSTART;
    }

    /* Make a closing node, and hook it on the end. */
    if (paren) {
	if (parno < WRE_NSUBEXP_BUILTIN) {
	    ender = regnode(CLOSE+parno);
	} else {
	    ender = regnode(CLOSE);
	    if (ender)
		Klparnum_add(CLOSE, parno, regcode - 3);
	}
    } else {
	ender = regnode(END);	
    }
    regtail(ret, ender);

    /* Hook the tails of the branches to the closing node. */
    for (br = ret; br != NULL; br = regnext(br))
	regoptail(br, ender);

    /* Check for proper termination. */
    if (paren && *regparse++ != ')') {
	FAIL("unmatched ()");
    } else if (!paren && *regparse != '\0') {
	if (*regparse == ')') {
	    FAIL("unmatched ()");
	} else
	    FAIL("junk on end");	/* "Can't happen". */
	/* NOTREACHED */
    }

    return(ret);
}

/*
 - regbranch - one alternative of an | operator
 *
 * Implements the concatenation operator.
 */
static WREchar *
regbranch(flagp)
int *flagp;
{
    register WREchar *ret;
    register WREchar *chain;
    register WREchar *latest;
    int flags;

    *flagp = WORST;		/* Tentatively. */

    ret = regnode(BRANCH);
    chain = NULL;
    while (*regparse != '\0' && *regparse != '|' && *regparse != ')') {
	latest = regpiece(&flags);
	if (latest == NULL)
	    return(NULL);
	*flagp |= flags&HASWIDTH;
	if (chain == NULL)	/* First piece. */
	    *flagp |= flags&SPSTART;
	else
	    regtail(chain, latest);
	chain = latest;
    }
    if (chain == NULL)	/* Loop ran zero times. */
	(void) regnode(NOTHING);

    return(ret);
}

/*
 - regpiece - something followed by possible [*+?]
 *
 * Note that the branching code sequences used for ? and the general cases
 * of * and + are somewhat optimized:  they use the same NOTHING node as
 * both the endmarker for their branch list and the body of the last branch.
 * It might seem that this node could be dispensed with entirely, but the
 * endmarker role is not redundant.
 */
static WREchar *
regpiece(flagp)
int *flagp;
{
    register WREchar *ret;
    register WREchar op;
    register WREchar *next;
    int flags;

    ret = regatom(&flags);
    if (ret == NULL)
	return(NULL);

    op = *regparse;
    if (!IS_MULT(op)) {
	*flagp = flags;
	return(ret);
    }

    if (!(flags&HASWIDTH) && op != '?')
	FAIL("*+ operand could be empty");
    *flagp = (op != '+') ? (WORST|SPSTART) : (WORST|HASWIDTH);

    if (op == '*' && (flags&SIMPLE))
	reginsert(STAR, ret);
    else if (op == '*') {
	/* Emit x* as (x&|), where & means "self". */
	reginsert(BRANCH, ret);			/* Either x */
	regoptail(ret, regnode(BACK));		/* and loop */
	regoptail(ret, ret);			/* back */
	regtail(ret, regnode(BRANCH));		/* or */
	regtail(ret, regnode(NOTHING));		/* null. */
    } else if (op == '+' && (flags&SIMPLE))
	reginsert(PLUS, ret);
    else if (op == '+') {
	/* Emit x+ as x(&|), where & means "self". */
	next = regnode(BRANCH);			/* Either */
	regtail(ret, next);
	regtail(regnode(BACK), ret);		/* loop back */
	regtail(next, regnode(BRANCH));		/* or */
	regtail(ret, regnode(NOTHING));		/* null. */
    } else if (op == '?') {
	/* Emit x? as (x|) */
	reginsert(BRANCH, ret);			/* Either x */
	regtail(ret, regnode(BRANCH));		/* or */
	next = regnode(NOTHING);		/* null. */
	regtail(ret, next);
	regoptail(ret, next);
    }
    regparse++;
    if (IS_MULT(*regparse))
	FAIL("nested *?+");

    return(ret);
}

/*
 - regatom - the lowest level
 *
 * Optimization:  gobbles an entire sequence of ordinary WREcharacters so that
 * it can turn them into a single node, which is smaller to store and
 * faster to run.  Backslashed WREcharacters are exceptions, each becoming a
 * separate node; the code is simpler that way and it's not worth fixing.
 */
static WREchar *
regatom(flagp)
int *flagp;
{
    register WREchar *ret;
    int flags;

    *flagp = WORST;		/* Tentatively. */

    switch (*regparse++) {
    case '^':
	ret = regnode(BOL);
	break;
    case '$':
	ret = regnode(EOL);
	break;
    case '.':
	ret = regnode(ANY);
	*flagp |= HASWIDTH|SIMPLE;
	break;
    case '[': {
	register int cclass;
	register int classend;

	if (*regparse == '^') {	/* Complement of range. */
	    ret = regnode(ANYBUT);
	    regparse++;
	} else
	    ret = regnode(ANYOF);
	if (*regparse == ']' || *regparse == '-')
	    regc(*regparse++);
	while (*regparse != '\0' && *regparse != ']') {
	    if (*regparse == '-') {
		regparse++;
		if (*regparse == ']' || *regparse == '\0')
		    regc('-');
		else {
		    cclass = UCHARAT(regparse-2)+1;
		    classend = UCHARAT(regparse);
		    if (cclass > classend+1)
			FAIL("invalid [] range");
		    for (; cclass <= classend; cclass++)
			regc(cclass);
		    regparse++;
		}
	    } else
		regc(*regparse++);
	}
	regc('\0');
	if (*regparse != ']')
	    FAIL("unmatched []");
	regparse++;
	*flagp |= HASWIDTH|SIMPLE;
    }
	break;
    case '(':
	ret = reg(1, &flags);
	if (ret == NULL)
	    return(NULL);
	*flagp |= flags&(HASWIDTH|SPSTART);
	break;
    case '\0':
    case '|':
    case ')':
	FAIL("internal urp");	/* Supposed to be caught earlier. */
	break;
    case '?':
    case '+':
    case '*':
	FAIL("?+* follows nothing");
	break;
    case '\\':
	if (*regparse == '\0')
	    FAIL("trailing \\");
	ret = regnode(EXACTLY);
	regc(*regparse++);
	regc('\0');
	*flagp |= HASWIDTH|SIMPLE;
	break;
    default: {
	register int len;
	register WREchar ender;

	regparse--;
	len = WREstrcspn(regparse, META);
	if (len <= 0)
	    FAIL("internal disaster");
	ender = *(regparse+len);
	if (len > 1 && IS_MULT(ender))
	    len--;		/* Back off clear of ?+* operand. */
	*flagp |= HASWIDTH;
	if (len == 1)
	    *flagp |= SIMPLE;
	ret = regnode(EXACTLY);
	while (len > 0) {
	    regc(*regparse++);
	    len--;
	}
	regc('\0');
    }
	break;
    }

    return(ret);
}

/*
 - regnode - emit a node
 */
static WREchar *			/* Location. */
regnode(op)
WREchar op;
{
    register WREchar *ret;
    register WREchar *ptr;

    ret = regcode;
    if (ret == &regdummy) {
	regsize += 3;
	return(ret);
    }

    ptr = ret;
    *ptr++ = op;
    *ptr++ = '\0';		/* Null "next" pointer. */
    *ptr++ = '\0';
    regcode = ptr;

    return(ret);
}

/*
 - regc - emit (if appropriate) a byte of code
 */
static void
regc(b)
WREchar b;
{
    if (regcode != &regdummy)
	*regcode++ = b;
    else
	regsize++;
}

/*
 - reginsert - insert an operator in front of already-emitted operand
 *
 * Means relocating the operand.
 */
static void
reginsert(op, opnd)
WREchar op;
WREchar *opnd;
{
    register WREchar *src;
    register WREchar *dst;
    register WREchar *place;

    if (regcode == &regdummy) {
	regsize += 3;
	return;
    }

    src = regcode;
    regcode += 3;
    dst = regcode;
    while (src > opnd)
	*--dst = *--src;

    place = opnd;		/* Op node, where operand used to be. */
    *place++ = op;
    *place++ = '\0';
    *place++ = '\0';
}

/*
 - regtail - set the next-pointer at the end of a node chain
 */
static void
regtail(p, val)
WREchar *p;
WREchar *val;
{
    register WREchar *scan;
    register WREchar *temp;
    register int offset;

    if (p == &regdummy)
	return;

    /* Find last node. */
    scan = p;
    for (;;) {
	temp = regnext(scan);
	if (temp == NULL)
	    break;
	scan = temp;
    }

    if (OP(scan) == BACK)
	offset = scan - val;
    else
	offset = val - scan;
    *(scan+1) = (offset>>8)&0377;
    *(scan+2) = offset&0377;
}

/*
 - regoptail - regtail on operand of first argument; nop if operandless
 */
static void
regoptail(p, val)
WREchar *p;
WREchar *val;
{
    /* "Operandless" and "op != BRANCH" are synonymous in practice. */
    if (p == NULL || p == &regdummy || OP(p) != BRANCH)
	return;
    regtail(OPERAND(p), val);
}

/*
 * regexec and friends
 */

/*
 * Global work variables for regexec().
 */
static WREchar *reginput;		/* String-input pointer. */
static WREchar *regbol;		/* Beginning of input, for ^ check. */
static WREchar **regstartp;	/* Pointer to startp array. */
static WREchar **regendp;		/* Ditto for endp. */

/*
 * Forwards.
 */
STATIC int regtry();
STATIC int regmatch();
STATIC int regrepeat();

#ifdef DEBUG
int wregnarrate = 0;
void Klwregdump();
STATIC char *regprop();
#endif

/*
 - regexec - match a WREregexp against a string
 */
int
Klwregexec(prog, string)
register WREregexp *prog;
register WREchar *string;
{
    register WREchar *s;

    /* Be paranoid... */
    if (prog == NULL || string == NULL) {
	Kl_regerror("NULL parameter");
	return(0);
    }

    /* Check validity of program. */
    if (UCHARAT(prog->program) != WREREGMAGIC) {
	Kl_regerror("corrupted program");
	return(0);
    }

    /* If there is a "must appear" string, look for it. */
    if (prog->regmust != NULL) {
	s = string;
	while ((s = WREstrchr(s, prog->regmust[0])) != NULL) {
	    if (WREstrncmp(s, prog->regmust, prog->regmlen) == 0)
		break;	/* Found it. */
	    s++;
	}
	if (s == NULL)	/* Not present. */
	    return(0);
    }

    /* Mark beginning of line for ^ . */
    regbol = string;

    /* Simplest case:  anchored match need be tried only once. */
    if (prog->reganch)
	return(regtry(prog, string));

    /* Messy cases:  unanchored match. */
    s = string;
    if (prog->regstart != '\0')
	/* We know what WREchar it must start with. */
	while ((s = WREstrchr(s, prog->regstart)) != NULL) {
	    if (regtry(prog, s))
		return(1);
	    s++;
	}
    else
	/* We don't -- general case. */
	do {
	    if (regtry(prog, s))
		return(1);
	} while (*s++ != '\0');

    /* Failure. */
    return(0);
}

/*
 - regtry - try match at specific point
 */
static int			/* 0 failure, 1 success */
regtry(prog, string)
WREregexp *prog;
WREchar *string;
{
    register int i;
    register WREchar **sp;
    register WREchar **ep;

    reginput = string;
    regstartp = prog->startp;
    regendp = prog->endp;

    sp = prog->startp;
    ep = prog->endp;
    for (i = prog->nsubexp; i > 0; i--) { /* COLAS: was NSUBEXP */
	*sp++ = NULL;
	*ep++ = NULL;
    }
    if (prog->o_parnums) {
	regcode_start = prog->program + 1;
	o_parnums = prog->o_parnums;
	c_parnums = prog->c_parnums;
#ifdef DEBUG
	nsubexp = prog->nsubexp;
#endif
    }
    if (regmatch(prog->program + 1)) {
	prog->startp[0] = string;
	prog->endp[0] = reginput;
	return(1);
    } else
	return(0);
}

/*
 - regmatch - main matching routine
 *
 * Conceptually the strategy is simple:  check to see whether the current
 * node matches, call self recursively to see whether the rest matches,
 * and then act accordingly.  In practice we make some effort to avoid
 * recursion, in particular by going through "ordinary" nodes (that don't
 * need to know whether the rest of the match failed) by a loop instead of
 * by recursion.
 */
static int			/* 0 failure, 1 success */
regmatch(prog)
WREchar *prog;
{
    register WREchar *scan;	/* Current node. */
    WREchar *next;		/* Next node. */

    scan = prog;
#ifdef DEBUG
    if (scan != NULL && wregnarrate)
	fprintf(stderr, "%s(\n", regprop(scan));
#endif
    while (scan != NULL) {
#ifdef DEBUG
	if (wregnarrate)
	    fprintf(stderr, "%s...\n", regprop(scan));
#endif
	next = regnext(scan);

	switch (OP(scan)) {
	case BOL:
	    if (reginput != regbol)
		return(0);
	    break;
	case EOL:
	    if (*reginput != '\0')
		return(0);
	    break;
	case ANY:
	    if (*reginput == '\0')
		return(0);
	    reginput++;
	    break;
	case EXACTLY: {
	    register int len;
	    register WREchar *opnd;

	    opnd = OPERAND(scan);
	    /* Inline the first WREcharacter, for speed. */
	    if (*opnd != *reginput)
		return(0);
	    len = WREstrlen(opnd);
	    if (len > 1 && WREstrncmp(opnd, reginput, len) != 0)
		return(0);
	    reginput += len;
	}
	    break;
	case ANYOF:
	    if (*reginput == '\0' || WREstrchr(OPERAND(scan), *reginput) == NULL)
		return(0);
	    reginput++;
	    break;
	case ANYBUT:
	    if (*reginput == '\0' || WREstrchr(OPERAND(scan), *reginput) != NULL)
		return(0);
	    reginput++;
	    break;
	case NOTHING:
	    break;
	case BACK:
	    break;
	case OPEN:
	case OPEN+1:
	case OPEN+2:
	case OPEN+3:
	case OPEN+4:
	case OPEN+5:
	case OPEN+6:
	case OPEN+7:
	case OPEN+8:
	case OPEN+9: {
	    register int no;
	    register WREchar *save;

	    if (!(no = OP(scan) - OPEN))
		no = Klparnum_find_o(scan - regcode_start);
	    save = reginput;

	    if (regmatch(next)) {
					/*
					 * Don't set startp if some later
					 * invocation of the same parentheses
					 * already has.
					 */
		if (regstartp[no] == NULL)
		    regstartp[no] = save;
		return(1);
	    } else
		return(0);
	}
	    break;
	case CLOSE:
	case CLOSE+1:
	case CLOSE+2:
	case CLOSE+3:
	case CLOSE+4:
	case CLOSE+5:
	case CLOSE+6:
	case CLOSE+7:
	case CLOSE+8:
	case CLOSE+9: {
	    register int no;
	    register WREchar *save;

	    no = OP(scan) - CLOSE;
	    if (!(no = OP(scan) - CLOSE))
		no = Klparnum_find_c(scan - regcode_start);
	    save = reginput;

	    if (regmatch(next)) {
					/*
					 * Don't set endp if some later
					 * invocation of the same parentheses
					 * already has.
					 */
		if (regendp[no] == NULL)
		    regendp[no] = save;
		return(1);
	    } else
		return(0);
	}
	    break;
	case BRANCH: {
	    register WREchar *save;

	    if (OP(next) != BRANCH)		/* No choice. */
		next = OPERAND(scan);	/* Avoid recursion. */
	    else {
		do {
		    save = reginput;
		    if (regmatch(OPERAND(scan)))
			return(1);
		    reginput = save;
		    scan = regnext(scan);
		} while (scan != NULL && OP(scan) == BRANCH);
		return(0);
					/* NOTREACHED */
	    }
	}
	    break;
	case STAR:
	case PLUS: {
	    register WREchar nextch;
	    register int no;
	    register WREchar *save;
	    register int min;

	    /*
	     * Lookahead to avoid useless match attempts
	     * when we know what WREcharacter comes next.
	     */
	    nextch = '\0';
	    if (OP(next) == EXACTLY)
		nextch = *OPERAND(next);
	    min = (OP(scan) == STAR) ? 0 : 1;
	    save = reginput;
	    no = regrepeat(OPERAND(scan));
	    while (no >= min) {
					/* If it could work, try it. */
		if (nextch == '\0' || *reginput == nextch)
		    if (regmatch(next))
			return(1);
					/* Couldn't or didn't -- back up. */
		no--;
		reginput = save + no;
	    }
	    return(0);
	}
	    break;
	case END:
	    return(1);	/* Success! */
	    break;
	default:
	    Kl_regerror("memory corruption");
	    return(0);
	    break;
	}

	scan = next;
    }

    /*
     * We get here only if there's trouble -- normally "case END" is
     * the terminating point.
     */
    Kl_regerror("corrupted pointers");
    return(0);
}

/*
 - regrepeat - repeatedly match something simple, report how many
 */
static int
regrepeat(p)
WREchar *p;
{
    register int count = 0;
    register WREchar *scan;
    register WREchar *opnd;

    scan = reginput;
    opnd = OPERAND(p);
    switch (OP(p)) {
    case ANY:
	count = WREstrlen(scan);
	scan += count;
	break;
    case EXACTLY:
	while (*opnd == *scan) {
	    count++;
	    scan++;
	}
	break;
    case ANYOF:
	while (*scan != '\0' && WREstrchr(opnd, *scan) != NULL) {
	    count++;
	    scan++;
	}
	break;
    case ANYBUT:
	while (*scan != '\0' && WREstrchr(opnd, *scan) == NULL) {
	    count++;
	    scan++;
	}
	break;
    default:		/* Oh dear.  Called inappropriately. */
	Kl_regerror("internal foulup");
	count = 0;	/* Best compromise. */
	break;
    }
    reginput = scan;

    return(count);
}

/*
 - regnext - dig the "next" pointer out of a node
 */
static WREchar *
regnext(p)
register WREchar *p;
{
    register int offset;

    if (p == &regdummy)
	return(NULL);

    offset = NEXT(p);
    if (offset == 0)
	return(NULL);

    if (OP(p) == BACK)
	return(p-offset);
    else
	return(p+offset);
}

#ifdef DEBUG

STATIC char *regprop();

/*
 - regdump - dump a WREregexp onto stdout in vaguely comprehensible form
 */
void
Klwregdump(r)
WREregexp *r;
{
    register WREchar *s;
    register WREchar op = EXACTLY;	/* Arbitrary non-END op. */
    register WREchar *next;


    s = r->program + 1;
    while (op != END) {	/* While that wasn't END last time... */
	op = OP(s);
	printf("%2d%s", s-r->program, regprop(s));	/* Where, what. */
	next = regnext(s);
	if (next == NULL)		/* Next ptr. */
	    printf("(0)");
	else 
	    printf("(%d)", (s-r->program)+(next-s));
	s += 3;
	if (op == ANYOF || op == ANYBUT || op == EXACTLY) {
	    /* Literal string, where present. */
	    while (*s != '\0') {
		putchar(*s);
		s++;
	    }
	    s++;
	}
	putchar('\n');
    }

    /* Header fields of interest. */
    if (r->regstart != '\0')
	printf("start `%c' ", r->regstart);
    if (r->reganch)
	printf("anchored ");
    if (r->regmust != NULL)
	printf("must have \"%s\"", r->regmust);
    printf("\n");
}

/*
 - regprop - printable representation of opcode
 */
static char *
regprop(op)
WREchar *op;
{
    register char *p;
    static char buf[50];

    (void) strcpy(buf, ":");

    switch (OP(op)) {
    case BOL:
	p = "BOL";
	break;
    case EOL:
	p = "EOL";
	break;
    case ANY:
	p = "ANY";
	break;
    case ANYOF:
	p = "ANYOF";
	break;
    case ANYBUT:
	p = "ANYBUT";
	break;
    case BRANCH:
	p = "BRANCH";
	break;
    case EXACTLY:
	p = "EXACTLY";
	break;
    case NOTHING:
	p = "NOTHING";
	break;
    case BACK:
	p = "BACK";
	break;
    case END:
	p = "END";
	break;
    case OPEN:
    case OPEN+1:
    case OPEN+2:
    case OPEN+3:
    case OPEN+4:
    case OPEN+5:
    case OPEN+6:
    case OPEN+7:
    case OPEN+8:
    case OPEN+9:
    {
	int no = OP(op)-OPEN;
	if (!no) no = Klparnum_find_o(op - regcode_start);
	sprintf(buf+strlen(buf), "OPEN%d", no);
	p = NULL;
	break;
    }
    case CLOSE:
    case CLOSE+1:
    case CLOSE+2:
    case CLOSE+3:
    case CLOSE+4:
    case CLOSE+5:
    case CLOSE+6:
    case CLOSE+7:
    case CLOSE+8:
    case CLOSE+9:
    {
	int no = OP(op)-CLOSE;
	if (!no) no = Klparnum_find_c(op - regcode_start);
	sprintf(buf+strlen(buf), "CLOSE%d", no);
	p = NULL;
	break;
    }
    case STAR:
	p = "STAR";
	break;
    case PLUS:
	p = "PLUS";
	break;
    default:
	Kl_regerror("corrupted opcode");
	break;
    }
    if (p != NULL) {
	(void) strcat(buf, p);
    }
    return(buf);
}
#endif

/*
 * The following is provided for those people who do not have WREstrcspn() in
 * their C libraries.  They should get off their butts and do something
 * about it; at least one public-domain implementation of those (highly
 * useful) string routines has been published on Usenet.
 */
/*
 * strcspn - find length of initial segment of s1 consisting entirely
 * of WREcharacters not from s2
 */

static int
WREstrcspn(s1, s2)
    WREchar *s1;			/* unicode */
    unsigned char *s2;			/* normal 8bit string */
{
    register WREchar *scan1;
    register unsigned char *scan2;
    register int count;

    count = 0;
    for (scan1 = s1; *scan1 != '\0'; scan1++) {
	for (scan2 = s2; *scan2 != '\0';)	/* ++ moved down. */
	    if (*scan1 == *scan2++)
		return(count);
	count++;
    }
    return(count);
}

/* COLAS: management of paren levels > 10
 * we maintain a p-list of parenleven, offset in program
 */

static
Klparnum_add(type, parno, ptr)
    int type;				/* OPEN or CLOSE */
    int parno;
    WREchar * ptr;
{
    parno -= WRE_NSUBEXP_BUILTIN;
    if (type == OPEN) {
	if (!o_parnums_size) {
	    o_parnums_size = parno+1;
	    o_parnums = (short *) malloc(sizeof(short) * o_parnums_size);
	} else if (parno >= o_parnums_size) {
	    o_parnums_size = parno+1;
	    o_parnums = (short *) realloc(o_parnums, 
					 sizeof(short) * o_parnums_size);
	}
	o_parnums[parno] = ptr - regcode_start - 1;
    } else {
	if (!c_parnums_size) {
	    c_parnums_size = parno+1;
	    c_parnums = (short *) malloc(sizeof(short) * c_parnums_size);
	} else if (parno >= c_parnums_size) {
	    c_parnums_size = parno+1;
	    c_parnums = (short *) realloc(c_parnums, 
					 sizeof(short) * c_parnums_size);
	}
	c_parnums[parno] = ptr - regcode_start - 1;
    }
}

static int
Klparnum_find_o(offset)
    int offset;
{
    short *p = o_parnums;
#ifdef DEBUG
    for (;*p != offset; p++)
	if (p >= (o_parnums + nsubexp))
	    stop_if_in_dbx("garbled o_parnums!");
#else
    while (*p != offset)
	p++;
#endif
    return p - o_parnums + WRE_NSUBEXP_BUILTIN;
}

static int
Klparnum_find_c(offset)
    int offset;
{
    short *p = c_parnums;
#ifdef DEBUG
    for (;*p != offset; p++)
	if (p >= (c_parnums + nsubexp))
	    stop_if_in_dbx("garbled c_parnums!");
#else
    while (*p != offset)
	p++;
#endif
    return p - c_parnums + WRE_NSUBEXP_BUILTIN;
}

/********************************************************** wide chars utils */
/* WRE versions of some of the strings function of the unix C linrary */

int
WREstrlen(s)
    WREchar *s;
{
    WREchar *p = s;
    while (*p++)
	;
    return p - s - 1;
}

WREchar *
WREstrchr(s, c)
    WREchar *s;
    WREchar c;
{
    while (*s)
	if (*s++ == c)
	    return s - 1;
    return 0;
}

void
WREstrcpy(d, s)
    WREchar *d;
    WREchar *s;
{
    WREchar *dorig = d;
    while (*s)
	*d++ = *s++;
}

void
WREstrncpy(d, s, l)
    WREchar *d;
    WREchar *s;
{
    WREchar *dorig = d;
    while (*s && l--)
	*d++ = *s++;
}

int
WREstrncmp(s1, s2, l)
    WREchar *s1;
    WREchar *s2;
    int l;
{
    for (;;) {
	if (l-- <= 0)			/* l chars done */
	    return 0;
	if (*s1 != *s2)			/* differ */
	    return (*s1 < *s2 ? -1 : 1);
	if (!(*s1))			/* end of string */
	    return 0;
	s1++, s2++;			/* go on... */
    }
}

#endif /* WCHARS */
