/*****************************************************************************\
*                                                                             *
* 	  an exemple of a ses extension: access to the ndbm package           *
*                                                                             *
\*****************************************************************************/
/* compile by:
   klone_cc ses_ndbm.c seslib.o -o ses_ndbm -lgdbm
 */
#include <stdio.h>
#include <ndbm.h>
#include <errno.h>

#include "seslib.h"

/*****************************************************************************\
* 				    TYPES                                     *
\*****************************************************************************/

int DBM_ID;
SES_Connection sescon;

/* this "destructor" method will be called when a DBM_ID is released */

void
ses_dbm_close(db)
    DBM *db;
{
    dbm_close(db);			/* simple, isnt'it? Actually dbm_close
					   could have been used as-is */
}


/*****************************************************************************\
* 				   WRAPPERS                                   *
\*****************************************************************************/

void
ses_dbm_open(args)
    SES_Args args;
{
    DBM * result;
    char *file = SES_ArgString(args);
    int flags = SES_ArgInt(args);
    int mode = SES_ArgInt(args);

    result = dbm_open(file, flags, mode);

    if (result)
	SES_ReturnId(args, DBM_ID, result);
    else {
	char mess[1024];
	sprintf(mess, "(Errors:ErrorOpeningFile \"%s\" %d %d)", 
		file, mode, errno);
	SES_ReturnERROR(args, mess);
    }
}

void
ses_dbm_store(args)
    SES_Args args;
{
    int result;
    DBM *db = (DBM *) SES_ArgId(args, DBM_ID);
    char *key = SES_ArgString(args);
    char *content = SES_ArgString(args);
    int flags = SES_ArgInt(args);
    datum key_datum, content_datum;
    key_datum.dsize = strlen(key);
    key_datum.dptr = key;
    content_datum.dsize = strlen(content) + 1; /* include terminating 0 */
    content_datum.dptr = content;

    result = dbm_store(db, key_datum, content_datum, flags);

    SES_ReturnInt(args, result);
}

void
ses_dbm_fetch(args)
    SES_Args args;
{
    datum result;
    DBM *db = (DBM *) SES_ArgId(args, DBM_ID);
    char *key = SES_ArgString(args);
    datum key_datum;
    key_datum.dsize = strlen(key);
    key_datum.dptr = key;

    result = dbm_fetch(db, key_datum);

    SES_ReturnString(args, result.dptr);
}

/*****************************************************************************\
* 			     INIT, DECLARES, MAIN                             *
\*****************************************************************************/

main()
{
    sescon = SES_Init("DBM");		/* init */

					/* export functions */
    DBM_ID = SES_DeclareType(sescon, "DBM", ses_dbm_close);

    SES_DeclareFunction(sescon, "DBM:open", ses_dbm_open, 3, DBM_ID,
			SES_String, SES_Int, SES_Int, 0);
    SES_DeclareFunction(sescon, "DBM:store", ses_dbm_store, 4, SES_Int,
			DBM_ID, SES_String, SES_String, SES_Int, 0);
    SES_DeclareFunction(sescon, "DBM:fetch", ses_dbm_fetch, 2, SES_String,
			DBM_ID, SES_String, 0);

					/* mainloop */
    SES_MainLoop(sescon);
}
