/*****************************************************************************\
*                                                                             *
* 				  SES CODECs                                  *
*                                                                             *
\*****************************************************************************/
/* This is #included by both sides of the SES connection
 * before including it you must define SES__CODEC_ERROR();
 * additional parameters can be defined by SESCODECPAR and SESCODECPARD, see
 * beginning of file klsesi.c, actually using this file as a template.
 */

#ifndef SESCODECPAR
#define SESCODECPAR
#endif

#ifndef SESCODECPARD
#define SESCODECPARD
#endif

/*****************************************************************************\
* 			    encoding/decoding data                            *
\*****************************************************************************/
/* encodes / decodes data onto a byte stream at ptr, returns ptr afterwards
 * Supposes the room has already been reserved.
 * encoders are kept close to decoders for easier maintainance
 */

/*********************************************************************** Int */

uchar *
SES__EncodeInt(ptr, number)
    uchar *ptr;
    int number;
{
    SES_PS_Int ps = (SES_PS_Int) ptr;
    ps->type = SES_Int;
    ps->number = number;
    return ptr + SES_PS_Int_padlen;
}

uchar *
SES__DecodeInt(ptr, numberp SESCODECPAR)
    uchar *ptr;
    int *numberp;
    SESCODECPARD
{
    SES_PS_Int ps = (SES_PS_Int) ptr;
    if (ps->type != SES_Int) {SES__CODEC_ERROR(SES_Int, ps->type);}
    *numberp = ps->number;
    return ptr + SES_PS_Int_padlen;
}

/******************************************************************** String */

uchar *
SES__EncodeString(ptr, namelen, name)
    uchar *ptr;
    int namelen;
    char *name;
{
    SES_PS_String ps = (SES_PS_String) ptr;
    ps->type = SES_String;
    ps->length = namelen;
    memcpy(ps->string, name, namelen+1);
    return ptr + SES_PS_String_padlen(namelen);
}

/* warning: namep points on volatile storage copy or use before next decode */
uchar *
SES__DecodeString(ptr, namelenp, namep SESCODECPAR)
    uchar *ptr;
    int *namelenp;
    char **namep;
    SESCODECPARD
{
    SES_PS_String ps = (SES_PS_String) ptr;
    if (ps->type != SES_String) {SES__CODEC_ERROR(SES_String, ps->type);}
    *namelenp = ps->length;
    *namep = (char **) ps->string;
    return ptr + SES_PS_String_padlen(*namelenp);
}

/********************************************************************** Atom */

uchar *
SES__EncodeAtom(ptr, name)
    uchar *ptr;
    char *name;
{
    SES_PS_Atom ps = (SES_PS_Atom) ptr;
    ps->type = SES_Atom;
    ps->length = strlen(name);
    memcpy(ps->string, name, ps->length+1);
    return ptr + SES_PS_Atom_padlen(ps->length);
}

/* warning: namep points on volatile storage copy or use before next decode */
uchar *
SES__DecodeAtom(ptr, namelenp, namep SESCODECPAR)
    uchar *ptr;
    int *namelenp;
    char **namep;
    SESCODECPARD
{
    SES_PS_Atom ps = (SES_PS_Atom) ptr;
    if (ps->type != SES_Atom) {SES__CODEC_ERROR(SES_Atom, ps->type);}
    *namelenp = ps->length;
    *namep = (char **) ps->string;
    return ptr + SES_PS_Atom_padlen(*namelenp);
}

/************************************************************************ Id */

uchar *
SES__EncodeId(ptr, idtype, value)
    uchar *ptr;
    int idtype;
    void *value;
{
    SES_PS_Id ps = (SES_PS_Id) ptr;
    ps->type = SES_Id;
    ps->idtype = idtype;
    ps->value = value;
    return ptr + SES_PS_Id_padlen;
}

uchar *
SES__DecodeId(ptr, idtype, valuep SESCODECPAR)
    uchar *ptr;
    int idtype;				/* must be of this type */
    void **valuep;
    SESCODECPARD
{
    SES_PS_Id ps = (SES_PS_Id) ptr;
    if (ps->type != SES_Id) {SES__CODEC_ERROR(SES_Id, ps->type);}
    if (ps->idtype != idtype) {SES__CODEC_ERROR(idtype, ps->idtype);}
    *valuep = ps->value;
    return ptr + SES_PS_Id_padlen;
}

/* may be useful
 * parses any id type, and returns the actual type read
 */
uchar *
SES__DecodeAnyId(ptr, idtypep, valuep SESCODECPAR)
    uchar *ptr;
    int *idtypep;			/* returns type here */
    void **valuep;
    SESCODECPARD
{
    SES_PS_Id ps = (SES_PS_Id) ptr;
    if (ps->type != SES_Id) {SES__CODEC_ERROR(SES_Id, ps->type);}
    *idtypep = ps->idtype;
    *valuep = ps->value;
    return ptr + SES_PS_Id_padlen;
}

/******************************************************************* Boolean */

uchar *
SES__EncodeBoolean(ptr, number)
    uchar *ptr;
    int number;
{
    SES_PS_Boolean ps = (SES_PS_Boolean) ptr;
    ps->type = SES_Boolean;
    ps->number = number;
    return ptr + SES_PS_Boolean_padlen;
}

uchar *
SES__DecodeBoolean(ptr, numberp SESCODECPAR)
    uchar *ptr;
    int *numberp;
    SESCODECPARD
{
    SES_PS_Boolean ps = (SES_PS_Boolean) ptr;
    if (ps->type != SES_Boolean) {SES__CODEC_ERROR(SES_Boolean, ps->type);}
    *numberp = ps->number;
    return ptr + SES_PS_Boolean_padlen;
}

/********************************************************** Reals (Doubles) */

uchar *
SES__EncodeReal(ptr, number)
    uchar *ptr;
    double number;
{
    SES_PS_Real ps = (SES_PS_Real) ptr;
    ps->type = SES_Real;
    ps->number = number;
    return ptr + SES_PS_Real_padlen;
}

uchar *
SES__DecodeReal(ptr, numberp SESCODECPAR)
    uchar *ptr;
    double *numberp;
    SESCODECPARD
{
    SES_PS_Real ps = (SES_PS_Real) ptr;
    if (ps->type != SES_Real) {SES__CODEC_ERROR(SES_Real, ps->type);}
    *numberp = ps->number;
    return ptr + SES_PS_Real_padlen;
}

/*********************************************************************** Ptr */

uchar *
SES__EncodePtr(ptr, anyptr)
    uchar *ptr;
    void *anyptr;
{
    SES_PS_Ptr ps = (SES_PS_Ptr) ptr;
    ps->type = SES_Ptr;
    ps->ptr = anyptr;
    return ptr + SES_PS_Ptr_padlen;
}

uchar *
SES__DecodePtr(ptr, anyptrp SESCODECPAR)
    uchar *ptr;
    void **anyptrp;
    SESCODECPARD
{
    SES_PS_Ptr ps = (SES_PS_Ptr) ptr;
    if (ps->type != SES_Ptr) {SES__CODEC_ERROR(SES_Ptr, ps->type);}
    *anyptrp = ps->ptr;
    return ptr + SES_PS_Ptr_padlen;
}


/*********************************************************************** END */
/* to add a codec for type XXX:
 * - add above the codec function pair SES__EncodeXXX  SES__DecodeXXX
 * - define the data structure SES_PS_XXX in ses.h
 * - add the SES_XXX tag value in ses.h, increment SES_MAX_Builtins
 */
