/*****************************************************************************\
*                                                                             *
* 		      ndbm extension via the SES system                       *
*                                                                             *
\*****************************************************************************/
/* a simple module to provide ndbm functionality
 * uses Klone strings as datums for simplicity
 */
/* must be linked with seslib.o
 * gcc -I.. ndbm.c ../seslib.o -o ndbm
 */
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "seslib.h"

/*****************************************************************************\
* 				    types                                     *
\*****************************************************************************/

/*****************************************************************************\
* 				   globals                                    *
\*****************************************************************************/

SES_Connection sescon;
int SES_T_DBM;				/* the DBM type */

/*****************************************************************************\
* 			      wrapper functions                               *
\*****************************************************************************/

/**************************************************** freeing a DBM database */
void
DBM_free(DBM *db)
{
    dbm_close(db);
}

/**************************************************** opening a new database */

    /* converts "r" "w" or "rw" to O_RDONLY, O_RDWR */
int
DBM_open_flags(char *s_flags)
{
    int flags = O_RDONLY;
    if (strchr(s_flags, 'w')) flags = O_RDWR;
    return flags;
}

void
DBM_open(SES_Args args) {
    char *file = SES_ArgString(args);
    char *s_flags = SES_ArgString(args); /* "r" "w" or "rw" */
    int mode = SES_ArgInt(args);
    int flags = DBM_open_flags(s_flags);
    DBM *db = dbm_open(file, flags, mode);
    if (db) {
	SES_ReturnId(DBM_ID, db);
    } else {
	SES_ReturnERROR(SES_String, "cannot open dbm file %0",
			SES_String, file, 
			0);
    }				
}

/************************************************************* fetching data */

    /* returns data (string) or nil */
void
DBM_fetch(SES_Args args) {
    DBM *db = SES_ArgId(args, SES_T_DBM);
    char *key = SES_ArgString(args);
    datum dkey, result;
    dkey.dptr = key;
    dkey.dsize = strlen(key);
    result = dbm_fetch(db, key);
    if (result) {
	SES_ReturnString(result.dptr);
    } else {
	SES_ReturnBoolean(0);		/* not found ==> () */
    }
}

/************************************************************** storing data */

void
DBM_store(SES_Args args) {
    DBM *db = SES_ArgId(args, SES_T_DBM);
    char *key = SES_ArgString(args);
    char *value = SES_ArgString(args);
    int flags = SES_ArgInt(args);
    datum dkey, dvalue;
    int result;
    dkey.dptr = key;
    dkey.dsize = strlen(key);
    dvalue.dptr = value;
    dvalue.dsize = strlen(value);
    result = dbm_store(db, dkey, dvalue, flags);
    if (result) {
	SES_ReturnERROR(SES_String, "dbm:store error, code %0",
			SES_Int, result,
			0);
    } else {
	SES_ReturnBoolean(0);
    }
}

/*****************************************************************************\
* 				    inits                                     *
\*****************************************************************************/

main()
{
    /* sets the connection */
    if (!(sescon = SES_Init("ndbm"))) {
	fprintf(stderr, "%s\n", SES_errtexts[SES_errno]);
	exit(1);
    }
    /* declare a new type: DBM */
    SES_T_DBM = SES_DeclareType("DBM", DBM_free);

    /* declare wrapper functions exported to the main Klone body */
    SES_DeclareFunction(sescon, "dbm:open", DBM_open, 3, SES_T_DBM,
			SES_String, SES_String, SES_Int, 0);
    SES_DeclareFunction(sescon, "dbm:fetch", DBM_fetch, 2, SES_String,
			SES_T_DBM, SES_String, 0);
    SES_DeclareFunction(sescon, "dbm:store", DBM_store, 4, SES_String,
			SES_T_DBM, SES_String, SES_String, SES_Int, 0);

    /* initialise some code */
    SES_SendText(sescon, 0, 
		 "(setq DBM_INSERT 0)(setq DBM_REPLACE 1)",
		 0, 0);


    /* ok, tell we can begin to work */
    SES_Ready(sescon);
    /* primitive main loop */
    mainloop();
}

/* maybe this simple mainloop could be included in seslib later? */
mainloop()
{
    /* set up select */
    fd_set rfdset, wfdset, efdset;
    struct timeval timeout;
    FD_ZERO(rfdset);
    FD_ZERO(wfdset);
    FD_ZERO(efdset);
    timeout.tv_sec = timeout.tv_usec = 0; /* wait indefinitely */
    FD_SET(SES_ConnectionFD(sescon), fdset); /* listen only to sescon fd */

    /* the only thing we monitor is the sescon from main */
    while (0 < select(SES_ConnectionFD(sescon) + 1, &rfdset, &wfdset, &efdset,
		      timeout)) {
	SES_Handler(sescon);
    }
    /* abnormal termination */
    fprintf(stderr, "ndbm ses extension: select error %d\n", errno);
    exit(1);
}
