/*****************************************************************************\
*                                                                             *
* 				  SES lib                                     *
*                                                                             *
\*****************************************************************************/
/* Stubs to be included in each extension (not in main body)
 * you will also need ses.h
 * C-only, does not need Klone
 */

#ifndef SESLIB_H
#define SESLIB_H

#include "kclist.h"
#include "ses.h"
					/* types */
    /* a connection is a data structure in the extension */
typedef struct _SES_Connection {
    /* general parameters */
    int version;	 /* version of the ses protocol used (0) */
    int transport;	 /* transport layer type (pipe = 0, shm = 1) */
    int key_shm;	 /* the key of the IPC queue */
    int key_sem;	 /* the key of the IPC semaphores to the queue */
    int fd_in;	         /* file descriptor main->extension (unique) */
    int fd_out;	 	 /* file descriptor extension->main (shared) */
    int pid;	 	 /* the process ID of the parent application */
    int id;		 /* ID (byte) of the extension */
    KCList IDs;	 /* the array of defined types, indexed by IDs */
    int errhandler_reply;	/* answer of the Klone errhandler */
    /* connection-type specific parameters */
					/* --- input --- */
    uchar *inbuf;			/* buffer */
    uchar *inlimit;			/* its allocated size */
    uchar *inend;			/* end marker in it */
    uchar *inptr;			/* current (start) pointer in it */
					/* --- output --- */
    uchar *outbuf;			/* buffer */
    uchar *outlimit;			/* its allocated size */
    uchar *outptr;			/* start of freespace pointer in it,
					 semantically equal to outend */
} *SES_Connection;

typedef void (*sescon_functions_type)(); /* SES_Connection->functions */

					/* constants */
 
/*****************************************************************************\
* 			ANSI C  K&R compatibilty layer                        *
\*****************************************************************************/
/* defines a _P_ macro to include ANSI declaration prototypes */
#ifdef _P_
#undef _P_
#endif
#if defined(__STDC__) || defined(__cplusplus)
#   define _USING_PROTOTYPES_ 1
#   define _P_(x)	x
#   ifdef __cplusplus
#       define VARARGS (...)
#   else
#       define VARARGS ()
#   endif
#else /* K&R C */
#   define _P_(x)	()
#endif /* K&R C */

#ifdef __cplusplus
#   define EXTERN extern "C"
#else
#   define EXTERN extern
#endif
#ifndef va_start
# if __STDC__
# include <stdarg.h>
# else
# include <varargs.h>
# endif
#endif

/*****************************************************************************\
* 				 declarations                                 *
\*****************************************************************************/

/********************************************************************* types */
typedef struct _SES_TypeDesc {
    unsigned char *(*encoder)();
    unsigned char *(*decoder)();
} *SESTypeDesc;


typedef struct _SES_Args {		/* arguments parsing context */
    uchar *ptr;				/* ptr to start of next arg */
    int position;			/* its position (arity) */
    struct _SES_Connection *sescon;	/* global connection context */
    struct _SES_PS_Header *request;	/* header of the current request */
    SES_PS_Header return_result;	/* where to return data (or 0) */
} *SES_Args;

typedef int SES_InputId;
typedef int SES_TimerId;
typedef void (*SES_Callback) _P_((void *, int, SES_InputId));
typedef void (*SES_TimerCallback) _P_((void *, SES_TimerId));

/***************************************************************** variables */
extern int SES_errno;
extern char *SES_errtexts[];
extern void (*SES_FatalIOErrorHandler) _P_((SES_Connection sescon, int num));

SESTypeDesc SES_typedescs;

/********************************* internal functions, dont use, may change! */
#define SESWriteBE32(p, n) (*(p)++ = (n)>>24, *(p)++ = ((n)>>16)&0xff, \
     *(p)++ = ((n)>>8)&0xff,  *(p)++ = (n)&0xff)

/*****************************************************************************\
* 				     API                                      *
\*****************************************************************************/

#define SES_ConnectionFD(con) ((con)->fd_in)

					/* functions */

extern SES_Connection SES_Init _P_((char *name));

extern void SES_MainLoop _P_((SES_Connection sescon));

extern void SES_MainPreSelect _P_((SES_Connection sescon));

void SES_Ready _P_((SES_Connection sescon));

extern int SES_DeclareFunction _P_((SES_Connection sescon, char *name,  
				    void (*func)(SES_Args args), 
				    int arity, int return_type, ...));

extern int SES_DeclareType _P_((SES_Connection sescon, 
				char *name, void (*free)()));

extern void SES_Handler _P_((SES_Connection sescon));

extern int SES_SendText _P_((SES_Connection sescon, void (*callback)(), 
			char *text, int length, void *closure, int blocking));

extern void SES_Flush _P_((SES_Connection sescon));

extern SES_InputId SES_AddInput _P_((SES_Connection sescon,
				     int source, int mask,
				     SES_Callback callback, void *client_data));

extern void SES_RemoveInput _P_((SES_InputId inputid));

extern SES_TimerId SES_AddTimeOut _P_((SES_Connection sescon,
				  unsigned long interval,
				  SES_TimerCallback callback,
				  void *client_data));

extern void SES_RemoveTimeOut _P_((SES_TimerId timerid));

					/* SES_ArgXxx */
extern int SES_ArgTypeOf _P_((SES_Args args));
extern int SES_ArgNumberOf _P_((SES_Args args));

extern int SES_ArgInt _P_((SES_Args args));
extern char *SES_ArgString _P_((SES_Args args));
extern char *SES_ArgAtom _P_((SES_Args args));
extern void *SES_ArgId _P_((SES_Args args, int IdType));
extern int SES_ArgBoolean _P_ ((SES_Args args));
extern double SES_ArgReal _P_ ((SES_Args args));
extern void *SES_ArgPtr _P_ ((SES_Args args));
					/* SES_ReturnXxx */
extern void SES_ReturnInt _P_((SES_Args args, int n));
extern void SES_ReturnString _P_((SES_Args args, char *string));
extern void SES_ReturnAtom _P_((SES_Args args, char *string));
extern void SES_ReturnId _P_((SES_Args args, int IdType, void *value));
extern void SES_ReturnBoolean _P_((SES_Args args, int n));
extern void SES_ReturnReal _P_((SES_Args args, double number));
extern void SES_ReturnPtr _P_((SES_Args args, void *ptr));
extern void SES_ReturnERROR _P_((SES_Args args, char *error_argument_list));


/******************************************** obsolete calls for back-compat */

#endif /* SESLIB_H */
