/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */
/*****************************************************************************\
*                                                                             *
* 			     WPPB parser routines                             *
*                                                                             *
\*****************************************************************************/
#include <sys/types.h>
#include <sys/stat.h>

#define SML_LIGHT

/* COLAS: here are some patches for testing this in wool... */
#ifndef KOALA
#include "sml.h"
#else
#include "EXTERN.h"
#include "wool.h"
#include "wl_number.h"
#include "wl_atom.h"
#include "wl_list.h"
#include "wl_coll.h"
#include "wl_string.h"
#include "wl_stream.h"
#include "wl_func.h"
#include "wlparser.h"

#define SmlObject WlO
#define SmlString WlString
#define SmlClassString WlStringType
#define SmlStream WlStream
#define SmlClassInt WlNumberType
#define SmlTRUE TRU
#define SmlCValue WlO
#define SmlFunctionDefine(name, func, arity) WlDeclareSubr(func, name, arity)

SmlConstantSetCValue(name, wltype, value)
    char *name;
    WlType wltype;
    void *value;
{
    WlO val;
    if (wltype == WlNumberType)
	val = (WlO) WlNumberMake(value);
    else if (wltype == WlStringType)
	val = (WlO) WlStringMake(value);
    else 
	WlBadArgument(wltype, 0, "string or int");
    WlConstantMake(name, val);
}

#define SmlArgumentMustBe(pos, obj, t) WlArgumentMustBe(obj, pos, t)
#define SmlObjectGetCValue(o) ((WlNumber) o)->number

    void sml_display_warning_on_tty(char *foo) {}
void (*sml_display_warning) (char *);
    int ResetSession(int i) {return 1;}
char *SmlWarningMessageMagicMI = "SML WARNING";
char *SmlErrorMessageMI = "SML ERROR";
#endif /* KOALA */

#ifdef WPPB
#include "wlwppb.h"

extern int (*SmlCheckPointer) (void);
extern unsigned char *WlWPPBReadEval(unsigned char *, WlO *);
extern unsigned char *WlWPPBReadSet(unsigned char *, WlO *);
extern unsigned char *WlWPPBReadEvalUnparse(unsigned char *, WlO *);
extern unsigned char *WlWPPBReadSetUnparse(unsigned char *, WlO *);

struct WlWPPBTable_mod {
    int index;
    WlWPPBReadFunction func;
};

static WlO WlWPPBWriteNull(WlO, WlStream);
static int WlWriteByte(unsigned int, WlStream);
static int WlCountItems(WlO *, int);
static void WlWPPBUnparseBufferRestore(struct _WlWPPBContext *,
				       struct WlWPPBTable_mod *);
static int WlWPPBUnparseBuffer(unsigned char *, unsigned char *,
			       WlStream, int, WlO);


/*****************************************************************************\
* 			     Some Write Routines                              *
\*****************************************************************************/
/* not many, just for convenience since for writing no performance is needed
 */

static void SmlEncrypt (int len, char *s)
{
  int i;
  int car, key = 7;
  for (i=0; i<len; i++) {
    car = (s[i] ^= key);
    key = car;
  }
}

static SmlObject SmlWPPBCryptString(SmlObject obj)
{
  SmlString string = (SmlString) obj;

  SmlArgumentMustBe(0, obj, SmlClassString);
  SmlEncrypt(WlStringLength(string), string->string);
  return (SmlObject) string;
}

static SmlObject SmlWriteBE32(SmlObject integer, SmlStream stream)
{
  unsigned int value = SmlObjectGetCValue (integer);
  unsigned int neg = 128;
  int i;

  SmlArgumentMustBe(0, integer, SmlClassInt);
  WlMustBeStream(stream, 1);
  WlStreamCanWRITE(stream);
  if (value < 0)
    WlSPutc((value >> 24) + neg, stream);
  else
    WlSPutc((value >> 24), stream); 
  WlSPutc((value >> 16), stream);
  WlSPutc((value >> 8), stream);
  WlSPutc(value, stream);
  return SmlTRUE;
}

static SmlObject SmlWriteBE24(SmlObject integer, SmlStream stream)
{
  unsigned int value = SmlObjectGetCValue (integer);
  int i;

  SmlArgumentMustBe(0, integer, SmlClassInt);
  WlMustBeStream(stream, 1);
  WlStreamCanWRITE(stream);
  WlSPutc((value >> 16),  stream);
  WlSPutc((value >> 8), stream);
  WlSPutc(value, stream);
  return SmlTRUE;
}

static SmlObject SmlWriteBE16(SmlObject integer, SmlStream stream)
{
  unsigned int value = SmlObjectGetCValue (integer);
  int i;

  SmlArgumentMustBe(0, integer, SmlClassInt);
  WlMustBeStream(stream, 1);
  WlStreamCanWRITE(stream);
  WlSPutc((value >> 8), stream);
  WlSPutc(value, stream);
  return SmlTRUE;
}

static SmlObject SmlWriteBE8(SmlObject integer, SmlStream stream)
{
  unsigned int value = SmlObjectGetCValue (integer);

  SmlArgumentMustBe(0, integer, SmlClassInt);
  WlMustBeStream(stream, 1);
  WlStreamCanWRITE(stream);
  WlSPutc(value, stream);
  return SmlTRUE;
}

/* THE main dispatch function is: (wppb:write obj stream)
 * here is a C function to call it
 * must return t if OK
 */

static WlO
WlWPPBWrite(obj, stream)
    WlO obj;
    WlStream stream;
{
    WlMustBeStream(stream, 1);
    return obj ? WlApply2(WlA_WPPB_write, obj, stream)
	: WlWPPBWriteNull(obj, stream);
}

static WlO
WlWPPBWriteNull(obj, stream)
    WlO obj;				/* 0 */
    WlStream stream;
{
    WlSPutc(WlWPPBType_Null, stream);
    return TRU;
}

/* (wppb:write-expr func stream) 
 * WARNING: if () is returned, the write was aborted, you MUST step back and 
 * write the func as a RAW item (normally t is retuned if OK)
 */

static WlO
WlWPPBWriteExpr(obj, stream)
    WlExpr obj;
    WlStream stream;
{
#define PutByte(b) if (WlWriteByte(byte = (b), stream)) return NIL
    int byte, i;
    if (!WlIsALambda(obj))
	WlMustBeExpr(obj, 0);
    WlMustBeStream(stream, 1);
    
    WlWriteByte(WlWPPBType_Expr, stream); /* header */
    if (obj->type == WlFExprType) { 
	WlSPutc(1, stream);
    } else if (obj->type == WlMExprType) { 
	WlSPutc(2, stream);
    } else {
	WlSPutc(0, stream);
    }

    PutByte(obj->arity);		/* base args */
    for (i = 0; i < obj->arity; i++)
	WlWPPBWrite(obj->parameters[i], stream);
    PutByte((obj->body_size) >> 8);	/* body */
    PutByte((obj->body_size) & 0xff);
    for (i = 0; i < obj->body_size; i++)
	WlWPPBWrite(obj->body[i], stream);
    if (obj->lambdalist) {		/* fancy args &... */
	WlSPutc(1, stream);
	PutByte(obj->lambdalist->nfixed);
	PutByte(obj->lambdalist->variable_arity);
	PutByte(WlCountItems(obj->lambdalist->optionals, 1));
	for (i = 0; i < byte; i++)
	    WlWPPBWrite(obj->lambdalist->optionals[i], stream);
	if (obj->lambdalist->rest) {
	    WlSPutc(1, stream);
	    WlWPPBWrite(obj->lambdalist->rest, stream);
	} else {
	    WlSPutc(0, stream);
	}
	if (obj->lambdalist->whole) {
	    WlSPutc(1, stream);
	    WlWPPBWrite(obj->lambdalist->whole, stream);
	} else {
	    WlSPutc(0, stream);
	}
	PutByte(WlCountItems((WlO *) obj->lambdalist->key, 
			     sizeof(struct _WlKeyDecls) / WLSO));
	for (i = 0; i < byte; i++) {
	    WlWPPBWrite(obj->lambdalist->key[i].key, stream);
	    WlWPPBWrite(obj->lambdalist->key[i].init, stream);
	}
	PutByte(WlCountItems(obj->lambdalist->aux, 1));
	for (i = 0; i < byte; i++)
	    WlWPPBWrite(obj->lambdalist->aux[i], stream);
    } else {
	WlSPutc(0, stream);
    }
    return TRU;
}

/* write a byte but aborts if number dont fit in a byte */
static int
WlWriteByte(b, stream)
    unsigned int b;
    WlStream stream;
{
    return (b > 255) ? 1 : (WlSPutc(b, stream), 0);
}

/* count null-terminated items */
static int
WlCountItems(p, size)
    WlO *p;
    int size;				/* in WLSO increments */
{
    int i;
    if (!p) return 0;
    for (i = 0; *p; i++) p += size;
    return i;
}

/*****************************************************************************\
* 				  deparsing                                   *
\*****************************************************************************/
/* deparsing is like parsing, but with more checks (speed is not important)
 * and instead of evaluating on the fly the read items, just dump them
 * into the output file
 * For modularity, we just use the parse routines, modifying only the 
 * side-effects ones such as EVAL and SETQ
 */


/* Unparse a (pre-opened) WPPB file into a (pre-opened) text file
 * formats optionally with the given pretty-printer
 * returns 1 = success, 0 = error
 */

static int
WlUnparseWPPBFile(filename, fp, out, verbose, pp)
    char *filename;			/* name of input file */
    FILE *fp;				/* opened input file (fp) */
    WlStream *out;			/* wool stream, not fp */
    int verbose;			/* flag */
    WlO pp;				/* WlO or NULL=raw */
{
    struct stat buf;
    int status = 1;
    int version;
    int stackptr;
    int checksum = 0;

    if (!fstat(WlFp2Fd(fp), &buf) && (buf.st_mode & S_IFREG)) {
	/* gobble whole file into a buffer */
	WlGCMark();
	WlString string = WlStringNMake(buf.st_size);
	unsigned char *buffer = (unsigned char *) string->string;
	int size = fread(buffer, 1, buf.st_size, fp);
	unsigned char *p = buffer, *last = p + size;
	int i;
	struct _WlWPPBContext context;

	fclose(fp);			/* buffer is gobbled, file not needed */
#ifdef ISM
        /* there, we must decrypt the buffer in place and return size, which 
         must be lesser or equal to original size */
	size = WlWPPBDecrypt(buffer, size);
	last = buffer + size;
	buf.st_size = size;
#endif

#ifdef DEBUG
	WlWPPBContextSave(&context, filename, buffer, last);
#else
	WlWPPBContextSave(&context, filename, buffer);
#endif

	buffer[size] = '\0';
#ifdef ISM
	if (size < 38) {		/* verify that header is there */
#else
	if (size < 14) {		/* verify that header is there */
#endif
	    if (size == 0)
		WlLoadWPPBError(p, "empty file!%s", (long) "");
	    else
		WlLoadWPPBError(p, "file too short to be a .wlp (%d bytes)", 
				(long) size);
	    goto error;
	}
	/* check header info */
	if (WlWPPB_MAGIC != (i = WlIntBE32Read(p))) {
	    WlLoadWPPBError(p, "bad magic number %d", (long) i);
	    goto error;
	}
	p += 4;
#ifdef ISM
	p += 20;
	if (SmlCheckPointer != NULL)
	  if ((*SmlCheckPointer)() < (i = WlIntBE32Read(p))) 
	    fprintf (stderr,
		     "Warning: File %s have bad Sml Version Number %d\n",
		     WlWPPBCurrentContext->filename, (long) i);
	p += 4;
#endif
	if (size != WlIntBE32Read(p) || size != buf.st_size) {
	    WlLoadWPPBError(p, "bad size %d", (long) size);
	    goto error;
	}
	/* here we check the checksum, we have time */
	p += 4;
	if (checksum = WlIntBE32Read(p)) { /* check=0 means file non checked */
	    /* remove checksum form count */
	    int comp_checksum = - (*p + p[1] + p[2] + p[3]);
	    unsigned char *q;
	    for (q = buffer; q < last; q++)
		comp_checksum += *q;
	    comp_checksum += size;
	    if (comp_checksum != checksum)
		WlLoadWPPBError(p, "bad checksum, computed %d", 
				(long) comp_checksum);
	}
	p += 5;				/* version read by WlWPPBContextSave */
	/* main loop: read items */
	stackptr = WlStackPtr;
	WlStackFramePush(1, &WlA_star_package, NIL, NIL);
	{
	    /* modify parse table not to do side-effects */
	    /* tracks mods to be undone by WlWPPBUnparseBufferRestore */
	    struct WlWPPBTable_mod WlWPPBTable_mods[3];
	    WlWPPBTable_mods[0].index = WlWPPBType_Eval;
	    WlWPPBTable_mods[0].func = WlWPPBReadEval;
	    WlWPPBTable_mods[1].index = WlWPPBType_Set;
	    WlWPPBTable_mods[1].func = WlWPPBReadSet;
	    WlWPPBTable_mods[2].func = 0;

	    WlWPPBTable[WlWPPBType_Eval] = WlWPPBReadEvalUnparse;
	    WlWPPBTable[WlWPPBType_Set] = WlWPPBReadSetUnparse;

	    WlUnwindProtectStatement
		(status = WlWPPBUnparseBuffer(p, last, (WlStream) out,
					      verbose, pp),
		 WlWPPBUnparseBufferRestore, &context, WlWPPBTable_mods);
	}
	WlStackFramePopNormal(stackptr);
	WlGC();
	return status;
    error:
	WlWPPBContextRestore(&context);
	WlGC();
	return 0;
    } else {
	return 0;
    }
}

static void
WlWPPBUnparseBufferRestore(contextp, table)
    struct _WlWPPBContext *contextp;
    struct WlWPPBTable_mod *table;
{
    int i;
    for (i = 0; table[i].func; i++) {
	WlWPPBTable[table[i].index] = table[i].func;
    }
    WlWPPBContextRestore(contextp);
}

/* unparsing a buffer */

static int					/* 1 = success, 0 = error */
WlWPPBUnparseBuffer(p, last, out, verbose, pp)
    unsigned char *p, *last;
    WlStream out;
    int verbose;
    WlO pp;
{
    WlO obj;				/* discarded at toplevel */
    jmp_buf jbuf;
    int res;

    if (res = setjmp(WlWPPBCurrentContext->jbuf)) { /* EOF */
	if (verbose == 1)
	    fprintf(WlStreamWriteFd(WlStderr), "\n");
	return res - 1;			/* longjmp 2 = success, 1 = error */
    } else {				/* main loop */
	for (;;) {
	    if (p >= last) {
		WlLoadWPPBError(p, "reading past buffer end%s!", "");
		return 1;
	    }
	    if (*p >= WlWPPBTableSize || !WlWPPBTable[*p]) {
		WlLoadWPPBError(p, "bad item code %d", (long) *p);
		return 0;
	    }
	    obj = 0;
	    p = (*(WlWPPBTable[*p]))(p+1, &obj);
	    /* now, difference from parse: we print p on out */
	    /* we strip top-level evals */
	    if (!obj)
		continue;		/* some functions dont fill &obj */

	    /* simplifications */
	    if (WlIsACollection(obj)) {
		WlCollection col = (WlCollection) obj;
		if (col->list[0] == (WlO) WlA_identity)
		    obj = col->list[1];
		else 
		    obj = (WlO) WlListMake(col);
	    }
	    if (WlIsAList(obj)		/* setq lambda ==> defun */
		&& ((WlList)obj)->size == 3
		&& WlIsALambda(((WlList)obj)->list[2])
		&& ((WlList)obj)->list[0] == (WlO) WlA_setq) {
		WlAtom defun = 0;
		WlList ll = (WlList) WlListCoerce(WlListType, 
						  ((WlList)obj)->list[2]);
		WlAtom lambda = (WlAtom) ll->list[0];
		if (lambda == WlA_lambda) 
		    defun = WlIntern("defun");
		else if (lambda == WlA_lambdaq)
		    defun = WlIntern("defunq");
		else if (lambda == WlA_lambdam)
		    defun = WlIntern("defmacro");
		if (defun) {
		    int i;
		    WlList new = WlListNMake(ll->size + 1);
		    WlListStore(new, 0, defun);
		    WlListStore(new, 1, ((WlList)obj)->list[1]);
		    for (i = 1; i < ll->size; i++)
			WlListStore(new, i+1, ll->list[i]);
		    obj = (WlO) new;
		}
	    }

	    /* if verbose, print what we are doing */
	    switch (verbose) {
	    case 0: break;
	    case 2: {
		WlString s = (WlString) WlApply3
		    (WlIntern("print-format")->c_val, WlStringType,
		     WlStringMake(" -> %0\n"), obj);
		if (s->size > 70)
		    s->string[69] = '\n', s->string[70] = '\0';
		fprintf(WlStreamWriteFd(WlStderr), s->string);
		fflush(WlStreamWriteFd(WlStderr));
		break;
	    }
	    default:
		fprintf(WlStreamWriteFd(WlStderr), "."),
		fflush(WlStreamWriteFd(WlStderr));
	    }
	    if (pp) {			/* use pretty-printer */
		WlStackFramePushValuesV(1, WlA_stdout, out);
		WlApply1(pp, obj);
		WlStackFramePop();
	    } else {
		WlStackFramePushValuesV(1, WlA_print_readably, TRU);
		WlSend_print(obj, out);
		WlStackFramePop();
	    }
	    WlSPutc('\n', out);
	}
    }
}


/* High-level Wool interface to WlUnparseWPPBFile:
 * (wppb:unpreparse [:print pretty-printer] files...)
 * for each file, generates its unparsed form in a <filename>.unp file
 * to avoid clobbing original file if present
 */

static WlO
WlWPPBUnparse(argc, argv)
    int argc;
    WlString *argv;
{
    WlString filename;
    WlString outname;
    WlString args[2];
    int i;
    FILE *in, *out;
    WlStream wl_out;
    WlO pretty = 0;
    int verbose = 0;

    for (i = 0; i < argc; i++) {
	filename = argv[i];
	if (WlIsAKeyword(filename)) {
	    if (filename == (WlString) WlK_print) {
		i++;
		pretty = WlTrueP(argv[i]) ? (WlO) argv[i] : 0;
		continue;
	    } else if (filename == (WlString) WlK_verbose) {
		i++;
		if (argv[i] == (WlString) WlIntern(":debug"))
		    verbose = 2;
		else 
		    verbose = WlTrueP(argv[i]);
		continue;
	    } else {
		return WlError2(WlE_INVALID_KEYWORD, argv[i],
				WlListMakeV(2, WlK_print, WlK_verbose));
	    }
	}
	WlMustBeString(filename, i);
	
	args[0] = filename;
	args[1] = WlStringMake(".unp");	/* do it each time, for GC? */
	outname = (WlString) WlStringAdd(2, args); /* concat */
	if (verbose)
	    fprintf(WlStreamWriteFd(WlStderr), "unpreparsing %s into %s\n",
		    filename->string, outname->string);
	if (!(in = fopen(filename->string, "r"))) 
	    return (WlO) WlOpenError(0, filename, WlK_input);
	if (!(out = fopen(outname->string, "w"))) 
	    return (WlO) WlOpenError(0, outname, WlK_output);
	wl_out = WlStreamMake(out, 2, outname->string);

	WlUnparseWPPBFile(filename->string, in, wl_out, verbose, pretty);
    }
    return NIL;
}

/*****************************************************************************\
* 			 wppb:write-obj misc methods                          *
\*****************************************************************************/
/* COLAS BEGIN 
 * conversion of all the  defmethod wppb:write-obj of sml_preparse.sml
 * into C code for securtity (crypting) reasons.
 * these routines are declared at the end of the custom_init() function
 */

/**************************************************************** misc utils */

WlAtom WlA_WPPB_write;
WlO WlWPPB_WriteObj_Atom();

/* wppb:write-eval 
 */

WlO
WlWPPB_WriteEval(obj, stream)
    WlO obj;
    WlStream stream;
{
    WlSPutc(WlWPPBType_Eval, stream);
    WlApply2(WlSend_eval(WlA_WPPB_write), obj, stream);
    return TRU;
}

/*  wppb:write-undefined
 */

WlO
WlWPPB_WriteUndefined(stream)
    WlStream stream;
{
    return WlWPPB_WriteObj_Atom(WlUndef, stream);
}

/* first, some byte-writing functions, taking a C int as argument
 */

WlWriteBE32(value, stream)
    unsigned int value;
    WlStream stream;
{
    WlSPutc(0xff & (value >> 24), stream); 
    WlSPutc(0xff & (value >> 16), stream);
    WlSPutc(0xff & (value >> 8), stream);
    WlSPutc(0xff & value, stream);
}

WlWriteBE24(value, stream)
    unsigned int value;
    WlStream stream;
{
    WlSPutc(0xff & (value >> 16), stream);
    WlSPutc(0xff & (value >> 8), stream);
    WlSPutc(0xff & value, stream);
}

WlWriteBE16(value, stream)
    unsigned int value;
    WlStream stream;
{
    WlSPutc(0xff & (value >> 8), stream);
    WlSPutc(0xff & value, stream);
}

WlWriteBE8(value, stream)
    unsigned int value;
    WlStream stream;
{
    WlSPutc(0xff & value, stream);
}


/********************************************************************** Atom */

WlO
WlWPPB_WriteObj_Atom(obj, stream)
    WlAtom obj;
    WlStream stream;
{
    if (obj == (WlAtom) TRU) {
	WlSPutc(WlWPPBType_True, stream);
    } else {
	WlString name = WlStringMake(obj->p_name);
	if (name->size > 255) {
	    WlSPutc(WlWPPBType_AtomL, stream);
	    WlWriteBE32(name->size, stream);
	} else {
	    WlSPutc(WlWPPBType_Atom, stream);
	    WlWriteBE8(name->size, stream);
	}
	WlWriteBE32(WlStringHash(name, stream));
	SmlWPPBCryptString((WlO) name);
	WlSPutBytes(name->size, name->string, stream);
    }
    return TRU;
}

/******************************************************************** String */

WlO
WlWPPB_WriteObj_String(obj, stream)
    WlString obj;
    WlStream stream;
{
    WlString new;
    if (obj->size > 255) {
	WlSPutc(WlWPPBType_StringL, stream);
	WlWriteBE32(obj->size, stream);
    } else {
	WlSPutc(WlWPPBType_String, stream);
	WlWriteBE8(obj->size, stream);
    }
    new = WlStringCopy(obj);
    SmlWPPBCryptString((WlO) new);
    WlSPutBytes(new->size, new->string, stream);
    return TRU;
}

/********************************************************************** List */

WlO
WlWPPB_WriteObj_List(obj, stream)
    WlList obj;
    WlStream stream;
{
    if (obj == (WlList) NIL) {
	WlSPutc(WlWPPBType_Nil, stream);	
    } else {
	int i;

	if (obj->size > 255) {
	    WlSPutc(WlWPPBType_ListML, stream);
	    WlWriteBE32(obj->size, stream);
	    if (WlIsAVector(obj))
		WlWriteBE8(1, stream);
	    else if (WlIsALocator(obj))
		WlWriteBE8(2, stream);
	    else
		WlWriteBE8(0, stream);
	} else if (obj->type == WlListType) {
	    WlSPutc(WlWPPBType_List, stream);
	    WlWriteBE8(obj->size, stream);
	} else {
	    WlSPutc(WlWPPBType_ListM, stream);
	    WlWriteBE8(obj->size, stream);
	    if (WlIsAVector(obj))
		WlWriteBE8(1, stream);
	    else if (WlIsALocator(obj))
		WlWriteBE8(2, stream);
	    else
		WlWriteBE8(0, stream);
	}
					/* recurse print on elements */
	for (i = 0; i < obj->size; i++) {
	    WlApply2(WlSend_eval(WlA_WPPB_write), obj->list[i], stream);
	}
    }
    return TRU;
}

/******************************************************************* Inlines */

WlO
WlWPPB_WriteObj_ListInline(obj, stream)
    WlList obj;
    WlStream stream;
{
    WlList l = (WlList) WlListCopy(obj);
    l->type = WlListType;
    return WlWPPB_WriteEval(l, stream);
}

/**************************************************************** QuotedExpr */

WlO
WlWPPB_WriteObj_QuotedExpr(obj, stream)
    WlQuotedExpr obj;
    WlStream stream;
{
    int type = 0;

    WlSPutc(WlWPPBType_QuotedExpr, stream);
    if (obj->symbol == WlA_quote) {
	type = 0;
    } else if (obj->symbol == WlA_backquote) {
	type = 1;
    } else if (obj->symbol == WlA_unquote) {
	type = 2;
    } else if (obj->symbol == WlA_unquotesplicing) {
	type = 3;
    }
    WlWriteBE8(type, stream);
    if (obj->expr == WlUndef) 
	WlWPPB_WriteUndefined(stream);
    else
	WlApply2(WlSend_eval(WlA_WPPB_write), obj->expr, stream);
    return TRU;
}

/*********************************************************************** Int */

WlO
WlWPPB_WriteObj_Int(obj, stream)
    WlNumber obj;
    WlStream stream;
{
    if (obj->number < -2) {
	WlSPutc(WlWPPBType_IntN, stream);
	WlWriteBE32(- obj->number, stream);
    } else if (obj->number > 20) {
	WlSPutc(WlWPPBType_IntP, stream);
	WlWriteBE32(obj->number, stream);
    } else {
	WlSPutc(WlWPPBType_IntS, stream);
	WlWriteBE8(obj->number + 2, stream);
    }
    return TRU;
}

/********************************************************************** Real */

WlO
WlWPPB_WriteObj_Real(obj, stream)
    WlReal obj;
    WlStream stream;
{
    int l;
    char s[256];

    sprintf(s, "%.20g", obj->real);
#ifdef ISM
    {
	/* Substitute foreign decimal character (,) by dot.
	   In fact, the decimal character is returned by a call to 
	   o nl_langinfo(RADIXCHAR) or,
	   o localeconv()->decimal_point
	   but both solutions failed on my system. */
	char *ptr;
	if(ptr = strchr(s, ','))
	    *ptr = '.';
    }
#endif /* ISM */
    if (!strchr(s, '.') && !strchr(s, 'e')) { /* prints as an integer */
	strcat(s, ".0");		/* adds a dot to show it is a real */
    }

    WlSPutc(WlWPPBType_Real, stream);
    l = strlen(s);
    WlWriteBE8(l, stream);
    WlSPutBytes(l, s, stream);
    WlWriteBE8(0, stream);
    return TRU;
}

/********************************************************************** Expr */

WlO
WlWPPB_WriteObj_Expr(obj, stream)
    WlExpr obj;
    WlStream stream;
{
    WlNumber pos = (WlNumber) WlFseek(1, &stream);
    if (WlFalseP(WlWPPBWriteExpr(obj, stream))) {
	WlO argv[2];
	argv[0] = (WlO) stream;
	argv[1] = (WlO) pos;
	WlFseek(2, argv);
	WlWPPB_WriteEval(WlListCoerce(WlListType, obj), stream);
    }
    return TRU;
}

#ifdef UNSIGNED
/****************************************************************** Unsigned */

WlO
WlWPPB_WriteObj_Unsigned(obj, stream)
    WlUnsigned obj;
    WlStream stream;
{
    WlSPutc(WlWPPBType_IntU, stream);
    WlWriteBE32(obj->number, stream);
    return TRU;
}

#endif /* UNSIGNED */    

/* COLAS END 
 */

/**************************************************************************\
* 				     Init                                  *
\**************************************************************************/
static void sml_check_session_toolkit (void)
{
  extern int ResetSession();
  extern void (*sml_display_warning) (char *);
  extern char *SmlWarningMessageMagicMI;
  extern char *SmlErrorMessageMI;
  extern void sml_display_warning_on_tty (char *);
  static int warning_seen = 0;

  sml_display_warning = sml_display_warning_on_tty;

  /* First ask for the given Marketting identifier. If not present ...
     ask for the joker */
  if (ResetSession (67) != 1) { /* MI TOOLKIT = 67 */
    if (ResetSession (0) == 1) {
      /* Display a warning about the use of the Magic MI (= 0)*/
      /* ISM98: display only once the warningMessageMagicMI */
      if (!warning_seen) {
	(*sml_display_warning) (SmlWarningMessageMagicMI);
	warning_seen = 1;
      }
    } else {
      fprintf (stderr, SmlErrorMessageMI);
      putc ('\n', stderr);
      fflush (stderr);
      exit (10);
    }
  }
}

void custom_init()
{
  sml_check_session_toolkit();
/*
  SmlExoAddMethod(SmlClassAtom, "wppb:write-obj", SmlWPPBWriteAtom, 2);
  SmlExoAddMethod(SmlClassInt, "wppb:write-obj", SmlWPPBWriteInt, 2);
*/
  SmlFunctionDefine("wppb:writeBE32", SmlWriteBE32, 2);
  SmlFunctionDefine("wppb:writeBE24", SmlWriteBE24, 2);
  SmlFunctionDefine("wppb:writeBE16", SmlWriteBE16, 2);
  SmlFunctionDefine("wppb:writeBE8", SmlWriteBE8, 2);
  SmlFunctionDefine("wppb:write-expr", WlWPPBWriteExpr , 2);
/*  SmlFunctionDefine("wppb:crypt-string", SmlWPPBCryptString, 1);*/
  SmlFunctionDefine("wppb:unpreparse", WlWPPBUnparse, NARY);
  SmlConstantSetCValue ("wppb:type:XType", SmlClassInt,
			(SmlCValue) WlWPPBType_XType);
  SmlConstantSetCValue ("wppb:type:Raw", SmlClassInt,
			(SmlCValue) WlWPPBType_Raw);
  SmlConstantSetCValue ("wppb:type:Set", SmlClassInt,
			(SmlCValue) WlWPPBType_Set);
  SmlConstantSetCValue ("wppb:type:SetFunc", SmlClassInt,
			(SmlCValue) WlWPPBType_SetFunc);
  SmlConstantSetCValue ("wppb:type:Eval", SmlClassInt,
			(SmlCValue) WlWPPBType_Eval);
  SmlConstantSetCValue ("wppb:type:WPPDInclude", SmlClassInt,
			(SmlCValue) WlWPPBType_WPPDInclude);
  SmlConstantSetCValue ("wppb:type:WPPDRef", SmlClassInt,
			(SmlCValue) WlWPPBType_WPPDRef);
  SmlConstantSetCValue ("wppb:type:WPPDRefL", SmlClassInt,
			(SmlCValue) WlWPPBType_WPPDRefL);
  SmlConstantSetCValue ("wppb:type:Atom", SmlClassInt,
			(SmlCValue) WlWPPBType_Atom);
  SmlConstantSetCValue ("wppb:type:AtomL", SmlClassInt,
			(SmlCValue) WlWPPBType_AtomL);
  SmlConstantSetCValue ("wppb:type:String", SmlClassInt,
			(SmlCValue) WlWPPBType_String);
  SmlConstantSetCValue ("wppb:type:StringL", SmlClassInt,
			(SmlCValue) WlWPPBType_StringL);
  SmlConstantSetCValue ("wppb:type:List", SmlClassInt,
			(SmlCValue) WlWPPBType_List);
  SmlConstantSetCValue ("wppb:type:ListM", SmlClassInt,
			(SmlCValue) WlWPPBType_ListM);
  SmlConstantSetCValue ("wppb:type:ListML", SmlClassInt,
			(SmlCValue) WlWPPBType_ListML);
  SmlConstantSetCValue ("wppb:type:IntS", SmlClassInt,
			(SmlCValue) WlWPPBType_IntS);
  SmlConstantSetCValue ("wppb:type:IntP", SmlClassInt,
			(SmlCValue) WlWPPBType_IntP);
  SmlConstantSetCValue ("wppb:type:IntN", SmlClassInt,
			(SmlCValue) WlWPPBType_IntN);
  SmlConstantSetCValue ("wppb:type:IntU", SmlClassInt,
			(SmlCValue) WlWPPBType_IntU);
  SmlConstantSetCValue ("wppb:type:Real", SmlClassInt,
			(SmlCValue) WlWPPBType_Real);
  SmlConstantSetCValue ("wppb:type:QuotedExpr", SmlClassInt,
			(SmlCValue) WlWPPBType_QuotedExpr);
  SmlConstantSetCValue ("wppb:type:Expr", SmlClassInt,
			(SmlCValue) WlWPPBType_Expr);
  SmlConstantSetCValue ("wppb:type:Nil", SmlClassInt,
			(SmlCValue) WlWPPBType_Nil);
  SmlConstantSetCValue ("wppb:type:True", SmlClassInt,
			(SmlCValue) WlWPPBType_True);
  SmlConstantSetCValue ("wppb:type:WPPDHere", SmlClassInt,
			(SmlCValue) WlWPPBType_WPPDHere);
  SmlConstantSetCValue ("wppb:type:Null", SmlClassInt,
			(SmlCValue) WlWPPBType_Null);
  SmlConstantSetCValue ("wppb:type:EOF", SmlClassInt,
			(SmlCValue) WlWPPBType_EOF);
  SmlConstantSetCValue ("wppb:extension-string", SmlClassString,
			(SmlCValue) WPPB_EXT);
  SmlConstantSetCValue ("wppb:extension-wld", SmlClassString,
			(SmlCValue) WPPD_EXT);
  SmlConstantSetCValue ("wppb:extension-wdd", SmlClassString,
			(SmlCValue) ".wdd");

/*  COLAS: performs the various defmethods for wppb:write-obj
 */
  {
      WlAtom selector = WlIntern("wppb:write-obj");
      unsigned int flags = WlGenericFunctionDeclare(selector);

      WlA_WPPB_write = WlIntern("wppb:write");

      WlTypeMethodAdd(WlAtomType, selector, 
		      WlSubrMake(WlSubrType, WlWPPB_WriteObj_Atom, 2));
      WlTypeMethodAdd(WlStringType, selector,
		      WlSubrMake(WlSubrType, WlWPPB_WriteObj_String, 2));
      WlTypeMethodAdd(WlListType, selector,
		      WlSubrMake(WlSubrType, WlWPPB_WriteObj_List, 2));
      WlTypeMethodAdd(WlListInlineType, selector,
		      WlSubrMake(WlSubrType, WlWPPB_WriteObj_ListInline, 2));
      WlTypeMethodAdd(WlQuotedExprType, selector,
		      WlSubrMake(WlSubrType, WlWPPB_WriteObj_QuotedExpr, 2));
      WlTypeMethodAdd(WlNumberType, selector,
		      WlSubrMake(WlSubrType, WlWPPB_WriteObj_Int, 2));
      WlTypeMethodAdd(WlRealType, selector,
		      WlSubrMake(WlSubrType, WlWPPB_WriteObj_Real, 2));
      WlTypeMethodAdd(WlExprType, selector,
		      WlSubrMake(WlSubrType, WlWPPB_WriteObj_Expr, 2));
#ifdef UNSIGNED
      WlTypeMethodAdd(WlUnsignedType, selector,
		      WlSubrMake(WlSubrType, WlWPPB_WriteObj_Unsigned, 2));
#endif
  }
}

void custom_end (void)
{
  return;
}

void init_ISM (void)
{
  /* add initialisation for ISMs' application */

  return;
}

void end_ISM (void)
{
  /* add end features for ISMs' application */

  return;
}

#endif /* WPPB */

