/* Copyright (c) 1995 by Groupe Bull.  All Rights Reserved */
/* $Id: bind.c,v 1.1 95/04/20 18:04:31 leon Exp $ */



#include <X11/Intrinsic.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <KlLib.h>
#include "kl_widget.h"


static KlHash bind_table = NULL; /* global bindings table */

#define KlMustBeExecutable(o, n) if(KlIsUndefinedMethod(o,KlSelExecute )) \
KlArgumentMustHaveTrait(o, n, KlTrait_function)



void
ExecuteKloneAction(Widget w, XEvent *event, String *params, Cardinal *num)
{
    KlO code;
    XrmQuark id;
    int n;
    if(1 != *num) {
	fprintf(stderr, "ExecuteKloneAction: wrong number of params\n");
	return;
    }
    id = atoi(params[0]);


    code = KlHashGet(bind_table, (KlO)id, 0);
    if(NULL != code) {
	KlArgumentMustHaveTrait(code, 0, KlTrait_list);
	KlSetLastEvent(event);
	KlSend_eval(code);
    }
}


/* description: implements the 'bind' action

 * returns: 

 */
KlWidget
KlBindAction(KlWidget klw, KlString desc, KlList value)
{
    char cbname[256], ttable[256];
    XrmQuark id;
    
    assert(desc);

    KlArgumentMustBeValidWidget(klw, 0);
    KlArgumentMustHaveTrait(desc, 1, KlTrait_string);
    KlMustBeExecutable(value, 2);

    if(NULL == bind_table) {
	bind_table = KlHashMake(NIL);
	KlIncRef(bind_table);
	bind_table->ref_counted &= ~(KlHashRefK); /* do not IncRef */
    }
    /* cos we cannot use directly a translation description as parameter in an
     * action proc, we use an indirection: we generate a unique ID (n) that is
     * the value of the binding name. This unique ID is used as parameter in
     * the action proc and its value is the lisp form to be executed */
    
    sprintf(cbname, "%x%s", klw->w, desc->string);
    id = XrmStringToQuark(cbname);
    
    KlHashPut(bind_table, id, value); /* name's value is binding */
    
    sprintf(ttable, "%s: ExecuteKloneAction(%d)", desc->string, (int)id);

    XtOverrideTranslations(klw->w, XtParseTranslationTable(ttable));
    
    return klw;
}



/* description: implements the 'unbind' action

 * returns: 

 */
KlWidget
KlUnbindAction(KlWidget klw, KlString desc)
{
    char cbname[256];
    XrmQuark id;
    
    assert(desc);

    KlArgumentMustBeValidWidget(klw, 0);
    KlArgumentMustHaveTrait(desc, 1, KlTrait_string);

    if(NULL == bind_table) {
	fprintf(stderr, "KlUnbindAction: unbind called before bind\n");
	return klw;
    }
    /* cos we cannot use directly a translation description as parameter in an
     * action proc, we use an indirection: we generate a unique ID (id) that is
     * the value of the binding name. This unique ID is used as parameter in
     * the action proc and its value is the lisp form to be executed */
    
    sprintf(cbname, "%x%s", klw->w, desc->string);
    id = XrmStringToQuark(cbname);
    
    KlHashDelete(bind_table, id); /* name's value is binding */
    
    return klw;
}
