/* Copyright (c) 1995 by Groupe Bull.  All Rights Reserved */
/* $Id: callback.c,v 1.3 96/02/02 15:48:50 leon Exp $ */


#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include "kl_widget.h"
#include "funcs.h"
#include "global.h"





/*
 * This global struct is used to pass data to the AddCallback handler, when the
 * callback name is undefined in the widget's callback list.
 */
static struct {
    KlWidget klw;
    char *sname;
    KlString name;
    KlList value;
    XtErrorMsgHandler old;
    Boolean success;
} cb_rec;


/*
 * This is called as WarningMsgHandler by the toolkit, but parameters are
 * contained in the cb_Rec record.
 */
static void
BindPseudoCallback(String name, String type, String class, String defaultp,
		   String *params, Cardinal *num_params)
{
    KlO setpcb = NIL;
    cb_rec.success = False;
    if(NIL == setpcb) {
	setpcb = (KlO)KlIntern("kl-add-pseudo-callback");
    }
    if(SameStrings(name, "invalidCallbackList") &&
       (SameStrings(type, "xtAddCallback"))) {
	int n;
	char *cname;
	KlPseudoClass pc = cb_rec.klw->pclass;
	if(pc && pc->callbacks) {
	    for(n = 0; n < pc->callbacks->size; n++) {
		cname = GetNameString(pc->callbacks->list[n]);
		if(SameStrings(cb_rec.sname, cname)) {
		    /* callback found : pseudo one */
		    KlApply3(setpcb, cb_rec.klw, cb_rec.name, cb_rec.value);
		    cb_rec.success = True;
		    return;
		}
	    }
	}
    }
    /* warning not recovered, call old handler */
    (cb_rec.old)(name, type, class, defaultp, params, num_params);
}



    
static void
ExecuteCallbackProc(Widget w, XtPointer client_data, XtPointer call_data)
{
    XmAnyCallbackStruct *cs = (XmAnyCallbackStruct *)call_data;
    KlList func = (KlList)client_data;
    
    KlArgumentMustHaveTrait(func, 0, KlTrait_list);
    if(cs->event)
	KlSetLastEvent(cs->event);
    KlSend_eval(func);
}


/* description: implements the 'add-callback' function.

 * Add the given lisp form 'value' as callback on the specified widget/callback
 * name.

 * example : (add-callback w 'activate `(print ,x))

 * returns: The Lisp form is returned, so that it can be used with rm-callback

 */
KlO
KlBindCallback(KlWidget klw, KlString name, KlList value)
{
    char cbname[64];
    XtAppContext app;
    
    KlArgumentMustBeValidWidget(klw, 0);
    KlArgumentMustHaveTrait(name, 1, KlTrait_string);
    KlMustBeExecutable(value, 2);
    sprintf(cbname, "%sCallback", GetNameString((KlO)name));
    KlIncRef(value);
    cb_rec.klw = klw;
    cb_rec.sname = GetNameString((KlO)name);
    cb_rec.value = value;
    cb_rec.name = name;
    cb_rec.success = True;
    app = XtWidgetToApplicationContext(klw->w);
    cb_rec.old =
	XtAppSetWarningMsgHandler(app, BindPseudoCallback);
    XtAddCallback(klw->w, cbname, ExecuteCallbackProc, value);
    XtAppSetWarningMsgHandler(app, cb_rec.old);
    if( ! cb_rec.success) {
	KlDecRef(value);
    }
    return (KlO)value;
}


/* description: implements the 'rm-callback' function.

 * Remove the given callback from the widget.

 * example : (rm-callback w 'activate form) where 'form' must be a returned
 * expression from add-callback.

 * returns: The widget is returned.

 */
KlO
KlUnbindCallback(KlWidget klw, KlString name, KlList value)
{
    char cbname[64];
    XtAppContext app;
    
    KlArgumentMustBeValidWidget(klw, 0);
    KlArgumentMustHaveTrait(name, 1, KlTrait_string);
    KlMustBeExecutable(value, 2);
    sprintf(cbname, "%sCallback", GetNameString((KlO)name));
    
    XtRemoveCallback(klw->w, cbname, ExecuteCallbackProc, value);
    KlDecRef(value);
    return (KlO)klw;
}

