/* Copyright 1994 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: cmdl.c,v 1.1 95/04/20 18:04:32 leon Exp $ */


#include <stdio.h>
#include "cmdl.h"



typedef enum _CmdlError {
    CmdlErrorUnknown = 0,
    CmdlErrorNotFound = 1,
} CmdlError;


static char * cmdlErrors[] = {
    "unrecognized option",
    "option not found",
};




void
CmdlUsage(char *name, CmdlOptionRec *cmdlOptions, int number)
{
    int i;

    /* regular arguments */
    fprintf(stderr, "Usage: %s", name);
    for(i = 0; i < number; i++) {
	if(cmdlOptions[i].type == CmdlOptionIsArg) {
	    fprintf(stderr, " %s", cmdlOptions[i].name);
	}
    }
    fprintf(stderr, "\n");

    
    for(i = 0; i < number; i++) {
	if(cmdlOptions[i].mode == CmdlOptionMandatory) {
	    switch(cmdlOptions[i].type) {
	    case CmdlOptionOneArg:
		fprintf(stderr, "\t%s <arg>: %s\n", cmdlOptions[i].name, 
			cmdlOptions[i].info);
		break;
	    case CmdlOptionNoArg: 
	    case CmdlOptionIsArg: 
		fprintf(stderr, "\t%s: %s\n", cmdlOptions[i].name, 
			cmdlOptions[i].info);
		break;
	    }
	}
    }
    fprintf(stderr, "Optionnal:\n");
    for(i = 0; i < number; i++) {
	if(cmdlOptions[i].mode == CmdlOptionOptionnal) {
	    switch(cmdlOptions[i].type) {
	    case CmdlOptionOneArg:
		fprintf(stderr, "\t%s <arg>: %s\n", cmdlOptions[i].name, 
			cmdlOptions[i].info);
		break;
	    case CmdlOptionNoArg: 
	    case CmdlOptionIsArg: 
		fprintf(stderr, "\t%s: %s\n", cmdlOptions[i].name, 
			cmdlOptions[i].info);
		break;
	    }
	}
    }
}



void static
CmdlPrintError(CmdlError error, char *option)
{
    fprintf(stderr, "Error: %s '%s'\n", cmdlErrors[error], option);
}

void static
CmdlPrintWarning(CmdlError error, char *option)
{
    fprintf(stderr, "Warning: %s '%s'\n", cmdlErrors[error], option);
}



void 
CmdlScan(CmdlOptionRec *cmdlOptions, int number, int argc, char **argv,
	 CmdlOptionFlag error, CmdlOptionFlag warning)
{
    int arg = 1, i, h;
    int found, is_arg=1;
    for(i = 0; i < number; i++) {
	*(cmdlOptions[i].value) = 0;
    }
    while(arg < argc) {
	found = 0;
	for(i = 0; i < number; i++) {
	    if(!strcmp(cmdlOptions[i].name, argv[arg])) {
		found = 1;
		if(cmdlOptions[i].type == CmdlOptionOneArg) {
		    *(cmdlOptions[i].value) = argv[arg+1];
		    arg++;
		}
		else {
		    *(cmdlOptions[i].value) = cmdlOptions[i].name;
		}
		break;
	    }
	    /* this is for IsArg arguments */
	    if(is_arg && (cmdlOptions[i].type == CmdlOptionIsArg) &&
	       (argv[arg][0] != '-') && ! *(cmdlOptions[i].value)) {
		found = 1;
		*(cmdlOptions[i].value) = argv[arg];
		break;
	    }
	}
	if(! found) {
	    static char *help [] = {"-?", "-h", "-help"};
	    for(h = 0; h < 3; h++) {
		if(!strcmp(argv[arg], help[h])) {
		    CmdlUsage(argv[0], cmdlOptions, number);
		    exit(0);
		}
	    }
	}
	if(! found) {
	    /* IsArg must be at the beginning */
	    is_arg = 0;
	}
	if(!found && (error & CmdlOptionUnknown)) {
	    CmdlPrintError(CmdlErrorUnknown, argv[arg]);
	    CmdlUsage(argv[0], cmdlOptions, number);
	    exit(0);
	}
	if(!found && (warning & CmdlOptionUnknown)) {
	    CmdlPrintWarning(CmdlErrorUnknown, argv[arg]);
	}
	arg++;
    }
    /* checks that all mandatory options have been found */
    for(i = 0; i < number; i++) {
	if((CmdlOptionMandatory == cmdlOptions[i].mode) &&
	   (NULL == cmdlOptions[i].value)) {

	    if(error & CmdlOptionNotFound) {
		CmdlPrintError(CmdlErrorNotFound, cmdlOptions[i].name);
		CmdlUsage(argv[0], cmdlOptions, number);
		exit(1);
	    }
	    if(warning & CmdlOptionNotFound) {
		CmdlPrintWarning(CmdlErrorNotFound, cmdlOptions[i].name);
	    }
	}
    }
}



