/* Copyright (c) 1995 by Groupe Bull.  All Rights Reserved */
/* $Id: cursor.c,v 1.2 95/06/07 14:46:47 leon Exp $ */


#include <X11/Intrinsic.h>
#include <X11/cursorfont.h>
#include "kl_widget.h"

struct _KlCursorRec {
    char *name;
    int   id;
    Cursor cursor;
} _KlCursorRec;

static struct _KlCursorRec cursor_table[] = { 
    {"X_cursor", XC_X_cursor, None},
    {"arrow", XC_arrow, None},
    {"based_arrow_down", XC_based_arrow_down, None},
    {"based_arrow_up", XC_based_arrow_up, None},
    {"boat", XC_boat, None},
    {"bogosity", XC_bogosity, None},
    {"bottom_left_corner", XC_bottom_left_corner, None},
    {"bottom_right_corner", XC_bottom_right_corner, None},
    {"bottom_side", XC_bottom_side, None},
    {"bottom_tee", XC_bottom_tee, None},
    {"box_spiral", XC_box_spiral, None},
    {"center_ptr", XC_center_ptr, None},
    {"circle", XC_circle, None},
    {"clock", XC_clock, None},
    {"coffee_mug", XC_coffee_mug, None},
    {"cross", XC_cross, None},
    {"cross_reverse", XC_cross_reverse, None},
    {"crosshair", XC_crosshair, None},
    {"diamond_cross", XC_diamond_cross, None},
    {"dot", XC_dot, None},
    {"dotbox", XC_dotbox, None},
    {"double_arrow", XC_double_arrow, None},
    {"draft_large", XC_draft_large, None},
    {"draft_small", XC_draft_small, None},
    {"draped_box", XC_draped_box, None},
    {"exchange", XC_exchange, None},
    {"fleur", XC_fleur, None},
    {"gobbler", XC_gobbler, None},
    {"gumby", XC_gumby, None},
    {"hand1", XC_hand1, None},
    {"hand2", XC_hand2, None},
    {"heart", XC_heart, None},
    {"icon", XC_icon, None},
    {"iron_cross", XC_iron_cross, None},
    {"left_ptr", XC_left_ptr, None},
    {"left_side", XC_left_side, None},
    {"left_tee", XC_left_tee, None},
    {"leftbutton", XC_leftbutton, None},
    {"ll_angle", XC_ll_angle, None},
    {"lr_angle", XC_lr_angle, None},
    {"man", XC_man, None},
    {"middlebutton", XC_middlebutton, None},
    {"mouse", XC_mouse, None},
    {"pencil", XC_pencil, None},
    {"pirate", XC_pirate, None},
    {"plus", XC_plus, None},
    {"question_arrow", XC_question_arrow, None},
    {"right_ptr", XC_right_ptr, None},
    {"right_side", XC_right_side, None},
    {"right_tee", XC_right_tee, None},
    {"rightbutton", XC_rightbutton, None},
    {"rtl_logo", XC_rtl_logo, None},
    {"sailboat", XC_sailboat, None},
    {"sb_down_arrow", XC_sb_down_arrow, None},
    {"sb_h_double_arrow", XC_sb_h_double_arrow, None},
    {"sb_left_arrow", XC_sb_left_arrow, None},
    {"sb_right_arrow", XC_sb_right_arrow, None},
    {"sb_up_arrow", XC_sb_up_arrow, None},
    {"sb_v_double_arrow", XC_sb_v_double_arrow, None},
    {"shuttle", XC_shuttle, None},
    {"sizing", XC_sizing, None},
    {"spider", XC_spider, None},
    {"spraycan", XC_spraycan, None},
    {"star", XC_star, None},
    {"target", XC_target, None},
    {"tcross", XC_tcross, None},
    {"top_left_arrow", XC_top_left_arrow, None},
    {"top_left_corner", XC_top_left_corner, None},
    {"top_right_corner", XC_top_right_corner, None},
    {"top_side", XC_top_side, None},
    {"top_tee", XC_top_tee, None},
    {"trek", XC_trek, None},
    {"ul_angle", XC_ul_angle, None},
    {"umbrella", XC_umbrella, None},
    {"ur_angle", XC_ur_angle, None},
    {"watch", XC_watch, None},
    {"xterm", XC_xterm, None},
};



/* Affect a new style for the cursor in a specified widget.
 * 
 */
KlO KlDefineCursor(KlWidget klw, KlO klcursor)
{
    char *cname;
    int i;
    KlArgumentMustBeValidWidget(klw, 0);

    if(! XtIsRealized(klw->w))
	return NIL;

    if( ! KlHasTrait(klcursor, KlTrait_string)) {
	/* argument is not a symbol or string, this means 'undefine' */
	XUndefineCursor(XtDisplay(klw->w), XtWindow(klw->w));
	XFlush(XtDisplay(klw->w));
	return (KlO)klw;
    }

    cname = ((KlString)klcursor)->string;
    
    for(i = 0; i < XtNumber(cursor_table); i++) {
	if(!strcmp(cname, cursor_table[i].name)) {
	    if(None == cursor_table[i].cursor) {
		cursor_table[i].cursor =
		    XCreateFontCursor(XtDisplay(klw->w), cursor_table[i].id);
	    }
	    XDefineCursor(XtDisplay(klw->w), XtWindow(klw->w),
			  cursor_table[i].cursor);
	    XFlush(XtDisplay(klw->w));
	    return (KlO)klw;
	}
    }
    return NIL;
}
	
