/* Copyright 1995 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: getval.c,v 1.5 96/10/23 17:12:31 leon Exp $ */


#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/RepType.h>
#include <assert.h>
#include "kl_widget.h"
#include "kl_classes.h"
#include "kl_convert.h"


char *Tm_Resource;		/* required for some converters */

typedef KlO (*KlConverterProc)(Widget w, XtPointer value);
typedef KlO (*KlWidgetProc)(Widget w);

typedef struct _KlConverterRec {
    String type;		/* representation type */
    String from;		/* representation used for SetVal conversion */
    String to;			/* representation used for GetVal conversion */
    KlConverterProc proc;	/* post conversion proc */
} KlConverterRec, *KlConverter;


typedef struct _KlPseudoResourceRec {
    String name;
    KlWidgetProc proc;
} KlPseudoResourceRec, KlPseudoResource;

 


KlO
CvtStringToKlBoolean(Widget w, XtPointer val)
{
    String value = (String)val;
    if ( ! strcmp (value, "true"))
	return TRU;
    else
	return NIL;
}


/* some converters returns a KlO. use this post conversion function to return
   the value "as is" */
KlO
CvtAsIs(Widget w, XtPointer val)
{
    String value = (String)val;
    return (KlO)value;
}    

/*
 * Check if the numerical value represents an unspecified pixmap and returns
 * NIL if this is true. */
static KlO
CheckPixmapValue(Widget w, XtPointer val)
{
    KlNumber value = (KlNumber)val;
    if(XmUNSPECIFIED_PIXMAP == value->number) {
	return NIL;
    }
    else {
	return (KlO)value;
    }
}    



/* special converters table. Could be a Hash table maybe. */
/* This table contains the get and set representation types that are expected.
   from is unused. to contains the representation type that must be returned by
   XtGetValue. The default type (if a reprensentation is not found) is
   XtString. The postConverter field specifies a conversion routine that may be
   call after XtGetValue and before returning result. */
static KlConverterRec converters[] = {
    {	XmRDimension,		0,	XtRKlNumber,	0    },
    {	XmRHorizontalDimension,	0,	XtRKlNumber,	0    },
    {	XmRVerticalDimension,	0,	XtRKlNumber,	0    },    
    {	XmRInt,			0,	XtRKlNumber,	0    },
    {	XmRCardinal,		0,	XtRKlNumber,	0    },
    {	XmRHorizontalInt,	0,	XtRKlNumber,	0    },
    {	XmRVerticalInt,		0,	XtRKlNumber,	0    },
    {	XmRPosition,		0,	XtRKlNumber,	0    },
    {	XmRHorizontalPosition,	0,	XtRKlNumber,	0    },
    {	XmRVerticalPosition,	0,	XtRKlNumber,	0    },
    {	XmRTextPosition,	0,	XtRKlNumber,	0    },
    {	XmRShort,		0,	XtRKlNumber,	0    },
    {	XmRBooleanDimension,	0,	XtRKlNumber,	0    },
    {	XmRPixmap,	0,	XtRKlNumber,		CheckPixmapValue     },

    {	XmRBoolean, 	XtRString, 	XtRString, 	CvtStringToKlBoolean },
#if XmVersion >= 2000
    {	XmRDynamicPixmap,	0,	XtRKlNumber,	CheckPixmapValue     },
    {	XmRSet, 	XtRString, 	XtRString, 	CvtStringToKlBoolean },
#endif
    
    {	XmRXmStringTable, XtRString, 	XtRString, CvtAsIs  },
    {	XmRWidgetList, XtRString, 	XtRKlList, CvtAsIs  },
    
    {	"KlO", 		0, 		"KlO", 		0  },

};



static KlO
GetEventMask(Widget w)
{
    return (XtIsRealized(w) ? (KlO)KlNumberMake(XtBuildEventMask(w)) : NIL);
}



static KlO
GetGeometry(Widget w)
{
    Arg args[4];
    int n = 0;
    Position x, y;
    Dimension width, height;
    char buf[64];

    XtSetArg(args[n], XmNx, &x); n++;
    XtSetArg(args[n], XmNy, &y); n++;
    XtSetArg(args[n], XmNwidth, &width); n++;
    XtSetArg(args[n], XmNheight,& height); n++;

    XtGetValues(w, args, n);
    sprintf(buf, "%ux%u%c%d%c%d", width, height,
	    ((x>=0) ? '+' : '-'), x, ( (y>=0) ? '+' : '-'), y);
    return (KlO)KlStringMake(buf);
}


/* pseudo resources table: These resource are used to access special values
   from widgets */
static KlPseudoResourceRec pseudo_resources[] = {
    {	"name", 	KlWidgetGetName,	},
    {	"parent", 	KlWidgetGetParent,	},
    {	"class", 	KlWidgetGetClass,	},
    {	"window", 	KlWidgetGetWindow,	},
    {	"managed", 	KlWidgetIsManaged,	},
    {	"realized", 	KlWidgetIsRealized,	},
    {	"event-mask", 	GetEventMask,		},
    {	"geometry", 	GetGeometry,		},
};




static void
FetchResourceList(KlWidgetClass wc, KlWidget klw)
{
    XtResourceList resources;
    Cardinal num;
    
    assert(NULL != wc);
    
    XtGetResourceList(klw->w->core.widget_class,
		      &resources, &num);
    if(resources) {
	wc->resources = resources;
	wc->num = num;
    }
}    




KlO
KlWidgetGetValue(KlWidget klw, char *name)
{
    XtPointer value = NULL;
    Cardinal i;
    KlWidgetClass wc = NULL;
    XtResource *res;
    KlConverter conv = NULL;
    XtResourceList resources = NULL;
    Cardinal num = 0;

    wc = (KlWidgetClass)GetWidgetClass(klw->w);
    if(wc) {
	if(! wc->resources) {
	    /* We cache the per class resource list, so that we do it only once
	       for each class. The resource list is kept in the class
	       record. */
	    FetchResourceList(wc, klw);
	}
	num = wc->num;
	resources = wc->resources;
    } 
    res = NULL;
    /* first search the resource desc in the class's resource list */
    for(i = 0; i < num; i++) {
	if ( ! strcmp (resources[i].resource_name, name)) {
	    res = &(resources[i]);
	}
    }
    if(! res) {
	/* if the resource is not found, It may be a klm pseudo resource */
	for(i = 0; i < XtNumber(pseudo_resources); i++) {
	    if ( ! strcmp (pseudo_resources[i].name, name)) {
		return (*(pseudo_resources[i].proc))(klw->w);
	    }
	}
    }
    if(res) {
	/* Is there a special converter for this rep type ? */
	conv = NULL;
	for(i = 0; i < XtNumber(converters); i++) {
	    if ( ! strcmp (converters[i].type, res->resource_type)) {
		conv = &(converters[i]);
		break;
	    }
	}
	/* Is this an enumerated type ??? */
	{
	    XmRepTypeId rtid;
	    XmRepTypeEntry rte;
	    int intvalue;
	    unsigned char value;
	    int i;
	    /* When the representation type is an enumerated registred type, we
	       do manually the val to string conversion */
	    rtid = XmRepTypeGetId(res->resource_type);
	    if(XmREP_TYPE_INVALID != rtid) {
		rte = XmRepTypeGetRecord(rtid);
		if(sizeof(int) == res->resource_size) {
		    /* the EditMode resource from XmText uses a RepType on an
                       int field instead of unsigned char. that is *silly* */
		    XtVaGetValues(klw->w, name, &intvalue, 0);
		    value = (unsigned char)intvalue;
		}
		else {
		    assert(sizeof(unsigned char) == res->resource_size);
		    XtVaGetValues(klw->w, name, &value, 0);
		}
		XtVaGetValues(klw->w, name, &value, 0);
		for(i = 0; i < (int)rte->num_values; i++) {
		    if(value == rte->values[i]) {
			return (KlO)KlIntern(rte->value_names[i]);
		    }
		}
	    }
	}
    }
    Tm_Resource = name;
    XtVaGetValues(klw->w, XtVaTypedArg,
		  name, (conv ? conv->to : XtRString),
		  (XtArgVal)&value, sizeof(char *),
		  0);
    if(! value)
	return NIL;
    if(conv && conv->proc) {
	/* there is a post converter, call it */
	return (*conv->proc)(klw->w, value);
    }
    if(!conv || (! strcmp(conv->to, XtRString))) {
	/* If There is no special converter, or if the returned type is a
           String, build a KlO */
	return (KlO)KlStringMake(value);
    }
    /* else, the returned value is already a KlO */
    return (KlO)value;
}    





