/* Copyright 1995 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: kl_convert.c,v 1.5 96/10/23 17:12:33 leon Exp $ */

#include <X11/IntrinsicP.h>
#include <Xm/Xm.h>
#include <KlLib.h>
#include "kl_convert.h"
#include <X11/Xproto.h>
#include <assert.h>


typedef struct ResAssocRec {
    String name_res;
    String num_res;
    int size;
} ResAssocRec, *ResAssoc;



static ResAssocRec res_assocs [] = {
    {	XmNchildren, 	XmNnumChildren,	sizeof(Cardinal)	}
};



/* returns the size of a list resource.

 * According to the association between a resource name and the related count
 * resource, this functions returns the list size.

 * returns: list size*/
int
GetListSize(Widget w, String res)
{
    int i;
    Arg args[1];
    for(i = 0; i <XtNumber(res_assocs); i++) {
	if(!strcmp(res, res_assocs[i].name_res)) {
	    if(sizeof(Cardinal) == res_assocs[i].size) {
		Cardinal count;
		XtSetArg(args[0], res_assocs[i].num_res, &count);
		XtGetValues(w, args, 1);
		return (int)count;
	    }
	    else if(sizeof(int) == res_assocs[i].size) {
		int count;
		XtSetArg(args[0], res_assocs[i].num_res, &count);
		XtGetValues(w, args, 1);
		return count;
	    }
	    XtAppWarning(XtWidgetToApplicationContext(w),
			 "Cannot compute liste size.");
	}
    }
    return 0;
}


#define done(address, type)                               \
    {                                                   \
        if(toVal->addr)                                 \
        {                                               \
            if(toVal->size < sizeof(type))              \
            {                                           \
                toVal->size = sizeof(type);             \
                return( FALSE) ;                        \
            }                                           \
            else                                        \
            {   *((type *) (toVal->addr)) = *address;   \
            }                                           \
        }                                               \
        else                                            \
        {                                               \
            toVal->addr = (XPointer) address;           \
        }                                               \
        toVal->size = sizeof(type);                     \
    }



Boolean
CvtWidgetListToKlList(display, args, num_args, fromVal, toVal, destructor_data)
    Display *display;
    XrmValuePtr args;
    Cardinal *num_args;
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
    XtPointer *destructor_data;
{
    static KlList list = NULL;
    int n=0;
    Widget *widgets = *(Widget **) fromVal->addr;
    Widget item;
    int count;
    extern char *Tm_Resource;
    Widget w = *((Widget *) args[0].addr);
    
    count = GetListSize(w, Tm_Resource);
    list = KlListNMake(count);
    
    for(n = 0; n < count; n++) {
	item = widgets[n];
	list->list[n] = (KlO)KlWidgetMake(item);
	KlIncRef(list->list[n]);
    }

    done(&list, void *);

    return True;
}


#if XmVersion < 2000
#define XmStringGenerate(s, tag, chrset, n) XmStringCreateSimple(s)
#endif

/*
 * Builds an XmString from a string that may contain tags like "<hello>".
 */
static XmString
GetXmString(char *from)
{
    char *p = NULL, *str = from;
    XmString xms;
    if('<' == *str) {
	/* there is a tag */
	str++;
	p = strchr(str, '>');
    }
    if(NULL != p) {
	/* these is a tag starting the string */
	char c = *p;
	/* As Klone share identical strings, we have to restore the original
           state */
	*p = '\0';
	xms = XmStringGenerate((char *)(p+1), str,
			       XmCHARSET_TEXT, NULL);
	*p = c;
    }
    else {
	/* there is no tag */
	xms = XmStringGenerate((char *)from,
			       XmFONTLIST_DEFAULT_TAG,
			       XmCHARSET_TEXT, NULL);
    }
    return xms;
}



Boolean
KlCvtStringToXmString(Display *dpy,
		      XrmValuePtr args,
		      Cardinal *num_args,
		      XrmValuePtr fromVal,
		      XrmValuePtr toVal,
		      XtPointer *destructor_data)
{
    char *str;
    XmString xms;
    
    str= (char *)fromVal->addr;
    
    if(str) {
	xms = GetXmString(str);
        if(xms) {
	    done(&xms, void *);
	    return True;
	}
    }
    XtDisplayStringConversionWarning(dpy, ((char *) fromVal->addr), 
				     XmRXmString) ;
    return False;
}


Boolean
KlCvtKlListToXmStringTable(Display *dpy,
			   XrmValuePtr args,
			   Cardinal *num_args,
			   XrmValuePtr fromVal,
			   XrmValuePtr toVal,
			   XtPointer *destructor_data)
{
    KlList list = *(KlList*)fromVal->addr;
    XmString *table;
    KlO item;
    char buf[16];
    char *svalue;
    int i;

    assert(list);
    table = (XmString *)XtMalloc(sizeof(XmString) * list->size);
    for(i = 0; i < list->size; i++) {
	item = (KlO)list->list[i];
	/* we accept strings and numbers */
	if(KlHasTrait(item, KlTrait_number)) {
	    sprintf(buf, "%ld", ((KlNumber)item)->number);
	    svalue = buf;
	}
	else if (KlHasTrait(item, KlTrait_string)) {
	    svalue = ((KlString)item)->string;
	}
	else {
	    svalue = "";
	}
	table[i] = GetXmString(svalue);
    }
    done(&table, XmString *);
    return True;
}




Boolean
KlCvtInt32ToKlNumber(Display *dpy,
			  XrmValuePtr args,
			  Cardinal *num_args,
			  XrmValuePtr fromVal,
			  XrmValuePtr toVal,
			  XtPointer *destructor_data)
{
    static KlNumber number;
    INT32 dimVal;
 
    dimVal = *(INT32 *) (fromVal->addr);

    number = KlNumberMake();
    number->number = (Int) dimVal;
    done(&number, KlNumber);
    return True;
}


Boolean
KlCvtInt16ToKlNumber(Display *dpy,
			  XrmValuePtr args,
			  Cardinal *num_args,
			  XrmValuePtr fromVal,
			  XrmValuePtr toVal,
			  XtPointer *destructor_data)
{
    static KlNumber number;
    INT16 dimVal;
 
    dimVal = *(INT16 *) (fromVal->addr);

    number = KlNumberMake();
    number->number = (Int) dimVal;
    done(&number, KlNumber);
    return True;
}





Boolean
KlCvtCard32ToKlNumber(Display *dpy,
			  XrmValuePtr args,
			  Cardinal *num_args,
			  XrmValuePtr fromVal,
			  XrmValuePtr toVal,
			  XtPointer *destructor_data)
{
    static KlNumber number;
    CARD32 dimVal;
 
    dimVal = *(CARD32 *) (fromVal->addr);

    number = KlNumberMake();
    number->number = (Int) dimVal;
    done(&number, KlNumber);
    return True;
}


Boolean
KlCvtCard16ToKlNumber(Display *dpy,
			  XrmValuePtr args,
			  Cardinal *num_args,
			  XrmValuePtr fromVal,
			  XrmValuePtr toVal,
			  XtPointer *destructor_data)
{
    static KlNumber number;
    CARD16 dimVal;
 
    dimVal = *(CARD16 *) (fromVal->addr);

    number = KlNumberMake();
    number->number = (Int) dimVal;
    done(&number, KlNumber);
    return True;
}







#define SetInt16Converter(name)\
    XtSetTypeConverter(\
		name, XtRKlNumber,\
		KlCvtInt16ToKlNumber,\
		NULL,\
		0,\
		XtCacheNone,\
		NULL)

#define SetInt32Converter(name)\
    XtSetTypeConverter(\
		name, XtRKlNumber,\
		KlCvtInt32ToKlNumber,\
		NULL,\
		0,\
		XtCacheNone,\
		NULL)



#define SetCard16Converter(name)\
    XtSetTypeConverter(\
		name, XtRKlNumber,\
		KlCvtCard16ToKlNumber,\
		NULL,\
		0,\
		XtCacheNone,\
		NULL)

#define SetCard32Converter(name)\
    XtSetTypeConverter(\
		name, XtRKlNumber,\
		KlCvtCard32ToKlNumber,\
		NULL,\
		0,\
		XtCacheNone,\
		NULL)



#define SetNumberConverter(name, reptype, signed)		\
    if(signed) {						\
    	if(sizeof(name) == sizeof(INT16)) {			\
		SetInt16Converter(reptype);			\
    	}							\
    	else if (sizeof(name) == sizeof(INT32)) {		\
		SetInt32Converter(reptype);			\
    	}							\
	else {							\
		XtAppWarning(app, "Cannot register converter.");\
    	}							\
    }								\
    else {							\
	if(sizeof(name) == sizeof(CARD16)) {	\
		SetCard16Converter(reptype);			\
    	}							\
    	else if (sizeof(name) == sizeof(CARD32)) {		\
		SetCard32Converter(reptype);			\
    	}							\
    	else {							\
		XtAppWarning(app, "Cannot register converter.");\
    	}							\
    }

    
    
void
RegisterKloneConverters(XtAppContext app)
{
    XtConvertArgRec convert_arg;
    
    SetNumberConverter( Dimension, 	XmRDimension 	 	, 0);
    SetNumberConverter( Dimension, 	XmRVerticalDimension 	, 0);
    SetNumberConverter( Dimension, 	XmRHorizontalDimension 	, 0);
    SetNumberConverter( Dimension, 	XmRBooleanDimension 	, 0);
    SetNumberConverter( int, 		XmRInt 	 	 	, 1);
    SetNumberConverter( Cardinal, 	XmRCardinal 	 	, 0);
    SetNumberConverter( int, 		XmRHorizontalInt 	, 1);
    SetNumberConverter( int, 		XmRVerticalInt 	 	, 1);
    SetNumberConverter( Position, 	XmRPosition 	 	, 1);
    SetNumberConverter( Position, 	XmRHorizontalPosition 	, 1);
    SetNumberConverter( Position, 	XmRVerticalPosition 	, 1);
    SetNumberConverter( XmTextPosition, XmRTextPosition 	, 0);
    SetNumberConverter( short, 		XmRShort 	 	, 1);
    SetNumberConverter( Pixmap, 	XmRPixmap 	 	, 0);
#if (XmVERSION >= 2)
    SetNumberConverter( Pixmap, 	XmRDynamicPixmap 	, 0);
#endif
    convert_arg.address_mode = XtWidgetBaseOffset;
    convert_arg.address_id = (XtPointer) XtOffsetOf(CoreRec, core.self);
    convert_arg.size = sizeof(CoreWidget);
    XtSetTypeConverter(XmRWidgetList,
		       XtRKlList,
		       CvtWidgetListToKlList,
		       &convert_arg,
		       1,
		       XtCacheNone,
		       NULL);

    XtSetTypeConverter(
		XtRString, XmRXmString,
		KlCvtStringToXmString,
		NULL,
		0,
		XtCacheNone,
		NULL);

    XtSetTypeConverter(
		"KlList", XmRXmStringTable,
		KlCvtKlListToXmStringTable,
		NULL,
		0,
		XtCacheNone,
		NULL);
   

}



