/* Copyright (c) 1995 by Groupe Bull.  All Rights Reserved */
/* $Id: kl_kno.c,v 1.3 96/02/02 15:48:22 leon Exp $ */

/***
   NAME
     kl_kno
   PURPOSE
     contains specific code for KnOs used with klm.
   NOTES
     
   HISTORY
     jml - Feb 7, 1995: Created.
***/

#include <X11/Intrinsic.h>
#include <stdio.h>
#include "kl_widget.h"
#include "kl_knvas.h"
#include "kl_tag.h"
#include <Kn/KnvasP.h>

typedef KlO (*KnOProc)(KnO kno);
typedef void (*KnOAccessor)(KnO kno, KlO value);

typedef struct _KnOPseudoResourceRec {
    String name;
    KnOProc get;
    KnOAccessor set;
} KnOPseudoResourceRec, KnOPseudoResource;




/*
 * Get accessors
 */

static KlO
KnOGetX(KnO kno)
{
    Position x, y;

    KnGetPosition(GetKnvas(kno), kno, &x, &y);
    return (KlO) KlNumberMake(x);
}

static KlO
KnOGetY(KnO kno)
{
    Position x, y;

    KnGetPosition(GetKnvas(kno), kno, &x, &y);
    return (KlO) KlNumberMake(y);
}


static KlO
KnOGetTag(KnO kno)
{
    return (KlO) KlTagMake(KnGetTag(GetKnvas(kno), kno));
}


static KlO
KnOGetInteractor(KnO kno)
{
    if(kno->interactor)
	return (KlO) (kno->interactor);
    else
	return NIL;
}  

static KlO
KnOGetLabel(KnO kno)
{
    KlArgumentMustBe((KlO)kno, 0, KnLabelClass);
    return (KlO)KlStringMake(((KnLabel)kno)->s);
}


static Boolean
FreezeUpdate(Widget knvas)
{
    Boolean status;
    XtVaGetValues(knvas, XtNautoUpdate, &status, NULL);
    XtVaSetValues(knvas, XtNautoUpdate, False, NULL);
    return status;
}


static void
UnfreezeUpdate(Widget knvas, Boolean status)
{
    XtVaSetValues(knvas, XtNautoUpdate, status, NULL);
    if(status) {
	KnvasUpdateDisplay(knvas);
    }
}



/*
 * Set accessors
 */

static void
KnOSetX(KnO kno, KlO value)
{
    Boolean status;
    Widget knvas = GetKnvas(kno);
    Position x, y;
    KnGetPosition(knvas, kno, &x, &y);
    status = FreezeUpdate(knvas);
    KnMove(knvas, kno, ((KlNumber)value)->number, y);
    UnfreezeUpdate(knvas, status);
}   

static void
KnOSetY(KnO kno, KlO value)
{
    Boolean status;
    Widget knvas = GetKnvas(kno);
    Position x, y;
    KnGetPosition(knvas, kno, &x, &y);
    status = FreezeUpdate(knvas);
    KnMove(knvas, kno, kno->x, ((KlNumber)value)->number);
    UnfreezeUpdate(knvas, status);
}    

static void
KnOSetTag(KnO kno, KlO value)
{
    Widget knvas = GetKnvas(kno);
    KnSetTag(knvas, kno, ((KlTag)value)->tag);
    
}     

static void
KnOSetInteractor(KnO kno, KlO value)
{
    Widget knvas = GetKnvas(kno);
    KnSetInteractor(knvas, kno, (KnInteractor)value);
    
}



static KnOPseudoResourceRec pseudo_resources[] = {
    {	"x",		KnOGetX,		KnOSetX			},
    {	"y",		KnOGetY,		KnOSetY			},
    {	"tag",		KnOGetTag,		KnOSetTag		},
    {	"interactor",	KnOGetInteractor,	KnOSetInteractor	},
    {	"label",	KnOGetLabel,		NULL			},
};


static KlO
KnOGet(KnO kno, char *name)
{
    int i;
    for(i = 0; i < XtNumber(pseudo_resources); i++) {
	if(!strcmp(name, pseudo_resources[i].name)) {
	    if(pseudo_resources[i].get)
		return (*pseudo_resources[i].get)(kno);
	    else {
		fprintf(stderr, "Error: Cannot get '%s' for this object.\n",
			name);
		return NIL;
	    }
	}
    }
    fprintf(stderr, "Error: '%s' is not a KnO attribute.\n", name);
    return NIL;
}    


static void
KnOSet(KnO kno, char *name, KlO value)
{
    int i;
    for(i = 0; i < XtNumber(pseudo_resources); i++) {
	if(!strcmp(name, pseudo_resources[i].name)) {
	    if(pseudo_resources[i].set) {
		(*pseudo_resources[i].set)(kno, value);
		return ;
	    }
	    else {
		fprintf(stderr, "Error: Cannot set '%s' for this object.\n",
			name);
		return ;
	    }
	}
    }
    fprintf(stderr, "Error: '%s' is not a KnO attribute.\n", name);
}    


KlO
KnOExecute(KnO kno, KlList list)
{
    KlO result;
    KlString name;
    KlO value;
    char *svalue;
    Int i;
    int argc = list->size - 1;
    KlO *argv = list->list + 1;


    if(argc == 1) {
	/* Get Value */
	name = (KlString)argv[0];
	return KnOGet(kno, &(name->string[1]));
    }
    
    if (argc > 1) {
	KlO *eval_list = KlAlloca(sizeof(KlO)*argc);
	KlO *dest = eval_list;
	KlO *last = argv + argc;

	while (argv < last) {
	    /* do not eval first argument (selector) */
	    *dest = *argv;
	    dest++, argv++;
	    /* eval selector's value */
	    *dest = KlSend_eval(*argv);
	    dest++, argv++;
	}
	
	for(i = 0; i < argc; i+=2) {
	    name = (KlString) eval_list[i];
	    value = (KlO) eval_list[i+1];

	    KnOSet(kno, &(name->string[1]), value);
	}
    }
    return (KlO) kno;
}


void KnOInitHook()
{
    KlDeclareMethod(KnOClass, KlSelExecute, KnOExecute);
}
