/* Copyright (c) 1995 by Groupe Bull.  All Rights Reserved */
/* $Id: kl_knvas.c,v 1.4 96/10/23 17:12:35 leon Exp $ */

/***
   NAME
     kl_knvas
   PURPOSE
     Things related to the Knvas widget.
   NOTES
     
   HISTORY
     jml - Feb 7, 1995: Created.
***/



#include "kl_widget.h"
#include "kl_knvas.h"
#include <Kn/KnvasP.h>
#include "kl_tag.h"
#include <Kn/Tag.h>
#include "funcs.h"


static void
SetTarget(Widget w, XtPointer client, XtPointer call)
{
    InteractorCallbackStruct *cs = (InteractorCallbackStruct *)call;
    KlO target = NULL;

    if(NULL == target) {
	target = (KlO)KlIntern("*target*");
	KlIncRef(target);
    }
    KlSend_setq(target, cs->target);
}



Widget
GetKnvas(KnO kno)
{
    if(kno->parent)
	return GetKnvas(kno->parent);
    else {
	KnLayer l =(KnLayer)kno;
	return (Widget)l->parents->list[0];
    }
}    

/* to simplify things, we automatically create a layer, so that user does not
   have to care about this : he can create KnOs child of the knvas widget */
KlWidget KlKnvas(Int argc, KlO argv[])
{
    Widget w;
    KlWidget res, parent;
    KlString wname;
    KlArgumentMustBe(argv[0], 0, KlWidgetType);
    KlArgumentMustHaveTrait(argv[1], 1, KlTrait_string);
    parent = (KlWidget)argv[0];
    wname = (KlString)argv[1];
    
    w = KnCreateKnvas(parent->w, GetNameString((KlO)wname), 0, 0);
    XtAddCallback(w, XtNinteractorCallback, SetTarget, 0);
    res = KlWidgetMake(w);
    KnCreateLayer(w);		/* default layer */
    if(argc > 2) {
	KlWidgetSetValues(res, argc-2, &(argv[2]));
    }
    XtManageChild(w);
    return res;
}


KlO KlLayer(KlWidget canvas)
{
    return (KlO) KnCreateLayer(canvas->w);
}


#define KNO_CHECK_PARAMS(parent)\
KnO kno;\
Widget pw;\
KnLayer layer;\
if(KlIsAWidget(parent)) {\
	  pw = ((KlWidget)parent)->w;\
          layer = KnvasGetDefaultLayer(pw);\
}\
else {\
	  KlArgumentMustHaveTrait(parent, 0, KnOTrait_kno);\
	  pw=GetKnvas(parent);\
          layer = (KnLayer)parent;\
}



#define UPDATE_AND_RETURN(w, kno)\
KlSend_clear(kno, w);\
if(XtIsRealized(w)&&((KnvasWidget)w)->knvas.autoUpdate)KnvasUpdateDisplay(w);\
return (KlO)kno

KlO KlLine(KnO parent, KlNumber x, KlNumber y, KlNumber w, KlNumber h)
{
    KNO_CHECK_PARAMS(parent);
    KlArgumentMustHaveTrait(x, 1, KlTrait_number);
    KlArgumentMustHaveTrait(y, 2, KlTrait_number);
    KlArgumentMustHaveTrait(w, 3, KlTrait_number);
    KlArgumentMustHaveTrait(h, 4, KlTrait_number);
    kno = KnCreateLine(pw, (KnO)layer,
		       x->number, y->number, w->number, h->number);
    UPDATE_AND_RETURN(pw, kno);
}


KlO KlRect(KnO parent, KlNumber x, KlNumber y, KlNumber w, KlNumber h)
{
    KNO_CHECK_PARAMS(parent);
    KlArgumentMustHaveTrait(x, 1, KlTrait_number);
    KlArgumentMustHaveTrait(y, 2, KlTrait_number);
    KlArgumentMustHaveTrait(w, 3, KlTrait_number);
    KlArgumentMustHaveTrait(h, 4, KlTrait_number);
    kno =  KnCreateRect(pw, (KnO)layer,
			      x->number, y->number, w->number, h->number);
    UPDATE_AND_RETURN(pw, kno);
}


KlO KlFRect(KnO parent, KlNumber x, KlNumber y, KlNumber w, KlNumber h)
{
    KNO_CHECK_PARAMS(parent);
    KlArgumentMustHaveTrait(x, 1, KlTrait_number);
    KlArgumentMustHaveTrait(y, 2, KlTrait_number);
    KlArgumentMustHaveTrait(w, 3, KlTrait_number);
    KlArgumentMustHaveTrait(h, 4, KlTrait_number);
    kno =  KnCreateFRect(pw, (KnO)layer,
			       x->number, y->number, w->number, h->number);
    UPDATE_AND_RETURN(pw, kno);
}


KlO KlCircle(KnO parent, KlNumber x, KlNumber y, KlNumber w, KlNumber h)
{
    KNO_CHECK_PARAMS(parent);
    KlArgumentMustHaveTrait(x, 1, KlTrait_number);
    KlArgumentMustHaveTrait(y, 2, KlTrait_number);
    KlArgumentMustHaveTrait(w, 3, KlTrait_number);
    KlArgumentMustHaveTrait(h, 4, KlTrait_number);
    kno =  KnCreateCircle(pw, (KnO)layer,
				x->number, y->number, w->number, h->number);
    UPDATE_AND_RETURN(pw, kno);
}


KlO KlFCircle(KnO parent, KlNumber x, KlNumber y, KlNumber w, KlNumber h)
{
    KNO_CHECK_PARAMS(parent);
    KlArgumentMustHaveTrait(x, 1, KlTrait_number);
    KlArgumentMustHaveTrait(y, 2, KlTrait_number);
    KlArgumentMustHaveTrait(w, 3, KlTrait_number);
    KlArgumentMustHaveTrait(h, 4, KlTrait_number);
    kno =  KnCreateFCircle(pw, (KnO)layer,
				 x->number, y->number, w->number, h->number);
    UPDATE_AND_RETURN(pw, kno);
}


KlO KlIcon(KnO parent, KlString name, KlNumber x, KlNumber y)
{
    KNO_CHECK_PARAMS(parent);
    KlArgumentMustHaveTrait(name, 1, KlTrait_string);
    KlArgumentMustHaveTrait(x, 2, KlTrait_number);
    KlArgumentMustHaveTrait(y, 3, KlTrait_number);
    kno =  KnCreateIcon(pw, (KnO)layer, name->string,
				 x->number, y->number);
    UPDATE_AND_RETURN(pw, kno);
}


KlO KlSticker(KnO parent, KlString name, KlNumber x, KlNumber y)
{
    KNO_CHECK_PARAMS(parent);
    KlArgumentMustHaveTrait(name, 1, KlTrait_string);
    KlArgumentMustHaveTrait(x, 2, KlTrait_number);
    KlArgumentMustHaveTrait(y, 3, KlTrait_number);
    kno =  KnCreateLabel(pw, (KnO)layer, x->number, y->number, name->string);
    UPDATE_AND_RETURN(pw, kno);
}


KlO KlCreateTag(KlWidget parent, KlString name)
{
    KnTag tag;
    KlArgumentMustBeValidWidget(parent, 0);
    KlArgumentMustHaveTrait(name, 1, KlTrait_string);

    tag = KnCreateTag(parent->w, name->string);
    return (KlO)KlTagMake(tag);
}

KlO KlCreateAnchor(KnO parent, KnO master, KlNumber x, KlNumber y)
{
    KNO_CHECK_PARAMS(parent);
    KlArgumentMustHaveTrait(master, 1, KnOTrait_kno);
    KlArgumentMustHaveTrait(x, 2, KlTrait_number);
    KlArgumentMustHaveTrait(y, 3, KlTrait_number);
    
    kno =  KnCreateAnchor(pw, (KnO)layer, x->number, y->number, False);
    KnAttach(pw, master, (KnAnchor)kno);
    
    UPDATE_AND_RETURN(pw, kno);
}

KlO KlCreateLink(KnO parent, KnO from, KnO to)
{
    KNO_CHECK_PARAMS(parent);
    KlArgumentMustHaveTrait(from, 1, KnOTrait_kno);
    KlArgumentMustHaveTrait(to, 2, KnOTrait_kno);
    
    kno =  KnCreateLink(pw, (KnO)layer, from, to);
    
    UPDATE_AND_RETURN(pw, kno);
}


KlO KlMoveInteractor(KlWidget knvas, KlString name)
{
    KlArgumentMustBeValidWidget(knvas, 0);
    KlArgumentMustHaveTrait(name, 1, KlTrait_string);

    return (KlO) KnCreateMoveInteractor(knvas->w, name->string);
}


KlO KlResizeInteractor(KlWidget knvas, KlString name)
{
    KlArgumentMustBeValidWidget(knvas, 0);
    KlArgumentMustHaveTrait(name, 1, KlTrait_string);
    return (KlO) KnCreateResizeInteractor(knvas->w, name->string);
}



KlO KlSelectInteractor(KlWidget knvas, KlString name)
{
    KlArgumentMustBeValidWidget(knvas, 0);
    KlArgumentMustHaveTrait(name, 1, KlTrait_string);
    return (KlO) KnCreateSelectInteractor(knvas->w, name->string);
}

		    

