/* Copyright 1995 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: kl_motif.c,v 1.4 96/10/23 17:12:36 leon Exp $ */


#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <Xm/XmAll.h>
#include "kl_widget.h"
#include "kl_motif.h"
#include "funcs.h"



#define CheckConstructorArguments() \
    Widget w;\
    KlWidget res, parent;\
    KlString wname;\
    if(argc < 2) {return (KlWidget)KlError1i(KlE_BAD_NUMBER_OF_ARGS, (2));};\
    parent = (KlWidget)argv[0];\
    KlArgumentMustBeValidWidget(parent, 0);\
    KlArgumentMustHaveTrait(argv[1], 1, KlTrait_string);\
    wname = (KlString)argv[1];

#define SetCreationTimeValues() \
    if(argc > 2) {\
        if(argc % 2) {\
	    KlBadArgument(argv[argc-1], argc-1, "value");\
        }\
	KlWidgetSetValues(res, argc-2, &(argv[2]));\
    }


#define DeclareNAryManagedConstructor(fname, motiffunc)\
KlWidget fname (Int argc, KlO argv[])\
{\
    CheckConstructorArguments();\
    w = motiffunc (parent->w, wname->string, 0, 0);\
    res = KlWidgetMake(w);\
    SetCreationTimeValues();\
    XtManageChild(w);\
    return res;\
}


#define DeclareNAryUnmanagedConstructor(fname, motiffunc)\
KlWidget fname (Int argc, KlO argv[])\
{\
    CheckConstructorArguments();\
    w = motiffunc (parent->w, wname->string, 0, 0);\
    res = KlWidgetMake(w);\
    SetCreationTimeValues();\
    return res;\
}




DeclareNAryManagedConstructor( KlArrowButton, XmCreateArrowButton);
DeclareNAryManagedConstructor( KlBulletinBoard, XmCreateBulletinBoard);
DeclareNAryManagedConstructor( KlCascadeButton, XmCreateCascadeButton );
DeclareNAryManagedConstructor( KlCommand, XmCreateCommand );
DeclareNAryManagedConstructor( KlDialogShell, XmCreateDialogShell );
DeclareNAryManagedConstructor( KlDragIcon, XmCreateDragIcon );
DeclareNAryManagedConstructor( KlDrawingArea, XmCreateDrawingArea );
DeclareNAryManagedConstructor( KlDrawnButton, XmCreateDrawnButton );
DeclareNAryManagedConstructor( KlFileSelectionBox, XmCreateFileSelectionBox );
DeclareNAryManagedConstructor( KlForm, XmCreateForm );
DeclareNAryManagedConstructor( KlScrollBar, XmCreateScrollBar );
DeclareNAryManagedConstructor( KlSelectionBox, XmCreateSelectionBox );
DeclareNAryManagedConstructor( KlSeparator, XmCreateSeparator );
DeclareNAryManagedConstructor( KlText, XmCreateText );
DeclareNAryManagedConstructor( KlTextField, XmCreateTextField );
DeclareNAryManagedConstructor( KlToggleButton, XmCreateToggleButton );
DeclareNAryManagedConstructor( KlWorkArea, XmCreateWorkArea );
DeclareNAryManagedConstructor( KlRadioBox, XmCreateRadioBox );
DeclareNAryManagedConstructor( KlMenuBar, XmCreateMenuBar );
DeclareNAryManagedConstructor( KlPanedWindow, XmCreatePanedWindow );
DeclareNAryManagedConstructor( KlPushButton, XmCreatePushButton );
DeclareNAryManagedConstructor( KlRowColumn, XmCreateRowColumn );
DeclareNAryManagedConstructor( KlFrame, XmCreateFrame );
DeclareNAryManagedConstructor( KlLabel, XmCreateLabel );
DeclareNAryManagedConstructor( KlItemList, XmCreateList );
DeclareNAryManagedConstructor( KlScrolledList, XmCreateScrolledList );
DeclareNAryManagedConstructor( KlMenuShell, XmCreateMenuShell );
DeclareNAryManagedConstructor( KlMessageBox, XmCreateMessageBox );
DeclareNAryManagedConstructor( KlScale, XmCreateScale );





DeclareNAryUnmanagedConstructor( KlPopupMenu,
				 XmCreatePopupMenu );
DeclareNAryUnmanagedConstructor ( KlBulletinBoardDialog,
				  XmCreateBulletinBoardDialog );
DeclareNAryUnmanagedConstructor ( KlCommandDialog,
				  XmCreateCommandDialog );
DeclareNAryUnmanagedConstructor ( KlFileSelectionDialog,
				  XmCreateFileSelectionDialog );
DeclareNAryUnmanagedConstructor ( KlFormDialog,
				  XmCreateFormDialog );
DeclareNAryUnmanagedConstructor ( KlMessageDialog,
				  XmCreateMessageDialog );
DeclareNAryUnmanagedConstructor ( KlErrorDialog,
				  XmCreateErrorDialog );
DeclareNAryUnmanagedConstructor ( KlInformationDialog,
				  XmCreateInformationDialog);
DeclareNAryUnmanagedConstructor ( KlQuestionDialog,
				  XmCreateQuestionDialog );
DeclareNAryUnmanagedConstructor ( KlWarningDialog,
				  XmCreateWarningDialog );
DeclareNAryUnmanagedConstructor ( KlWorkingDialog,
				  XmCreateWorkingDialog );
DeclareNAryUnmanagedConstructor ( KlTemplateDialog,
				  XmCreateTemplateDialog );
DeclareNAryUnmanagedConstructor ( KlSelectionDialog,
				  XmCreateSelectionDialog );
DeclareNAryUnmanagedConstructor ( KlPromptDialog,
				  XmCreatePromptDialog );


/* local hack to be able to manage construction time arguments properly for
   Scrolled & MainWindow. should use a more general split algorithm instead */
static void
ToUpper(char *buf)
{
  int i;
  for(i = 0; i < strlen(buf); i++) {
    buf[i] = toupper(buf[i]);
  }
}

static Boolean
FindCreationArgument(char *name, char *value, Int argc, KlO argv[])
{
  char buf[64];
  int i;
  for(i = 2; i <argc; i+=2) {
      if( ! (KlHasTrait(argv[i], KlTrait_string) &&
	     KlHasTrait(argv[i+1], KlTrait_string))) {
	  continue;
      }
    sprintf(buf, "%s", GetNameString(argv[i+1]));
    ToUpper(buf);
    if(!strcmp(GetNameString(argv[i]), name) &&
       !strcmp(buf, value)) {
      return True;
    }
  }
  return False;
}





KlWidget KlScrolledWindow(Int argc, KlO argv[])
{
    Arg arg;
    CheckConstructorArguments () ;
    XtSetArg(arg, XmNscrollingPolicy, XmAUTOMATIC);
    if(FindCreationArgument(XmNscrollingPolicy, "APPLICATION_DEFINED",
			    argc, argv)) {
      XtSetArg(arg, XmNscrollingPolicy, XmAPPLICATION_DEFINED);
    }
    w = XmCreateScrolledWindow(parent->w, wname->string, &arg, 1);
    res = KlWidgetMake(w);
    SetCreationTimeValues();
    XtManageChild(w);
    return  res;
}



KlWidget KlMainWindow(Int argc, KlO argv[])
{
    Arg arg;
    CheckConstructorArguments () ;
    XtSetArg(arg, XmNscrollingPolicy, XmAUTOMATIC);
    if(FindCreationArgument(XmNscrollingPolicy, "APPLICATION_DEFINED",
			    argc, argv)) {
      XtSetArg(arg, XmNscrollingPolicy, XmAPPLICATION_DEFINED);
    }
    w = XmCreateMainWindow(parent->w, wname->string, &arg, 1);
    res = KlWidgetMake(w);
    SetCreationTimeValues();
    XtManageChild(w);
    return  res;
}




KlWidget KlHorizontalPacker(Int argc, KlO argv[])
{
    CheckConstructorArguments () ;
    w = XmCreateRowColumn(parent->w, wname->string, 0, 0);
    res = KlWidgetMake(w);
    SetCreationTimeValues();
    XtVaSetValues(w, XmNpacking, XmPACK_COLUMN,
		  XmNorientation, XmHORIZONTAL,
		  0);
    XtManageChild(w);
    return  res;
}



KlWidget KlVerticalPacker(Int argc, KlO argv[])
{
    CheckConstructorArguments () ;
    w = XmCreateRowColumn(parent->w, wname->string, 0, 0);
    res = KlWidgetMake(w);
    SetCreationTimeValues();
    XtVaSetValues(w, XmNpacking, XmPACK_COLUMN,
		  XmNorientation, XmVERTICAL,
		  0);
    XtManageChild(w);
    return  res;
}

KlWidget KlOptionMenu(Int argc, KlO argv[])
{
    Widget menu;
    Arg args[1];
    char menuname[64];
    
    CheckConstructorArguments () ;
    
    sprintf(menuname, "%s_menu", wname->string);
    menu = XmCreatePulldownMenu(parent->w, menuname, 0, 0);
    XtSetArg(args[0], XmNsubMenuId, menu);
    w = XmCreateOptionMenu(parent->w, wname->string, args, 1);
    res = KlWidgetMake(menu);
    SetCreationTimeValues();
    XtManageChild(w);
    return  res;
}

KlWidget KlPulldownMenu(Int argc, KlO argv[])
{
    Widget menu;
    Arg args[1];
    char menuname[64];
    
    CheckConstructorArguments () ;
    

    sprintf(menuname, "%s_menu", wname->string);
    menu = XmCreatePulldownMenu(parent->w, menuname, 0, 0);
    XtSetArg(args[0], XmNsubMenuId, menu);
    w = XmCreateCascadeButton(parent->w, wname->string, args, 1);
    res = KlWidgetMake(menu);
    SetCreationTimeValues();
    XtManageChild(w);
    return  res;
}

KlWidget KlScrolledText(Int argc, KlO argv[])
{
    Arg arg;
    
    CheckConstructorArguments () ;
    

    XtSetArg(arg, XmNeditMode, XmMULTI_LINE_EDIT);
    w = XmCreateScrolledText(parent->w, wname->string, &arg, 1);
    res = KlWidgetMake(w);
    SetCreationTimeValues();
    XtManageChild(w);
    return res;
}













#if ( XmVERSION >= 2 )

DeclareNAryManagedConstructor( KlNotebook, XmCreateNotebook );

DeclareNAryManagedConstructor( KlSpinBox, XmCreateSpinBox );

DeclareNAryManagedConstructor( KlComboBox, XmCreateComboBox );

DeclareNAryManagedConstructor( KlIconGadget, XmCreateIconGadget );

DeclareNAryManagedConstructor( KlContainer, XmCreateContainer );
#ifdef __hpux
DeclareNAryManagedConstructor( KlCSText, XmCreateText );
#else
DeclareNAryManagedConstructor( KlCSText, XmCreateCSText );
#endif
#endif /* XmVersion */

