/* Copyright 1995 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: kl_widget.c,v 1.2 95/06/07 14:47:07 leon Exp $ */


#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <assert.h>
#include "kl_widget.h"
#include "kl_classes.h"
#include "kl_convert.h"
#include "funcs.h"

/* This is private in the r5 distrib */
#if (XtSpecificationRelease < 6)
typedef struct {
    String      name;   /* resource name */
    String      type;   /* representation type name */
    XtArgVal    value;  /* representation */
    int         size;   /* size of representation */
} XtTypedArg, *XtTypedArgList;
#endif /*  (XtSpecificationRelease < 6) */


KlType KlWidgetType;
static KlHash widgets_table;	/* Hash table containing widgets */


KlAtom KlA_BAD_WIDGET;


extern KlO
KlWidgetGetValue(KlWidget klw, char *name);

extern KlO
KlWidgetSetValues(KlWidget klw, int argc, KlO *argv);


KlO
KlWidgetGetParent(Widget w)
{
    Widget parent = XtParent(w);
    if(parent)
	return (KlO)KlWidgetMake(parent);
    else
	return NIL;
}



KlO
KlWidgetGetName(Widget w)
{
    return (KlO)KlStringMake(XtNewString(XtName(w)));
}


KlO
KlWidgetGetClass(Widget w)
{
    WidgetClass wc = XtClass(w);

    return (KlO)KlStringMake(XtNewString(wc->core_class.class_name));
}



KlO
KlWidgetGetWindow(Widget w)
{
    return (XtIsRealized(w) ? (KlO)KlNumberMake(XtWindow(w)) : NIL);
}




KlO
KlWidgetIsManaged(Widget w)
{
    return XtIsManaged(w) ? TRU : NIL;
}


KlO
KlWidgetIsRealized(Widget w)
{
    return XtIsRealized(w) ? TRU : NIL;
}



void
KlDestroyWidgetCB(Widget w, XtPointer client, XtPointer call)
{
    /* This callbacks set the widget to NULL, so that any further
       use of klw->w, tested with KlArgumentMustBeValidWidget will
       trigger the KlA_BAD_WIDGET error */
    KlWidget klw = (KlWidget)client;
    KlHashDelete(widgets_table, (KlO)w);
    klw->w = NULL;
}


KlO
KlWidgetFree(KlO klo)
{
    KlWidget klw = (KlWidget)klo;
    /* we should never free an unreferenced widget */
    assert(NULL == klw->w);
    KlDecRef(klw->pclass);
    KlDecRef(klw->p_list);
    return TRU;
}    




/* returns the Klone object associated to the widget.

 * If there is always a klone object that holds the specified widget, returns
 * this objects, or create one and add it in the table.

 * returns: The previous or a new KlWidget*/
KlWidget
KlWidgetMake(Widget w)
{
    KlWidget res;

    assert(w);
    
    res = (KlWidget)KlHashGet(widgets_table, (KlO)w, (KlO)NIL);
    if(NIL == (KlO)res) {		/* not found */
	res = (KlWidget)KlOMake(KlWidgetType);
	res->pclass = NULL;
	res->w = w;
	res->p_list = 0;
	KlHashPut(widgets_table, w, (KlO)res);
	XtAddCallback(w, XtNdestroyCallback, KlDestroyWidgetCB, res);
    }
    return res;
}





    
KlO
KlWidgetSetOrGet(KlWidget klw, KlList list)
{
    int argc = list->size - 1;
    KlO *argv = list->list + 1;

    if(argc == 1) {
	/* Get Value */
	KlAtom selector;
	char *sname;
	/* Here, we accept any atom sub-type */
	selector = (KlAtom)KlSend_eval(argv[0]);
	sname = GetNameString((KlO)selector);

	if(NULL != klw->pclass) {
	    return (KlO)KlWidgetGetPseudoValue(klw,  sname, selector);
	}
	else {
	    return KlWidgetGetValue(klw,  sname);
	}
    }
    /* Set Values */
    if (argc > 1 && !(argc %2)) {/* check for even list */
	/* the execute method is responsible for evaluating its arguments */
	KlO *eval_list = KlAlloca(sizeof(KlO)*argc);
	KlO *dest = eval_list;
	KlO *last = argv + argc;
	
	while (argv < last) {
	    /* first eval arguments by 2 */
	    /* first one must have the Trait_string once evaluated */
	    *dest = KlSend_eval(*argv);
	    KlArgumentMustHaveTrait(*dest, (dest-eval_list), KlTrait_string);
	    dest++, argv++;
	    *dest = KlSend_eval(*argv);
	    dest++, argv++;
	}
	if((NULL != klw->pclass)) {
	    /* if the widget is linked to a pseudo class, call this */
	    KlWidgetSetPseudoValues(klw, argc, eval_list);
	}
	else {
	    /* else the standard SetValues proc */
	    KlWidgetSetValues(klw, argc, eval_list);
	}
	/* trigger spies */
	{
	    static KlO trigger = NULL;
	    if(NULL == trigger) {
		trigger = (KlO)KlIntern("trigger-spies");
		KlIncRef(trigger); /* incref to never lose the atom */
	    }
	    if(NULL != klw->p_list) {
		KlList rlist;
		rlist = KlListNMake(0);
		rlist->size = argc;
		rlist->list=eval_list;
		KlIncRef(rlist);
		KlApply2(trigger, klw, rlist);
		rlist->size = 0;
		rlist->list=NULL;
		KlDecRef(rlist);
		
	    }
	}
    }
    else {
	/* odd number of arguments */
	KlBadArgument(argv[argc-1], argc-1, "value");
    }
    
    XFlush(XtDisplayOfObject(klw->w));
    return (KlO)klw;
}




KlO
KlWidgetExecute(KlWidget klw, KlList list)
{
    KlArgumentMustBeValidWidget(klw, 0);
    return KlWidgetSetOrGet(klw, list);
}




KlO
KlWidgetApply(KlWidget klw, KlList list)
{
    KlO result;
    KlString name, value;
    Int i;
    int argc = list->size - 1;
    KlO *argv = list->list + 1;
    
    KlArgumentMustBeValidWidget(klw, 0);
    for(i = 0; i < argc; i+=2) {
	name = (KlString) argv[i];
	value = (KlString) argv[i+1];
	KlArgumentMustHaveTrait(name, 0, KlTrait_string);
	    KlArgumentMustHaveTrait(value, 0, KlTrait_string);
	XtVaSetValues(klw->w, XtVaTypedArg,
		      &(name->string[1]), XtRString,
		      (XtArgVal)value->string, strlen(value->string),
		      0);
    }
    XFlush(XtDisplayOfObject(klw->w));
    return (KlO)klw;
}



void
KlWidgetImport()
{
    
}



    
void
KlWidgetClassInitialize()
{
    KlList list;
    
    /* declare KlWidget type */
    KlDeclareType(&KlWidgetType, "KlWidget", sizeof(struct KlWidgetRec));

    KlDeclarePlistSlot(KlWidgetType, KlWidget, p_list);
    
    /* declare methods */
    KlDeclareMethod1(KlWidgetType, KlSelExecute, KlWidgetExecute);
    KlDeclareMethod1(KlWidgetType, KlSelApply, KlWidgetApply);
    KlDeclareMethod1(KlWidgetType, KlSelFree, KlWidgetFree);


    
    list = KlListNMake(0);
    widgets_table = KlHashMake(list);
    KlIncRef(widgets_table);
    widgets_table->ref_counted &= ~(KlHashRefK); /* do not IncRef keys */
}

