/* Copyright 1995 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: main.c,v 1.4 96/10/23 17:12:38 leon Exp $ */

#include "flags.h"
#include "revision.h"
#include "cmdl.h"
#include "global.h"



KlmOptionRec options;		/* the global options record */

static CmdlOptionRec optionsdesc[] = {
    {	"[file]", CmdlOptionIsArg, CmdlOptionOptionnal, &(options.file),
	"Starts in batch mode (no prompt) and load the given file."
    },
    {	"-f", CmdlOptionOneArg, CmdlOptionOptionnal, &(options.loadfile),
	"Starts in prompt mode and load the given file."
    },
    {	"-v", CmdlOptionNoArg, CmdlOptionOptionnal, &(options.verbose),
	"When specified, turn on the verbose mode."
    },
    {	"-d", CmdlOptionNoArg, CmdlOptionOptionnal, &(options.debug),
	"When specified, turn on the debug mode."
    },
    {	"-t", CmdlOptionNoArg, CmdlOptionOptionnal, &(options.top),
	"When specified, auto creates a Toplevel widget named 'top'."
    },
    {	"-s", CmdlOptionNoArg, CmdlOptionOptionnal, &(options.server),
	"Turn on server ability."
    },
};




extern MkloneInitialize();
extern MkloneToplevel();

#include <X11/Intrinsic.h>
#include <Xm/Xm.h>

#define KlPROFILE_USER_VARIABLE "KLONEPROFILE"
#define KlPATH_SHELL_VARIABLE   "KLONEPATH"
#ifndef KLPATH
#ifndef AMIGA
#define KLPATH ".:/usr/local/lib/klone:/usr/local/lib/klm:kl:../klone/kl"
#else
#define KLPATH ":kl:/klone"
#endif
#endif

#include <signal.h>
#include <stdio.h>
#ifdef SYSV
#include <sys/types.h>
#endif                                  /* SYSV */
#include <sys/file.h>

#ifdef c_plus_plus
extern "C" {
#endif
#include "EXTERN.h"
#include "klone.h"
#include "kl_atom.h"
#include "kl_coll.h"
#include "kl_func.h"
#include "kl_list.h"
#include "kl_number.h"
#include "kl_string.h"
#include "kl_stream.h"
#ifdef c_plus_plus
}

#endif

#include "klone-ic.h"


DECLARE_strchr;

int noisy = 1; 
/* char *input_file = 0; */
int read_stdin = 1;
int use_built_in = 1;
int script = 0;
int KlArgc;
char **KlArgv;



int toplevel();
int KlPoolToplevel();

main_init()
{
    KlActivePointerToIntMake("stack-level", &KlStackPtr);

    /*
     * execution the built-in klone code. made here to be done before the user
     * profile is read
     */
    if (use_built_in)
        KlExecuteStringNoReturn(KlStartupCode);
    
    
    /* make klone list of args */
    KlDeclareCommandLineArguments(KlArgc, KlArgv);
}


int toplevel();

/* these are defined and called in the Knvas lib */
typedef void(*InitHook)(void);
extern InitHook kno_init_hook;

extern void KnOInitHook();


main(argc, argv)
    int argc;
    char *argv[];

{
    char *s;
    extern char *KlFixPath();
    KlO result;
    int normal;
    int (*toplevel_evaluator)();
    char *emacs_indicator = "-*-";
  

#ifdef MONITOR
    moncontrol(0);                      /* do not trace inits */
#endif /* MONITOR */
    KlMallocInit();

    KlUserProfileName = ".klonerc";
    KlTextExtension = ".kl";
    KlApplicationName = "klone";

    CmdlScan(optionsdesc, XtNumber(optionsdesc), argc, argv,
	     CmdlOptionNever, CmdlOptionNever);

    /* initialize paths (put .:$HOME before built-ins) */
    KlPath = KLPATH;
    if ((s = (char *) getenv(KlPROFILE_USER_VARIABLE)) && (s[0] != '\0'))
	KlUserProfileName = s;
    if ((s = (char *) getenv(KlPATH_SHELL_VARIABLE)) && (s[0] != '\0'))
	KlPath = s;

    /* set up command line arguments */
    KlArgc = argc - script;
    KlArgv = argv + script;

    /* first, initialize klone */
    KlDeclareExtension("main", 0, 0, main_init);
#ifdef USE_CANVAS
    KnInitialize(False);
    kno_init_hook = KnOInitHook;
#endif

    
    KlInit();

#ifdef KLMATH				/* add proto of math lib */
    KlMathInit();
#endif
#ifdef KLSES
    KlSesInit();
    KlSesInitComms(0);
#endif /* KLSES */

    /* print version number */
    if (options.verbose)
	KlPrintBanner();


    /* handle the special case of a script given as argument */
    if (options.file) {
	FILE *fd;
	
	noisy =0;
	script = 1;
	KlUserProfileName = 0;
	read_stdin = 0;
	
	if (fd = fopen(options.file, "r")) {
	    KlyySkipExecLine(fd);
	    KlStdyyRedirect(KlStreamMake(fd, 1, options.file));
	} else {
	    fprintf(stderr, "Cannot open %s\n", options.file);
	    exit(1);
	}
    }
    
    /* define a catch-all point */
    KlSetNonCatchedErrorPoint(1); 	/* = 0 for set */

    /* check if we closed stdout. in this case, abort */
    if (noisy && !KlStdout->direction)
	KlExit(0);

    
    MkloneInitialize(argc, argv);


    if(options.debug) {
	KlString name;
	name = KlStringMake("kdb");
	KlLoad(name);
    }

    if(options.loadfile) {
	KlString name;
	name = KlStringMake(options.loadfile);
	KlLoad(name);
    }
    
    if (read_stdin)
        toplevel_evaluator = MkloneToplevel;
    else
	toplevel_evaluator = toplevel;

    for (;;) {
	KlCatch(KlA_ERROR, CFAPPLY(toplevel_evaluator, ()), result, normal);

	if (script && !KlContinueReadingOnError)
	    KlExit(1);

	if (noisy && !KlStdout->direction)
	    KlExit(0);
    }

}

int
toplevel()
{
    /* main routine: read/eval/print */
    for (;;) {
	KlO KlEvalExpr;
	KlGCMark();
	if (!KlRead())			/* EOF */
	    break;
	if (KlEvalExpr = KlSend_eval(KlReadExpr)) {
	    if (noisy) {
		KlPuts(" = ");
		KlToplevelPrint(KlEvalExpr);
		KlPutchar('\n');
		KlFlush(0);
	    }
	}
	KlGC();				/* do a GC after each eval */
    }
    
    MkloneToplevel(0);
}

KlPrintBanner()
{
    printf("KLM toplevel %s\n", rcsid);
    fprintf(stderr, "X11R%d, %s\n",
	    XtSpecificationRelease, XmVERSION_STRING);
}
