/* Copyright 1995 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: mklone.c,v 1.5 96/10/23 17:12:39 leon Exp $ */

#include "flags.h"
#include "global.h"
#include <Xm/PushB.h>
#include <Xm/SelectioB.h>
#include <Xm/MenuShell.h>
#include <Xm/MessageB.h>
#include <X11/IntrinsicP.h>
#include <X11/cursorfont.h>
#include <X11/Xmu/Editres.h>
#include <stdio.h>
#include <assert.h>
#include <KlLib.h>
#include "kl_motif.h"
#ifdef USE_CANVAS
#include "kl_knvas.h"
#endif /* USE_CANVAS */
#include "kl_widget.h"
#include "kl_classes.h"
#include "kl_convert.h"
#ifdef USE_TCP
#include "ipc/TcpServer.h"
#endif
#include "funcs.h"
#include "boot-klm.h"
#include "top-klm.h"
#include <syslog.h>


#ifdef KLSES

#ifdef SYSV
#include <sys/types.h>
#endif					/* SYSV */
#include <sys/file.h>
#include <fcntl.h>
#include <sys/time.h>
#ifdef NEED_SELECT_H
#include <sys/select.h>
#endif
#include <errno.h>

#include "EXTERN.h"
#include "klone.h"
#include "kl_number.h"
#include "kl_string.h"
#include "kl_atom.h"
#include "kl_list.h"
#include "kl_stream.h"
#include "kl_func.h"
#include "ses.h"
#include "klsesi.h"

#include "INTERN.h"
#endif /* KLSES */

#define STDIN_FD 0


GlobalRec klm;			/* global record */



void
KlExecuteTimerProc(XtPointer client_data, XtIntervalId *id)
{
    KlList func = (KlList)client_data;
    
    assert(func);
    KlArgumentMustHaveTrait(func, 0, KlTrait_list);
    KlSend_eval(func);
}




void
KlExecuteInput(XtPointer client_data, int *source, XtInputId *id)
{
    KlList func = (KlList)client_data;
    
    assert(func);
    KlArgumentMustHaveTrait(func, 0, KlTrait_list);
    KlSend_eval(func);
}


void
KloneInput(XtPointer cdata, int *fd, XtInputId *id)
{
    KlPoolToplevel();
    while(READ_DATA_PENDING(stdin)) {
	KlPoolToplevel();
    }
}


typedef struct _PoolRec {
    int plevel;
    Boolean first;
} PoolRec, *Pool;


#ifdef USE_TCP
void
PoolExternInput(Widget w, XtPointer client, XtPointer call)
{
    Pool pool = (Pool)client;
    TcpServerCallbackStruct *cs = (TcpServerCallbackStruct*)call;
    char *buffer;
    int input_len;
    
    buffer = cs->buffer;
    input_len = cs->n;

    if(pool->first) {
	pool->plevel = KlPool(NULL);
    } else {
	/* ensures that if we jump out of here in the middle we redo inits */
	pool->first = 1;
    }


    /* Make KlPool gobble all this */
    if (!(pool->plevel = KlPool(buffer))) {

	KlExecuteStringNoReturn(KlStdpool->klstring->string);
	KlPool(NULL);

    }

    pool->first = 0;
}    
#endif


KlWidget
KlToplevelMake(KlString name, KlString class)
{
    Widget toplevel, server;
    KlList list;
    Int i;
    KlString str;
    String *fallbacks;
    String shell_name;
    Arg args[8];
    char rdv[32];
    KlWidget result;
    Boolean connected;
    KlAtom KlA_falllback_resources = KlIntern("*fallback-resources*");

    
    list = (KlList)KlA_falllback_resources->c_val;
    
    assert(list);
    assert(class);
    if(list->type == KlListType) {
	fallbacks = (String *)KlAlloca(sizeof(String) * (list->size+1));
	for(i = 0; i <list->size; i++) {
	    str = (KlString)list->list[i];
	    assert(str);
	    KlArgumentMustBe(str, 0, KlStringType);
	    fallbacks[i] = str->string;
	}
	fallbacks[list->size] = NULL;
	XtAppSetFallbackResources(klm.app, fallbacks);
    }
    /* well, openning multiple displays is not really necessary ... but
       it is required if we wanna load the fallback resources */
    klm.dpy = XtOpenDisplay(klm.app, (char *)NULL, NULL, class->string, 
			0, 0, &(klm.argc), klm.argv);

    if(NULL == klm.dpy) {
	return (KlWidget)NIL;
    }

    shell_name = name->string;
    if(0 == strlen(shell_name)) {
	shell_name = 0;
    }
    toplevel = 	
	XtAppCreateShell(shell_name, class->string, 
			 applicationShellWidgetClass, klm.dpy, 0, 0);
    XtAddEventHandler(toplevel, 0, True, _XEditResCheckMessages, NULL);
    {
	/* register here the converters, so that we can override some converter
           added by Motif (StringToXmString) */
	static Boolean need_register = True;
	Tm_RegisterConverters(NULL, klm.app);
	RegisterKloneConverters(klm.app);
	need_register = False;
    }
#ifdef USE_TCP
    if(options.server)  {
	int n = 0;
	Pool pool;
	pool = (Pool)XtMalloc(sizeof(PoolRec));
	pool->plevel = 0;
	pool->first = True;
	sprintf(rdv, "/%s", shell_name);

	XtSetArg(args[n], XtNconnected, True); n++;
	XtSetArg(args[n], XtNrendezVous, rdv); n++;
	server =
	    XtCreateWidget(rdv, tcpServerObjectClass, toplevel, args, n);
	XtAddCallback(server, XtNreceiveCallback, PoolExternInput, pool); 
	
	XtVaGetValues(server, XtNconnected, &connected, 0);
	if(! connected) {
	    printf("Server initialization failed.\n");
	}
    }
#endif
    
    result =  KlWidgetMake(toplevel);
    KlListAppend(klm.toplevels, result);
    return result;

}

static struct {
  XtInputId input_id;
  KlStream kl_stream;
  KlO kl_lisp;
} Observe[1024];
static int ObserveIdx=0;

KlO
KlObserveInput(KlStream stream, KlO lisp)
{
    KlIncRef(lisp);
    KlIncRef(stream);

    Observe[ObserveIdx].kl_stream = stream;
    Observe[ObserveIdx].kl_lisp = lisp;
    Observe[ObserveIdx].input_id =     XtAppAddInput(klm.app,
		  fileno(stream->fd),
		  (XtPointer)XtInputReadMask,
		  (XtInputCallbackProc)KlExecuteInput,
		  (XtPointer)lisp);
    ObserveIdx++;
    return (KlO)stream;
}    

KlO
KlUnobserveInput(KlStream stream)
{

  int i;
  for (i=0;i<ObserveIdx;i++)
    if (Observe[i].kl_stream == stream) {
      XtRemoveInput(Observe[i].input_id);
      KlDecRef(Observe[i].kl_stream);
      KlDecRef(Observe[i].kl_lisp);
      Observe[i] = Observe[--ObserveIdx];
      return TRU;
    }
  return NIL;
}    



KlO
KlWidgetRealize(KlWidget klw)
{
    KlArgumentMustBeValidWidget(klw, 0);
    
    XtRealizeWidget(klw->w);
    XSync(XtDisplayOfObject(klw->w), False);
    XmUpdateDisplay(klw->w);
    return (KlO)klw;
}




KlO
KlWidgetUpdate(KlWidget klw)
{
    KlArgumentMustBeValidWidget(klw, 0);
    
#ifdef USE_CANVAS
    if(XtIsSubclass(klw->w, knvasWidgetClass)) {
      KnvasUpdateDisplay(klw->w);
    }
#endif	    
    XmUpdateDisplay(klw->w);
    return (KlO)klw;
}



KlO
KlWidgetDestroy(Int argc, KlO argv[])
{
    int i;
    KnO kno;
    Display *dpy;

    for(i = 0; i < argc; i++) {
	kno = (KnO)argv[i];
#ifdef USE_CANVAS
	if(KlIsOfType(kno, KnLayerClass)) {
	    KnvasRemoveLayer(GetKnvas(kno), kno);
	}
	else
	if(KlIsOfType(kno, KnOClass)) {
	    KnRemove(GetKnvas(kno), kno);
	}
	else
#endif	    
        {
	    KlWidget klw = (KlWidget)kno;

	    KlArgumentMustBeValidWidget(klw, i);
	    XtDestroyWidget(klw->w);
	}
    }
    return TRU;
}



KlO
KlWidgetManage(Int argc, KlO argv[])
{
    extern XEvent last_event;	/* defined in event.c */
    KlWidget klw = NULL;
    int i;

    for(i = 0; i < argc; i++) {
	klw = (KlWidget)argv[i];
	KlArgumentMustBeValidWidget(klw, i);

	if(XtIsShell(klw->w)) {
	    /* let's try to unify show & build */
	    if(! XtIsRealized(klw->w)) {
		KlWidgetRealize(klw);
	    }
	}
	else {
	    if(XmIsMenuShell(XtParent(klw->w))) {
		XmMenuPosition(klw->w, &last_event);
	    }
	    XtManageChild(klw->w);
	}
    }
    if(klw) {
	XFlush(XtDisplayOfObject(klw->w));
    }
    return TRU;
}



KlO
KlWidgetUnmanage(Int argc, KlO argv[])
{
    int i;
    KlWidget klw = NULL;
    for(i = 0; i < argc; i++) {
	klw = (KlWidget)argv[i];
	KlArgumentMustBeValidWidget(klw, i);
	if(XtParent(klw->w)) {	/* do *not* unmanage a toplevel shell */
	    XtUnmanageChild(klw->w);
	}
    }
    if(klw) {
	XFlush(XtDisplayOfObject(klw->w));
    }
    return TRU;
}


KlWidget
KlTrackWidget(KlWidget klw, KlString cursor)
{
    Widget res;
    XEvent e;
    Cursor cross;
    /* display may change from one call to another so we cannot keep the cursor
       in a static variable */
    KlArgumentMustBeValidWidget(klw, 0);
    
    cross = XCreateFontCursor(XtDisplay(klw->w), XC_cross);
    res = XmTrackingEvent(klw->w, cross, False, &e);
    XFreeCursor(XtDisplay(klw->w), cross);
    return (res ? KlWidgetMake(res) : (KlWidget)NIL);
}    


static void
PrintResourceList(KlWidgetClass wc)
{
    Cardinal i;
    assert(wc);
    printf("Resources for class %s:\n", wc->class);
    for(i = 0; i < wc->num; i++){
	printf("  %.32s of type %s\n", wc->resources[i].resource_name,
	       wc->resources[i].resource_type);
    }
}



KlWidget
KlWidgetListResources(KlWidget klw)
{
    KlWidgetClass wc = NULL;

    KlArgumentMustBeValidWidget(klw, 0);
    
    wc = (KlWidgetClass)KlGetWidgetClass(klw);
    if(! wc) {
	XtAppWarning(XtWidgetToApplicationContext(klw->w),
		     "Cannot find widget class.");
	return (KlWidget)NIL;
    }
    if(! wc->resources) {
	XtGetResourceList(klw->w->core.widget_class,
			  &(wc->resources), &(wc->num));
    }
    PrintResourceList(wc);
    return (KlWidget)TRU;
}





KlWidget
KlWidgetSetValue(KlWidget klw, KlString name, KlString value)
{

    assert(value);
    assert(name);
    
    KlArgumentMustBeValidWidget(klw, 0);
    KlArgumentMustHaveTrait(name, 1, KlTrait_string);
    KlArgumentMustHaveTrait(value, 2, KlTrait_string);
    XtVaSetValues(klw->w, XtVaTypedArg,
		  name->string, XtRString, (XtArgVal)value->string, 
		  strlen(value->string),
		  0);
    XFlush(XtDisplayOfObject(klw->w));
    return klw;
}


static void
SetBooleanCB(Widget w, XtPointer client, XtPointer call)
{
    Boolean *b = (Boolean *)client;
    *b = True;
}



KlO
KlLocalLoop(KlWidget klw)
{
    XEvent e;
    Widget shell;
    volatile Boolean end_loop = False;
    volatile Boolean result = False;

    KlArgumentMustBeValidWidget(klw, 0);
    if( ! (XmIsMessageBox(klw->w) || XmIsSelectionBox(klw->w))) {
	XtAppWarning(klm.app,
		   "localloop actually requires a MessageBox or SelectionBox subclass.");
	return NIL;
    }
    shell = XtParent(klw->w);
    XtAddCallback(shell, XmNpopdownCallback,
		  SetBooleanCB, (XtPointer)&end_loop);
    XtAddCallback(klw->w, XmNokCallback, SetBooleanCB, (XtPointer)&result);
		  
    while(! end_loop) {
	XtAppNextEvent(klm.app, &e);
	XtDispatchEvent(&e);
    }
    XtRemoveCallback(shell, XmNpopdownCallback, SetBooleanCB,
		     (XtPointer)&end_loop);
    XtRemoveCallback(klw->w, XmNokCallback, SetBooleanCB, (XtPointer)&result);

    return result ? TRU : NIL;
}    


KlNumber
KlAddTimeout(KlNumber tms, KlList value)
{
    XtIntervalId id;
    assert(tms);
    KlArgumentMustHaveTrait(tms, 0, KlTrait_number);
    KlMustBeExecutable(value, 1);
    
    KlIncRef(value);

    id = XtAppAddTimeOut(klm.app, tms->number, KlExecuteTimerProc, value);
    return KlNumberMake(id);
}



KlNumber
KlRemoveTimeout(KlNumber id)
{
    KlArgumentMustBe(id, 0, KlNumberType);
    XtRemoveTimeOut(id->number);
    return id;
}



KlWidget
KlGetWidgetFromName(KlWidget parent, KlString name)
{
    Widget root = NULL, w = NULL;
    String sname;
    Cardinal n;
    if((KlO)parent != NIL) {
	KlArgumentMustBeValidWidget(parent, 0);
	root = parent->w;	
    }
    KlArgumentMustHaveTrait(name, 0, KlTrait_string);
    sname = name->string;
    
    if(root) {
	w = XtNameToWidget(root, sname);
    }
    else  if(sname[0] == '^'){
	/* ABSOLUTE PATH NAME */
	String rest;
	
	sname = &(sname[1]);
	rest = strchr(sname, '.');
	if(rest) {
	    *rest = 0;
	    rest++;
	}
	for(n = 0; n < klm.toplevels->size; n++) {
	    parent = (KlWidget)klm.toplevels->list[n];
	    root = parent->w;
	    if(! strcmp(XtName(root), sname)) {
		if(rest)
		    w = XtNameToWidget(root, rest);
		else
		    w = root;
		if(w)
		    break;
	    }
	}
    }
    if(w) {
	return KlWidgetMake(w);
    }
    else {
	return (KlWidget)NIL;
    }
}

extern KlO KlDefineCursor(KlWidget widget, KlO cursor);


KlO
KlPutResource(KlString line)
{
    int i;
    KlWidget top;
    Widget w;
    XrmDatabase xdb;
    KlArgumentMustBe(line, 0, KlStringType);
    for(i = 0; i < klm.toplevels->size; i++) {
	top  = (KlWidget)klm.toplevels->list[i];
	w = top->w;
	xdb = XrmGetDatabase(XtDisplayOfObject(w));
	if(NULL != xdb) {
	    XrmPutLineResource(&xdb, line->string);
	}
    }
    return TRU;
}    

#ifdef KLSES
void SesMainDispatchEvent(XtPointer closure, int* source, XtInputId*id) {
  KlSesHandler(closure, 0, id);
  KlSesMainPreSelect();
}
#endif /* KLSES */


MkloneInitialize(int argc, char *argv[])
{
    klm.dpy = NULL;
    klm.app = NULL;
    klm.argc = 0;
    klm.argv = NULL;
    klm.toplevels = KlListNMake(0);
    KlIncRef(klm.toplevels);
    
    KlWidgetClassInitialize();
    KlPseudoClassStructInitialize();
    KlEventStructInitialize();
#ifdef USE_CANVAS
    KlTagClassInitialize();
#endif

    KlA_BAD_WIDGET = KlErrorCodeMakeS("BAD_WIDGET", "bad widget : %0");
    
    KlDeclareSubr(KlWidgetRealize, "build", 1);
    KlDeclareSubr(KlWidgetDestroy, "destroy", NARY);
    KlDeclareSubr(KlWidgetManage, "show", NARY);
    KlDeclareSubr(KlWidgetUnmanage, "hide", NARY);
    KlDeclareSubr(KlWidgetUpdate, "update", 1);
    KlDeclareSubr(KlBindCallback, "callback", 3);
    KlDeclareSubr(KlBindCallback, "add-callback", 3);
    KlDeclareSubr(KlUnbindCallback, "rm-callback", 3);
    KlDeclareSubr(KlBindAction, "bind", 3);
    KlDeclareSubr(KlUnbindAction, "unbind", 2);
    KlDeclareSubr(KlObserveInput, "observe", 2);
    KlDeclareSubr(KlUnobserveInput, "unobserve", 1);
    KlDeclareSubr(KlAddTimeout, "timeout", 2);
    KlDeclareSubr(KlAddTimeout, "untimeout", 1);
    KlDeclareSubr(KlGetWidgetFromName, "get-widget", 2);
    KlDeclareSubr(KlLocalLoop, "localloop", 1);
    KlDeclareSubr(KlTrackWidget, "track", 2);
    KlDeclareSubr(KlDefineCursor, "cursor", 2);
    KlDeclareSubr(KlPutResource, "put-resource", 1);
    
    
    /* could become a pseudo-resource */
    KlDeclareSubr(KlWidgetListResources, "resources", 1);

    KlRegisterWidgetClasses();
    
    KlDeclareSubr(KlToplevelMake, "Toplevel", 2);

    KlExecuteStringNoReturn(KlmBootCode);

    
    XtToolkitInitialize();
    klm.app = XtCreateApplicationContext();
    {
	/* record action rec */
	static XtActionsRec
	    actions[] = {
		"ExecuteKloneAction", ExecuteKloneAction,
	    };
	XtAppAddActions(klm.app, actions, XtNumber(actions));
    }

    klm.argc = argc;
    klm.argv = argv;

    if(options.top) {
	KlExecuteStringNoReturn(KlmTopCode);
    }
#ifdef KLSES
    XtAppAddInput(klm.app, KlSes_fd_main_r, (XtPointer)XtInputReadMask, 
		  SesMainDispatchEvent, 0);
#endif /* KLSES */
}




extern int read_stdin;
MkloneToplevel()
{
  XEvent e;
    static done = False;
    if(read_stdin && ! done) {
	XtAppAddInput(klm.app, STDIN_FD, (XtPointer)XtInputReadMask, 
		      KloneInput, 0);
	done = True;
	KlDrawPrompt( KlPool(NULL)); /* draw prompt the first time */
    }
    /*    XtAppMainLoop(klm.app);*/
    for(;;) {
#ifdef KLSES	
      KlSesMainPreSelect();
#endif /* KLSES */      
      XtAppNextEvent(klm.app, &e);
      XtDispatchEvent(&e);
    }

}


