/* Copyright (c) 1995 by Groupe Bull.  All Rights Reserved */
/* $Id: pclass.c,v 1.2 95/06/07 14:47:14 leon Exp $ */


#include "kl_widget.h"
#include "funcs.h"

KlStructClass KlPseudoClassType;




KlWidget
KlSetPseudoClass(KlWidget klw, KlStruct pclass)
{
    KlArgumentMustBeValidWidget(klw, 0);
    klw->pclass = (KlPseudoClass)pclass->ptr;
/*    KlIncRef(klw->pclass); */
    return klw;
}

KlO
KlPseudoClassMake()
{
    KlPseudoClass pc;
    KlO pseudoclass;
    pc = (KlPseudoClass)KlMalloc(sizeof(KlPseudoClassRec));
    pc->name = NIL; 
    pc->initialize = NIL;
    pc->destroy = NIL;
    pc->setvalues = NIL;
    pc->getvalue = NIL;
    pc->resources = (KlList)NIL;
    pc->callbacks = (KlList)NIL;
    pseudoclass = (KlO)KlStructMake(KlPseudoClassType, pc);
    KlIncRef(pseudoclass);
    return(pseudoclass);
}    



void
KlPseudoClassStructInitialize()
{
    KlPseudoClassType = KlStructClassMake("KlWidgetClass",
					  sizeof(KlPseudoClassRec),
					  7);
    
    KlDeclareStructClassSlot(KlPseudoClassType,
			     "name",
			     KlStructAccessorKlone,
			     KlOffsetOf (KlPseudoClassRec, name));
			     
    KlDeclareStructClassSlot(KlPseudoClassType,
			     "initialize",
			     KlStructAccessorKlone,
			     KlOffsetOf (KlPseudoClassRec, initialize));
    
    KlDeclareStructClassSlot(KlPseudoClassType,
			     "destroy",
			     KlStructAccessorKlone,
			     KlOffsetOf (KlPseudoClassRec, destroy));
    
    KlDeclareStructClassSlot(KlPseudoClassType,
			     "getvalue",
			     KlStructAccessorKlone,
			     KlOffsetOf (KlPseudoClassRec, getvalue));
    
    KlDeclareStructClassSlot(KlPseudoClassType,
			     "setvalues",
			     KlStructAccessorKlone,
			     KlOffsetOf (KlPseudoClassRec, setvalues));
    
    KlDeclareStructClassSlot(KlPseudoClassType,
			     "resources",
			     KlStructAccessorKlone,
			     KlOffsetOf (KlPseudoClassRec, resources));
    
    KlDeclareStructClassSlot(KlPseudoClassType,
			     "callbacks",
			     KlStructAccessorKlone,
			     KlOffsetOf (KlPseudoClassRec, callbacks));

    KlDeclareSubr(KlPseudoClassMake, "make-widget-class", 0);
    KlDeclareSubr(KlSetPseudoClass, "set-widget-class", 2);
}







void
KlWidgetSetPseudoValues(KlWidget klw, int argc, KlO *argv)
{
    Int i, n, nr;
    KlAtom name, res;
    KlString value;
    char *sname, *rname;
    KlPseudoClass pc;
    KlO *eval_list = argv;
    KlList pseudo_res;
    KlO *regular_res;
    Boolean is_pseudo;

    pseudo_res = KlListNMake(0);
    KlIncRef(pseudo_res);

    regular_res = KlAlloca(sizeof(KlO)* argc);
    nr = 0;
    pc = klw->pclass;

    /* in the loop, we first split the resource list into pseudo and regular
       resources */
    for(i = 0; i < argc; i+=2) {
	name = (KlAtom) eval_list[i];
	value = (KlString) eval_list[i+1];

	sname = GetNameString((KlO)name);
	
	is_pseudo = False;
	for(n = 0; n < pc->resources->size; n++) {
	    res = (KlAtom)pc->resources->list[n];
	    rname = GetNameString((KlO)res);
	    if(! strcmp (sname, rname)) {
		/* resource found : pseudo one*/
		KlListAppend(pseudo_res, name);
		KlListAppend(pseudo_res, value);
		is_pseudo = True;
		break;
	    }
	}
	if(! is_pseudo) {
	    /* regular resource */
	    regular_res[nr] = (KlO)name; nr++;
	    regular_res[nr] = (KlO)value; nr++;
	}
    }
    if(nr > 0) {
	KlWidgetSetValues(klw, nr, regular_res);
    }
    if(pseudo_res->size > 0) {
	KlO func[3];

	func[0] = (KlO)pc->setvalues;
	func[1] = (KlO)klw;
	func[2] = (KlO)pseudo_res;
	
	KlExecuteKloneNoReturn(3, func);
    }
    KlDecRef(pseudo_res);
}
    


KlO
KlWidgetGetPseudoValue(KlWidget klw, char *sname, KlO selector)
{
    KlPseudoClass pc;
    int n;
    KlAtom res;

    pc = klw->pclass;
    for(n = 0; n < pc->resources->size; n++) {
	res = (KlAtom)pc->resources->list[n];
	if(! strcmp (sname, GetNameString((KlO)res))) {
	    /* resource found : pseudo one. execute the getvalue hook */
	    KlO func[3];
	    
	    func[0] = (KlO)pc->getvalue;
	    func[1] = (KlO)klw;
	    func[2] = (KlO)selector;
	    
	    return KlExecuteKlone(3, func);
	}
    }
    return (KlO)KlWidgetGetValue(klw, sname);
}
