/*****************************************************************************\
*                                                                             *
* 	       an exemple of a ses extension: simple calculator               *
*                                                                             *
\*****************************************************************************/
/* compile by:
   klone_cc ses_calc.c seslib.o -o ses_calc
 */

#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#include <stdio.h>
#include <errno.h>

#include "seslib.h"

SES_Connection sescon;

static Display *display;

/*****************************************************************************\
* 				   WRAPPERS                                   *
\*****************************************************************************/

GC cleargc, drawgc;

void
drawinit(args)
    SES_Args args;
{
  Window win = SES_ArgInt(args);
  
  XSelectInput(display, win, ButtonPressMask|ButtonMotionMask|ButtonReleaseMask);
  XFlush(display);

  SES_ReturnInt(args, win);

}

void
drawrect(args)
    SES_Args args;
{
  Window win = SES_ArgInt(args);
    int x = SES_ArgInt(args);
    int y = SES_ArgInt(args);
    int w = SES_ArgInt(args);
    int h = SES_ArgInt(args);

    XFillRectangle(display, win, drawgc, x, y, w, h);
    XFlush(display);
  SES_ReturnInt(args, win);
}

void
clearrect(args)
    SES_Args args;
{
  Window win = SES_ArgInt(args);
    int x = SES_ArgInt(args);
    int y = SES_ArgInt(args);
    int w = SES_ArgInt(args);
    int h = SES_ArgInt(args);

    XFillRectangle(display, win, cleargc, x, y, w, h);
    XFlush(display);
  SES_ReturnInt(args, win);
}


/*****************************************************************************\
* 			     INIT, DECLARES, MAIN                             *
\*****************************************************************************/

void SesDispatchEvent(XtPointer closure, int* source, XtInputId*id) {
  SES_Handler(sescon);
  SES_MainPreSelect(sescon);
}

static char press[] = "press";
static char move[] = "move";
static char release[] = "release";
static char command[256];

void main(int argc, char *argv[])
{
  XtAppContext app;
  XEvent e;
    sescon = SES_Init("draw");		/* init */

					/* export functions */
    SES_DeclareFunction(sescon, "draw:init", drawinit, 1, SES_Int,
			SES_Int, 0);

    SES_DeclareFunction(sescon, "draw:rect", drawrect, 5, SES_Int,
			SES_Int,
			SES_Int, SES_Int,
			SES_Int, SES_Int, 0);
    SES_DeclareFunction(sescon, "draw:clear", clearrect, 5, SES_Int,
			SES_Int,
			SES_Int, SES_Int,
			SES_Int, SES_Int, 0);

    XtToolkitInitialize();
    app = XtCreateApplicationContext();
    display = XtOpenDisplay(app, NULL, "Ses_Draw", "ses_draw",
			    NULL, 0, &argc, argv);
    cleargc = XCreateGC(display, DefaultRootWindow(display), None, NULL);
    XSetFunction(display, cleargc, GXxor);
    XSetForeground(display, cleargc, 4);
    drawgc = cleargc;


    SES_Ready(sescon);

    XtAppAddInput(app, SES_ConnectionFD(sescon),
		  (XtPointer)XtInputReadMask, SesDispatchEvent, NULL);
    
    for(;;) {
      char *action = NULL;
      SES_MainPreSelect(sescon);
      XtAppNextEvent(app, &e);
      if(e.type == ButtonPress) {
	action = press;
      }
      else if(e.type == MotionNotify) {
	action = move;
      }
      else if(e.type == ButtonRelease) {
	action = release;
      }
      if(NULL != action) {
	XButtonEvent *me = (XButtonEvent*)&e;
	sprintf(command, "(sesdemo:%s %d %d)", action, me->x, me->y);
	SES_SendText(sescon, NULL, command, strlen(command), NULL, 1);
      }
    }
}
