/* Copyright (c) 1995 by Groupe Bull.  All Rights Reserved */
/* $Id: setval.c,v 1.3 95/10/24 15:07:59 leon Exp $ */


#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Kn/interactor.h>
#include "kl_widget.h"
#include "kl_classes.h"
#include "kl_convert.h"
#include "funcs.h"

/* This is private in the r5 distrib */
#if (XtSpecificationRelease < 6)
typedef struct {
    String      name;   /* resource name */
    String      type;   /* representation type name */
    XtArgVal    value;  /* representation */
    int         size;   /* size of representation */
} XtTypedArg, *XtTypedArgList;
#endif /*  (XtSpecificationRelease < 6) */

typedef void (*SetValProc)(KlWidget, /* the KlWidget */
			   KlO,	/* the new value as KlO */
			   XtTypedArg *); /* the new value converted */

typedef struct SetValRec {
    char *name;
    SetValProc proc;
} SetValRec;




static void
SetSensitive(KlWidget klw, KlO value, XtTypedArg *argument)
{
    Boolean target;
    XrmValue from;
    XrmValue to;

    from.addr = (void *)argument->value;
    from.size = argument->size;
    to.addr = &target;
    to.size = sizeof(Boolean);
    if(XtConvertAndStore(klw->w, argument->type, &from,
		      XtRBoolean, &to)) {
	XtSetSensitive(klw->w, target);
    }
}    


static void
SetEventMask(KlWidget klw, KlO value, XtTypedArg *argument)
{
    KlArgumentMustBe(value, 0, KlNumberType);
    if(XtIsRealized(klw->w)) {
	XSelectInput(XtDisplayOfObject(klw->w),
		     XtWindowOfObject(klw->w), ((KlNumber)value)->number);
    }
}


static void
SetInteractor(KlWidget klw, KlO value, XtTypedArg *unused)
{
    KlArgumentMustBe(value, 0, KnInteractorClass);
    KnvasAddInteractor(klw->w, (KnInteractor)value);
}


static void
SetGeometry(KlWidget klw, KlO value, XtTypedArg *unused)
{
    char *geometry;
    int mask = 0;
    Position x, y;
    Dimension width, height;
    int x_r, y_r;
    unsigned int width_r, height_r;
    Arg args[4];
    int n = 0;
    
    KlArgumentMustHaveTrait(value, 0, KlTrait_string);
    geometry = GetNameString(value);

    mask = XParseGeometry(geometry, &x_r, &y_r, &width_r, &height_r);
    /* be carefull here on size conversions by Xt */
    x = x_r;
    y = y_r;
    width = width_r;
    height = height_r;
    if(mask & XValue) {
	XtSetArg(args[n], XmNx, x); n++;
    }
    if(mask & YValue) {
	XtSetArg(args[n], XmNy, y); n++;
    }
    if(mask & WidthValue) {
	XtSetArg(args[n], XmNwidth, width); n++;
    }
    if(mask & HeightValue) {
	XtSetArg(args[n], XmNheight, height); n++;
    }
    XtSetValues(klw->w, args, n);
}

static void
SetSubject(KlWidget klw, KlO value, XtTypedArg *unused)
{
    static KlO ssub = NULL;
    if(NULL == ssub) {
	ssub = KlIntern("set-subject");
    }
    KlApply2(ssub, klw, value);
}

/* some resources require a special SetValue procedure */
static SetValRec special_set_table[] = {
    {"sensitive", SetSensitive},
    {"event-mask", SetEventMask},	/* pseudo-resource  */
    {"interactor", SetInteractor},
    {"geometry", SetGeometry},
    {"subject", SetSubject},
};

static SetValProc
NeedSpecialSetValue(char *name)
{
    int i;
    for(i = 0; i < XtNumber(special_set_table); i++) {
	if(!strcmp(name, special_set_table[i].name)) {
	    return special_set_table[i].proc;
	}
    }
    return NULL;
}    




KlO
KlWidgetSetValues(KlWidget klw, int argc, KlO *argv)
{
    KlO result;
    KlAtom name;
    KlString value;
    SetValProc setval;
    Int i;
    XtTypedArgList argument, nestedlist;
    char *svalue;

    KlO *eval_list = argv;
    

    /* Here, we use a nested list of arguments, because we cannot fill the
       arglist directly. This processes multiple SetValue in one call. */
    nestedlist =
	(XtTypedArgList)KlAlloca(sizeof(XtTypedArg) * ((argc/2) + 1));
    argument = nestedlist;
    for(i = 0; i < argc; i+=2) {
	name = (KlAtom) eval_list[i];
	value = (KlString) eval_list[i+1];

	KlArgumentMustHaveTrait(name, i, KlTrait_string);
	argument->name = GetNameString((KlO)name);

	

	/* WARNING: since we are using ALLOCA, the following code MUST NOT BE
           SEPARATED in a function */
	if(KlHasTrait(value, KlTrait_number)) {
	    /* numeric arguments are allowed */
	    svalue = (char *)KlAlloca(8);
	    sprintf(svalue, "%d", ((KlNumber)value)->number);
	    argument->type = (String)XtRString;
	    argument->value = (XtArgVal)svalue;
	    argument->size = (XtArgVal)strlen(svalue)+1; /* trailing 0 */
	}
	else if (KlHasTrait(value, KlTrait_list)) {
	    /* for some resources List argument may be usefull.
	       We provide the KlList -> XmStringTable converter */
	
	    argument->type = (String)"KlList";
	    argument->value = (XtArgVal)value;
	    argument->size = (XtArgVal)sizeof(KlList);
	}
	else if (value->type == KlWidgetType) {
	    /* special case: widgets may not be converted from strings */
	    /* WARNING !!! UGLY !!! USE A TABLE INSTEAD !!! */
	    if(!strcmp(argument->name, XmNmenuHelpWidget) ||
	       !strcmp(argument->name, XmNmenuHistory) ||
	       !strcmp(argument->name, XmNsubMenuId)) {
		argument->type = (String)XmRMenuWidget;
	    }
	    else {
		argument->type = (String)XmRWidget;
	    }
	    argument->value = (XtArgVal)((KlWidget)value)->w;
	    argument->size = (XtArgVal)sizeof(Widget);
	}
	else if (KlHasTrait(value, KlTrait_string)) {
	    svalue = GetNameString((KlO)value);
	    argument->type = (String)XtRString;
	    argument->value = (XtArgVal)svalue;
	    argument->size = (XtArgVal)strlen(svalue)+1; /* trailing 0 */
	}
	else {
	    argument->type = "KlO";
	    argument->value = (XtArgVal)value;
	    argument->size = (XtArgVal)sizeof(KlO);
	}
	
	
	if(setval = NeedSpecialSetValue(argument->name)) {
	    /* if there is a special setvalue procedure, call it */
	    (*setval)(klw, (KlO)value, argument);
	}
	else {
	    /* else, let Xt do its job */
	    argument++;
	}
    }
    argument->name =  NULL;
    argument->type =  NULL;
    argument->value =  (XtArgVal)0;
    argument->size =  0;
    XtVaSetValues(klw->w, XtVaNestedList, nestedlist, 0);
    
    XFlush(XtDisplayOfObject(klw->w));
    return (KlO)klw;
}
